
// brush.h


typedef struct
{
	int		numpoints;
	int		maxpoints;
	vec_t 	points[8][5];			// variable sized
} winding_t;


// the normals on planes point OUT of the brush
#define	MAXPOINTS	16
typedef struct face_s
{
	struct face_s	*next;
	vec3_t			planepts[3];
    texdef_t		texdef;
    plane_t			plane;
	winding_t		*face_winding;
	float			d_color[3];
	qtexture_t		*d_texture;
//	int				d_numpoints;
//	vec3_t		   *d_points;
	vec_t			c_s, c_t;
} face_t;

#define	MAX_FACES	16
typedef struct brush_s
{
	struct brush_s	*prev, *next;	// links in active/selected
	struct brush_s	*oprev, *onext;	// links in entity
	struct entity_s	*owner;
	vec3_t			mins, maxs;

	face_t			*brush_faces;
	visgroup_t		*group;
} brush_t;

void		Brush_AddToList				(brush_t *, brush_t *);
void		Brush_Build					(brush_t *);
void		Brush_BuildWindings			(brush_t *);
brush_t		*Brush_Clone				(brush_t *);
brush_t		*Brush_Create				(vec3_t, vec3_t, texdef_t *);
void		Brush_Draw					(brush_t *, BOOL);
void		Brush_DrawXY				(brush_t *);
void		Brush_Free					(brush_t *);
void		Brush_MakeSided				(int);
void		Brush_Move					(brush_t *, vec3_t);
brush_t		*Brush_Parse				(void);
face_t		*Brush_Ray					(vec3_t, vec3_t, brush_t *, vec_t *);
void		Brush_RemoveFromList		(brush_t *);
void		Brush_SelectFaceForDragging	(brush_t *, face_t *, qboolean);
void		Brush_SetTexture			(brush_t *, texdef_t *);
void		Brush_SideSelect			(brush_t *, vec3_t, vec3_t, qboolean);
void		Brush_Write					(brush_t *, FILE *);
void		Brush_RemoveEmptyFaces		(brush_t *);
int			AddPlanept					(vec_t *);
face_t		*Face_Clone					(face_t *);
void		Face_Draw					(face_t *);
winding_t	*MakeFaceWinding			(brush_t *, face_t *);
void		Brush_FitTexture			(brush_t *b, int nHeight, int nWidth);
void		Face_FitTexture				(face_t * face, int nHeight, int nWidth);
