
#include "qe3.h"

#define	PAGEFLIPS	2

void DrawPathLines (void);

camera_t		camera;
static int		pressx;
extern GLuint	skybox_display_list;

/*
============
Cam_Init
============
*/
void Cam_Init (void)
{
//	camera.draw_mode = cd_texture;
//	camera.draw_mode = cd_solid;
//	camera.draw_mode = cd_wire;

	camera.timing = false;

	camera.origin[0] = 0;
	camera.origin[1] = 20;
	camera.origin[2] = 46;

	camera.color[0] = (float)0.3;
	camera.color[1] = (float)0.3;
	camera.color[2] = (float)0.3;
}


//============================================================================

void Cam_BuildMatrix (void)
{
	float	xa, ya;
	float	matrix[4][4];
	int		i;

	xa = camera.angles[0]/180*Q_PI;
	ya = camera.angles[1]/180*Q_PI;

	// the movement matrix is kept 2d
	// *not anymore*

    camera.forward[0] = cos(ya);
    camera.forward[1] = sin(ya);
//    camera.forward[2] = 0;
	camera.forward[2] = sin(xa);
	camera.right[0] = camera.forward[1];
    camera.right[1] = -camera.forward[0];
	AngleVectors(camera.angles, camera.forward, camera.right, camera.up);
	camera.forward[2] = -camera.forward[2];

	glGetFloatv (GL_PROJECTION_MATRIX, &matrix[0][0]);

	for (i=0 ; i<3 ; i++)
	{
		camera.vright[i] = matrix[i][0];
		camera.vup[i] = matrix[i][1];
		camera.vpn[i] = matrix[i][2];
	}

	VectorNormalize (camera.vright);
	VectorNormalize (camera.vup);
	VectorNormalize (camera.vpn);
}

//===============================================

void FollowFloor(void)
{
	brush_t	*b;
	vec_t	d, bestd = 16384, current;
	vec3_t	start, dir;

	start[0] = camera.origin[0];
	start[1] = camera.origin[1];
	start[2] = 8192;
	dir[0] = dir[1] = 0;
	dir[2] = -1;

	current = 8192 - (camera.origin[2] - (globals.EyeHeight - globals.StepHeight - 1));

	for (b=active_brushes.next ; b != &active_brushes ; b=b->next)
	{
		if (FilterBrush (b))
			continue;
		if (b->owner->eclass->fixedsize)
			continue;
		if (!Brush_Ray (start, dir, b, &d))
			continue;
		if (d > current && d < bestd)
			bestd = d;
	}
	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
	{
		if (FilterBrush (b))
			continue;
		if (b->owner->eclass->fixedsize)
			continue;
		if (!Brush_Ray (start, dir, b, &d))
			continue;
		if (d > current && d < bestd)
			bestd = d;
	}

	if (bestd == 0 || bestd == 16384)
		return;

	camera.origin[2] += current + globals.StepHeight + 1 - bestd;
	Sys_UpdateWindows (W_CAMERA|W_Z);
}

/*
===============
Cam_ChangeFloor
===============
*/
void Cam_ChangeFloor (qboolean up)
{
	brush_t	*b;
	vec_t	d, bestd, current;
	vec3_t	start, dir;

	start[0] = camera.origin[0];
	start[1] = camera.origin[1];
	start[2] = 8192;
	dir[0] = dir[1] = 0;
	dir[2] = -1;

	current = 8192 - (camera.origin[2] - globals.EyeHeight);
	if (up)
		bestd = 0;
	else
		bestd = 16384;

	for (b=active_brushes.next ; b != &active_brushes ; b=b->next)
	{
		if (!Brush_Ray (start, dir, b, &d))
			continue;
		if (up && d < current && d > bestd)
			bestd = d;
		if (!up && d > current && d < bestd)
			bestd = d;
	}
	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
	{
		if (!Brush_Ray (start, dir, b, &d))
			continue;
		if (up && d < current && d > bestd)
			bestd = d;
		if (!up && d > current && d < bestd)
			bestd = d;
	}

	if (bestd == 0 || bestd == 16384)
		return;

	camera.origin[2] += current - bestd;
	Sys_UpdateWindows (W_CAMERA|W_Z);
}

//===============================================

int	cambuttonstate;
static	int	buttonx, buttony;
static	int	cursorx, cursory;

face_t	*side_select;

#define	ANGLE_SPEED	300
#define	MOVE_SPEED	400

/*
================
Cam_PositionDrag
================
*/
void Cam_PositionDrag (void)
{
	int		x, y;

	Sys_GetCursorPos (&x, &y);
	if (x != cursorx || y != cursory)
	{
		x -= cursorx;
		VectorMA (camera.origin, x, camera.vright, camera.origin);
		y -= cursory;
		camera.origin[2] -= y;

		Sys_SetCursorPos (cursorx, cursory);

		if(globals.LockXYZ2Cam)
		{
			globals.XY.origin[0] = globals.XZ.origin[0] = camera.origin[0];
			globals.XY.origin[1] = globals.YZ.origin[1] = camera.origin[1];
			globals.YZ.origin[2] = globals.XZ.origin[2] = z.origin[2] = camera.origin[2];
		}

		Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
	}
}

void PositionRotate(int x, int y, vec3_t origin)
{
	vec3_t		work, forward, dir, vecdist;
	vec_t		distance;
	int			i;


	for(i = 0; i < 3; i++)
	{
		vecdist[i] = fabs((camera.origin[i] - origin[i]));
		vecdist[i] *= vecdist[i];
	}

	camera.viewdistance = distance = sqrt(vecdist[0]+vecdist[1]+vecdist[2]);
	VectorSubtract (camera.origin, origin, work);
	VectorToAngles(work, camera.angles);

	if(camera.angles[PITCH] > 100)
		camera.angles[PITCH] -= 360;

	camera.angles[PITCH] -= y;
	
	if(camera.angles[PITCH] > 85)
		camera.angles[PITCH] = 85;
	if(camera.angles[PITCH] < -85)
		camera.angles[PITCH] = -85;

	camera.angles[YAW] += x;
	AngleVectors(camera.angles, forward, NULL, NULL);
	forward[2] = -forward[2];
	VectorMA (origin, distance, forward, camera.origin);
	VectorSubtract (origin, camera.origin, dir);
	VectorNormalize (dir);
	camera.angles[1] = atan2 (dir[1], dir[0])*180/3.14159;
	camera.angles[0] = asin (dir[2])*180/3.14159;
	Cam_BuildMatrix();
	Sys_SetCursorPos (cursorx, cursory);

	if(globals.LockXYZ2Cam)
	{
		globals.XY.origin[0] = globals.XZ.origin[0] = camera.origin[0];
		globals.XY.origin[1] = globals.YZ.origin[1] = camera.origin[1];
		globals.YZ.origin[2] = globals.XZ.origin[2] = z.origin[2] = camera.origin[2];
	}

	Sys_UpdateWindows (W_CAMERA|W_XY|W_XZ|W_YZ|W_Z);
}

void Cam_PositionRotate(void)
{
	int			x, y, i, i2;
	vec3_t		origin;
	brush_t		*b;
	vec3_t		mins, maxs, vecdist;
	facesel_t	*face;
	face_t		*f;

	Sys_GetCursorPos (&x, &y);

	if (x == cursorx && y == cursory)
		return;
	
	x -= cursorx;
	y -= cursory;

	if(selected_brushes.next != &selected_brushes)
	{
		mins[0] = mins[1] = mins[2] = 99999;
		maxs[0] = maxs[1] = maxs[2] = -99999;
		for(b = selected_brushes.next; b != &selected_brushes; b = b->next)
		{
			for(i = 0; i < 3; i++)
			{
				if(b->maxs[i] > maxs[i])
					maxs[i] = b->maxs[i];
				if(b->mins[i] < mins[i])
					mins[i] = b->mins[i];
			}
		}
		origin[0] = (mins[0] + maxs[0])/2;
		origin[1] = (mins[1] + maxs[1])/2;
		origin[2] = (mins[2] + maxs[2])/2;
	}
	else if(selected_faces)
	{
		mins[0] = mins[1] = mins[2] = 99999;
		maxs[0] = maxs[1] = maxs[2] = -99999;
		for(face = selected_faces; face; face = face->next)
		{
			f = face->face;
			for ( i2 = 0 ; i2 < f->face_winding->numpoints; i2++)
			{
				for(i = 0; i < 3; i++)
				{
					if(f->face_winding->points[i2][i] > maxs[i])
						maxs[i] = f->face_winding->points[i2][i];
					if(f->face_winding->points[i2][i] < mins[i])
						mins[i] = f->face_winding->points[i2][i];
				}
			}
		}
		origin[0] = (mins[0] + maxs[0])/2;
		origin[1] = (mins[1] + maxs[1])/2;
		origin[2] = (mins[2] + maxs[2])/2;
	}
	else
	{
		AngleVectors(camera.angles, camera.forward, NULL, NULL);
		camera.forward[2] = -camera.forward[2];
		VectorMA (camera.origin, camera.viewdistance, camera.forward, origin);
	}

	for(i = 0; i < 3; i++)
	{
		vecdist[i] = fabs((camera.origin[i] - origin[i]));
		vecdist[i] *= vecdist[i];
	}

	PositionRotate(x, y, origin);
	return;
}

void Cam_FreeLook(void)
{
	int			x, y;

	Sys_GetCursorPos (&x, &y);

	if (x == cursorx && y == cursory)
		return;
	
	x -= cursorx;
	y -= cursory;

	if(camera.angles[PITCH] > 100)
		camera.angles[PITCH] -= 360;

	camera.angles[PITCH] -= y;
	
	if(camera.angles[PITCH] > 85)
		camera.angles[PITCH] = 85;

	if(camera.angles[PITCH] < -85)
		camera.angles[PITCH] = -85;

	camera.angles[YAW] -= x;
	Sys_SetCursorPos (cursorx, cursory);
	Sys_UpdateWindows (W_CAMERA|W_XY|W_XZ|W_YZ);
}

/*
===============
Cam_MouseControl
===============
*/
void Cam_MouseControl (float dtime)
{
	float	yf;
	float	ycenter, sign;
	int		i, y_amt = 0, count;
	RECT	rect;
	vec3_t	forward;

	if (cambuttonstate != MK_RBUTTON)
		return;

	GetClientRect(globals.Camera.hwnd, &rect);
	ycenter = (rect.bottom - rect.top) * 0.5;
	count = abs(buttony - ycenter);

	if(buttony > ycenter)
		sign = 1;
	else
		sign = -1;

	for(i = 0; i < count; i++)
	{
		y_amt += sign * (count - i) * 0.025;
	}

	yf = (float)(y_amt) / ycenter;
	VectorCopy(camera.forward, forward);

	if(globals.OldCameraMovement)
		forward[2] = 0;

	VectorMA (camera.origin, yf*dtime*MOVE_SPEED*1.5f, forward, camera.origin);
	Sys_GetCursorPos (&buttonx, &buttony);
	camera.angles[YAW] -= (buttonx - pressx) * 0.5f;

	if(globals.followfloor)
		FollowFloor();
	
	if(globals.LockXYZ2Cam)
	{
			globals.XY.origin[0] = globals.XZ.origin[0] = camera.origin[0];
			globals.XY.origin[1] = globals.YZ.origin[1] = camera.origin[1];
			globals.YZ.origin[2] = globals.XZ.origin[2] = z.origin[2] = camera.origin[2];
		Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
	}
	else
		Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ | W_Z);

	Sys_SetCursorPos (pressx, buttony);
	buttonx = pressx;
}




/*
==============
Cam_MouseDown
==============
*/
void Cam_MouseDown (int x, int y, int buttons)
{
	vec3_t		dir;
	float		f, r, u;
	int			i;

	//
	// calc ray direction
	//
	u = (float)(y - camera.height/2) / (camera.width/2);
	r = (float)(x - camera.width/2) / (camera.width/2);
	f = 1;

	for (i=0 ; i<3 ; i++)
		dir[i] = camera.vpn[i] * f + camera.vright[i] * r + camera.vup[i] * u;
	VectorNormalize (dir);

	Sys_GetCursorPos (&cursorx, &cursory);

	cambuttonstate = buttons;
	buttonx = x;
	buttony = y;

	// LBUTTON = manipulate selection
	// shift-LBUTTON = select
	// middle button = grab texture
	// ctrl-middle button = set entire brush to texture
	// ctrl-shift-middle button = set single face to texture
	if ( (  buttons ==  MK_LBUTTON)
		|| (buttons == (MK_LBUTTON | MK_SHIFT))
		|| (buttons == (MK_LBUTTON | MK_CONTROL))
		|| (buttons == (MK_LBUTTON | MK_CONTROL | MK_SHIFT))
		|| (buttons ==  MK_MBUTTON)
		|| (buttons == (MK_MBUTTON|MK_CONTROL))
		|| (buttons == (MK_MBUTTON|MK_SHIFT|MK_CONTROL))
		|| (buttons == (MK_LBUTTON|MK_RBUTTON))
		|| (buttons == (MK_LBUTTON|MK_RBUTTON|MK_CONTROL))
		|| (buttons == (MK_LBUTTON|MK_RBUTTON|MK_SHIFT|MK_CONTROL)) )
	{
		Drag_Begin (x, y, buttons, 
			camera.vright, camera.vup,
			camera.origin, dir);
		return;
	}

	if (buttons == MK_RBUTTON)
	{
		pressx = cursorx;
		Cam_MouseControl ((float)0.0);
		return;
	}

	if(buttons == (MK_MBUTTON | MK_SHIFT) || buttons == (MK_LBUTTON | MK_RBUTTON | MK_SHIFT))
	{
		trace_t	t;
		
		t = Test_Ray (camera.origin, dir, false);
		
		if (t.brush)
		{
			face_t *f = t.face;
			int		j;

			for(i = 0; i < 3; i++)
			{
				for(j = 0; j < 3; j++)
				{
					globals.Clipper[i][j] = f->planepts[i][j];
				}
			}
			globals.ClipperNumPoints = 3;
			Sys_UpdateWindows(W_XY|W_XY|W_YZ|W_Z|W_CAMERA);
		}

	}
}

/*
==============
Cam_MouseUp
==============
*/
void Cam_MouseUp (int x, int y, int buttons)
{
	cambuttonstate = 0;
	Drag_MouseUp ();
}


/*
==============
Cam_MouseMoved
==============
*/
void Cam_MouseMoved (int x, int y, int buttons)
{
	trace_t		t;
	vec3_t		dir;
	float		f, r, u;
	int			i;

	cambuttonstate = buttons;
	if (!buttons)
	{
		//
		// calc ray direction
		//
		u = (float)(y - camera.height/2) / (camera.width/2);
		r = (float)(x - camera.width/2) / (camera.width/2);
		f = 1;

		for (i=0 ; i<3 ; i++)
			dir[i] = camera.vpn[i] * f + camera.vright[i] * r + camera.vup[i] * u;
		VectorNormalize (dir);
		t = Test_Ray (camera.origin, dir, false);
		if (t.brush)
		{
			if (t.brush->brush_faces->texdef.name[0] == '(')
				Sys_Status(0, "*Entity* %s", t.brush->owner->eclass->name);
			else
				Sys_Status(0, "%s", t.face->texdef.name);
		}
		else
			Sys_Status(0, "*No face*");

		return;
	}
	buttonx = x;
	buttony = y;

	if (buttons == (MK_RBUTTON|MK_CONTROL) )
	{
		Cam_PositionDrag ();
//		Sys_UpdateWindows (W_XY|W_CAMERA|W_Z);
		return;
	}

	if (buttons == (MK_RBUTTON|MK_SHIFT))
	{
		Cam_PositionRotate();
		return;
	}

	if (buttons == (MK_RBUTTON|MK_SHIFT|MK_CONTROL))
	{
		Cam_FreeLook();
		return;
	}

	Sys_GetCursorPos (&cursorx, &cursory);

	if (buttons & (MK_LBUTTON | MK_MBUTTON) || buttons == (MK_RBUTTON|MK_SHIFT) )
	{
		Drag_MouseMoved (x, y, buttons);
		Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_Z|W_CAMERA|W_Z);
	}
}

void Cam_DoubleClick(int x, int y, int buttons)
{
	trace_t		t;
	vec3_t		dir;
	float		f, r, u;
	int			i;

	if(buttons != MK_LBUTTON) // Only handle doubleclicks for left mouse alone.
	{
		Cam_MouseDown(x, y, buttons);
		Cam_MouseUp(x, y, buttons);
		return;
	}

	u = (float)(y - camera.height/2) / (camera.width/2);
	r = (float)(x - camera.width/2) / (camera.width/2);
	f = 1;
	
	for (i=0 ; i<3 ; i++)
		dir[i] = camera.vpn[i] * f + camera.vright[i] * r + camera.vup[i] * u;
	
	VectorNormalize (dir);
	t = Test_Ray (camera.origin, dir, false);
	
	if (t.brush)
	{
		if(buttons & MK_CONTROL)
			camera.viewdistance = t.dist;
		else
			VectorMA(camera.origin, t.dist - camera.viewdistance, dir, camera.origin);
	}
	else
	{
		Sys_Status(0, "*No face*");
	}
}

vec3_t	cull1, cull2;
int		cullv1[3], cullv2[3];

void InitCull (void)
{
	int		i;

	VectorSubtract (camera.vpn, camera.vright, cull1);
	VectorAdd (camera.vpn, camera.vright, cull2);

	for (i=0 ; i<3 ; i++)
	{
		if (cull1[i] > 0)
			cullv1[i] = 3+i;
		else
			cullv1[i] = i;
		if (cull2[i] > 0)
			cullv2[i] = 3+i;
		else
			cullv2[i] = i;
	}
}

qboolean CullBrush (brush_t *b)
{
	static vec3_t	point;
	static float	d;

	point[0] = b->mins[cullv1[0]] - camera.origin[0];
	point[1] = b->mins[cullv1[1]] - camera.origin[1];
	point[2] = b->mins[cullv1[2]] - camera.origin[2];

	d = DotProduct (point, cull1);
	if (d < -1)
		return true;

	point[0] = b->mins[cullv2[0]] - camera.origin[0];
	point[1] = b->mins[cullv2[1]] - camera.origin[1];
	point[2] = b->mins[cullv2[2]] - camera.origin[2];

	d = DotProduct (point, cull2);
	if (d < -1)
		return true;

	return DepthClip(b);
}
int Sys_Milliseconds (void);
void Draw_Skybox(void)
{
	glCullFace		(GL_FRONT);
	glEnable		(GL_CULL_FACE);
	glShadeModel	(GL_FLAT);
	glPolygonMode	(GL_FRONT_AND_BACK, GL_FILL);
	glEnable		(GL_TEXTURE_2D);
	glTexEnvf		(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glDisable		(GL_BLEND);
	glDisable		(GL_DEPTH_TEST);
	glColor3f		(1.0f, 1.0f, 1.0f);

	//rotation
	if(globals.AnimateSky && globals.skyrotate)
	{
		glPushMatrix ();
//		glTranslatef (-camera.origin[0], -camera.origin[1], -camera.origin[2]);
		glRotatef (Sys_Milliseconds()* 0.001f * globals.skyrotate, globals.skyaxis[0], globals.skyaxis[1], globals.skyaxis[2]);
	}

	glCallList		(skybox_display_list);
	if(globals.skyrotate)
		glPopMatrix();

	glEnable		(GL_DEPTH_TEST);
}

void DrawTrans(void)
{
    qtexture_t		*prev = 0;
	facesel_t		*currentface, *lastface;
	winding_t		*w;
	int				i;

	if(trans_faces_rscript)
	{
		prev = NULL;
		for(currentface = trans_faces_rscript; currentface; lastface = currentface, currentface = currentface->next, free(lastface))
		{
			RS_DrawPoly(currentface->face, false);
		}
		trans_faces_rscript = NULL;
	}

	if(trans_faces)
	{
		glEnable	(GL_BLEND);
		glBlendFunc	(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDepthMask(GL_FALSE);
		prev = NULL;
		for(currentface = trans_faces; currentface; lastface = currentface, currentface = currentface->next, free(lastface))
		{
			if ( currentface->face->d_texture != prev && camera.draw_mode == cd_texture)
			{
				prev = currentface->face->d_texture;
				glBindTexture( GL_TEXTURE_2D, currentface->face->d_texture->image->texture_number );
			}
			glColor4f(
				currentface->face->d_color[0],
				currentface->face->d_color[1],
				currentface->face->d_color[2],
				currentface->trans
				);
			w = currentface->face->face_winding;
			glBegin(GL_POLYGON);
		    for (i=0 ; i<w->numpoints ; i++)
			{
				if (camera.draw_mode == cd_texture)
					glTexCoord2dv( &w->points[i][3] );
				glVertex3dv(w->points[i]);
			}
			glEnd();
		}
		glDisable (GL_BLEND);
		trans_faces = NULL;
	}
}

extern BOOL SurfaceDlgActive;

/*
==============
Cam_Draw
==============
*/
void Cam_Draw (void)
{
    brush_t		*brush;
	face_t		*face, *lastdrawnface;
	facesel_t	*currentface;
	float		screenaspect;
	float		yfov;
	double		start, end;
	int			i;

	if (!active_brushes.next)
		return;	// not valid yet

	trans_faces  = trans_faces_rscript = NULL;

	if (camera.timing)
		start = Sys_DoubleTime ();

	//
	// clear
	//
	QE_CheckOpenGLForErrors("Cam_Draw()1", GL_NO_ERROR);

	glViewport(0, 0, camera.width, camera.height);
	glScissor(0, 0, camera.width, camera.height);
	glClearColor (
		globals.d_savedinfo.colors[COLOR_CAMERABACK][0],
		globals.d_savedinfo.colors[COLOR_CAMERABACK][1],
		globals.d_savedinfo.colors[COLOR_CAMERABACK][2],
		0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//
	// set up viewpoint
	//
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
    screenaspect = (float)camera.width/camera.height;
	yfov = 2*atan((float)camera.height/camera.width)*180/Q_PI;
    gluPerspective (yfov,  screenaspect,  2,  8192);
    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up
    glRotatef (camera.angles[0],  0, 1, 0);
    glRotatef (-camera.angles[1],  0, 0, 1);
	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();
	
	switch (camera.draw_mode)
	{
		case cd_wire:
			glPolygonMode	(GL_FRONT_AND_BACK, GL_LINE);
			glDisable		(GL_TEXTURE_2D);
			glDisable		(GL_TEXTURE_1D);
			glDisable		(GL_BLEND);
			glDisable		(GL_DEPTH_TEST);
			glColor3f		(1.0, 1.0, 1.0);
//			glEnable (GL_LINE_SMOOTH);
			break;
		
		case cd_solid:
			glCullFace		(GL_FRONT);
			glEnable		(GL_CULL_FACE);
			glShadeModel	(GL_FLAT);			
			glPolygonMode	(GL_FRONT_AND_BACK, GL_FILL);
			glDisable		(GL_TEXTURE_2D);			
			glDisable		(GL_BLEND);
			glEnable		(GL_DEPTH_TEST);
			glDepthFunc		(GL_LEQUAL);
			break;
		
		case cd_texture:
			glCullFace		(GL_FRONT);
			glEnable		(GL_CULL_FACE);			
			glShadeModel	(GL_FLAT);
			glPolygonMode	(GL_FRONT_AND_BACK, GL_FILL);
			glEnable		(GL_TEXTURE_2D);
			glTexEnvf		(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glDisable		(GL_BLEND);
			glEnable		(GL_DEPTH_TEST);
			glDepthFunc		(GL_LEQUAL);
#if 0
			{
				GLfloat fogColor[4] = {0.0, 1.0, 0.0, 0.25};
				
				glEnable		(GL_FOG);
				glFogi			(GL_FOG_MODE, GL_LINEAR);
				glHint			(GL_FOG_HINT, GL_NICEST);  /*  per pixel   */
				glFogf			(GL_FOG_START, -8192);
				glFogf			(GL_FOG_END, 65536);
				glFogfv			(GL_FOG_COLOR, fogColor);
			}
#endif
			break;
		
		case cd_blend:
			glCullFace		(GL_FRONT);
			glEnable		(GL_CULL_FACE);
			glShadeModel	(GL_FLAT);
			glPolygonMode	(GL_FRONT_AND_BACK, GL_FILL);
			glEnable		(GL_TEXTURE_2D);
			glTexEnvf		(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glDisable		(GL_DEPTH_TEST);
			glEnable		(GL_BLEND);
			glBlendFunc		(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			break;
	}

	if(globals.ShowSkybox && skybox_textures[0])
	{
		Draw_Skybox();
	}
    glTranslatef (-camera.origin[0],  -camera.origin[1],  -camera.origin[2]);

	Cam_BuildMatrix ();

	InitCull ();
	glGetFloatv (GL_MODELVIEW_MATRIX, globals.r_world_matrix);
	//
	// draw stuff
	//
	for (brush = active_brushes.next ; brush != &active_brushes ; brush=brush->next)
	{
		if (CullBrush (brush))
			continue;
		if (FilterBrush (brush))
			continue;

		Brush_Draw(brush, true);
	}

	DrawTrans();

	//
	// now draw selected brushes
	//

	glTranslatef( globals.d_select_translate[0], globals.d_select_translate[1], globals.d_select_translate[2]);

	// draw normally
	Clipper_RefreshBrushes();

	for (brush = selected_brushes.next ; brush != &selected_brushes ; brush=brush->next)
	{
		Brush_Draw(brush, true);
	}

	DrawTrans();

	glDepthMask(GL_TRUE);
	// blend on top

//	glColor4f((float)1.0, (float)0.0, (float)0.0, (float)0.3);
	glColor4f(globals.d_savedinfo.colors[COLOR_SEL_COLOR][0],
			globals.d_savedinfo.colors[COLOR_SEL_COLOR][1],
			globals.d_savedinfo.colors[COLOR_SEL_COLOR][2],
			globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
	glEnable (GL_BLEND);
	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_TEXTURE_2D);

	if(!globals.ClipperNumPoints)
	{
		for (brush = selected_brushes.next ; brush != &selected_brushes ; brush=brush->next)
		{
			for (face=brush->brush_faces ; face ; face=face->next)
			{
				if(SurfaceDlgActive && brush == selected_brushes.prev && face == brush->brush_faces)
					continue;
				
				Face_Draw( face );
			}
		}
	}
	else
	{
		if(globals.ClipperMode & 1)
			glColor4f(1.0f, 1.0f, 1.0f,	globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
		
		for (brush = clipper_brushes_front.next ; brush != &clipper_brushes_front ; brush=brush->next)
		{
			for (face=brush->brush_faces ; face ; face=face->next)
			{
				Face_Draw( face );
			}
		}

		if(globals.ClipperMode & 2)
			glColor4f(1.0f, 1.0f, 1.0f,	globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
		else
			glColor4f(globals.d_savedinfo.colors[COLOR_SEL_COLOR][0], globals.d_savedinfo.colors[COLOR_SEL_COLOR][1], globals.d_savedinfo.colors[COLOR_SEL_COLOR][2], globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);

		for (brush = clipper_brushes_back.next ; brush != &clipper_brushes_back ; brush=brush->next)
		{
			for (face=brush->brush_faces ; face ; face=face->next)
			{
				Face_Draw( face );
			}
		}
	}

	for(currentface = selected_faces; currentface; currentface = currentface->next)
	{
		if(SurfaceDlgActive && !currentface->next)
			lastdrawnface = currentface->face;
		else
			Face_Draw(currentface->face);
	}

	if(SurfaceDlgActive)
	{
		glColor4f(0, 0, 1, globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);

		if(selected_faces)
		{
			Face_Draw(lastdrawnface);
		}
		else
		{
			if(selected_brushes.prev != &selected_brushes)
				Face_Draw(selected_brushes.prev->brush_faces);
		}
	}
	// non-zbuffered outline

	glDisable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
//	glColor3f (1, 1, 1);
	glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);

	for (brush = selected_brushes.next ; brush != &selected_brushes ; brush=brush->next)
	{
		for (face=brush->brush_faces ; face ; face=face->next)
		{
			Face_Draw( face );
		}
	}
	
	if(!globals.ClipperNumPoints)
	{
		for(currentface = selected_faces; currentface; currentface = currentface->next)
		{
			Face_Draw(currentface->face);
		}
	}
	else
	{
		if(globals.ClipperMode & 1)
			glColor3f(1.0f, 1.0f, 0.0f);
		
		for (brush = clipper_brushes_front.next ; brush != &clipper_brushes_front ; brush=brush->next)
		{
			for (face=brush->brush_faces ; face ; face=face->next)
			{
				Face_Draw( face );
			}
		}

		if(globals.ClipperMode & 2)
			glColor3f(1.0f, 1.0f, 0.0f);
		else
			glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
		
		for (brush = clipper_brushes_back.next ; brush != &clipper_brushes_back ; brush=brush->next)
		{
			for (face=brush->brush_faces ; face ; face=face->next)
			{
				Face_Draw( face );
			}
		}
	}

	if(SurfaceDlgActive)
	{
		glColor4f(0, 1, 1, globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);

		if(selected_faces)
		{
			Face_Draw(lastdrawnface);
		}
		else
		{
			if(selected_brushes.prev != &selected_brushes)
				Face_Draw(selected_brushes.prev->brush_faces);
		}
	}

	// edge / vertex flags

	if (globals.d_select_mode == sel_vertex)
	{
		glPointSize (4);
		glColor3f (0,1,0);
		glBegin (GL_POINTS);
		for (i=0 ; i<globals.d_numpoints ; i++)
			glVertex3dv (globals.d_points[i]);
		glEnd ();
		glPointSize (1);
	}
	else if (globals.d_select_mode == sel_edge)
	{
		vec_t	*v1, *v2;

		glPointSize (4);
		glColor3f (0,0,1);
		glBegin (GL_POINTS);
		for (i=0 ; i<globals.d_numedges ; i++)
		{
			v1 = globals.d_points[globals.d_edges[i].p1];
			v2 = globals.d_points[globals.d_edges[i].p2];
			glVertex3f ( (v1[0]+v2[0])*0.5,(v1[1]+v2[1])*0.5,(v1[2]+v2[2])*0.5);
		}
		glEnd ();
		glPointSize (1);
	}

	//
	// draw pointfile
	//
	glEnable(GL_DEPTH_TEST);

	if(globals.ShowEntityConnections)
		DrawPathLines ();

	if (globals.d_pointfile_display_list)
	{
		Pointfile_Draw();
//		glCallList (globals.d_pointfile_display_list);
	}


	// bind back to the default texture so that we don't have problems
	// elsewhere using/modifying texture maps between contexts
	glBindTexture( GL_TEXTURE_2D, 0 );

    glFinish();
	QE_CheckOpenGLForErrors("Cam_Draw()2", GL_NO_ERROR);
//	Sys_EndWait();
	if (camera.timing)
	{
		end = Sys_DoubleTime ();
		Sys_Printf ("Camera: %i ms\n", (int)(1000*(end-start)));
	}
}

