#include "qe3.h"

face_t *Face_Alloc( void );
void Face_Free(face_t *f);

void CSG_SplitBrushByFace (brush_t *in, face_t *f, brush_t **front, brush_t **back);

void Clipper_FreeBrushes(void)
{
	while(clipper_brushes_front.next != &clipper_brushes_front)
	{
		Brush_Free(clipper_brushes_front.next);
	}

	while(clipper_brushes_back.next != &clipper_brushes_back)
	{
		Brush_Free(clipper_brushes_back.next);
	}
}

extern int IgnoreBrushNotLinked;
extern int SnapPlanepts;

void Clipper_RefreshBrushes(void)
{
	brush_t	*b, *front, *back;
	face_t	*f;
	vec3_t	t1, t2, t3;
	int		i;

	Clipper_FreeBrushes();

	if(globals.ClipperNumPoints < 2)
		return;

	f = Face_Alloc();
	memcpy(f->planepts, globals.Clipper, sizeof(f->planepts));
	for (i = 0; i < 3; i++)
	{
		t1[i] = f->planepts[0][i] - f->planepts[1][i];
		t2[i] = f->planepts[2][i] - f->planepts[1][i];
		t3[i] = f->planepts[1][i];
	}
	
	CrossProduct(t1,t2, f->plane.normal);
	
	if (VectorCompare (f->plane.normal, vec3_origin))
	{
		Face_Free(f);
		return;
	}
	
	VectorNormalize (f->plane.normal);
	f->plane.dist = DotProduct (t3, f->plane.normal);

	IgnoreBrushNotLinked = 1;
	SnapPlanepts = 0;

	for(b = selected_brushes.next; b != &selected_brushes; b = b->next)
	{
		CSG_SplitBrushByFace(b, f, &front, &back);

		if(front)
			Brush_AddToList(front, &clipper_brushes_front);

		if(back)
			Brush_AddToList(back, &clipper_brushes_back);
	}
	IgnoreBrushNotLinked = 0;
	SnapPlanepts = 1;
	Face_Free(f);
}

void Clipper_Off(void)
{
	globals.ClipperNumPoints = 0;
	Clipper_FreeBrushes();
}

extern int Orientations[3][3];

int Clipper_AddPoint(vec3_t pt, int orientation)
{
	globals.ClipperNumPoints++;

	if(globals.ClipperNumPoints == 4)
		globals.ClipperNumPoints = 1;

	VectorCopy(pt, globals.Clipper[globals.ClipperNumPoints-1]);

	if(globals.ClipperNumPoints == 2)
	{
		VectorCopy(globals.Clipper[1], globals.Clipper[2]);
		globals.Clipper[2][globals.XY_Orientation[2]] += globals.d_gridsize;
	}

	Clipper_RefreshBrushes();

	return globals.ClipperNumPoints-1;
}

void SetClipperMode(int mode)
{
	globals.ClipperMode = mode;

	CheckMenu(ID_SELECTION_CLIPPER_SPLIT, mode == 3);
	CheckMenu(ID_SELECTION_CLIPPER_FRONT, mode == 1);
	CheckMenu(ID_SELECTION_CLIPPER_BACK, mode == 2);
}

int FindClipperPoint(vec3_t org, vec3_t dir)
{
	int		i, besti;
	float	d, bestd;
	vec3_t	temp;

	// find the point closest to the ray
	besti = -1;
	bestd = 8;

	for (i=0 ; i<globals.ClipperNumPoints ; i++)
	{
		VectorSubtract (globals.Clipper[i], org, temp);
		d = DotProduct (temp, dir);
		VectorMA (org, d, dir, temp);
		VectorSubtract (globals.Clipper[i], temp, temp);
		d = VectorLength (temp);
		if (d < bestd)
		{
			bestd = d;
			besti = i;
		}
	}

	return besti;
}

void DoClip(void)
{
	brush_t	*b, *bnext;

	if(globals.ClipperNumPoints < 2)
		return;

	if(selected_brushes.next == &selected_brushes || (clipper_brushes_front.next == &clipper_brushes_front && clipper_brushes_back.next == &clipper_brushes_back))
		return;

	Select_Delete();

	if(globals.ClipperMode & 1)
	{
		for(b = clipper_brushes_front.next; b != &clipper_brushes_front; b = bnext)
		{
			bnext = b->next;
			Brush_RemoveFromList(b);
			Brush_AddToList(b, &selected_brushes);
		}
	}

	if(globals.ClipperMode & 2)
	{
		for(b = clipper_brushes_back.next; b != &clipper_brushes_back; b = bnext)
		{
			bnext = b->next;
			Brush_RemoveFromList(b);
			Brush_AddToList(b, &selected_brushes);
		}
	}
}
