#include "qe3.h"
#include "io.h"
#include "directory.h"

dir_t	*directories = NULL;

void DirectoryInit(void)
{
	directories = malloc(sizeof(dir_t));

	directories->name = strdup("");
	directories->files = NULL;
	directories->next = NULL;
	directories->subdirs = NULL;
}

dir_t *AddSubDir(dir_t *dir, char *name)
{
	dir_t	*d;

	for(d = dir->subdirs; d; d = d->next)
	{
		if(!stricmp(d->name, name))
			return d;
	}

	d = malloc(sizeof(dir_t));
	d->name = strdup(name);
	d->files = NULL;
	d->subdirs = NULL;
	d->next = dir->subdirs;
	dir->subdirs = d;
	d->parent = dir;

	return d;
}

dir_t *AddDir(dir_t *dir, char *name)
{
	dir_t	*d = dir;
	char	buf[512];
	char	*s1, *s2;

	setdirstr2(buf, name, "/", NULL);
	s1 = buf;

	while(s2 = strchr(s1, '/'))
	{
		*s2++ = 0;
		d = AddSubDir(d, s1);
		s1 = s2;
	}

	return d;
}

dir_t *FindSubDir(dir_t *dir, char *name)
{
	dir_t	*d;

	for(d = dir->subdirs; d; d = d->next)
	{
		if(!stricmp(d->name, name))
			return d;
	}

	return NULL;
}

dir_t *FindDir(dir_t *dir, char *name)
{
	dir_t	*d = dir;
	char	buf[512];
	char	*s1, *s2;

	setdirstr2(buf, name, "/", NULL);
	s1 = buf;

	while(s2 = strchr(s1, '/'))
	{
		*s2++ = 0;
		d = FindSubDir(d, s1);

		if(!d)
			return NULL;

		s1 = s2;
	}

	return d;
}

file_t *NewFile(char *name, FILE *file, int offset, int size)
{
	file_t	*f;
	
	f = malloc(sizeof(file_t));
	f->name = strdup(name);
	f->next = NULL;

	return f;
}

file_t *AddFileToDir(dir_t *dir, char *name, FILE *file, int offset, int size)
// This could be optimised a bit more. Lets just get it working first
{
	file_t	*f, *curr, *prev;

	for(f = dir->files; f; f = f->next)
	{
		if(!stricmp(f->name, name))
			break;
	}

	if(!f)
	{
		prev = NULL;
		curr = dir->files;
		f = NewFile(name, file, offset, size);

		while(curr != NULL && stricmp(name, curr->name) > 0)
		{
			prev = curr;
			curr = curr->next;
		}

		if(prev == NULL)
		{
			f->next = dir->files;
			dir->files = f;
		}
		else
		{
			prev->next = f;
			f->next = curr;
		}
	}

	f->file = file;
	f->offset = offset;
	f->size = size;

	return f;
}

file_t *AddFile(dir_t *head, char *name, FILE *file, int offset, int size)
{
	dir_t	*d;
	char	buf[512];
	char	*s = NULL;

	setdirstr2(buf, name, NULL);
	s = strrchr(buf, '/');

	if(s)
	{
		*s++ = 0;
		d = AddDir(directories, buf);
	}
	else
	{
		s = buf;
		d = directories;
	}

	return AddFileToDir(d, s, file, offset, size);
}

file_t *FindFile(char *name)
{
	dir_t	*d;
	file_t	*f;
	char	buf[512];
	char	*s = NULL;

	setdirstr2(buf, name, NULL);
	s = strrchr(buf, '/');

	if(s)
	{
		*s++ = 0;
		d = FindDir(directories, buf);
	}
	else
	{
		s = buf;
		d = directories;
	}

	if(d)
	{
		for(f = d->files; f; f = f->next)
		{
			if(!stricmp(f->name, s))
				return f;
		}
	}
	
	return NULL;
}

qfile_t *qfopen(char *name, char *flags, char *ext)
{
	qfile_t	*qfile;
	FILE	*f;
	char	buf[512];
	file_t	*file = NULL;

	setdirstr(buf, globals.modpath, "/", name, ext?ext:NULL, NULL);
	if(!globals.modpath[0] || !(f = fopen(buf, flags)))
	{
		setdirstr(buf, globals.basepath, "/", name, ext?ext:NULL, NULL);
		if(!(f = fopen(buf, flags)))
		{
			if(!(file = FindFile(name)))
				return NULL;
		}
	}

	qfile = malloc(sizeof(qfile_t));

	if(!file)
	{
		qfile->diskfile = f;
		qfile->pakfile = NULL;
	}
	else
	{
		qfile->diskfile = file->file;
		qfile->pakfile = file;
		fseek(file->file, file->offset, SEEK_SET);
	}

	qfile->flags = strdup(flags);

	return qfile;
}


int qfseek(qfile_t *file, int offset, int whence)
{
	if(file->pakfile)
	{
		switch(whence)
		{
			case SEEK_SET:
				return fseek(file->pakfile->file, file->pakfile->offset + offset, SEEK_SET);
				break;
			case SEEK_END:
				return fseek(file->pakfile->file, file->pakfile->offset + file->pakfile->size - offset, SEEK_SET);
				break;
			default:
				return fseek(file->pakfile->file, offset, SEEK_CUR);
				break;
		}
	}
	return fseek(file->diskfile, offset, whence);
}

int qftell(qfile_t *file)
{
	if(file->pakfile)
	{
		return ftell(file->pakfile->file) - file->pakfile->offset;
	}

	return ftell(file->diskfile);
}

void qfclose(qfile_t *file)
{
	if(!file->pakfile)
	{
		fclose(file->diskfile);
	}

	free(file->flags);
	free(file);
}

size_t qfread(void *ptr, size_t size, size_t nmemb, qfile_t *file)
{
	if(!file->pakfile)
		return fread(ptr, size, nmemb, file->diskfile);

	return fread(ptr, size, nmemb, file->pakfile->file);
}

/*
 * qLoadFile
 * ==========================================
 * Expects filenames relative to the basepath
 */
int qLoadFile (char *filename, void **bufferptr)
{
	qfile_t	*f;
	int		length;
	void	*buffer;

	f = qfopen (filename, "rb", NULL);
	if (!f)
	{
		*bufferptr = NULL;
		return -1;
	}

	if(!f->pakfile)
	{
		length = Q_filelength (f->diskfile);
	}
	else
	{
		length = f->pakfile->size;
		qfseek(f, 0, SEEK_SET);
	}

	buffer = qmalloc (length+1);
	((char *)buffer)[length] = 0;

	if ( (int)fread (buffer, 1, length, f->pakfile?f->pakfile->file:f->diskfile) != length)
		Error ("File read failure on %s", filename);

	qfclose (f);

	*bufferptr = buffer;
	return length;
}

int qLoadTextureFile(char *filename, void **bufferptr)
{
	char	buf[512];

	setdirstr(buf, "textures/", filename, NULL);
	return qLoadFile(buf, bufferptr);
}

void GetPakTextureDirs(DIRLIST **dirlist, dir_t *parent, char *parentname)
{
	dir_t	*d;
	char	buf[256];

	for(d = parent->subdirs; d; d = d->next)
	{
		if(parentname)
			sprintf(buf, "%s/%s", parentname, d->name);
		else
			strcpy(buf, d->name);

		AddToDirListAlphabetized(dirlist, buf, FROMPAK);

		if(d->subdirs)
			GetPakTextureDirs(dirlist, d, d->name);
	}
}

BOOL GetPackTextureDirs(DIRLIST **dirlist)
{
	dir_t	*d;

	d = FindDir(directories, "textures/");

	if(!d)
		return false;

	GetPakTextureDirs(dirlist, d, NULL);
	return true;
}

int GetPackFileList(FILELIST **filelist, char *pattern)
{
	dir_t	*d;
	file_t	*f;
	char	buf[512], *s, filename[512];
	int		retval = 0;

	setdirstr2(buf, "textures/", pattern, NULL);
	s = strrchr(buf, '/');

	if(s)
		*s++ = 0;

	d = FindDir(directories, buf);

	if(!d)
		return 0;

	for(f = d->files; f; f = f->next)
	{
		retval++;
		setdirstr2(filename, "/", f->name, NULL);
		AddToFileListAlphabetized(filelist, filename, f->offset, f->size);
	}
	return true;
}

void GetPakTextureFiles(FILELIST **list, dir_t *parent, char *parentname)
{
	dir_t	*d;
	file_t	*f;
	char	buf[256];

	for(d = parent->subdirs; d; d = d->next)
	{
		if(d->subdirs)
			GetPakTextureFiles(list, d, d->name);
		
		for(f = d->files; f; f = f->next)
		{
			if(parentname)
				sprintf(buf, "%s/%s/%s", parentname, d->name, f->name);
			else
				sprintf(buf, "%s/%s", d->name, f->name);
			AddToFileListAlphabetized(list, buf, f->offset, f->size);
		}
	}
}

void GetFullPackFileList(FILELIST **filelist)
{
	dir_t	*d;

	d = FindDir(directories, "textures/");

	if(!d)
		return;

	GetPakTextureFiles(filelist, d, NULL);
}

void ClearDirectory(dir_t **dirhead)
{
	dir_t	*d, *dnext;
	file_t	*f, *fnext;

	for(d = *dirhead; d; d = dnext)
	{
		dnext = d->next;
		ClearDirectory(&d->subdirs);

		for(f = d->files; f; f = fnext)
		{
			fnext = f->next;
			free(f->name);
			free(f);
		}
	}

	*dirhead = NULL;
}

void CopyPakDirectories(dir_t *source, dir_t **target)
{
	dir_t	*d, *td;

	if(!source)
		return;

	td = *target = malloc(sizeof(dir_t));

	td->name = strdup("");
	td->files = NULL;
	td->next = NULL;
	td->subdirs = NULL;

	for(d = source->subdirs; d; d = d->next)
	{
		td = AddSubDir(*target, d->name);
		CopyPakDirectories(td, &td);
	}
}

void CopyDiskDirectories(dir_t *dir, char *basedir)
{
	struct _finddata_t	fileinfo;
	int					handle;
	char				dirstring[512], name[512];
	dir_t				*subdir;

	if(globals.UseExternalTextures)
		setdirstr(dirstring, basedir, "/*.*", NULL);
	else if(globals.gametype == QUAKE2)
		setdirstr(dirstring, basedir, "/*.wal", NULL);
	else
		setdirstr(dirstring, basedir, "/*.m32", NULL);

	handle = _findfirst (dirstring, &fileinfo);
	if (handle != -1)
	{
		do
		{
			if (!(fileinfo.attrib & _A_SUBDIR))
				continue;
			if (fileinfo.name[0] == '.')
				continue;

			setdirstr(name, basedir, "/", fileinfo.name, NULL);
			subdir = AddSubDir(dir, fileinfo.name);
			CopyDiskDirectories(subdir, name);
		} while (_findnext( handle, &fileinfo ) != -1);
		_findclose (handle);
	}
}

void GetDirectories(dir_t **dir)
{
	char	buf[512];

	CopyPakDirectories(FindDir(directories, "textures/"), dir);
	setdirstr(buf, globals.basepath, "/textures/", NULL);
	CopyDiskDirectories(*dir, buf);

	if(globals.modpath[0])
	{
		setdirstr(buf, globals.modpath, "/textures/", NULL);
		CopyDiskDirectories(*dir, buf);
	}
}

