#include "qe3.h"

/*

  drag either multiple brushes, or select plane points from
  a single brush.

*/

qboolean	drag_ok;
vec3_t		drag_xvec;
vec3_t		drag_yvec;

static		int	buttonstate;
static		int	pressx, pressy;
static		vec3_t	pressdelta;
static		int	buttonx, buttony;

BOOL		TexturesLocked = false;
//int		num_move_points;
//float	*move_points[1024];

int		lastx, lasty;

qboolean	drag_first;
void Select_DragTexture(vec3_t move);


void	AxializeVector (vec3_t v)
{
	vec3_t	a;
	float	o;
	int		i;

	if (!v[0] && !v[1])
		return;
	if (!v[1] && !v[2])
		return;
	if (!v[0] && !v[2])
		return;

	for (i=0 ; i<3 ; i++)
		a[i] = fabs(v[i]);
	if (a[0] > a[1] && a[0] > a[2])
		i = 0;
	else if (a[1] > a[0] && a[1] > a[2])
		i = 1;
	else
		i = 2;

	o = v[i];
	VectorCopy (vec3_origin, v);
	if (o<0)
		v[i] = -1;
	else
		v[i] = 1;
	
}


BOOL FaceSelected(face_t *face);

/*
===========
Drag_Setup
===========
*/
void Drag_Setup (int x, int y, int buttons, vec3_t xaxis, vec3_t yaxis, vec3_t origin, vec3_t dir)
{
	trace_t		t;
	int			facedrag = 0;
	facesel_t	*f;

	if(GetFocus() == globals.Camera.hwnd)
	{
		if (selected_brushes.next == &selected_brushes)
		{
			if(!selected_faces)
			{
				Sys_Status(0, "No selection to drag\n");
				return;
			}
			if (buttons == MK_CONTROL)
				facedrag = 1;
			else
				facedrag = 2;
		}
	}
	else
	{
		if (selected_brushes.next == &selected_brushes)
		{
			Sys_Status(0, "No selection to drag\n");
			return;
		}
	}

	drag_first = true;
	globals.d_num_move_points = 0;
	VectorCopy (vec3_origin, pressdelta);
	pressx = x;
	pressy = y;

	VectorCopy (xaxis, drag_xvec);
	AxializeVector (drag_xvec);
	VectorCopy (yaxis, drag_yvec);
	AxializeVector (drag_yvec);

	if (globals.d_select_mode == sel_vertex)
	{
		SelectVertexByRay (origin, dir);	
		if (globals.d_num_move_points)
		{
			drag_ok = true;
			return;
		}
	}
	if (globals.d_select_mode == sel_edge)
	{
		SelectEdgeByRay (origin, dir);	
		if (globals.d_num_move_points)
		{
			drag_ok = true;
			return;
		}
	}

	//
	// check for direct hit first
	//
	t = Test_Ray (origin, dir, true);
	if (t.selected)
	{
		drag_ok = true;

		if (buttons == (MK_LBUTTON|MK_CONTROL) )
		{
			Sys_Printf ("Shear dragging face\n");
			Brush_SelectFaceForDragging (t.brush, t.face, true);
		}
		else
			Sys_Printf ("Dragging entire selection\n");
		
		return;
	}
	else
		switch(facedrag)
		{
			case 1:
			{
				drag_ok = true;
				Sys_Printf ("Dragging textures\n");
				return;
			}
			case 2:
			{
				drag_ok = true;
				for(f = selected_faces; f; f = f->next)
					Brush_SelectFaceForDragging (f->owner, f->face, true);
				return;
			}
//			if(FaceSelected(t.face))
//			{
//			}
		}

	if (globals.d_select_mode == sel_vertex || globals.d_select_mode == sel_edge)
		return;

	//
	// check for side hit
	//
/*	if (selected_brushes.next->next != &selected_brushes)
	{
		Sys_Printf ("Click isn't inside multiple selection\n");
		return;
	}
*/
	if (selected_brushes.next->owner->eclass->fixedsize)
	{
		Sys_Printf ("Can't stretch fixed size entities\n");
		return;
	}


	if (buttons & MK_CONTROL)
	{
		Brush_SideSelect (selected_brushes.next, origin, dir, true);
		Sys_Printf ("Side drag\n");
	}
	else
	{
		Brush_SideSelect (selected_brushes.next, origin, dir, false);
		Sys_Printf ("Side stretch\n");
	}


	drag_ok = true;
}

entity_t *peLink;

void UpdateTarget(vec3_t origin, vec3_t dir)
{
	trace_t	t;
	entity_t *pe;
	int i;
	char sz[128];

	t = Test_Ray (origin, dir, 0);

	if (!t.brush)
		return;

	pe = t.brush->owner;

	if (pe == NULL)
		return;

	// is this the first?
	if (peLink != NULL)
	{

		// Get the target id from out current target
		// if there is no id, make one

		i = IntForKey(pe, "target");
		if (i <= 0)
		{
			i = GetUniqueTargetId(1);
			sprintf(sz, "%d", i);

			SetKeyValue(pe, "target", sz);
		}

		// set the target # into our src

		sprintf(sz, "%d", i);
		SetKeyValue(peLink, "targetname", sz);

		Sys_UpdateWindows(W_ENTITY);

	}

	// promote the target to the src

	peLink = pe;
	
}

/*
===========
Drag_Begin
===========
*/
void Drag_Begin (int x, int y, int buttons,
		   vec3_t xaxis, vec3_t yaxis,
		   vec3_t origin, vec3_t dir)
{
	trace_t	t;
	drag_ok = false;
	VectorCopy (vec3_origin, pressdelta);

	drag_first = true;
	peLink = NULL;

	// shift LBUTTON = select entire brush
	if (buttons == (MK_LBUTTON | MK_SHIFT))
	{
		if (!dir[globals.XY_Orientation[0]] && !dir[globals.XY_Orientation[1]])
			Select_Ray (origin, dir, SF_ENTITIES_FIRST);	// hack for XY
		else
			Select_Ray (origin, dir, 0);
		return;
	}

	// ctrl-shift LBUTTON = select single face
	if (buttons == (MK_LBUTTON | MK_CONTROL | MK_SHIFT))
	{
		if(!globals.AllowFaceAndBrushSelection)
			Select_Deselect (false);
		Select_Ray (origin, dir, SF_SINGLEFACE);
		return;
	}

	// LBUTTON + all other modifiers = manipulate selection
	if (buttons & MK_LBUTTON && !(buttons&MK_RBUTTON))
	{
		Drag_Setup (x, y, buttons, xaxis, yaxis, origin, dir);
		return;
	}

	// middle button = grab texture
	if ((buttons == MK_MBUTTON) || (buttons == (MK_LBUTTON|MK_RBUTTON)))
	{
		t = Test_Ray (origin, dir, false);
		if (t.face)
		{
			VectorCopy(t.brush->mins, globals.d_new_brush_bottom);
			VectorCopy(t.brush->maxs, globals.d_new_brush_top);
			Texture_SetTexture (&t.face->texdef, true, false);
		}
		else
			Sys_Printf ("Did not select a texture\n");
		return;
	}

	// ctrl-middle button = set entire brush to texture
	if ((buttons == (MK_MBUTTON|MK_CONTROL)) || (buttons == (MK_LBUTTON|MK_RBUTTON|MK_CONTROL)))
	{
		t = Test_Ray (origin, dir, false);
		if (t.brush)
		{
			if (t.brush->brush_faces->texdef.name[0] == '(')
				Sys_Printf ("Can't change an entity texture\n");
			else
			{
				Brush_SetTexture (t.brush, &globals.d_texturewin.texdef);
				Sys_UpdateWindows (W_ALL);
			}
		}
		else
			Sys_Printf ("Didn't hit a brush\n");
		return;
	}

	// ctrl-shift-middle button = set single face to texture
	if ((buttons == (MK_MBUTTON|MK_SHIFT|MK_CONTROL)) || (buttons == (MK_LBUTTON|MK_RBUTTON|MK_SHIFT|MK_CONTROL)))
	{
		t = Test_Ray (origin, dir, false);
		if (t.brush)
		{
			if (t.brush->brush_faces->texdef.name[0] == '(')
				Sys_Printf ("Can't change an entity texture\n");
			else
			{
				t.face->texdef = globals.d_texturewin.texdef;
				Brush_Build( t.brush );
				Sys_UpdateWindows (W_ALL);
			}
		}
		else
			Sys_Printf ("Didn't hit a brush\n");
		return;
	}

}


brush_t TempList;
/*
===========
MoveSelection
===========
*/
void MoveSelection (vec3_t move)
{
	int					i, crushedface = 0;
	brush_t				*b, *bcopy;
	facesel_t			*f;

	if (!move[0] && !move[1] && !move[2])
		return;

	TempList.next = TempList.prev = &TempList;

	//
	// dragging only a part of the selection
	//
	if (globals.d_num_move_points)
	{
		int maxrange = globals.MaxGridExtent;

		for (i=0 ; i<globals.d_num_move_points ; i++)
			VectorAdd (globals.d_move_points[i], move, globals.d_move_points[i]);
		
		//First, run through the list and make sure no brushes become invalid
		for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
		{
			bcopy = Brush_Clone(b);
			Brush_AddToList(bcopy, &TempList);
			Brush_Build(bcopy);
			for (i=0 ; i<3 ; i++)
			{
				if (bcopy->mins[i] > bcopy->maxs[i]	||
//					bcopy->maxs[i] - bcopy->mins[i] > 4096 ||
					bcopy->maxs[i] > maxrange || bcopy->maxs[i] < -maxrange ||
					bcopy->mins[i] > maxrange || bcopy->mins[i] < -maxrange)
					break;	// dragged backwards or fucked up
			}

			Brush_Free(bcopy);

			if (i != 3)
			{
				crushedface = 1;
				break;
			}
		}
		for (f=selected_faces ; f ; f=f->next)
		{
			bcopy = Brush_Clone(f->owner);
			Brush_AddToList(bcopy, &TempList);
			Brush_Build(bcopy);
			for (i=0 ; i<3 ; i++)
			{
				if (bcopy->mins[i] > bcopy->maxs[i]	||
//					bcopy->maxs[i] - bcopy->mins[i] > 4096 ||
					bcopy->maxs[i] > maxrange || bcopy->maxs[i] < -maxrange ||
					bcopy->mins[i] > maxrange || bcopy->mins[i] < -maxrange)
					break;	// dragged backwards or fucked up
			}

			Brush_Free(bcopy);

			if (i != 3)
			{
				crushedface = 1;
				break;
			}
		}

		// if any of the brushes were crushed out of existance
		// cancel the entire move
		if (crushedface)
		{
			Sys_Printf ("Brush dragged backwards, move canceled\n");
			for (i=0 ; i<globals.d_num_move_points ; i++)
				VectorSubtract (globals.d_move_points[i], move, globals.d_move_points[i]);
		}
		
		for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
			Brush_Build( b );
		
		for (f = selected_faces ; f; f=f->next)
			Brush_Build( f->owner );
	}
	else
	{
		//
		// if there are lots of brushes selected, just translate instead
		// of rebuilding the brushes
		//
		if (drag_yvec[2] == 0 && selected_brushes.next->next != &selected_brushes)
		{
			VectorAdd (globals.d_select_translate, move, globals.d_select_translate);
		}
		else
		{
			Select_Move (move);
		}
	}
}

/*
===========
Drag_MouseMoved
===========
*/
void Drag_MouseMoved (int x, int y, int buttons)
{
	vec3_t	move, delta;
	int		i;
	float	h = 0, v = 0;
	float	scale = z.scale;
	HWND	hwnd = GetFocus();

	if (!buttons)
	{
		drag_ok = false;
		return;
	}
	if (!drag_ok)
		return;
	
	if(hwnd != globals.Camera.hwnd)
	{
		if(hwnd == globals.Z.hwnd)
		{
			if(y<0)
			{
				v = y;
				z.origin[2] += y;
			}
			else if(y>z.height)
			{
				v = y-z.height;
				z.origin[2] += v;
			}
		}
		else if(globals.Multiple2DViews)
		{
			if(hwnd == globals.XY.hwnd)
			{
				if(x<0)
				{
					h = x;
					globals.XY.origin[0] += h;
				}
				else if(x>globals.XY.width)
				{
					h = x-globals.XY.width;
					globals.XY.origin[0] += h;
				}
				if(y<0)
				{
					v = y;
					globals.XY.origin[1] += y;
				}
				else if(y>globals.XY.height)
				{
					v = y-globals.XY.height;
					globals.XY.origin[1] += v;
				}
				scale = globals.XY.scale;
			}
			else if(hwnd == globals.XZ.hwnd)
			{
				if(x<0)
				{
					h = x;
					globals.XZ.origin[0] += h;
				}
				else if(x>globals.XZ.width)
				{
					h = x-globals.XZ.width;
					globals.XZ.origin[0] += h;
				}
				if(y<0)
				{
					v = y;
					globals.XZ.origin[2] += y;
				}
				else if(y>globals.XZ.height)
				{
					v = y-globals.XZ.height;
					globals.XZ.origin[2] += v;
				}
				scale = globals.XZ.scale;
			}
			else if(hwnd == globals.YZ.hwnd)
			{
				if(x<0)
				{
					h = x;
					globals.YZ.origin[1] += h;
				}
				else if(x>globals.YZ.width)
				{
					h = x-globals.YZ.width;
					globals.YZ.origin[1] += h;
				}
				if(y<0)
				{
					v = y;
					globals.YZ.origin[2] += y;
				}
				else if(y>globals.YZ.height)
				{
					v = y-globals.YZ.height;
					globals.YZ.origin[2] += v;
				}
				scale = globals.YZ.scale;
			}
		}
		else
		{
			if(hwnd == globals.XY.hwnd)
			{
				if(x<0)
				{
					h = x;
					globals.XY.origin[globals.XY_Orientation[0]] += h;
				}
				else if(x>globals.XY.width)
				{
					h = x-globals.XY.width;
					globals.XY.origin[globals.XY_Orientation[0]] += h;
				}
				if(y<0)
				{
					v = y;
					globals.XY.origin[globals.XY_Orientation[1]] += y;
				}
				else if(y>globals.XY.height)
				{
					v = y-globals.XY.height;
					globals.XY.origin[globals.XY_Orientation[1]] += v;
				}
				scale = globals.XY.scale;
			}
		}
	}

	// clear along one axis
	if (buttons & MK_SHIFT)
	{
		drag_first = false;
		if (abs(x-pressx) > abs(y-pressy))
			y = pressy;
		else
			x = pressx;
	}

	for (i=0 ; i<3 ; i++)
	{
		move[i] = drag_xvec[i]*(x - pressx) 
				+ drag_yvec[i]*(y - pressy);
		move[i] = floor(move[i]/globals.d_gridsize+0.5)*globals.d_gridsize;
	}


	Sys_Status(0, "%i %i %i (%5.1f)", (int)move[0], (int)move[1], (int)move[2], VectorLength(move));

	VectorSubtract (move, pressdelta, delta);
	MoveSelection (delta);
	VectorCopy (move, pressdelta);
	if(h || v)
	{
		Sys_GetCursorPos(&x, &y);
		Sys_SetCursorPos(x-h, y+v);
		pressx -= h*scale;
		pressy -= v*scale;
	}
}

/*
===========
Drag_MouseUp
===========
*/
void Drag_MouseUp (void)
{
	Sys_Status(0, "drag completed.");
	if (globals.d_select_translate[0] || globals.d_select_translate[1] || globals.d_select_translate[2])
	{
		Select_Move (globals.d_select_translate);
		VectorCopy (vec3_origin, globals.d_select_translate);
		Sys_UpdateWindows (W_CAMERA);
	}
}
