#include "qe3.h"
#include "io.h"
#include "inc.h"
#include "fnmatch.h"
#include "directory.h"

PACKDirPtr				pakdirptr;
UInt16					dirsize;
wadlist_t				openwads;
char					currentwads[1024];
FILELIST				*availabletextures = NULL;
category_t				*categories;
paklist_t				*openpaks = NULL;

BOOL ReadBytes(FILE *file, void *addr, UInt32 size)
{
	while (size > 0x8000)
    {
		if (fread(addr, 1, 0x8000, file) != 0x8000)
			return FALSE;
		
		addr = (char *)addr + 0x8000;
		size -= 0x8000;
    }

	if (fread(addr, 1, size, file) != size)
		return FALSE;
	
	return TRUE;
}

int ReadMagic(FILE *file)
{
	UInt8 buf[4];
	
	if (ReadBytes(file, buf, 4) == FALSE)
		return FTYPE_ERROR;
	if (!strncmp(buf, "PACK", 4))
		return FTYPE_PACK;
	if (!strncmp(buf, "WAD2", 4))
		return FTYPE_WAD2;
	if (!strncmp(buf, "WAD3", 4))
		return FTYPE_WAD3;
	if (buf[0] == 0x17 && buf[1] == 0 && buf[2] == 0 && buf[3] == 0)
		return FTYPE_BSP;
	if (!strncmp(buf, "IDPO", 4))
		return FTYPE_MODEL;
	if (!strncmp(buf, "IDSP", 4))
		return FTYPE_SPRITE;
	if (buf[0] == 0x0a && buf[1] == 0x05 && buf[2] == 0x01 && buf[3] == 0x08)
		return FTYPE_PCX;

	return FTYPE_UNKNOWN;
}

FILE *OpenFileReadMagic(char *filename, int *ftype)
{
	FILE *f;
	
	*ftype = FTYPE_ERROR;

	if ((f = fopen(filename, "rb")) == NULL)
		return NULL;
	
	*ftype = ReadMagic(f);

	if (*ftype == FTYPE_ERROR)
    {
		fclose(f);
		return NULL;
    }
	
	return f;
}

char *ConvertFilePath(char *filename)
{
	char *cp;
	
	if (filename == NULL)
		Error("BUG: cannot convert a NULL pathname");
	
	for (cp = filename; *cp; cp++)
	{
		if (*cp == '/' || *cp == '\\')
		{
#ifdef QEU_DOS
			*cp = '\\';
#else
			*cp = '/';
#endif
		}
	}
	
	return filename;
}

/*
 * Read the PACK directory into memory.  The optional offset to the
 * start of the PACK file is given in "offset".  The number of files in
 * the directory is returned in *dirsize_r.
 */
PACKDirPtr ReadPACKDirectory(FILE *packfile, UInt32 offset, UInt16 *dirsize_r)
{
	PACKDirPtr dir;
	UInt32     pos, size;
	UInt16     max, i;
	
	*dirsize_r = 0;

	if (packfile == NULL)
		return NULL;
	
	if ((fseek(packfile, offset, SEEK_SET) < 0)
		|| (ReadMagic(packfile) != FTYPE_PACK)
		|| (ReadInt32(packfile, &pos) == FALSE)
		|| (ReadInt32(packfile, &size) == FALSE)
		|| (size == 0L)
		|| (size / sizeof(struct PACKDirectory) > 65535L)
		|| (fseek(packfile, offset + pos, SEEK_SET) < 0))
		return NULL;
	
	dir = (PACKDirPtr)qmalloc(size);
	max = (UInt16)(size / sizeof(struct PACKDirectory));
	
	for (i = 0; i < max; i++)
    {
		if (ReadBytes(packfile, &dir[i], sizeof(struct PACKDirectory)) == FALSE)
		{
			free(dir);
			return NULL;
		}

		ConvertFilePath(dir[i].name);
		dir[i].offset = SwapInt32(dir[i].offset);
		dir[i].size = SwapInt32(dir[i].size);
    }
	
	*dirsize_r = max;
	return dir;
}

void ClearFileList(FILELIST **list)
{
	FILELIST	*temp;

	while(*list)
	{
		temp = *list;
		*list = (*list)->next;
		free(temp);
	}
}

void ClearDirList(DIRLIST **list)
{
	DIRLIST	*temp;

	while(*list)
	{
		temp = *list;
		*list = (*list)->next;
		free(temp);
	}
}

directory_t *FindPakDir(directory_t *dir, char *name)
{
	directory_t	*currentPtr;

	for(currentPtr = dir; currentPtr; currentPtr = currentPtr->next)
	{
		if(!stricmp(name, currentPtr->name))
		{
			return currentPtr;
		}
	}
	return NULL;
}

BOOL AddToDirListAlphabetized(DIRLIST **list, char *dirname, int from)
{
	DIRLIST	*currentPtr, *previousPtr, *newPtr;

	for(currentPtr = *list; currentPtr; currentPtr = currentPtr->next)
	{
		if(!stricmp(dirname, currentPtr->dirname))
		{
			return false;
		}
	}
	previousPtr = NULL;
	currentPtr = *list;

        if((newPtr = (DIRLIST *)qmalloc(sizeof(DIRLIST))) == NULL)
		return false;

	strcpy(newPtr->dirname, dirname);
	newPtr->from = from;

	while(currentPtr != NULL && stricmp(dirname, currentPtr->dirname) > 0)
	{
		previousPtr = currentPtr;
		currentPtr = currentPtr->next;
	} //End while
	if(previousPtr == NULL)
	{
		newPtr->next = *list;
		*list = newPtr;
	} //End if
	else
	{
		previousPtr->next = newPtr;
		newPtr->next = currentPtr;
	} //End else
	return true;
}

BOOL AddToFileListAlphabetized(FILELIST **list, char *filename, UInt32 offset, UInt32 size)
{
	FILELIST	*currentPtr, *previousPtr, *newPtr;

	for(currentPtr = *list; currentPtr; currentPtr = currentPtr->next)
	{
		if(!stricmp(filename, currentPtr->filename))
		{
			currentPtr->offset = offset;
			currentPtr->size = size;
			return true;
		}
	}
	previousPtr = NULL;
	currentPtr = *list;

        if((newPtr = (FILELIST *)qmalloc(sizeof(FILELIST))) == NULL)
		return false;

	strcpy(newPtr->filename, filename);
	newPtr->offset = offset;
	newPtr->size = size;

	while(currentPtr != NULL && stricmp(filename, currentPtr->filename) > 0)
	{
		previousPtr = currentPtr;
		currentPtr = currentPtr->next;
	} //End while
	if(previousPtr == NULL)
	{
		newPtr->next = *list;
		*list = newPtr;
	} //End if
	else
	{
		previousPtr->next = newPtr;
		newPtr->next = currentPtr;
	} //End else
	return true;
}

directory_t *AddPakDir(directory_t **dir, char *name)
{
	directory_t	*currentPtr, *previousPtr, *newPtr;

	for(currentPtr = *dir; currentPtr; currentPtr = currentPtr->next)
	{
		if(!stricmp(name, currentPtr->name))
		{
			return currentPtr;
		}
	}
	previousPtr = NULL;
	currentPtr = *dir;

        if((newPtr = (directory_t *)qmalloc(sizeof(directory_t))) == NULL)
		return NULL;

	strcpy(newPtr->name, name);
	newPtr->files = NULL;

	while(currentPtr != NULL && stricmp(name, currentPtr->name) > 0)
	{
		previousPtr = currentPtr;
		currentPtr = currentPtr->next;
	}
	if(previousPtr == NULL)
	{
		newPtr->next = *dir;
		*dir = newPtr;
	}
	else
	{
		previousPtr->next = newPtr;
		newPtr->next = currentPtr;
	}
	return newPtr;
}

void OpenPakFile(char *filename)
{
	int			i;
	FILE		*pakfile;
	paklist_t	*newpak;
	int			f_type;

	if((pakfile = OpenFileReadMagic(filename, &f_type)) == NULL)
	{
		Sys_Printf("ERROR: Could not open %s", filename);
		return;
	}
	if(f_type != FTYPE_PACK)
	{
		Sys_Printf("ERROR: %s is not a valid pack file", filename);
		if(f_type != FTYPE_ERROR)
			fclose(pakfile);
		return;
	}
	pakdirptr = ReadPACKDirectory(pakfile, 0, &dirsize);
	if (pakdirptr == NULL)
	{
		Sys_Printf("ERROR: Could not read pack directory", filename);
		fclose(pakfile);
		return;
	}
	if (dirsize == 0)
	{
		fclose(pakfile);
		return;
	}

	newpak = malloc(sizeof(paklist_t));
	newpak->filehandle = pakfile;
	newpak->filename = strdup(filename);
	newpak->header = pakdirptr;
	newpak->dirsize = dirsize;
	newpak->next = openpaks;
	openpaks = newpak;

	for (i = 0; i < dirsize; i++)
		AddFile(directories, pakdirptr[i].name, pakfile, pakdirptr[i].offset, pakdirptr[i].size);
}

void GetFullPackFileList(FILELIST **filelist);

void GetTexturesInDir(FILELIST **list, char *dir, char *dirstring)
{
	struct _finddata_t	fileinfo;
	int					handle;
	char				name[512], *ext;

	handle = _findfirst (dirstring, &fileinfo);
	if (handle != -1)
	{
		do
		{
			if (fileinfo.attrib & _A_SUBDIR)
				continue;
			if (fileinfo.name[0] == '.')
				continue;

			ext = strrchr(fileinfo.name, '.');

			if(!ext)
				continue;

			if(globals.UseExternalTextures)
			{
				if(stricmp(ext, ".tga") && stricmp(ext, ".jpg") && stricmp(ext, ".wal") && stricmp(ext, ".m32"))
					continue;
			}

			*ext = 0;
			sprintf(name, "%s/%s", dir, fileinfo.name);
			AddToFileListAlphabetized(list, name, 0, 0);
		} while (_findnext( handle, &fileinfo ) != -1);
		_findclose (handle);
	}
}

void GetTextureList(FILELIST **list, char *dir)
{
	char				dirstring[512];

	if(globals.UseExternalTextures)
	{
		setdirstr(dirstring, globals.basepath, "/textures/", dir, "/*.*", NULL);
		GetTexturesInDir(list, dir, dirstring);
		
		if(globals.modpath[0])
		{
			setdirstr(dirstring, globals.modpath, "/textures/", dir, "/*.*", NULL);
			GetTexturesInDir(list, dir, dirstring);
		}
	}
	else if(globals.gametype == QUAKE2)
	{
		setdirstr(dirstring, globals.basepath, "/textures/", dir, "/*.wal", NULL);
		GetTexturesInDir(list, dir, dirstring);

		if(globals.modpath[0])
		{			
			setdirstr(dirstring, globals.modpath, "/textures/", dir, "/*.wal", NULL);
			GetTexturesInDir(list, dir, dirstring);
		}
	}
	else
	{
		setdirstr(dirstring, globals.basepath, "/textures/", dir, "/*.m32", NULL);
		GetTexturesInDir(list, dir, dirstring);

		if(globals.modpath[0])
		{
			setdirstr(dirstring, globals.modpath, "/textures/", dir, "/*.m32", NULL);
			GetTexturesInDir(list, dir, dirstring);
		}
	}
}

void UpdateListOfTextures(FILELIST **list)
{
	wadlist_t			*wad;
	struct _finddata_t	fileinfo;
	int					handle;
	char				dirstring[512];
	rscript_t			*rs;

	if(globals.gametype == QUAKE2 || globals.gametype == SOF)
	{
		GetFullPackFileList(list);
		setdirstr(dirstring, globals.basepath, "/textures/*.*", NULL);
		handle = _findfirst (dirstring, &fileinfo);
		if (handle != -1)
		{
			do
			{
				if (!(fileinfo.attrib & _A_SUBDIR))
					continue;
				if (fileinfo.name[0] == '.')
					continue;
				// add this directory to the list
				GetTextureList(list, fileinfo.name);
			} while (_findnext( handle, &fileinfo ) != -1);
			_findclose (handle);
		}

		if(globals.modpath[0])
		{
			setdirstr(dirstring, globals.modpath, "/textures/*.*", NULL);
			handle = _findfirst (dirstring, &fileinfo);
			if (handle != -1)
			{
				do
				{
					if (!(fileinfo.attrib & _A_SUBDIR))
						continue;
					if (fileinfo.name[0] == '.')
						continue;
					// add this directory to the list
					GetTextureList(list, fileinfo.name);
				} while (_findnext( handle, &fileinfo ) != -1);
				_findclose (handle);
			}
		}

	}
	else
	{
		for(wad = openwads.next; wad != &openwads; wad = wad->next)
		{
			GetWadFileList(list, wad->filename);
		}
	}

	for(rs = rootscript; rs; rs = rs->next)
	{
//		Sys_Printf("Adding %s to file list\n", s->name);
		AddToFileListAlphabetized(list, rs->name, FROMSHADER, 0);
	}
}

void FillComboBoxWithTextures(HWND h)
{
	FILELIST		*temp;

	SendMessage(h, CB_RESETCONTENT, 0 , 0);
		
	for(temp = availabletextures; temp; temp = temp->next)
	{
		if(globals.gametype == QUAKE2 || globals.gametype == SOF)
			StripExtension (temp->filename);

		SendMessage(h, CB_ADDSTRING, 0, (LPARAM)temp->filename);
	}
	SendMessage(h, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
}

void OpenMultiplePakFiles(char *filenames)
{
	char		*temp, buf[1024];
	FILE		*outfile;

	if(!filenames || !*filenames)
		return;

	strcpy(buf, filenames);

	outfile = fopen("out.txt", "w");
	for (temp = strtok(buf, ";"); temp; temp=strtok(0, ";"))
	{
		if(outfile)
			fprintf(outfile, "\"%s\"\n", temp);
		OpenPakFile(temp);
	}
	if(outfile)
		fclose(outfile);
}
