#ifndef _INC_H_
#define _INC_H_
#include "qe3.h"

typedef char           Int8;
typedef short          Int16;
typedef long           Int32;
typedef unsigned char  UInt8;
typedef unsigned short UInt16;
typedef unsigned long  UInt32;
typedef float          Float32;
typedef double         Float64;
#define MAX(a, b)              (((a) > (b)) ? (a) : (b))
#define MIN(a, b)              (((a) < (b)) ? (a) : (b))
#define RANDOM(x)              (random() % (x))
#define RANDOMIZE()             srand((int) time(NULL))

#define FTYPE_UNKNOWN 0
#define FTYPE_PACK    3    /* .pak  "PACK" */
#define FTYPE_WAD2    4    /* .wad  "WAD2" */
#define FTYPE_BSP     10   /* .bsp  (0x17 0x00 0x00 0x00) */
#define FTYPE_MODEL   11   /* .mdl  "IDPO" */
#define FTYPE_SPRITE  12   /* .spr  "IDSP" */
#define FTYPE_PCX     38   /* .pcx  (0x0a 0x05 0x01 0x08) */
#define FTYPE_WAD3     39  /* .wad  "WAD3" */
#define FTYPE_ERROR   -1

#ifdef FAT_ENDIAN
Bool	ReadInt16		(FILE *file, UInt16 huge *x);
Bool	ReadInt32		(FILE *file, UInt32 huge *x);
Bool	ReadFloat32		(FILE *file, Float32 huge *x);
Bool	WriteInt16		(FILE *file, UInt16 huge *x);
Bool	WriteInt32		(FILE *file, UInt32 huge *x);
Bool	WriteFloat32	(FILE *file, Float32 huge *x);
UInt16	SwapInt16		(UInt16 x);
UInt32	SwapInt32		(UInt32 x);
Float32	SwapFloat32		(Float32 x);
#else
#define ReadInt16(f, p)		ReadBytes((f), (p), 2L)
#define ReadInt32(f, p)		ReadBytes((f), (p), 4L)
#define ReadFloat32(f, p)	ReadBytes((f), (p), 4L)
#define WriteInt16(f, p)	WriteBytes((f), (p), 2L)
#define WriteInt32(f, p)	WriteBytes((f), (p), 4L)
#define WriteFloat32(f, p)	WriteBytes((f), (p), 4L)
#define SwapInt16(x)		(x)
#define SwapInt32(x)		(x)
#define SwapFloat32(x)		(x)
#endif /* FAT_ENDIAN */

typedef struct WAD2Directory *WAD2DirPtr;
struct WAD2Directory
{
   UInt32 offset;               /* offset to start of data */
   UInt32 size;                 /* byte size of data in WAD2 file */
   UInt32 dsize;                /* byte size of data when decompressed */
   UInt8  dtype;                /* type of data:
				   @ = raw data
				   B = bitmap with header
				   E = raw bitmap (no header)
				   */
   UInt8  compression;          /* type of compression:
				   0 = no compression
				   */
   UInt16 padding;              /* not used (always 0) */
   char   name[16];             /* name of data block */
};

#define FROMDISK	-1
#define FROMPAK		-2
#define FROMWAD		-3
#define FROMSHADER	-4

struct PACKDirectory
{
   char   name[56];             /* name of file */
   UInt32 offset;               /* offset to start of data */
   UInt32 size;                 /* byte size of data */
};
typedef struct PACKDirectory *PACKDirPtr;

typedef struct paklist_s
{
	char 				*filename;
	FILE				*filehandle;
	PACKDirPtr			header;
	int					dirsize;
	struct paklist_s	*next;
}paklist_t;

typedef struct DirListStruct
{
	char					dirname[256];
	int						from;
	struct	DirListStruct	*next;
} DIRLIST;

typedef struct FileListStruct
{
	char					filename[64];
	UInt32					offset;
	UInt32					size;
	struct	FileListStruct	*next;
} FILELIST;

typedef struct directory_s
{
	char				name[64];
	FILELIST			*files;
	struct directory_s	*next;
} directory_t;

typedef struct wadfilelist_s
{
	char					name[17];
	char					loaded; // Set to 0 when created
	struct wadlist_s		*owner;
	UInt32					offset;
	UInt32					size;
	struct wadfilelist_s	*next;
}wadfilelist_t;

typedef struct wadlist_s
{
	char 				*filename;
	char				*longname;
	FILE				*filehandle;
	wadfilelist_t		*files;
	struct wadlist_s	*prev;
	struct wadlist_s	*next;
}wadlist_t;

typedef struct categorytextures_s
{
	char						*name;
	char						*from;
	struct categorytextures_s	*next;
}categorytextures_t;

typedef struct category_s
{
	char				*name;
	int					identifier;		//to find this from the menu.
	categorytextures_t	*textures;
	struct category_s	*subcategories;
	struct category_s	*next;
}category_t;

extern wadlist_t	openwads;
extern char			currentwads[1024];
extern FILELIST		*availabletextures;
extern category_t	*categories;
extern paklist_t	*openpaks;

BOOL		ReadBytes					(FILE *file, void *addr, UInt32 size);
int			ReadMagic					(FILE *file);
FILE		*OpenFileReadMagic			(char *filename, int *ftype);
void		ClearFileList				(FILELIST **);
void		ClearDirList				(DIRLIST **);
int			GetPackFileList				(FILELIST **, char *);
int			GetPackTextureDirs			(DIRLIST **);
BOOL		AddToDirListAlphabetized	(DIRLIST **, char *, int);
BOOL		AddToFileListAlphabetized	(FILELIST **t, char *, UInt32, UInt32);
void		OpenPakFile					(char *);
void		OpenWad2					(char *);
BOOL		WadLoadFile					(char *, void **, byte *);
void		CloseWadFile				(char *);
int			GetWadFileList				(FILELIST **, char *);
wadlist_t	*WadIsOpen					(char *);
wadlist_t	*OpenWad					(char *);
void		CloseExtinctWads			(char *);
void		FillComboBoxWithTextures	(HWND);
void		UpdateListOfTextures		(FILELIST **);
void		ParseTextureCategoryFile	(void);
void		CreateCategoryMenus			(HMENU, category_t *, int);
category_t	*FindCategory				(int, category_t *);
void		OpenMultiplePakFiles		(char *);
int			TryDocking					(HWND, long, LPRECT, int);


#endif
