// lbmlib.c

#include "qe3.h"
#include "inc.h"
#pragma warning(disable : 4142)
#pragma warning(disable : 4005)
#include "directory.h"
#include <il/il.h>

ILuint	imageid = -1;

/*
============================================================================

LOAD PCX

============================================================================
*/
typedef struct
{
    char	manufacturer;
    char	version;
    char	encoding;
    char	bits_per_pixel;
    unsigned short	xmin,ymin,xmax,ymax;
    unsigned short	hres,vres;
    unsigned char	palette[48];
    char	reserved;
    char	color_planes;
    unsigned short	bytes_per_line;
    unsigned short	palette_type;
    char	filler[58];
    unsigned char	data;			// unbounded
} pcx_t;


/*
==============
LoadPCX
==============
*/
int LoadPCX (char *filename, byte **pic, byte **palette, int *width, int *height)
{
	byte	*raw;
	pcx_t	*pcx;
	int		x, y;
	int		len;
	int		dataByte, runLength;
	byte	*out, *pix;

	if (pic)
		*pic = NULL;
	if (palette)
		*palette = NULL;
	if (width)
		*width = 0;
	if (height)
		*height = 0;

	//
	// load the file
	//
	len = qLoadFile(filename, (void **)&raw);

	if (len == -1)
		return false;

	//
	// parse the PCX file
	//
	pcx = (pcx_t *)raw;
	raw = &pcx->data;

	pcx->xmin = LittleShort(pcx->xmin);
	pcx->ymin = LittleShort(pcx->ymin);
	pcx->xmax = LittleShort(pcx->xmax);
	pcx->ymax = LittleShort(pcx->ymax);
	pcx->hres = LittleShort(pcx->hres);
	pcx->vres = LittleShort(pcx->vres);
	pcx->bytes_per_line = LittleShort(pcx->bytes_per_line);
	pcx->palette_type = LittleShort(pcx->palette_type);

	if (pcx->manufacturer != 0x0a
		|| pcx->version != 5
		|| pcx->encoding != 1
		|| pcx->bits_per_pixel != 8
		|| pcx->xmax >= 640
		|| pcx->ymax >= 480)
		Error ("Bad pcx file %s", filename);
	
	if (palette)
	{
		*palette = qmalloc(768);
		memcpy (*palette, (byte *)pcx + len - 768, 768);
	}

	if (width)
		*width = pcx->xmax+1;
	if (height)
		*height = pcx->ymax+1;

	if (!pic)
		return false;

        out = qmalloc ( (pcx->ymax+1) * (pcx->xmax+1) );
	if (!out)
	{
		Sys_Printf("LoadPCX(): Couldn't allocate memory");
		return false;
	}

	*pic = out;

	pix = out;

	for (y=0 ; y<=pcx->ymax ; y++, pix += pcx->xmax+1)
	{
		for (x=0 ; x<=pcx->xmax ; )
		{
			dataByte = *raw++;

			if((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				dataByte = *raw++;
			}
			else
				runLength = 1;

			while(runLength-- > 0)
				pix[x++] = dataByte;
		}

	}

	if ( raw - (byte *)pcx > len)
	{
		free(pcx);
		free(pix);
		Sys_Printf("PCX file %s was malformed", filename);
		return false;
	}
	free (pcx);
	return true;
}
/*
============================================================================

LOAD IMAGE

============================================================================
*/

/*
==============
Load256Image

Will load either an lbm or pcx, depending on extension.
Any of the return pointers can be NULL if you don't want them.
==============
*/
void Load256Image (char *name, byte **pixels, byte **palette,
				   int *width, int *height)
{
	LoadPCX (name, pixels, palette, width, height);
}

int ilEnd(void)
{
	if(imageid == -1)
		return false;

	ilDeleteImages(1, &imageid);
	imageid = -1;
	return true;
}

void ilStart(void)
{
	if(imageid != -1)
		ilEnd();

	ilGenImages(1, &imageid);
	ilBindImage(imageid);
}

int GetImage(char *name, byte **pixels, int *width, int *height, int type)
{
	byte	*buf;
	int		size;

	ilStart();
	size = qLoadFile(name, &buf);

	if(size == -1)
		return false;

	if(!ilLoadL(type, (ILvoid *)buf, (ILint)size))
	{
		free(buf);
		return false;
	}

	free(buf);

	if(!ilConvertImage(IL_RGBA, IL_UNSIGNED_BYTE))
	{
		ilEnd();
		return false;
	}

	*width = ilGetInteger(IL_IMAGE_WIDTH);
	*height = ilGetInteger(IL_IMAGE_HEIGHT);
	*pixels = ilGetData();

	return true;
}

int LoadPNG(char *name, byte **pixels, int *width, int *height)
{
	return GetImage(name, pixels, width, height, IL_PNG);
}

int LoadTGA (char *name, byte **pixels, int *width, int *height)
{
	return GetImage(name, pixels, width, height, IL_TGA);
}

int LoadJPG (char *name, byte **pixels, int *width, int *height)
{
	return GetImage(name, pixels, width, height, IL_JPG);
}

