#include "qe3.h"
#include "md2.h"
#include "directory.h"
int			frame, oldframe, backlerp;
md2model_t	*currententity;

MD2_t *NewMD2(void)
{
	MD2_t	*newmd2;

	newmd2 = malloc(sizeof(MD2_t));

	newmd2->stateStart = IDLE1_START;
	newmd2->stateEnd = IDLE1_END;
	newmd2->numGlCommands = newmd2->frameSize = newmd2->numFrames = 0;
	newmd2->currentFrame = IDLE1_START;
	newmd2->nextFrame = newmd2->currentFrame+1;
	newmd2->endFrame = IDLE1_END;
	newmd2->interpolation = 0.0f;
	newmd2->glCommands = NULL;
	newmd2->frames = NULL;

	return newmd2;
}

void DeleteMD2(MD2_t *MD2)
{
	if(MD2->glCommands)
		free(MD2->glCommands);

	if(MD2->frames)
		free(MD2->frames);
}

void *MD2Load(char *filename)
{
	qfile_t		*file = NULL;
	MD2_HEADER	header;
	MD2_t		*MD2;
	char		*frames;
	long		*glCommands;
	int			offset = 0;

	if((file= qfopen(filename, "rb", NULL))==NULL)
	{
		return NULL;
	}
	
	qfread(&header, sizeof(MD2_HEADER), 1, file);
	
	if(header.magic!= 844121161)
	{
//		Sys_Printf("%s is not a valid .md2\n", filename);
		qfclose(file);
		return NULL;
	}
	
	if(header.version!=8)
	{
//		Sys_Printf("%s is not a valid .md2\n", filename);
		qfclose(file);
		return NULL;
	}
	
	frames= malloc(header.frameSize*header.numFrames);
	
	if(frames==NULL)
	{
		qfclose(file);
		return NULL;
	}
	
	qfseek(file, offset + header.offsetFrames, SEEK_SET);
	qfread(frames, header.frameSize*header.numFrames, 1, file);	
	glCommands= malloc(header.numGlCommands*sizeof(long));
	
	if(glCommands==NULL)
	{
		free(frames);
		qfclose(file);
		return NULL;
	}
	
	qfseek(file, offset + header.offsetGlCommands, SEEK_SET);
	qfread(glCommands, header.numGlCommands*sizeof(long), 1, file);

	if(offset)
		qfclose(file);

	MD2 = NewMD2();
	MD2->frames = frames;
	MD2->glCommands = glCommands;
	MD2->numFrames	= header.numFrames;
	MD2->numGlCommands	= header.numGlCommands;
	MD2->frameSize	= header.frameSize;
	MD2->numTriangles	= header.numTriangles;
	MD2SetState(MD2, IDLE1);

//	Sys_Printf("Loaded %s correctly\n", filename);
	return (void *)MD2;
}

void MD2Render(MD2_t *MD2, int numFrame)
{
	MD2_FRAME_PTR			currentFrame;
	vec3_t					v1;
	long					*command;
	float					texcoord[2];
	int						loop;
	int						vertIndex;
	int						type;
	int						numVertex;
	int						index;
	
	currentFrame= (MD2_FRAME*) ((char*)MD2->frames+MD2->frameSize*numFrame);
	command      = MD2->glCommands;
	while((*command)!=0)
	{
		if(*command>0)		// This Is A Triangle Strip
		{
			numVertex= *command; 
			command++; 
			type= 0;
		}
		else			// This Is A Triangle Fan
		{
			numVertex= - *command; 
			command++; 
			type= 1;
		}
		
		if(numVertex<0)
			numVertex= -numVertex;
		
		index= 0;
		
		for(loop=0; loop<numVertex; loop++)
		{
			MD2->vertList[index].u= *((float*)command); 
			command++;
			MD2->vertList[index].v= *((float*)command); 
			command++;
			
			vertIndex= *command; 
			command++;
			
			MD2->vertList[index].x= ( (currentFrame->vertices[vertIndex].vertex[0]* currentFrame->scale[0])+ currentFrame->translate[0]);
			MD2->vertList[index].z= -((currentFrame-> vertices[vertIndex].vertex[1]* currentFrame->scale[1])+ currentFrame->translate[1]);
			MD2->vertList[index].y= ( (currentFrame->vertices[vertIndex].vertex[2]* currentFrame->scale[2])+ currentFrame->translate[2]);
			index++;
		}
		
		if(type==0)
		{
			glBegin(GL_TRIANGLE_STRIP);
			for(loop=0; loop<index; loop++)
			{
				v1[0]=(MD2->vertList[loop].x);
				v1[1]=(MD2->vertList[loop].y);
				v1[2]=(MD2->vertList[loop].z);
				
				texcoord[0]= MD2->vertList[loop].u;
				texcoord[1]= MD2->vertList[loop].v;
				
				glTexCoord2fv(texcoord);
				glVertex3dv(v1);
			}
			glEnd();
		}
		else
		{
			glBegin(GL_TRIANGLE_FAN);
			for(loop=0; loop<index; loop++)			
			{
				v1[0]=(MD2->vertList[loop].x);
				v1[1]=(MD2->vertList[loop].y);
				v1[2]=(MD2->vertList[loop].z);
				
				texcoord[0]= MD2->vertList[loop].u;
				texcoord[1]= MD2->vertList[loop].v;
				
				glTexCoord2fv(texcoord);
				glVertex3dv(v1);
			}
			glEnd();
		}
	}
}


void MD2Animate(MD2_t *MD2, int startFrame, int EndFrame, float percent)
{
	if(MD2->currentFrame == -1 || MD2->currentFrame > MD2->endFrame)
	{
		MD2->currentFrame= startFrame;
		MD2->nextFrame	= MD2->currentFrame+1;
		MD2->endFrame	= EndFrame;
	}
	
	//If the interpolation value is above one, then we advance to the next frame
	//of animation for the model
	if(MD2->interpolation > 1.0f)
	{
		MD2->interpolation= 0.0f;
		MD2->currentFrame = MD2->nextFrame;
		MD2->nextFrame	 = MD2->currentFrame+1;
	}
	
	MD2->interpolation += percent;
	MD2Render(MD2, MD2->currentFrame);
}

void MD2SetState(MD2_t *MD2, int state)
{
	if(MD2->numFrames <= 2)
	{
		MD2->stateStart = MD2->stateEnd = 0;
		return;
	}

	switch(state)
	{
		case IDLE1:
			MD2->stateStart	 = IDLE1_START;
			MD2->stateEnd	 = IDLE1_END;
			MD2->currentFrame = -1;
			break;
		case RUN:
			MD2->stateStart	 = RUN_START;
			MD2->stateEnd	 = RUN_END;
			MD2->currentFrame = -1;
			break;
		case SHOT_STAND:
			MD2->stateStart	 = SHOT_STAND_START;
			MD2->stateEnd	 = SHOT_STAND_END;
			MD2->currentFrame = -1;
			break;
		case SHOT_SHOULDER:
			MD2->stateStart	 = SHOT_SHOULDER_START;
			MD2->stateEnd	 = SHOT_SHOULDER_END;
			MD2->currentFrame = -1;
			break;
		case JUMP:
			MD2->stateStart	 = JUMP_START;
			MD2->stateEnd	 = JUMP_END;
			MD2->currentFrame = -1;
			break;
		case IDLE2:
			MD2->stateStart	 = IDLE2_START;
			MD2->stateEnd	 = IDLE2_END;
			MD2->currentFrame = -1;
			break;
		case SHOT_FALLDOWN:
			MD2->stateStart	 = SHOT_FALLDOWN_START;
			MD2->stateEnd	 = SHOT_FALLDOWN_END;
			MD2->currentFrame = -1;
			break;
		case IDLE3:
			MD2->stateStart	 = IDLE3_START;
			MD2->stateEnd	 = IDLE3_END;
			MD2->currentFrame = -1;
			break;
		case IDLE4:
			MD2->stateStart	 = IDLE4_START;
			MD2->stateEnd	 = IDLE4_END;
			MD2->currentFrame = -1;
			break;
		case CROUCH:
			MD2->stateStart	 = CROUCH_START;
			MD2->stateEnd	 = CROUCH_END;
			MD2->currentFrame = -1;
			break;
		case CROUCH_CRAWL:
			MD2->stateStart	 = CROUCH_CRAWL_START;
			MD2->stateEnd	 = CROUCH_CRAWL_END;
			MD2->currentFrame = -1;
			break;
		case CROUCH_IDLE:
			MD2->stateStart	 = CROUCH_IDLE_START;
			MD2->stateEnd	 = CROUCH_IDLE_END;
			MD2->currentFrame = -1;
			break;
		case CROUCH_DEATH:
			MD2->stateStart	 = CROUCH_DEATH_START;
			MD2->stateEnd	 = CROUCH_DEATH_END;
			MD2->currentFrame = -1;
			break;
		case DEATH_FALLBACK:
			MD2->stateStart	 = DEATH_FALLBACK_START;
			MD2->stateEnd	 = DEATH_FALLBACK_END;
			MD2->currentFrame = -1;
			break;
		case DEATH_FALLFORWARD:
			MD2->stateStart	 = DEATH_FALLFORWARD_START;
			MD2->stateEnd	 = DEATH_FALLFORWARD_END;
			MD2->currentFrame = -1;
			break;
		case DEATH_FALLBACKSLOW:
			MD2->stateStart	 = DEATH_FALLBACKSLOW_START;
			MD2->stateEnd	 = DEATH_FALLBACKSLOW_END;
			MD2->currentFrame = -1;
			break;
	}

//	if(MD2->stateStart > MD2->numFrames - 1)
//		MD2->stateStart = 0;
//
//	if(MD2->stateEnd > MD2->numFrames - 1)
//		MD2->stateEnd = 0;

}

int LoadMD2Model(char *name)
{
	return true;
}
