#include "qe3.h"
#include "md2.h"

model_t	*models = NULL;

void LoadModel(char *classname, char *modelname, char *skinname, int type);

/*
	This is the file where all qe5 <-> model interfacing happens.
*/


//	Model Definition File Format:
//	classname "model filename" "model skin filename" type
char *get_token(char **src, char *token_sep);
void SkipSpace(char **s)
{
	while(*s && (**s == ' ' || **s == '\t'))
		*s++;
}

int GetModelType(char *s)
{
	if(!stricmp(s, "MD2"))
		return TYPE_MD2;

	return 0;
}

void ParseModelDefinitionFile(char *filename)
{
	FILE	*f;
	char	buf[1024], *s;
	char	*classname, *modelname, *skinname, *typestr;
	int		line = 0, type;

	if(!(f = fopen(filename, "r")))
	{
		Sys_Printf("No Model Definition File\n");
		return;
	}

	while(!feof(f))
	{
		line++;
		fgets(buf, 1024, f);
		s = strrchr(buf, '\n');

		if(s)
			*s = 0;

		s = buf;
		SkipSpace(&s);

		if(!*s || (*s == '/' && *(s+1) == '/'))
			continue;

		classname	= get_token(&s, " \t");
		modelname	= get_token(&s, " \t\"");
		skinname	= get_token(&s, " \t\"");
		typestr		= get_token(&s, " \t");

		if(classname && *classname && modelname && *modelname && skinname && *skinname && typestr && *typestr)
		{
			type = GetModelType(typestr);

			if(!type)
			{
				Sys_Printf("Warning: Line %d of model definition file has unknown type %s\n", line, typestr);
				continue;
			}
			LoadModel(classname, modelname, skinname, type);
		}
		else
		{
			Sys_Printf("Warning: Line %d of model definition file is incomplete\n", line);
		}
	}
}

model_t *FindModel(char *classname)
{
	model_t	*m;

	for(m = models; m; m = m->next)
		if(!stricmp(m->classname, classname))
			return m;

	return NULL;
}

void InitModels(void)
{
	char		*basepath = ValueForKey(globals.d_project_entity, "basepath");
	char		buf[512];
	eclass_t	*e;
	model_t		*m;

	return;
	setdirstr2(buf, basepath?basepath:"", "scripts/models.qe5", NULL);
	ParseModelDefinitionFile(buf);

	for(m = models; m; m = m->next)
	{
		e = Eclass_ForName (m->classname, 0);
		if(!e->fixedsize)
			Sys_Printf("Warning: Attempted to set a model for brush-based entity: %s\n", m->classname);
		else
			e->model = m;
	}
}

void AnimateModels(void)
{
}

int SetupEntityForRendering(entity_t *e)
{
	vec_t	angle = FloatForKey(e, "angle");
	return 0;
}

void LoadModel(char *classname, char *modelname, char *skinname, int type)
{
	char	*basepath = ValueForKey(globals.d_project_entity, "basepath");
	void	*loadmodel = NULL;
	model_t	*newmodel;
	image_t	*image;

	if(newmodel = FindModel(classname))
		return;

	if(!(image = LoadTextureImage(skinname, IMAGE_TGA|IMAGE_JPG|IMAGE_PCX|IMAGE_WAL)))
	{
		Sys_Printf("Couldn't load skin %s\n", skinname);
		return;
	}

	switch(type)
	{
		case TYPE_MD2:
			{
				MD2_t	*md2;

				loadmodel = MD2Load(modelname);
				md2 = (MD2_t *)loadmodel;
				if(loadmodel)
					md2->skin = image;
			}
			break;
	}

	if(!loadmodel)
	{
		Sys_Printf("Couldn't load %s\n", modelname);
		return;
	}

	newmodel = malloc(sizeof(model_t));
	newmodel->classname = strdup(classname);
	newmodel->model = loadmodel;
	newmodel->image = image;
	newmodel->type = type;
	newmodel->next = models;
	models = newmodel;
}
void MD2Animate(MD2_t *MD2, int startFrame, int EndFrame, float percent);


void DrawModel(brush_t *b)
{
	model_t *m = b->owner->eclass->model;
	image_t	*skin;
	float	angle = FloatForKey(b->owner, "angle");
	vec3_t	origin;

	VectorSubtract (b->mins, b->owner->eclass->mins, origin);
	glPushMatrix ();
	glTranslatef(origin[0], origin[1], origin[2]);
	glRotatef(90,  1, 0, 0);
	glRotatef(angle+90, 0, 1, 0);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	switch(m->type)
	{
		case TYPE_MD2:
			{
				MD2_t	*md2 = (MD2_t *)m->model;
				
				// select skin
				skin = md2->skin;
				if (!skin)
					skin = notexture->image;	// fallback...
				glBindTexture(GL_TEXTURE_2D, skin->texture_number);
				if(md2->numFrames > 2)
					MD2Animate(md2, md2->stateStart, md2->stateEnd, (rs_realtime - m->time)*8.0);
				else
					MD2Render(md2, 0);//md2->currentFrame);
			}
			break;
	}
	m->time = rs_realtime;
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glPopMatrix ();
}
