/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_script.c - scripted texture rendering - MrG
// Areas commented 'Heffo' done by Heffo (funny that)

//#include "gl_local.h"
#include "../qe3.h"
#include "../inc.h"
#include "gl_cin.h"

#include <io.h>

void CIN_FreeCin (int texnum);

#define m_malloc(x) malloc(x)
#define m_free(x) free(x)

float rs_realtime = 0;

#define TOK_DELIMINATORS "\r\n\t "

rscript_t	*rootscript = NULL;

#if 0
void RS_ListScripts()
{
	rscript_t	*rs;
	int			i=0;

	rs=rootscript;
	while (rs != NULL) {
		ri.Con_Printf(PRINT_ALL,">> %s\n",rs->name);
		i++;
		rs=rs->next;
	}
}
#endif

char *COM_SkipPath (char *pathname)
{
	char	*last;
	
	last = pathname;
	while (*pathname)
	{
		if (*pathname=='/')
			last = pathname+1;
		pathname++;
	}
	return last;
}

int RS_Animate(rs_stage_t *stage)
{
	anim_stage_t *anim;
	float	time, rt;

	rt=rs_realtime * 1000;
	time = rt - (stage->last_anim_time+stage->anim_delay);
	anim=stage->last_anim;
	while ((stage->last_anim_time*1000) < rt) {
		anim=anim->next;
		if (!anim)
			anim=stage->anim_stage;
		stage->last_anim_time+=stage->anim_delay;
	}
	stage->last_anim = anim;
	return anim->texture->texture_number;
}

/*
int RS_Animate(rs_stage_t *stage)
{
	int rt;
	anim_stage_t *anim;

	rt = rs_realtime * 1000;
	anim=stage->last_anim;
	if (rt < (stage->anim_delay+stage->last_anim_time))
		return stage->last_anim->texture->texnum;
	while (rt >= (stage->anim_delay+stage->last_anim_time)) {
		anim=anim->next;
		if (!anim)
			anim=stage->anim_stage;
		stage->last_anim_time+=stage->anim_delay;
	}
	stage->last_anim=anim;
	stage->last_anim_time = rt;
	return anim->texture->texnum;
}
*/

void RS_ResetScript(rscript_t *rs)
{
	rs_stage_t		*stage, *tmp_stage;
	anim_stage_t	*anim, *tmp_anim;

	strcpy(rs->name,"");
	stage = rs->stage;
	while (stage != NULL) {
		if (stage->anim_count) {
			anim=stage->anim_stage;
			while (anim != NULL) {
				tmp_anim=anim;
				if (anim->texture)
					if (anim->texture->is_cin)
//						CIN_FreeCin(anim->texture->texnum);
						CIN_FreeCin(anim->texture->texture_number);
				anim=anim->next;
				m_free(tmp_anim);
			}
		}

		tmp_stage=stage;
		stage=stage->next;
		m_free(tmp_stage);
	}
	rs->stage = NULL;
	rs->dontflush = false;
	rs->subdivide = 0;				// Heffo - surface subdivision
	rs->warpdist = 0.0f; // Heffo - vertex warping
	rs->warpsmooth = 0.0f; // Heffo - vertex warping
	rs->ready = false;
}

void RS_ClearStage(rs_stage_t *stage)
{
	anim_stage_t	*anim, *tmp_anim;

	stage->alphashift.max=stage->alphashift.min=stage->alphashift.speed=0;
	stage->anim_delay=0;

	anim=stage->anim_stage;
	while (anim != NULL) {
		tmp_anim=anim;
		anim=anim->next;
		m_free(tmp_anim);
	}
	stage->anim_stage = NULL;
		
	stage->blendfunc.blend=false;
	stage->blendfunc.dest=stage->blendfunc.source=0;
	stage->last_anim=0;
	stage->last_anim_time=0;
	stage->anim_count=0;

	stage->rot_speed=0;
	stage->scroll.speedX=0;
	stage->scroll.speedY=0;
	stage->scroll.typeX=0;
	stage->scroll.typeY=0;
	stage->texture=NULL;
	stage->envmap=false;
	stage->lightmap=true;
	stage->alphamask=false;
}

rscript_t *RS_NewScript(char *name)
{
	rscript_t *rs;

	if (rootscript == NULL)
	{
		rootscript=(rscript_t *)m_malloc(sizeof(rscript_t));
		rs=rootscript;
	}
	else
	{
		rs=rootscript;

		while (rs->next != NULL)
			rs=rs->next;
		rs->next=(rscript_t *)m_malloc(sizeof(rscript_t));
		rs=rs->next;
	}

	if(!strnicmp(name, "textures/", 9))
	{
		name += 9;
		strcpy(rs->name,name);
	}
	else
		rs->name[0] = 0;	//We don't use textures that aren't in the textures/ directory.

	rs->stage = NULL;
	rs->next = NULL;
	rs->dontflush = false;
	rs->subdivide = 0;				// Heffo - surface subdivision
	rs->warpdist = 0.0f; // Heffo - vertex warping
	rs->warpsmooth = 0.0f; // Heffo - vertex warping
	rs->ready = false;
	rs->surfaceflags = rs->contentflags = rs->lightamount = 0;
	return rs;
}

rs_stage_t *RS_NewStage(rscript_t *rs)
{
	rs_stage_t	*stage;

	if (rs->stage == NULL) {
		rs->stage=(rs_stage_t *)m_malloc(sizeof(rs_stage_t));
		stage=rs->stage;
	} else {
		stage=rs->stage;
		while (stage->next != NULL)
			stage=stage->next;
		stage->next=(rs_stage_t *)m_malloc(sizeof(rs_stage_t));
		stage=stage->next;
	}
	strcpy(stage->name,"pics/noimage.tga");
	stage->anim_stage = NULL;
	stage->next = NULL;
	stage->last_anim = NULL;
	RS_ClearStage(stage);
	return stage;
}

void RS_FreeAllScripts()
{
	rscript_t *rs, *tmp_rs;
	
	rs=rootscript;
	while (rs != NULL) {
		tmp_rs=rs->next;
		RS_ResetScript(rs);
		m_free(rs);
		rs=tmp_rs;
	}
}

void RS_FreeScript(rscript_t *rs)
{
	rscript_t	*tmp_rs;

	if (!rs)
		return;

	if (rootscript == rs) {
		rootscript=rootscript->next;
		RS_ResetScript(rs);
		m_free(rs);
		return;
	}

	tmp_rs=rootscript;
	while (tmp_rs->next != rs)
		tmp_rs=tmp_rs->next;
	tmp_rs->next=rs->next;
	RS_ResetScript(rs);
	m_free(rs);
}

void RS_FreeUnmarked()
{
	rscript_t *rs, *tmp_rs;
	
	rs=rootscript;
	while (rs != NULL) {
		tmp_rs=rs->next;
		if (!rs->dontflush) {
			RS_FreeScript(rs);
		}
		rs=tmp_rs;
	}
}

float RS_BlendID(char *blend)
{
	if (!strcmp(blend,""))
		return 0;
	if (!_stricmp(blend,"GL_ZERO"))
		return GL_ZERO;
	if (!_stricmp(blend,"GL_ONE"))
		return GL_ONE;
	if (!_stricmp(blend,"GL_DST_COLOR"))
		return GL_DST_COLOR;
	if (!_stricmp(blend,"GL_ONE_MINUS_DST_COLOR"))
		return GL_ONE_MINUS_DST_COLOR;
	if (!_stricmp(blend,"GL_SRC_ALPHA"))
		return GL_SRC_ALPHA;
	if (!_stricmp(blend,"GL_ONE_MINUS_SRC_ALPHA"))
		return GL_ONE_MINUS_SRC_ALPHA;
	if (!_stricmp(blend,"GL_DST_ALPHA"))
		return GL_DST_ALPHA;
	if (!_stricmp(blend,"GL_ONE_MINUS_DST_ALPHA"))
		return GL_ONE_MINUS_DST_ALPHA;
	if (!_stricmp(blend,"GL_SRC_ALPHA_SATURATE"))
		return GL_SRC_ALPHA_SATURATE;
	if (!_stricmp(blend,"GL_SRC_COLOR"))
		return GL_SRC_COLOR;
	if (!_stricmp(blend,"GL_ONE_MINUS_SRC_COLOR"))
		return GL_ONE_MINUS_SRC_COLOR;
	return 0;
}

rscript_t *RS_FindScript(char *name)
{
	rscript_t	*rs;

	rs=rootscript;
	while (rs != NULL) {
		if (!_stricmp(rs->name,name)) {
			return rs;
		}
		rs=rs->next;
	}
	return NULL;
}

image_t *RS_LoadTexture(char *name)
{
	char		ext[64], *s;
	image_t		*image;

	if(!strnicmp(name, "textures/", 9))
		s = name+9;
	else
		s = NULL;

	for(image = globals.images; image; image = image->next)
	{
		if(!stricmp(image->name, name) || (s && !stricmp(image->name, s)))
		{
			return image;
		}
	}

//	Sys_Printf("Loading script image file: %s\n", name);
	s = strrchr(name, '.');
	strcpy(ext, s?s+1:"");

	if (!stricmp(ext, "cin"))
	{
		return LoadTextureImage(name, IMAGE_CIN);
	}
	else if(!stricmp(ext, "png"))
	{
		return LoadTextureImage(name, IMAGE_JPG);
	}
	else if(!stricmp(ext, "tga"))
	{
		return LoadTextureImage(name, IMAGE_TGA);
	}
	else if(!stricmp(ext, "jpg"))
	{
		return LoadTextureImage(name, IMAGE_JPG);
	}
	else if(!stricmp(ext, "pcx"))
	{
		return LoadTextureImage(name, IMAGE_PCX);
	}
	else if(!stricmp(ext, "wal"))
	{
		return LoadTextureImage(name, IMAGE_WAL);
	}
	else
		Sys_Printf("Unsupported file type \"%s\"\n", ext);

	return NULL;
}

void RS_ReadyScript(rscript_t *rs)
{
	rs_stage_t		*stage;
	anim_stage_t	*anim;

	if (rs->ready)
		return;

//	mode = (rs->dontflush) ? it_pic : it_wall;
	stage = rs->stage;
	while (stage != NULL) {
		anim=stage->anim_stage;
		while (anim != NULL) {
			anim->texture=RS_LoadTexture(anim->name);
			if (!anim->texture)
				anim->texture=notexture->image;//Texture_ForName("pics/noimage.tga");
			anim=anim->next;
		}

		if (strlen(stage->name))
		{
			if(!stricmp(stage->name, "$lightmap"))
			{
				stage->lm_only = true;
				stage->texture = 0;
			}
			else
			{
				stage->lm_only = false;
				stage->texture=RS_LoadTexture(stage->name);
			}

			if (!stage->texture)
				stage->texture=notexture->image;//Texture_ForName("pics/noimage.tga");
		}
		stage=stage->next;
	}
	rs->ready = true;
}
#if 0
void RS_UpdateRegistration()
{
	rscript_t		*rs;
	rs_stage_t		*stage;
	anim_stage_t	*anim;
	
	rs=rootscript;
	while (rs != NULL) {
		stage=rs->stage;
		while (stage != NULL) {
			anim=stage->anim_stage;
			while (anim != NULL) {
				anim->texture=Texture_ForName(anim->name);
				if (!anim->texture)
					anim->texture=notexture;//Texture_ForName("pics/noimage.tga");
				anim=anim->next;
			}
			if (stage->texture) {
				stage->texture=Texture_ForName(stage->name);
				if (!stage->texture)
					stage->texture=notexture;//Texture_ForName("pics/noimage.tga");
			}
			stage=stage->next;
		}
		rs=rs->next;
	}
}
#endif
int RS_ScrollName(char *text)
{
	if (!_stricmp(text,"static"))		// static
		return 0;
	else if (!_stricmp(text,"sine"))	// sine wave
		return 1;
	else if (!_stricmp(text,"cosine"))	// cosine wave
		return 2;
	return 0;
}

/*
scriptname
{
	subdivide <size>
	vertexwarp <speed> <distance> <smoothness>
	safe
	{
		map <texturename>
		scroll <xtype> <xspeed> <ytype> <yspeed>
		blendfunc <source> <dest>
		alphashift <speed> <min> <max>
		anim <delay> <tex1> <tex2> <tex3> ... end
		envmap
		nolightmap
		alphamask
	}
}
*/
int RS_FuncName (char *text)
{
	if (!_stricmp (text, "static"))			// static
		return 0;
	else if (!_stricmp (text, "sine"))		// sine wave
		return 1;
	else if (!_stricmp (text, "cosine"))	// cosine wave
		return 2;

	return 0;
}

void RS_LoadScript(char *script)
{
	qboolean		inscript = false, instage=false;
	char			ignored = 0;
	char			intoolblock = 0;
	char			*token, *fbuf, *buf;
	rscript_t		*rs = NULL;
	rs_stage_t		*stage;
	anim_stage_t	*anim;
	unsigned char	tcmod = 0;
	int				len;

	len=LoadFileNoCrash (script, (void **)&fbuf);
	if (!fbuf || len < 16) {
		Sys_Printf("Could not load script %s\n",script);
		return;
	}
//	buf=(char *)malloc(len+1);
	buf = fbuf;
//	memcpy(buf,fbuf,len);
//	buf[len]=0;
//	free(fbuf);

//	Sys_Printf("Script File: %s\n", script);
//	Sys_Printf("%s\n", buf);
	for(token=strtok(buf,TOK_DELIMINATORS); token; token=strtok(NULL, TOK_DELIMINATORS)) {
		if (!_stricmp(token,"/*"))
			ignored++;
		else if (!_stricmp(token,"*/"))
			ignored--;
		else if(!_stricmp(token,"["))
			intoolblock++;
		if(intoolblock)
		{
//			if (!_stricmp(token,"editorimage"))
//			{
//				token=strtok(NULL, TOK_DELIMINATORS);
//
//				if(!strnicmp(token, "textures/", 9))
//					editorimage = strdup(token);
//				else
//					Sys_Printf("Rscript shader %s has bad editor image %s.\nTexture filename in shaders must start with \"textures/\"", shadername, token);
//			}
//			else if (!_stricmp(token,"editoralpha"))
//			{
//				token=strtok(NULL, TOK_DELIMINATORS);
//				trans = atof(token);
//				if(trans == 1.0f)
//					trans = 0.20f;
//			}
			/*else*/ if (!_stricmp(token,"editornocarve"))
			{
//				nocarve = 1;
			}
			else if(!stricmp(token, "surfaceflags"))	//		surfaceflags <flags (int)>
			{
				token=strtok(NULL, TOK_DELIMINATORS);
				rs->surfaceflags = atoi(token);
			}
			else if(!stricmp(token, "contentflags"))	//		contentflags <flags (int)>
			{
				token=strtok(NULL, TOK_DELIMINATORS);
				rs->contentflags = atoi(token);
			}
			else if(!stricmp(token, "lightamount"))		//		lightamount <value>
			{
				token=strtok(NULL, TOK_DELIMINATORS);
				rs->lightamount = atoi(token);
			}
			else if(!_stricmp(token,"]"))
				intoolblock--;

			continue;
		}

		if (!inscript && !ignored)
		{
			if (!_stricmp(token,"{"))
			{
				inscript=true;
			}
			else
			{
				rs = RS_FindScript(token);
				if (!rs)
				{
					rs=RS_NewScript(token);
				}
				else
				{
					RS_FreeScript(rs);
					rs=RS_NewScript(token);
				}
			}
		}
		else if (inscript && !ignored)
		{
			if (!_stricmp(token,"}"))
			{
				if (instage)
				{
					instage=false;
				}
				else
				{
					inscript=false;
				}
			}
			else if (!_stricmp(token,"{"))
			{
				if (!instage)
				{
					instage=true;
					stage=RS_NewStage(rs);
				}
			}
			else
			{
				if (instage && !ignored)
				{
					if (!_stricmp(token,"map"))
					{					// map
						token=strtok(NULL, TOK_DELIMINATORS);
						strcpy(stage->name,token);
					}
					else if (!_stricmp(token,"scroll"))
					{			// scroll Xtype Xspeed Ytype Yspeed
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->scroll.typeX=RS_ScrollName(token);
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->scroll.speedX=atof(token);
		
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->scroll.typeY=RS_ScrollName(token);
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->scroll.speedY=atof(token);
					}
					else if (!_stricmp(token,"blendfunc"))
					{		// blendfunc
						stage->blendfunc.blend=true;
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->blendfunc.source = RS_BlendID(token);
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->blendfunc.dest = RS_BlendID(token);
					}
					else if (!_stricmp(token,"alphashift"))
					{		// alphashift
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->alphashift.speed=(float)atof(token);
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->alphashift.min=(float)atof(token);
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->alphashift.max=(float)atof(token);
					}
					else if (!_stricmp(token, "anim"))
					{			// anim
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->anim_delay=(float)atof(token);
						anim=(anim_stage_t *)m_malloc(sizeof(anim_stage_t));
						stage->anim_stage=anim;
						stage->last_anim=anim;
						token=strtok(NULL, TOK_DELIMINATORS);
						while (_stricmp(token,"end"))
						{
							stage->anim_count++;
//							Sys_Printf("%s : %d\n", token, strlen(token));
							strcpy(anim->name,token);
							anim->texture=NULL;
							token=strtok(NULL, TOK_DELIMINATORS);
							if (!_stricmp(token,"end"))
							{
								anim->next = NULL;
								break;
							}
							anim->next = (anim_stage_t *)m_malloc(sizeof(anim_stage_t));
							anim = anim->next;
						}
					}
					else if (!_stricmp(token, "envmap"))
					{		// envmap
						stage->envmap=true;
					}
					else if (!_stricmp(token, "nolightmap"))
					{	// nolightmap - unused
						stage->lightmap=false;
					}
					else if (!_stricmp(token, "alphamask"))
					{		// alphamask
						stage->alphamask=true;
					}
					else if (!_stricmp(token,"rotate"))
					{			// rotate
						token=strtok(NULL, TOK_DELIMINATORS);
						stage->rot_speed=(float)atof(token);
					}
					else if (!_stricmp(token,"scale"))
					{			// scale
						token = strtok (NULL, TOK_DELIMINATORS);
						stage->scale.typeX = RS_FuncName(token);
						token = strtok (NULL, TOK_DELIMINATORS);
						stage->scale.scaleX = atof(token);
						
						token = strtok (NULL, TOK_DELIMINATORS);
						stage->scale.typeY = RS_FuncName(token);
						token = strtok (NULL, TOK_DELIMINATORS);
						stage->scale.scaleY = atof(token);
					}
				}
				else
				{
					if (!_stricmp(token,"safe"))
					{				// safe
						rs->dontflush=true;
					}
					else if (!_stricmp(token, "subdivide"))
					{ // Heffo - surface subdivision
						int divsize, p2divsize;;
						token=strtok(NULL, TOK_DELIMINATORS);
						divsize = atoi(token);
						if(divsize >= 128)		// cap max & min subdivide sizes
							divsize = 128;
						if(divsize <= 8)
							divsize = 8;
						for( p2divsize = 2; p2divsize <= divsize ; p2divsize <<= 1 ); // find the next smallest valid ^2 size, if not already one
						p2divsize >>= 1;
						rs->subdivide = (char)p2divsize;
					}
					else if (!_stricmp(token, "vertexwarp"))
					{ // Heffo - vertex warping
						token=strtok(NULL, TOK_DELIMINATORS);
						rs->warpspeed = atof(token);
						token=strtok(NULL, TOK_DELIMINATORS);
						rs->warpdist = atof(token);
						token=strtok(NULL, TOK_DELIMINATORS);
						rs->warpsmooth = atof(token);
						if(rs->warpsmooth <= 0.001f)
							rs->warpsmooth = 0.001f;
						if(rs->warpsmooth >= 1.0f)
							rs->warpsmooth = 1.0f;
					}
					else if (!_stricmp(token, "mirror"))
					{
						// Not implemented in BeefQuake R5 or QE5, but handle the token anyway
					}
				}
			}			
		}
	}
	free(buf);
}

void RS_ScanPathForScripts(char *dir)
{
	char			script[MAX_PATH];
	char			dirstring[1024], *c;
	int				handle;
	struct			_finddata_t fileinfo;

//	sprintf (dirstring,"%s/scripts/*.txt", dir);
	setdirstr(dirstring, dir, "/scripts/*.txt", NULL);
	handle = _findfirst (dirstring, &fileinfo);
	if (handle != -1) {
		do {
			if (fileinfo.name[0] == '.')
				continue;
			c=COM_SkipPath(fileinfo.name);
//			sprintf(script,"scripts/%s\n",c);
			setdirstr(script, dir, "/scripts/", c, NULL);
//			Sys_Printf("Loading %s/scripts/%s\n",dir,c);
			Sys_Printf("Loading %s\n",script);
			RS_LoadScript(script);
		} while (_findnext( handle, &fileinfo ) != -1);
		_findclose (handle);
	}
}


//extern float	r_turbsin[];
//#define TURBSCALE (256.0 / (2 * M_PI))

/*void RotateST(float is, float it, float degrees, float *os, float *ot)
{
	float s,t;

	is-=0.5f;
	it-=0.5f;

	s = cos(degrees)*is - sin(degrees)*it;
	t = sin(degrees)*is + cos(degrees)*it;
	
	s+=0.5f;
	t+=0.5f;

	*os=s;
	*ot=t;
}
*/
_inline void RS_RotateST (float *os, float *ot, float degrees, face_t *fa)
{
	float cost = cos(degrees), sint = sin(degrees);
	float is = *os, it = *ot;

	*os = cost * (is - fa->c_s) + sint * (fa->c_t - it) + fa->c_s;
	*ot = cost * (it - fa->c_t) + sint * (is - fa->c_s) + fa->c_t;
}

_inline void RS_SetEnvmap (vec3_t v, float *os, float *ot)
{
	vec3_t vert;
	
	vert[0] = v[0]*globals.r_world_matrix[0]+v[1]*globals.r_world_matrix[4]+v[2]*globals.r_world_matrix[8] +globals.r_world_matrix[12];
	vert[1] = v[0]*globals.r_world_matrix[1]+v[1]*globals.r_world_matrix[5]+v[2]*globals.r_world_matrix[9] +globals.r_world_matrix[13];
	vert[2] = v[0]*globals.r_world_matrix[2]+v[1]*globals.r_world_matrix[6]+v[2]*globals.r_world_matrix[10]+globals.r_world_matrix[14];
	
	VectorNormalize (vert);
	
	*os = vert[0];
	*ot = vert[1];
}

void RS_SetTexcoords (rs_stage_t *stage, float *os, float *ot, face_t *fa)
{
	float	txm = 0, tym = 0;

	// scale
	if (stage->scale.scaleX) {
		switch (stage->scale.typeX) {
		case 0:	// static
			*os *= stage->scale.scaleX;
			break;
		case 1:	// sine
			*os *= stage->scale.scaleX*sin(rs_realtime*0.05);
			break;
		case 2:	// cosine
			*os *= stage->scale.scaleX*cos(rs_realtime*0.05);
			break;
		}
	}

	if (stage->scale.scaleY) {
		switch (stage->scale.typeY) {
		case 0:	// static
			*ot *= stage->scale.scaleY;
			break;
		case 1:	// sine
			*ot *= stage->scale.scaleY*sin(rs_realtime*0.05);
			break;
		case 2:	// cosine
			*ot *= stage->scale.scaleY*cos(rs_realtime*0.05);
			break;
		}
	}

	// rotate
	if (stage->rot_speed)
		RS_RotateST (os, ot, -stage->rot_speed * rs_realtime * 0.0087266388888888888888888888888889, fa);
}

void RS_DrawPoly (face_t *surf, qboolean lightmap)
{
	winding_t	*p = surf->face_winding;
	vec_t		*v;
	int			i, nv = surf->face_winding->numpoints;
	vec3_t		wv;
	rscript_t	*rs=(rscript_t *)surf->d_texture->script;
	rs_stage_t	*stage = rs->stage;
	float		os, ot, alpha;
	float		scale, time = rs_realtime * rs->warpspeed, txm, tym;

	do
	{
		if(stage->lm_only)
			continue;

		if (stage->anim_count)
			glBindTexture( GL_TEXTURE_2D, RS_Animate(stage));
		else
			glBindTexture( GL_TEXTURE_2D, stage->texture->texture_number);

		// sane defaults
		alpha = 1.0f;

		if (stage->blendfunc.blend) {
			glBlendFunc(stage->blendfunc.source,stage->blendfunc.dest);
			glEnable(GL_BLEND);
		} else {
			glDisable(GL_BLEND);
		}

		if (stage->alphashift.min || stage->alphashift.speed) {
			if (!stage->alphashift.speed && stage->alphashift.min > 0) {
				alpha = stage->alphashift.min;
			} else if (stage->alphashift.speed) {
				alpha = sin (rs_realtime * stage->alphashift.speed);
				alpha = (alpha + 1)*0.5f;
				if (alpha > stage->alphashift.max) 
					alpha = stage->alphashift.max;
				if (alpha < stage->alphashift.min) 
					alpha = stage->alphashift.min;
			}
		} else
			alpha = 1.0f;

		if (stage->scroll.speedX) {
			switch (stage->scroll.typeX) {
			case 0:	// static
				txm = rs_realtime*stage->scroll.speedX;
				break;
			case 1:	// sine
				txm = sin (rs_realtime*stage->scroll.speedX);
				break;
			case 2:	// cosine
				txm = cos (rs_realtime*stage->scroll.speedX);
				break;
			}
		} else
			txm=0;
	
		if (stage->scroll.speedY) {
			switch (stage->scroll.typeY) {
			case 0:	// static
				tym = rs_realtime*stage->scroll.speedY;
				break;
			case 1:	// sine
				tym = sin (rs_realtime*stage->scroll.speedY);
				break;
			case 2:	// cosine
				tym = cos (rs_realtime*stage->scroll.speedY);
				break;
			}
		} else
			tym=0;

		glColor4f (1, 1, 1, alpha);

//		if (stage->envmap)
//			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		if (stage->alphamask) {
			glAlphaFunc(GL_GREATER,0.1f);
			glEnable(GL_ALPHA_TEST);
		}

#if 0
		if (rs->subdivide)
		{
			glpoly_t *bp;
			int i;

			for (bp = surf->polys; bp; bp = bp->next)
			{
				p = bp;

				if (stage->envmap) {
//					qglTexGenf(GL_S,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
//					qglTexGenf(GL_T,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
//					GLSTATE_ENABLE_TEXGEN
				}
				qglBegin(GL_TRIANGLE_FAN);
				for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
				{
					if (stage->envmap) {
						RS_SetEnvmap (v, &os, &ot);
					} else {
						os = v[3];
						ot = v[4];

						RS_SetTexcoords (stage, &os, &ot, surf);
					}

					if (lightmap) {
						qglMTexCoord2fSGIS(GL_TEXTURE0, os+txm, ot+tym);
						qglMTexCoord2fSGIS(GL_TEXTURE1, v[5], v[6]);
					} else {
						qglTexCoord2f (os, ot);
					}

					if (!rs->warpsmooth)
						qglVertex3fv (v);
					else {
						scale = rs->warpdist * sin(v[0]*rs->warpsmooth+time)*sin(v[1]*rs->warpsmooth+time)*sin(v[2]*rs->warpsmooth+time);
						VectorMA (v, scale, surf->plane->normal, wv);
						qglVertex3fv (wv);
					}
				}

				qglEnd();
			}
		} else {
#endif
//				if (stage->envmap) {
//					qglTexGenf(GL_S,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
//					qglTexGenf(GL_T,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
//					GLSTATE_ENABLE_TEXGEN
//				}
				glBegin (GL_TRIANGLE_FAN);

				for (i = 0; i < nv; i++)
				{
					v = p->points[i];

					if (stage->envmap) {
						RS_SetEnvmap (v, &os, &ot);
					} else {
						os = v[3];
						ot = v[4];

						RS_SetTexcoords (stage, &os, &ot, surf);
					}

					glTexCoord2f (os+txm, ot+tym);

					if (!rs->warpsmooth)
						glVertex3dv (v);
					else {
						scale = rs->warpdist * sin(v[0]*rs->warpsmooth+time)*sin(v[1]*rs->warpsmooth+time)*sin(v[2]*rs->warpsmooth+time);
						VectorMA (v, scale, surf->plane.normal, wv);
						glVertex3dv (wv);
					}
				}

				glEnd ();

//		if (stage->envmap)
//			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		glColor4f (1, 1, 1, 1);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_BLEND);
		glDisable(GL_ALPHA_TEST);
	} while (stage = stage->next);
}

int Rscript_GetFileList(FILELIST **filelist, char *dir)
{
	rscript_t	*s;
	int			len, found = 0;

	len = strlen(dir);

	for(s = rootscript; s; s = s->next)
	{
		if(!strnicmp(s->name, dir, len))
		{
			found++;
//			Sys_Printf("Adding %s to file list\n", s->name);
			AddToFileListAlphabetized(filelist, s->name, FROMDISK, 0);
		}
	}

	return found;
}

void Rscript_GetTextureDirs(DIRLIST **dirlist)
{
	rscript_t	*rs;
	char		*s, buf[512];
	int			founddir;

	for(rs = rootscript; rs; rs = rs->next)
	{
		if(!rs->name[0])
			continue;

		strcpy(buf, rs->name);
		founddir = 0;

//		if(!strnicmp(buf, "textures/", 9))
//			strcpy(buf, buf+9);
//		else
//			continue;
//		Sys_Printf("%s\n", buf);
		for(s = buf; *s; s++)
		{
			if(*s == '/')
			{
				*s = 0;
				founddir = 1;
				break;
			}
		}
		if(founddir)
		{
			AddToDirListAlphabetized(dirlist, buf, FROMSHADER);
		}
		else
			Sys_Printf("Warning: Didn't find a subdirectory for script %s\n", rs->name);
	}
}

void RScript_CopyToClipboard(int what)
{
	char		buf[512];
	HGLOBAL		hglbCopy;
	LPTSTR		lptstrCopy;
	int			surfaceflags = globals.d_texturewin.texdef.flags;
	int			contentflags = globals.d_texturewin.texdef.contents;
	int			lightamount = globals.d_texturewin.texdef.value;

	if (!OpenClipboard(NULL))
	{
		MessageBox(globals.Main.hwnd, "Unable to open the clipboard!", "Error", MB_OK | MB_ICONERROR );
		return;
	}

	switch(what)
	{
		case ID_EDIT_RSCRIPT_MAKESURFACEFLAGSFIELDFROMCURRENTTEXTURE:
			sprintf(buf, "surfaceflags %d", surfaceflags);
			break;
		case ID_EDIT_RSCRIPT_MAKECONTENTFLAGSFIELDFROMCURRENTTEXTURE:
			sprintf(buf, "contentflags %d", contentflags);
			break;
		case ID_EDIT_RSCRIPT_MAKELIGHTAMOUNTFIELDFROMCURRENTTEXTURE:
			sprintf(buf, "lightamount %d", lightamount);
			break;
		case ID_EDIT_RSCRIPT_MAKEALL:
			sprintf(buf, "\t\tsurfaceflags %d\n\t\tcontentflags %d\n\t\tlightamount %d\n", surfaceflags, contentflags, lightamount);
			break;
	}

	EmptyClipboard();
	hglbCopy = GlobalAlloc(GMEM_MOVEABLE, (strlen(buf) + 1) * sizeof(TCHAR));
	if (hglbCopy == NULL)
	{
		CloseClipboard();
		return;
	}
	lptstrCopy = GlobalLock(hglbCopy);
	strcpy(lptstrCopy, buf);
	GlobalUnlock(hglbCopy);
	SetClipboardData(CF_TEXT, hglbCopy);
}

void InitRscript(void)
{
	RS_ScanPathForScripts(ValueForKey(globals.d_project_entity, "basepath"));
}

void RS_DrawQuad(rscript_t *rs, float x, float y, float h, float v)
{
	rs_stage_t	*stage;
	float txm=0,tym=0,alpha;

	float /*scale,*/ time;
	time = (rs_realtime * rs->warpspeed);

	stage=rs->stage;
	while (stage)
	{
		if (stage->anim_count)
			glBindTexture( GL_TEXTURE_2D, RS_Animate(stage));
		else
			glBindTexture( GL_TEXTURE_2D, stage->texture->texture_number);
		
		if (stage->scroll.speedX)
		{
			switch(stage->scroll.typeX)
			{
				case 0:	// static
					txm=rs_realtime*stage->scroll.speedX;
					break;
				case 1:	// sine
					txm=sin(rs_realtime)*stage->scroll.speedX;
					break;
				case 2:	// cosine
					txm=cos(rs_realtime)*stage->scroll.speedX;
					break;
			}
		} else
			txm=0;
		
		if (stage->scroll.speedY)
		{
			switch(stage->scroll.typeY)
			{
				case 0:	// static
					tym=rs_realtime*stage->scroll.speedY;
					break;
				case 1:	// sine
					tym=sin(rs_realtime)*stage->scroll.speedY;
					break;
				case 2:	// cosine
					tym=cos(rs_realtime)*stage->scroll.speedY;
					break;
			}
		} else
			tym=0;
		
		if (stage->blendfunc.blend)
		{
			glBlendFunc(stage->blendfunc.source,stage->blendfunc.dest);
			glEnable(GL_BLEND);
		}
		else
		{
			glDisable(GL_BLEND);
		}

		if (stage->alphashift.min || stage->alphashift.speed)
		{
			if (!stage->alphashift.speed && stage->alphashift.min > 0)
			{
				alpha=stage->alphashift.min;
			}
			else if (stage->alphashift.speed)
			{
				alpha=sin(rs_realtime * stage->alphashift.speed);
				alpha=(alpha+1)*0.5f;
				if (alpha > stage->alphashift.max)
					alpha=stage->alphashift.max;
				if (alpha < stage->alphashift.min)
					alpha=stage->alphashift.min;
			}
		}
		else
			alpha=1.0f;
		
		glColor4f(1,1,1,alpha);
		
//		if (stage->envmap)
//		{
//			glTexGenf(GL_S,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
//			glTexGenf(GL_T,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
//			glEnable(GL_TEXTURE_GEN_S);
//			glEnable(GL_TEXTURE_GEN_T);
//		}
		if (stage->alphamask)
		{
			glEnable(GL_ALPHA_TEST);
		}

		glBegin (GL_QUADS);
		glTexCoord2f(0.0+txm, 0.0+tym);
		glVertex2f(x, y);
		glTexCoord2f(1.0+txm, 0.0+tym);
		glVertex2f(h, y);
		glTexCoord2f(1.0+txm, 1.0+tym);
		glVertex2f(h, v);
		glTexCoord2f(0.0+txm, 1.0+tym);
		glVertex2f(x, v);
		glEnd ();
		
		glColor4f(1,1,1,1);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_BLEND);
		glDisable(GL_ALPHA_TEST);
		stage=stage->next;
	}
}
