
#include "qe3.h"

globals_t  globals;

void Select_ShiftTexture(int x, int y);
void Select_ScaleTexture(int x, int y);
void Select_RotateTexture(int amt);
void SaveAsDialog (void);

void QE_CheckOpenGLForErrors(char *function_name, int ignore)
{
    int	    i;

    while ( ( i = glGetError() ) != GL_NO_ERROR )
    {
		if(i != i)//gnore)
		{
			char buffer[100];

			sprintf( buffer, "OpenGL Error: %s\nDetected in function %s", gluErrorString( i ), function_name);

			if(MessageBox(globals.d_hwndMain, buffer, "QuakeEd Error", MB_OKCANCEL | MB_ICONERROR /* MB_OK */ ) == IDOK)
			{
				if(modified)
					if(MessageBox(globals.d_hwndMain, "Map has been modified. Do you want to save your changes?", "Save Changes?", MB_YESNO|MB_ICONQUESTION ) == IDYES)
						SaveAsDialog();

				exit( 1 );
			}
		}
    }
}


char *ExpandReletivePath (char *p)
{
	static char	temp[1024];
	char	*base;

	if (!p || !p[0])
		return NULL;
	if (p[0] == '/' || p[0] == '\\')
		return p;

	base = ValueForKey(globals.d_project_entity, "basepath");
	sprintf (temp, "%s/%s", base, p);
	return temp;
}



void *qmalloc (int size)
{
	void *b;
        if((b = malloc(size)) == NULL)
                Error("Could not allocate memory");
	memset (b, 0, size);
	return b;
}

char *copystring (char *s)
{
	char	*b;
        b = qmalloc(strlen(s)+1);
	strcpy (b,s);
	return b;
}

/*
===============
QE_CheckAutoSave

If five minutes have passed since making a change
and the map hasn't been saved, save it out.
===============
*/
void QE_CheckAutoSave( void )
{
	static clock_t s_start;
	clock_t        now;

	now = clock();

	if ( modified != 1 || !s_start)
	{
		s_start = now;
		return;
	}

	if ( now - s_start > ( CLOCKS_PER_SEC * 60 * QE_AUTOSAVE_INTERVAL ) )
	{
		Sys_Printf ("Autosaving...\n");
		Sys_Status(0, "Autosaving...");

		Map_SaveFile (ValueForKey(globals.d_project_entity, "autosave"), false);

		Sys_Status(0, "Autosaving...Saved.");
		modified = 2;
		s_start = now;
	}
}



/*
===========
QE_LoadProject
===========
*/
qboolean QE_LoadProject (char *projectfile)
{
	char	*data;
	char	buf[256];

	Sys_Printf ("QE_LoadProject (%s)\n", projectfile);
	GetCurrentDirectory(255, buf);
	Sys_Printf("Current Directory: %s\n", buf);

	if ( LoadFileNoCrash (projectfile, (void *)&data) == -1)
		return false;
	StartTokenParsing (data);
	globals.d_project_entity = Entity_Parse (true);
	if (!globals.d_project_entity)
		Error ("Couldn't parse %s", projectfile);
	free (data);

	Eclass_InitForSourceDirectory (ValueForKey (globals.d_project_entity, "entitypath"));

	FillClassList ();		// list in entity window
	FillBSPMenu ();

	return true;
}

void FollowFloor(void);
/*
===========
QE_KeyDown
===========
*/
#define	SPEED_MOVE	32
#define	SPEED_TURN	22.5
void SetupXYOrientation(int Orientation);

qboolean QE_KeyDown (int key)
{
//	Sys_Printf("%d\n", key);
	switch (key)
	{
	case 'K':
		PostMessage( globals.d_hwndMain, WM_COMMAND, ID_MISC_SELECTENTITYCOLOR, 0 );
		break;

	case 'L':	
		PostMessage( globals.d_hwndMain, WM_COMMAND, ID_SELECTION_LOCKTEXTURES, 0 );
		break;

	case 'F':
		if(GetKeyState(VK_CONTROL) < 0)
			PostMessage( globals.d_hwndMain, WM_COMMAND, ID_SELECTION_FITTEXTURES, 0 );
		else
			PostMessage( globals.d_hwndMain, WM_COMMAND, ID_MISC_FOLLOWFLOOR, 0 );
		break;

	case VK_UP:
		if(GetKeyState(VK_SHIFT) < 0)
		{
			if(GetKeyState(VK_CONTROL) < 0)				
				Select_RotateTexture(-1);
			else
				Select_ShiftTexture(0, globals.d_gridsize);
		}
		else if(GetKeyState(VK_CONTROL) < 0)
			Select_ScaleTexture(0, globals.d_gridsize);
		else
		{
			VectorMA (camera.origin, SPEED_MOVE, camera.forward, camera.origin);
			if(globals.followfloor)
				FollowFloor();
			if(globals.LockXYZ2Cam)
			{
				globals.XY.origin[0] = globals.XZ.origin[0] = camera.origin[0];
				globals.XY.origin[1] = globals.YZ.origin[1] = camera.origin[1];
				globals.YZ.origin[2] = globals.XZ.origin[2] = z.origin[2] = camera.origin[2];
				Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			}
			else
				Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
		}
		break;
	case VK_DOWN:
		if(GetKeyState(VK_SHIFT) < 0)
		{
			if(GetKeyState(VK_CONTROL) < 0)				
				Select_RotateTexture(1);
			else
				Select_ShiftTexture(0, 0 - globals.d_gridsize);
		}
		else if(GetKeyState(VK_CONTROL) < 0)
			Select_ScaleTexture(0, 0 - globals.d_gridsize);
		else
		{
			VectorMA (camera.origin, -SPEED_MOVE, camera.forward, camera.origin);
			if(globals.followfloor)
				FollowFloor();
			if(globals.LockXYZ2Cam)
			{
				globals.XY.origin[0] = globals.XZ.origin[0] = camera.origin[0];
				globals.XY.origin[1] = globals.YZ.origin[1] = camera.origin[1];
				globals.YZ.origin[2] = globals.XZ.origin[2] = z.origin[2] = camera.origin[2];
				Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			}
			else
				Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
		}
		break;
	case VK_LEFT:
		if(GetKeyState(VK_SHIFT) < 0)
		{
			if(GetKeyState(VK_CONTROL) < 0)				
				Select_RotateTexture(-15);
			else
				Select_ShiftTexture(0 - globals.d_gridsize, 0);
		}
		else if(GetKeyState(VK_CONTROL) < 0)
			Select_ScaleTexture(0 - globals.d_gridsize, 0);
		else
		{
			camera.angles[1] += SPEED_TURN;
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
		}
		break;
	case VK_RIGHT:
		if(GetKeyState(VK_SHIFT) < 0)
		{
			if(GetKeyState(VK_CONTROL) < 0)				
				Select_RotateTexture(15);
			else
				Select_ShiftTexture(globals.d_gridsize, 0);
		}
		else if(GetKeyState(VK_CONTROL) < 0)
			Select_ScaleTexture(globals.d_gridsize, 0);
		else
		{
			camera.angles[1] -= SPEED_TURN;
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
		}
		break;
	case VK_TAB:
		if(GetKeyState(VK_CONTROL) < 0)
		{
			if(globals.Multiple2DViews)
				DestroyExtra2DViews();
			else
				CreateExtra2DViews();

			globals.Multiple2DViews ^= 1;
		}
		else if(!globals.Multiple2DViews)
		{
			if(GetKeyState(VK_SHIFT) < 0)
			{
				globals.XY_Orientation[3] -= 1;
				if(globals.XY_Orientation[3] < 0)
					globals.XY_Orientation[3] = 2;
			}
			else
			{
				globals.XY_Orientation[3] += 1;
				if(globals.XY_Orientation[3] > 2)
					globals.XY_Orientation[3] = 0;
			}
				SetupXYOrientation(globals.XY_Orientation[3]);
				Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_Z | W_Z);
		}
		break;
	case 'D':
		camera.origin[2] += SPEED_MOVE;
		Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
		break;
	case 'C':
		camera.origin[2] -= SPEED_MOVE;
		Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
		break;
	case 'A':
		camera.angles[0] += SPEED_TURN;
		if (camera.angles[0] > 85)
			camera.angles[0] = 85;
		Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
		break;
	case 'Z':
		camera.angles[0] -= SPEED_TURN;
		if (camera.angles[0] < -85)
			camera.angles[0] = -85;
		Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
		break;
	case VK_COMMA:
		VectorMA (camera.origin, -SPEED_MOVE, camera.right, camera.origin);
		if(globals.followfloor)
			FollowFloor();
		if(globals.LockXYZ2Cam)
		{
			globals.XY.origin[0] = globals.XZ.origin[0] = camera.origin[0];
			globals.XY.origin[1] = globals.YZ.origin[1] = camera.origin[1];
			globals.YZ.origin[2] = globals.XZ.origin[2] = z.origin[2] = camera.origin[2];
		}

		Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
		break;
	case VK_PERIOD:
		VectorMA (camera.origin, SPEED_MOVE, camera.right, camera.origin);
		if(globals.followfloor)
			FollowFloor();
		if(globals.LockXYZ2Cam)
		{
			globals.XY.origin[0] = globals.XZ.origin[0] = camera.origin[0];
			globals.XY.origin[1] = globals.YZ.origin[1] = camera.origin[1];
			globals.YZ.origin[2] = globals.XZ.origin[2] = z.origin[2] = camera.origin[2];
		}

		Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
		break;
		
	case 'X':
		{
			globals.DepthClipping ^= 1;
			
			if(!globals.DepthClipping)
				Sys_Status(2, "Depth Clip OFF");
			else
				Sys_Status(2, "Clip Distance: %d", globals.DepthClipDistance);
			
			Sys_UpdateWindows (W_CAMERA);
		}
		break;
	case 'U':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_MISC_ALWAYSUPDATECAMERA, 0);
		break;
	case '0':
		globals.d_showgrid = !globals.d_showgrid;
		Sys_UpdateWindows(W_XY|W_XZ|W_YZ|W_Z);
		break;
	case '1':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_GRID_1, 0);
		break;	
	case '2':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_GRID_2, 0);
		break;
	case '3':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_GRID_4, 0);
		break;
	case '4':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_GRID_8, 0);
		break;
	case '5':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_GRID_16, 0);
		break;
	case '6':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_GRID_32, 0);
		break;
	case '7':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_GRID_64, 0);
		break;
		
	case 'E':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_SELECTION_DRAGEDGES, 0);
		break;
	case 'V':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_SELECTION_DRAGVERTECIES, 0);
		break;
		
	case 'N':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_VIEW_ENTITY, 0);
		break;
	case 'O':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_VIEW_CONSOLE, 0);
		break;
	case 'T':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_VIEW_TEXTURE, 0);
		break;
	case 'S':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_TEXTURES_INSPECTOR, 0);
		break;
		
	case ' ':
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_SELECTION_CLONE, 0);
		break;
		
	case VK_BACK:
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_SELECTION_DELETE, 0);
		break;
	case VK_ESCAPE:
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_SELECTION_DESELECT, 0);
		break;
	case VK_END:
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_VIEW_CENTER, 0);
		break;
		
	case VK_DELETE:
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_VIEW_ZOOMIN, 0);
		break;
	case VK_INSERT:
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_VIEW_ZOOMOUT, 0);
		break;
		
	case VK_NEXT:
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_VIEW_DOWNFLOOR, 0);
		break;
	case VK_PRIOR:
		PostMessage (globals.d_hwndMain, WM_COMMAND, ID_VIEW_UPFLOOR, 0);
		break;

	case VK_HOME:
		camera.origin[0] = camera.origin[1] = camera.origin[2] = 0;
		globals.XY.origin[0] = globals.XY.origin[1] = globals.XY.origin[2] = 0;
		globals.XZ.origin[0] = globals.XZ.origin[1] = globals.XZ.origin[2] = 0;
		globals.YZ.origin[0] = globals.YZ.origin[1] = globals.YZ.origin[2] = 0;
		Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_Z | W_CAMERA);

	case 'Q':
		if(GetKeyState(VK_CONTROL) < 0)
		{
			if(GetKeyState(VK_SHIFT) < 0)
			{
				globals.ClipperMode--;
				if(globals.ClipperMode < 1)
				{
					globals.ClipperMode = 3;
				}
			}
			else
			{
				globals.ClipperMode++;
				if(globals.ClipperMode > 3)
				{
					globals.ClipperMode = 1;
				}
			}
		}
		else
		{
			PostMessage (globals.d_hwndMain, WM_COMMAND, ID_SELECTION_CLIPPER_CLIP, 0);
		}

		SetClipperMode(globals.ClipperMode);
		Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_Z | W_CAMERA);
		break;

	default:
		return false;
		
	}

	return true;
}

/*
===============
ConnectEntities

Sets target / targetname on the two entities selected
from the first selected to the secon
===============
*/
void ConnectEntities (void)
{
	entity_t	*e1, *e2, *e;
	char		*target, *tn;
	int			maxtarg, targetnum;
	char		newtarg[32];

	if (globals.d_select_count != 2)
	{
		Sys_Status(0, "Must have two brushes selected.");
		Sys_Beep ();
		return;
	}

	e1 = globals.d_select_order[0]->owner;
	e2 = globals.d_select_order[1]->owner;

	if (e1 == world_entity || e2 == world_entity)
	{
		Sys_Status(0, "Can't connect to the world.");
		Sys_Beep ();
		return;
	}

	if (e1 == e2)
	{
		Sys_Status(0, "Brushes are from same entity.");
		Sys_Beep ();
		return;
	}

	target = ValueForKey (e1, "target");
	if (target && target[0])
		strcpy (newtarg, target);
	else
	{
		target = ValueForKey (e2, "targetname");
		if (target && target[0])
			strcpy (newtarg, target);
		else
		{
			// make a unique target value
			maxtarg = 0;
			for (e=entities.next ; e != &entities ; e=e->next)
			{
				tn = ValueForKey (e, "targetname");
				if (tn && tn[0])
				{
					targetnum = atoi(tn+1);
					if (targetnum > maxtarg)
						maxtarg = targetnum;
				}
			}
			sprintf (newtarg, "t%i", maxtarg+1);
		}
	}

	SetKeyValue (e1, "target", newtarg);
	SetKeyValue (e2, "targetname", newtarg);
	Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_CAMERA);

	Select_Deselect(true);
	Select_Brush (globals.d_select_order[1], true);
}

qboolean QE_SingleBrush (void)
{
	if ( (selected_brushes.next == &selected_brushes)
		|| (selected_brushes.next->next != &selected_brushes) )
	{
		Sys_Printf ("Error: you must have a single brush selected\n");
		return false;
	}
	if (selected_brushes.next->owner->eclass->fixedsize)
	{
		Sys_Printf ("Error: you cannot manipulate fixed size entities\n");
		return false;
	}

	return true;
}

void QE_Init (void)
{
	/*
	** initialize variables
	*/
	globals.d_gridsize = 8;
	globals.d_showgrid = true;

	/*
	** other stuff
	*/
	Cam_Init ();
	XY_Init ();
	Z_Init ();
	Texture_Init ();
}

void QE_ConvertDOSToUnixName( char *dst, const char *src )
{
	while ( *src )
	{
		if ( *src == '\\' )
			*dst = '/';
		else
			*dst = *src;
		dst++; src++;
	}
	*dst = 0;
}

int g_numbrushes, g_numentities;

void QE_CountBrushesAndUpdateStatusBar( void )
{
	static int      s_lastbrushcount, s_lastentitycount;
	static qboolean s_didonce;
	
	entity_t   *e;
	brush_t	   *b, *next;

	g_numbrushes = 0;
	g_numentities = 0;
	
	if ( active_brushes.next != NULL )
	{
		for ( b = active_brushes.next ; b != NULL && b != &active_brushes ; b=next)
		{
			next = b->next;
			if (b->brush_faces )
			{
				if ( !b->owner->eclass->fixedsize)
					g_numbrushes++;
				else
					g_numentities++;
			}
		}
	}

	if ( entities.next != NULL )
	{
		for ( e = entities.next ; e != &entities && g_numentities != MAX_MAP_ENTITIES ; e = e->next)
		{
			g_numentities++;
		}
	}

	if ( ( ( g_numbrushes != s_lastbrushcount ) || ( g_numentities != s_lastentitycount ) ) || ( !s_didonce ) )
	{
		Sys_UpdateStatusBar();

		s_lastbrushcount = g_numbrushes;
		s_lastentitycount = g_numentities;
		s_didonce = true;
	}
}

void SetHeightsForGame(void)
{
	switch(globals.gametype)
	{
		case HALFLIFE:
			globals.EyeHeight = 52;
			globals.TotalHeightMinusEyeHeight = 8;
			break;
		default:
			globals.EyeHeight = 48;
			globals.TotalHeightMinusEyeHeight = 8;
			break;
	}
}
