#ifndef __QE3_H__
#define __QE3_H__

// disable data conversion warnings for gl
#pragma warning(disable : 4244)     // MIPS
#pragma warning(disable : 4136)     // X86
#pragma warning(disable : 4051)     // ALPHA
 
#include <windows.h>

#include <gl/gl.h>
#include <gl/glu.h>
//#include <gl/glaux.h>
#include "glingr.h"
#include <math.h>
#include <stdlib.h>

#include "cmdlib.h"
#include "mathlib.h"
#include "parse.h"
#include "lbmlib.h"

#include <commctrl.h>
#include "afxres.h"
#include "resource.h"

#include "qedefs.h"

typedef struct
{
    vec3_t	normal;
    double	dist;
    int		type;
} plane_t;

#include "qfiles.h"

#include "textures.h"
#include "visgroup.h"
#include "brush.h"
#include "entity.h"
#include "map.h"
#include "select.h"

#include "camera.h"
#include "z.h"
#include "mru.h"
#include "mrg/gl_script.h"
#include "model.h"

typedef struct
{
	int		p1, p2;
	face_t	*f1, *f2;
} pedge_t;

typedef struct
{
	int			iSize;
	int			iTexMenu;		// nearest, linear, etc
	float		fGamma;			// gamma for textures
	char		szProject[256];	// last project loaded
	vec3_t		colors[COLOR_LAST];
	vec_t		alpha[COLOR_LAST];
	qboolean	show_names,
				show_coordinates,
				no_console_texture,
				no_console_scanfile;
	int			exclude;
} SavedInfo_t;

//
// system functions
//
void    Sys_UpdateStatusBar	(void);
void    Sys_UpdateWindows	(int bits);
void    Sys_Beep			(void);
void    Sys_ClearPrintf		(void);
void    Sys_Printf			(char *text, ...);
void	glPrintf			(float x, float y, float z, char *text, ...);
double	Sys_DoubleTime		(void);
void    Sys_GetCursorPos	(int *x, int *y);
void    Sys_SetCursorPos	(int x, int y);
void    Sys_SetTitle		(char *text);
void    Sys_BeginWait		(void);
void    Sys_EndWait			(void);
void    Sys_Status			(int part, char *format, ...);

typedef struct
{
	HGLRC					hglrc;
	HDC						hdc;
	HWND					hwnd;
    PIXELFORMATDESCRIPTOR	pfd;
	int						fontlist;
	HFONT					font;
	vec3_t					origin;
	int						width;
	int						height;
	float					scale;
	int						Orientation;
}WInfo_t;
#include "xy.h"

/*
** most of the QE globals are stored in this structure
*/
typedef struct
{
	qboolean		d_showgrid;
	int				d_gridsize;

	int				d_num_entities;

	entity_t		*d_project_entity;

	vec3_t			d_new_brush_bottom,
					d_new_brush_top;

	HINSTANCE		d_hInstance;

	HWND			d_hwndMain;
	WInfo_t			Main;
	WInfo_t			Camera;
	WInfo_t			Edit;
	WInfo_t			Entity;
	WInfo_t			Texture;
	WInfo_t			TextureTree;
	WInfo_t			XY;
	WInfo_t			XZ;
	WInfo_t			YZ;
	WInfo_t			Z;
	HWND			d_hwndStatus;

	vec3_t			d_points[MAX_POINTS];
	int				d_numpoints;
	pedge_t			d_edges[MAX_EDGES];
	int				d_numedges;

	int				d_num_move_points;
	vec_t			*d_move_points[1024];

	qtexture_t		*d_qtextures;
	image_t			*images;

	texturewin_t	d_texturewin;

	int				d_pointfile_display_list;

	xy_t			d_xy;

	LPMRUMENU		d_lpMruMenu;

	SavedInfo_t		d_savedinfo;

	int				d_workcount;

	// connect entities uses the last two brushes selected
	int				d_select_count;
	brush_t			*d_select_order[2];
	vec3_t			d_select_translate;    // for dragging w/o making new display lists
	select_t		d_select_mode;

	int				d_parsed_brushes;

	qboolean		show_blocks;
	qboolean		Camera_Initialized;
	qboolean		DepthClipping;
	int				DepthClipDistance;
	BOOL			transactive;
	BOOL			trans33;
	BOOL			trans66;
	BOOL			nodraw;
	int				gametype;
	BOOL			followfloor;
	int				StepHeight;
	BOOL			UpdateCamera;
	BOOL			LockXYZ2Cam;
	BOOL			Lock2DOrigins;
	BOOL			Lock2DZoom;
	BOOL			LockCamTo2DOrigins;
	int				AllowFaceAndBrushSelection;
	int				ShowEntityConnections;
	int				ShowSkybox;
	float			skyrotate;
	vec3_t			skyaxis;
	int				ShowTextureNames;
	int				Clip2XYZ;
	int				ShowHint;
	int				WriteMapFloats;
	int				StippleSelection;
	HWND			d_hwndToolbar;
	HWND			d_hwndToolbarTextureCombo;
	HWND			d_hwndToolbarEntityCombo;
	char			WorkingDir[512];
	int				EyeHeight;
	int				TotalHeightMinusEyeHeight;
	int				UseExternalTextures;
	int				XY_Orientation[4];
	int				AlwaysUseDefaultWindowPlacement;
	int				DrawVisGroupColors;
	int				ShowFullSurfaceEditor;
	int				OffsetCopiedBrushes;
	int				Multiple2DViews;
	int				HideZWithMultiple2DViews;
	int				HideTextureView;
	int				AnimateCameraRScripts;
	int				AnimateTextureRScripts;
	int				AnimateSky;
	int				AnimationSpeed;
	int				HaveTextureShaderExtension;
	char			basepath[512];
	char			modpath[512];
	float			r_world_matrix[16];
	int				MaxGridExtent;
	HMENU			Menu;
	vec3_t			Clipper[3];
	int				ClipperNumPoints;
	int				ClipperMode;	// 1 = front; 2 = back; 1|2 = split
	int				OldCameraMovement;
} globals_t;

void *qmalloc (int size);
char *copystring (char *s);
char *ExpandReletivePath (char *p);

void Pointfile_Delete	(void);
void Pointfile_Check	(void);
void Pointfile_Next		(void);
void Pointfile_Prev		(void);
void Pointfile_Clear	(void);
void Pointfile_Draw		(void);
void Pointfile_Load		(void);
void SetHeightsForGame	(void);
//
// drag.c
//
void Drag_Begin (int x, int y, int buttons,
		   vec3_t xaxis, vec3_t yaxis,
		   vec3_t origin, vec3_t dir);
void Drag_MouseMoved (int x, int y, int buttons);
void Drag_MouseUp (void);

//
// csg.c
//
void CSG_MakeHollow (void);
void CSG_Subtract (void);

//
// vertsel.c
//

void SetupVertexSelection (void);
void SelectEdgeByRay (vec3_t org, vec3_t dir);
void SelectVertexByRay (vec3_t org, vec3_t dir);

void ConnectEntities (void);

extern	int	update_bits;

extern	int	screen_width;
extern	int	screen_height;

extern	HANDLE	bsp_process;

char	*TranslateString (char *buf);

void ProjectDialog (void);

void FillTextureMenu (void);
void FillBSPMenu (void);

BOOL CALLBACK Win_Dialog (
    HWND hwndDlg,	// handle to dialog box
    UINT uMsg,	// message
    WPARAM wParam,	// first message parameter
    LPARAM lParam 	// second message parameter
);


//
// win_cam.c
//
void WCam_Create (HINSTANCE hInstance);


//
// win_xy.c
//
void WXY_Create (HINSTANCE hInstance, char *WindowName, char *LoadName, WInfo_t *winfo);
void DestroyExtra2DViews(void);
void CreateExtra2DViews(void);

//
// win_z.c
//
void WZ_Create (HINSTANCE hInstance);

//
// win_ent.c
//


//
// win_main.c
//
void Main_Create (HINSTANCE hInstance);
void SaveClientWindowStates(void);
extern BOOL SaveWindowState(HWND hWnd, const char *pszName);
extern BOOL LoadWindowState(HWND hWnd, const char *pszName);

extern BOOL SaveRegistryInfo(const char *pszName, void *pvBuf, long lSize);
extern BOOL loadRegistryInfo(const char *pszName, void *pvBuf, long *plSize);

//
// entityw.c
//
BOOL CreateEntityWindow(HINSTANCE hInstance);
void FillClassList (void);
BOOL UpdateEntitySel(eclass_t *pec);	
void SetInspectorMode(int iType);
int DrawTexControls(HWND hWnd);
void SetSpawnFlags(void);
void GetSpawnFlags(void);
void SetKeyValuePairs(void);
extern void BuildGammaTable(float g);


// win_dlg.c

void	DoFind			(void);
void	DoRotate		(void);
void	DoSides			(void);
void	DoAbout			(void);
void	DoSurface		(void);
void	DoLight			(void);
int		DoCreateVisGroup(void);
void	DoSetVisGroup	(void);
void	DoChangeFlags	(void);

// win_util.c
void CheckMenu(int id, int value);
void ToggleAndCheck(int id, int *value);

// clipper.c
void	SetClipperMode			(int mode);
int		Clipper_AddPoint		(vec3_t pt, int orientation);
void	Clipper_Off				(void);
void	Clipper_RefreshBrushes	(void);
void	Clipper_FreeBrushes		(void);
int		FindClipperPoint		(vec3_t org, vec3_t dir);
void	DoClip					(void);


/*
** QE function declarations
*/
void     QE_CheckAutoSave( void );
void     QE_ConvertDOSToUnixName( char *dst, const char *src );
void     QE_CountBrushesAndUpdateStatusBar( void );
void     QE_CheckOpenGLForErrors(char *function_name, int ignore);
void     QE_ExpandBspString (char *bspaction, char *out, char *mapname);
void     QE_Init (void);
qboolean QE_KeyDown (int key);
qboolean QE_LoadProject (char *projectfile);
qboolean QE_SingleBrush (void);

/*
** QE Win32 function declarations
*/
int		QEW_SetupPixelFormat	(PIXELFORMATDESCRIPTOR *pfd, HDC *hDC, qboolean zbuffer );
void	QEW_StopGL				( HWND *hWnd, HGLRC *hGLRC, HDC *hDC );
void	SetUpFont				(WInfo_t *winfo);

BOOL DepthClip(brush_t *b);
trace_t Test_Ray (vec3_t origin, vec3_t dir, int flags);
void Select_Ray (vec3_t origin, vec3_t dir, int flags);

/*
** extern declarations
*/
extern	globals_t		globals;
extern	int				NewWindow;
extern	UINT			MOUSEWHEEL;
extern	char			*Surface_FlagNames[32];
extern	char			*Content_FlagNames[32];

#define QUAKE2		0x00
#define QUAKE1		0x01
#define HALFLIFE	0x02
#define SOF			0x03
#endif
