// select.c

#include "qe3.h"
#include "fnmatch.h"

#define	DIST_START	999999

extern BOOL TexturesLocked;
void Select_DragTexture(vec3_t move);

void ClearSelectedFaces(facesel_t **head)
{
	facesel_t *a = *head, *b;

	while(a)
	{
		b = a;
		a = a->next;
		free(b);
	}
	*head = NULL;
}

BOOL FaceSelected(face_t *face)
{
	facesel_t *a = selected_faces;

	while(a)
	{
		if(face == a->face)
			return true;
		a = a->next;
	}
	return false;
}

void DeselectFace(face_t *face)
{
	facesel_t *current, *last = NULL;

	for(current = selected_faces; current; last = current, current = current->next)
	{
		if(current->face == face)
		{
			if(last)
				last->next = current->next;
			else
				selected_faces = current->next;

			free(current);
			return;
		}
	}
}

void SelectFace(facesel_t **head, face_t *face, brush_t *brush, float trans)
{
	facesel_t *a;

        if((a = (facesel_t *)qmalloc(sizeof(facesel_t))) == NULL)
		Error("Unable to allocate memory in SelectFace()");
	else
	{
		a->face = face;
		a->owner = brush;
		a->trans = trans;
		a->next = *head;
		*head = a;
	}
}

BOOL DepthClip(brush_t *b)
{
	vec3_t	mins;
	vec3_t	maxs;
	vec3_t	view;
	int		maxdist;

	if(globals.DepthClipping)
	{
		maxdist = globals.DepthClipDistance;
		mins[0] = b->mins[0] - maxdist;
		mins[1] = b->mins[1] - maxdist;
		mins[2] = b->mins[2] - maxdist;
		maxs[0] = b->maxs[0] + maxdist;
		maxs[1] = b->maxs[1] + maxdist;
		maxs[2] = b->maxs[2] + maxdist;
		VectorCopy(camera.origin, view);
		if(view[0] < mins[0] ||
			view[1] < mins[1] ||
			view[2] < mins[2] ||
			view[0] > maxs[0] ||
			view[1] > maxs[1] ||
			view[2] > maxs[2])
			return true;
	}
	return false;
}

/*
===========
Test_Ray
===========
*/
trace_t Test_Ray (vec3_t origin, vec3_t dir, int flags)
{
	brush_t	*brush;
	face_t	*face;
	vec_t	dist;
	trace_t	t;

	memset (&t, 0, sizeof(t));
	t.dist = DIST_START;

	if (! (flags & SF_SELECTED_ONLY) )
		for (brush = active_brushes.next ; brush != &active_brushes ; brush=brush->next)
		{
			if ( (flags & SF_ENTITIES_FIRST) && brush->owner == world_entity)
				continue;
			if (FilterBrush (brush))
				continue;
			if (dir[0] && dir[1] && DepthClip(brush))
				continue;
			face = Brush_Ray (origin, dir, brush, &dist);
			if (dist > 0 && dist < t.dist)
			{
				t.dist = dist;
				t.brush = brush;
				t.face = face;
				t.selected = false;
			}
		}
	for (brush = selected_brushes.next ; brush != &selected_brushes ; brush=brush->next)
	{
		if ( (flags & SF_ENTITIES_FIRST) && brush->owner == world_entity)
			continue;
		if (FilterBrush (brush))
			continue;
			if (dir[0] && dir[1] && DepthClip(brush))
			continue;
		face = Brush_Ray (origin, dir, brush, &dist);
		if (dist > 0 && dist < t.dist)
		{
			t.dist = dist;
			t.brush = brush;
			t.face = face;
			t.selected = true;
		}
	}

	// if entites first, but didn't find any, check regular

	if ( (flags & SF_ENTITIES_FIRST) && t.brush == NULL)
		return Test_Ray (origin, dir, flags - SF_ENTITIES_FIRST);

	return t;
}


/*
============
Select_Brush

============
*/
void Select_Brush (brush_t *brush, int SelectEntireEntity)
{
	brush_t	*b;
	entity_t	*e;

	ClearSelectedFaces(&selected_faces);
	if (globals.d_select_count < 2)
		globals.d_select_order[globals.d_select_count] = brush;
	globals.d_select_count++;

	for(e = entities.next; e != &entities && e != brush->owner; e = e->next);
	if(e == &entities && brush->owner != world_entity)
		Sys_Printf("Entity for brush not found in list\n");

	e = brush->owner;
	if (e)
	{
		// select complete entity on first click
		if (e != world_entity && SelectEntireEntity)
		{
			for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
				if (b->owner == e)
					goto singleselect;
			for (b=e->brushes.onext ; b != &e->brushes ; b=b->onext)
			{
				Brush_RemoveFromList (b);
				Brush_AddToList (b, &selected_brushes);
			}
		}
		else
		{
singleselect:
			Brush_RemoveFromList (brush);
			Brush_AddToList (brush, &selected_brushes);
		}

		if (e->eclass)
		{
			UpdateEntitySel(brush->owner->eclass);
		}
	}
}

/*
============
Select_Ray

If the origin is inside a brush, that brush will be ignored.
============
*/
void Select_Ray (vec3_t origin, vec3_t dir, int flags)
{
	trace_t	t;

	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
	t = Test_Ray (origin, dir, flags);

	if (!t.brush)
		return;

	if (flags == SF_SINGLEFACE)
	{
		if(FaceSelected(t.face))
			DeselectFace(t.face);
		else
			SelectFace(&selected_faces, t.face, t.brush, 0);
		globals.d_select_mode = sel_brush;
		return;

	}

	// move the brush to the other list

	globals.d_select_mode = sel_brush;

	if (t.selected)
	{		
		Brush_RemoveFromList (t.brush);
		Brush_AddToList (t.brush, &active_brushes);
	} else
	{
		Select_Brush (t.brush, GetKeyState(VK_MENU) >= 0);
	}
}


void Select_Delete (void)
{
	brush_t		*brush, *b;
	face_t		*f, *f2;
	facesel_t	*face;
	facesel_t	*adjoining = NULL;
	winding_t	*w;
	float		d;
	vec_t		v;
	vec3_t		mins, maxs, center, move[2];
	int			i, j, numptsonplane, numadjoining = 0, numfaces = 0;

	globals.d_select_mode = sel_brush;

	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);

	globals.d_select_count = 0;
	globals.d_num_move_points = 0;
	if(selected_brushes.next != &selected_brushes)
	{
		while (selected_brushes.next != &selected_brushes)
		{
			brush = selected_brushes.next;
			Brush_Free (brush);
		}
		return;
	}

	if(!selected_faces)
		return;

	if(selected_faces->next)
	{
		Sys_Printf("Can only delete a single face\n");
		return;
	}

	/* What we want to do here is:
		0: Precalculate the center of the face to be deleted
		1: Create a list of faces touching the face to be deleted (facesel_t)
		2: Go through the list and make sure none of the windings has 3 points (error msg and return if so)
		3: Go through the list again and set up for an edge drag and calculate the vector offset from the center of the delface's center to the center of the edge
		4: Move edge so center of edge is at center of delface
		5: Delete the face and rebuild its brush
	*/	
	mins[0] = mins[1] = mins[2] = 99999;
	maxs[0] = maxs[1] = maxs[2] = -99999;
	f = selected_faces->face;
	b = selected_faces->owner;
	w = MakeFaceWinding (b, f);
	
//	Sys_Printf("Calculating center point:\n");
	for(i = 0; i < w->numpoints; i++)
	{
//		Sys_Printf("Winding point: %f %f %f\n", w->points[i][0], w->points[i][1], w->points[i][2]);
		for(j = 0; j < 3; j++)
		{
			v = w->points[i][j];
			if(v > maxs[j])
				maxs[j] = v;
			if(v < mins[j])
				mins[j] = v;
		}
	}

//	Sys_Printf("Mins: %f %f %f\n", mins[0], mins[1], mins[2]);
//	Sys_Printf("Maxs: %f %f %f\n", maxs[0], maxs[1], maxs[2]);
	
	for(i = 0; i < 3; i++)
		center[i] = (maxs[i] + mins[i]) * 0.5;

//	Sys_Printf("Center: %f %f %f\n", center[0], center[1], center[2]);
	
	free(w);

	for (f2 = b->brush_faces ; f2 ; f2 = f2->next)
	{
		numfaces++;

		if (f2 == f)
			continue;

		w = MakeFaceWinding (b, f2);

		if (!w)
			continue;
		
		// any points on f will become new control points
		for (i=0 ; i<w->numpoints ; i++)
		{
			d = DotProduct (w->points[i], f->plane.normal) - f->plane.dist;
			if (d > -ON_EPSILON && d < ON_EPSILON)
			{
				SelectFace(&adjoining, f2, b, false);
				numadjoining++;
				break;
			}
		}
		free(w);
	}
	if(numfaces <= 5)
	{
		ClearSelectedFaces(&adjoining);
		Sys_Printf("Cannot delete faces from a brush with 5 sides or less\n");
		return;
	}

//	VectorScale (f->plane.normal, f->plane.dist, center);
	
	for (face = adjoining ; face ; face = face->next)
	{
		f2 = face->face;

		if (f2 == f)
			continue;
		w = MakeFaceWinding (b, f2);
		if (!w)
			continue;
		
		// any points on f will become new control points
		for (i=0 ; i<w->numpoints ; i++)
		{
			d = DotProduct (w->points[i], f->plane.normal) - f->plane.dist;
			if (d > -ON_EPSILON && d < ON_EPSILON)
			{
				break;
			}
		}
		
		//
		// if none of the points were on the plane,
		// leave it alone
		//
		if (i != w->numpoints)
		{
			numptsonplane = 0;
			if (i == 0)
			{	// see if the first clockwise point was the
				// last point on the winding
				d = DotProduct (w->points[w->numpoints-1], f->plane.normal) - f->plane.dist;
				if (d > -ON_EPSILON && d < ON_EPSILON)
					i = w->numpoints - 1;
			}
			
			VectorCopy (w->points[i], f2->planepts[0]);
			VectorCopy (f2->planepts[0], move[numptsonplane]);
			numptsonplane++;

			if (++i == w->numpoints)
				i = 0;
			
			// see if the next point is also on the plane
			d = DotProduct (w->points[i], f->plane.normal) - f->plane.dist;
			if (d > -ON_EPSILON && d < ON_EPSILON)
			{
				VectorCopy (w->points[i], move[numptsonplane]);
				numptsonplane++;
			}
			
			VectorCopy (w->points[i], f2->planepts[1]);
			if (++i == w->numpoints)
				i = 0;
			
			// the third point is never on the plane
			VectorCopy (w->points[i], f2->planepts[2]);
		}
		
		free(w);
		if(numptsonplane == 1)
		{
//			Sys_Printf("Moving single point: %5.3f %5.3f %5.3f to %5.3f %5.3f %5.3f\n", f2->planepts[0][0], f2->planepts[0][1], f2->planepts[0][2], center[0], center[1], center[2]);
			for(i = 0; i < 3; i++)
			{
//				f2->planepts[0][i] += center[i] - f2->planepts[0][i];
				f2->planepts[0][i] = center[i];
			}
		}
		else
		{
//			Sys_Printf("Moving two points: %5.3f %5.3f %5.3f and %5.3f %5.3f %5.3f\n", f2->planepts[0][0], f2->planepts[0][1], f2->planepts[0][2], f2->planepts[1][0], f2->planepts[1][1], f2->planepts[1][2]);
			for(i = 0; i < 3; i++)
			{ 
				maxs[i] = (f2->planepts[0][i] + f2->planepts[1][i]) * 0.5;
				mins[i] = center[i] - maxs[i];

				f2->planepts[0][i] += mins[i];
				f2->planepts[1][i] += mins[i];
			}
//			Sys_Printf("Point center: %5.3f %5.3f %5.3f\n", maxs[0], maxs[1], maxs[2]);
//			Sys_Printf("Offset: %5.3f %5.3f %5.3f\n", mins[0], mins[1], mins[2]);

		}
	}
	free(selected_faces->face->face_winding);
	selected_faces->face->face_winding = 0;
	Brush_RemoveEmptyFaces(b);
	Brush_Build(b);
	ClearSelectedFaces(&adjoining);
	ClearSelectedFaces(&selected_faces);

	// FIXME: remove any entities with no brushes
}

void Select_Deselect (BOOL clearfaces)
{
	brush_t	*b;
	int		i;

	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);

	if(globals.ClipperNumPoints)
	{
		globals.ClipperNumPoints = 0;
		return;
	}

	globals.d_workcount++;
	globals.d_select_count = 0;
	globals.d_num_move_points = 0;
	b = selected_brushes.next;

	if (b == &selected_brushes)
	{
		if (clearfaces && selected_faces)
			ClearSelectedFaces(&selected_faces);

		return;
	}

	ClearSelectedFaces(&selected_faces);
	globals.d_select_mode = sel_brush;

	// grab top / bottom height for new brushes
	for(i = 0; i < 3; i++)
	{
		if (b->mins[i] < b->maxs[i])
		{
			globals.d_new_brush_bottom[i] = b->mins[i];
			globals.d_new_brush_top[i] = b->maxs[i];
		}
	}

	selected_brushes.next->prev = &active_brushes;
	selected_brushes.prev->next = active_brushes.next;
	active_brushes.next->prev = selected_brushes.prev;
	active_brushes.next = selected_brushes.next;
	selected_brushes.prev = selected_brushes.next = &selected_brushes;	
}

/*
============
Select_Move
============
*/
void Select_Move (vec3_t delta)
{
	brush_t	*b;

// actually move the selected brushes
	if(!selected_faces)
	{
		for (b = selected_brushes.next ; b != &selected_brushes ; b=b->next)
			Brush_Move (b, delta);
	}
	else
	{
		Select_DragTexture(delta);
	}
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

/*
============
Select_Clone

Creates an exact duplicate of the selection in place, then moves
the selected brushes off of their old positions
============
*/
void Select_Clone (void)
{
	brush_t		*b, *b2, *n, *next, *next2;
	vec3_t		delta;
	entity_t	*e;
	HWND		h = GetFocus();

	globals.d_workcount++;
	globals.d_select_mode = sel_brush;

	if(h == globals.XZ.hwnd)
	{
		delta[0] = globals.d_gridsize;
		delta[1] = 0;
		delta[2] = globals.d_gridsize;
	}
	else if(h == globals.YZ.hwnd)
	{
		delta[0] = 0;
		delta[1] = globals.d_gridsize;
		delta[2] = globals.d_gridsize;
	}
	else
	{
		delta[0] = globals.d_gridsize;
		delta[1] = globals.d_gridsize;
		delta[2] = 0;
	}

	for (b=selected_brushes.next ; b != &selected_brushes ; b=next)
	{
		next = b->next;
		// if the brush is a world brush, handle simply
		if (b->owner == world_entity)
		{
			n = Brush_Clone (b);
			Brush_AddToList (n, &active_brushes);
			Entity_LinkBrush (world_entity, n);
			Brush_Build( n );

			if(globals.OffsetCopiedBrushes)
				Brush_Move (b, delta);
			continue;
		}

		e = Entity_Clone (b->owner);
		// clear the target / targetname
		DeleteKey (e, "target");
		DeleteKey (e, "targetname");

		// if the brush is a fixed size entity, create a new entity
		if (b->owner->eclass->fixedsize)
		{
			n = Brush_Clone (b);
			Brush_AddToList (n, &active_brushes);
			Entity_LinkBrush (e, n);
			Brush_Build( n );
			if(globals.OffsetCopiedBrushes)
				Brush_Move (b, delta);
			continue;
		}
        
		// brush is a complex entity, grab all the other ones now

		next = &selected_brushes;

		for ( b2 = b ; b2 != &selected_brushes ; b2=next2)
		{
			next2 = b2->next;
			if (b2->owner != b->owner)
			{
				if (next == &selected_brushes)
					next = b2;
				continue;
			}

			// move b2 to the start of selected_brushes,
			// so it won't be hit again
			Brush_RemoveFromList (b2);
			Brush_AddToList (b2, &selected_brushes);
			
			n = Brush_Clone (b2);
			Brush_AddToList (n, &active_brushes);
			Entity_LinkBrush (e, n);
			Brush_Build( n );
			if(globals.OffsetCopiedBrushes)
				Brush_Move (b2, delta);
		}

	}
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}



/*
============
Select_SetTexture
============
*/
void Select_SetTexture (texdef_t *texdef)
{
	brush_t		*b;
	facesel_t	*f;

	if (selected_faces)
	{
		for(f = selected_faces; f; f = f->next)
		{
			f->face->texdef = *texdef;
			Brush_Build(f->owner);
		}
	}
	else
	{
		for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
			if (!b->owner->eclass->fixedsize)
				Brush_SetTexture (b, texdef);
	}
	Sys_UpdateWindows (W_ALL);
}


/*
================================================================

  TRANSFORMATIONS

================================================================
*/

void Select_GetBounds (vec3_t mins, vec3_t maxs, brush_t *listhead)
{
	brush_t	*b;
	int		i;

	for (i=0 ; i<3 ; i++)
	{
		mins[i] = 99999;
		maxs[i] = -99999;
	}

	for (b=listhead->next ; b != listhead ; b=b->next)
		for (i=0 ; i<3 ; i++)
		{
			if (b->mins[i] < mins[i])
				mins[i] = b->mins[i];
			if (b->maxs[i] > maxs[i])
				maxs[i] = b->maxs[i];
		}
}

void Select_GetMid (vec3_t mid)
{
	vec3_t	mins, maxs;
	int		i;

	Select_GetBounds (mins, maxs, &selected_brushes);
	for (i=0 ; i<3 ; i++)
//		mid[i] = globals.d_gridsize*floor ( ( (mins[i] + maxs[i])*0.5 )/globals.d_gridsize );
		mid[i] = (mins[i] + maxs[i]) * 0.5;
}

vec3_t	select_origin;
vec3_t	select_matrix[3];
qboolean	select_fliporder;

void Select_AplyMatrix (void)
{
	brush_t	*b;
	face_t	*f;
	int		i, j;
	vec3_t	temp;

	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
	{
		for (f=b->brush_faces ; f ; f=f->next)
		{
			for (i=0 ; i<3 ; i++)
			{
				VectorSubtract (f->planepts[i], select_origin, temp);
				for (j=0 ; j<3 ; j++)
					f->planepts[i][j] = DotProduct(temp, select_matrix[j])
						+ select_origin[j];
			}
			if (select_fliporder)
			{
				VectorCopy (f->planepts[0], temp);
				VectorCopy (f->planepts[2], f->planepts[0]);
				VectorCopy (temp, f->planepts[2]);
			}
		}
		Brush_Build(b);
	}
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_ApplyMatrix (int nAxis, float fDeg)
{
	brush_t	*b;
	face_t	*f;
	int		i, j;
	vec3_t	temp;
	
	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
	{
		for (f=b->brush_faces ; f ; f=f->next)
		{
			for (i=0 ; i<3 ; i++)
			{
				VectorSubtract (f->planepts[i], select_origin, temp);
				for (j=0 ; j<3 ; j++)
					f->planepts[i][j] = DotProduct(temp, select_matrix[j]) + select_origin[j];
			}
			if (select_fliporder)
			{
				VectorCopy (f->planepts[0], temp);
				VectorCopy (f->planepts[2], f->planepts[0]);
				VectorCopy (temp, f->planepts[2]);
			}
		}
		Brush_Build(b);		
	}
}

void Select_FlipAxis (int axis)
{
	int		i;

	Select_GetMid (select_origin);
	for (i=0 ; i<3 ; i++)
	{
		VectorCopy (vec3_origin, select_matrix[i]);
		select_matrix[i][i] = 1;
	}
	select_matrix[axis][axis] = -1;

	select_fliporder = true;
	Select_AplyMatrix ();
}

void Clamp(float *f, int nClamp)
{
  float fFrac = *f - (int)*f;
  *f = (int)*f % nClamp;
  *f += fFrac;
}

void TextureAxisFromPlane(plane_t *, vec3_t, vec3_t);
void ProjectOnPlane(vec3_t normal,float dist,vec3_t ez, vec3_t p)
{
	if (fabs(ez[0]) == 1)
		p[0] = (dist - normal[1] * p[1] - normal[2] * p[2]) / normal[0];
	else if (fabs(ez[1]) == 1)
		p[1] = (dist - normal[0] * p[0] - normal[2] * p[2]) / normal[1];
	else
		p[2] = (dist - normal[0] * p[0] - normal[1] * p[1]) / normal[2];
}

void Back(vec3_t dir, vec3_t p)
{
	if (fabs(dir[0]) == 1)
		p[0] = 0;
	else if (fabs(dir[1]) == 1)
		p[1] = 0;
	else p[2] = 0;
}



// using scale[0] and scale[1]
void ComputeScale(vec3_t rex, vec3_t rey, vec3_t p, face_t *f)
{
	float px = DotProduct(rex, p);
	float py = DotProduct(rey, p);
  vec3_t aux;
	px *= f->texdef.scale[0];
	py *= f->texdef.scale[1];
  VectorCopy(rex, aux);
  VectorScale(aux, px, aux);
  VectorCopy(aux, p);
  VectorCopy(rey, aux);
  VectorScale(aux, py, aux);
  VectorAdd(p, aux, p);
}

void ComputeAbsolute(face_t *f, vec3_t p1, vec3_t p2, vec3_t p3)
{
	vec3_t ex,ey,ez;	        // local axis base
	vec3_t aux;
	vec3_t rex,rey;

  // compute first local axis base
  TextureAxisFromPlane(&f->plane, ex, ey);
  CrossProduct(ex, ey, ez);
	    
  VectorCopy(ex, aux);
  VectorScale(aux, -f->texdef.shift[0], aux);
  VectorCopy(aux, p1);
  VectorCopy(ey, aux);
  VectorScale(aux, -f->texdef.shift[1], aux);
  VectorAdd(p1, aux, p1);
  VectorCopy(p1, p2);
  VectorAdd(p2, ex, p2);
  VectorCopy(p1, p3);
  VectorAdd(p3, ey, p3);
  VectorCopy(ez, aux);
  VectorScale(aux, -f->texdef.rotate, aux);
  VectorRotate(p1, aux, p1);
  VectorRotate(p2, aux, p2);
  VectorRotate(p3, aux, p3);
	// computing rotated local axis base
  VectorCopy(ex, rex);
  VectorRotate(rex, aux, rex);
  VectorCopy(ey, rey);
  VectorRotate(rey, aux, rey);

  ComputeScale(rex,rey,p1,f);
	ComputeScale(rex,rey,p2,f);
	ComputeScale(rex,rey,p3,f);

	// project on normal plane
	// along ez 
	// assumes plane normal is normalized
	ProjectOnPlane(f->plane.normal,f->plane.dist,ez,p1);
	ProjectOnPlane(f->plane.normal,f->plane.dist,ez,p2);
	ProjectOnPlane(f->plane.normal,f->plane.dist,ez,p3);
};


void AbsoluteToLocal(plane_t normal2, face_t *f, vec3_t p1, vec3_t p2, vec3_t p3)
{
	vec3_t	ex,ey,ez;
	vec3_t	aux;
	vec3_t	rex,rey;
	vec_t	x;
	vec_t	y;
	
	// computing new local axis base
	TextureAxisFromPlane(&normal2, ex, ey);
	CrossProduct(ex, ey, ez);
	
	// projecting back on (ex,ey)
	Back(ez,p1);
	Back(ez,p2);
	Back(ez,p3);
	
	// rotation
	VectorCopy(p2, aux);
	VectorSubtract(aux, p1,aux);
	
	x = DotProduct(aux,ex);
	y = DotProduct(aux,ey);
	f->texdef.rotate = 180 * atan2(y,x) / Q_PI;
	
	// computing rotated local axis base
	VectorCopy(ez, aux);
	VectorScale(aux, f->texdef.rotate, aux);
	VectorCopy(ex, rex);
	VectorRotate(rex, aux, rex);
	VectorCopy(ey, rey);
	VectorRotate(rey, aux, rey);
	
	// scale
	VectorCopy(p2, aux);
	VectorSubtract(aux, p1, aux);
	f->texdef.scale[0] = DotProduct(aux, rex);
	VectorCopy(p3, aux);
	VectorSubtract(aux, p1, aux);
	f->texdef.scale[1] = DotProduct(aux, rey);
	
	// shift
	// only using p1
	x = DotProduct(rex,p1);
	y = DotProduct(rey,p1);
	x /= f->texdef.scale[0];
	y /= f->texdef.scale[1];
	
	VectorCopy(rex, p1);
	VectorScale(p1, x, p1);
	VectorCopy(rey, aux);
	VectorScale(aux, y, aux);
	VectorAdd(p1, aux, p1);
	VectorCopy(ez, aux);
	VectorScale(aux, -f->texdef.rotate, aux);
	VectorRotate(p1, aux, p1);
	f->texdef.shift[0] = -DotProduct(p1, ex);
	f->texdef.shift[1] = -DotProduct(p1, ey);
	
	// stored rot is good considering local axis base
	// change it if necessary
	f->texdef.rotate = -f->texdef.rotate;
	
	Clamp(&f->texdef.shift[0], f->d_texture->image->width);
	Clamp(&f->texdef.shift[1], f->d_texture->image->height);
	Clamp(&f->texdef.rotate, 360);
}

void RotateFaceTexture(face_t* f, int nAxis, float fDeg)
{
	vec3_t	p1,p2,p3, rota;   
	plane_t	normal2;
	vec3_t	vNormal;

	p1[0] = p1[1] = p1[2] = 0;
	VectorCopy(p1, p2);
	VectorCopy(p1, p3);
	VectorCopy(p1, rota);
	ComputeAbsolute(f, p1, p2, p3);
  
	rota[nAxis] = fDeg;
	VectorRotate2(p1, rota, select_origin, p1);
	VectorRotate2(p2, rota, select_origin, p2);
	VectorRotate2(p3, rota, select_origin, p3);

	vNormal[0] = f->plane.normal[0];
	vNormal[1] = f->plane.normal[1];
	vNormal[2] = f->plane.normal[2];
	VectorRotate(vNormal, rota, vNormal);
	normal2.normal[0] = vNormal[0];
	normal2.normal[1] = vNormal[1];
	normal2.normal[2] = vNormal[2];
	AbsoluteToLocal(normal2, f, p1, p2 ,p3);

}

void RotateTextures(int nAxis, float fDeg, vec3_t vOrigin)
{
	brush_t *b;
	face_t *f;

	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
	{
		for (f=b->brush_faces ; f ; f=f->next)
		{
			RotateFaceTexture(f, nAxis, fDeg);
			//++timo removed that call .. works fine .. ???????
			Brush_Build(b);
		}
		Brush_Build(b);
	}
}

void Select_RotateAxis (int axis, float deg)
{
	int		i;
	vec_t	c, s;

	while(deg >= 360)
		deg -= 360;
	while(deg < 0)
		deg += 360;

	if (deg == 0)
		return;

	Select_GetMid (select_origin);
	select_fliporder = false;
	
	deg = -deg;
	c = cos(deg * Q_PI / 180.0);
	s = sin(deg * Q_PI / 180.0);

	for (i=0 ; i<3 ; i++)
	{
		VectorCopy (vec3_origin, select_matrix[i]);
		select_matrix[i][i] = 1;
	}
	
	switch (axis)
	{
	case 0:
		select_matrix[1][1] = c;
		select_matrix[1][2] = -s;
		select_matrix[2][1] = s;
		select_matrix[2][2] = c;
			break;
	case 1:
		select_matrix[0][0] = c;
		select_matrix[0][2] = s;
		select_matrix[2][0] = -s;
		select_matrix[2][2] = c;
		break;
	case 2:
		select_matrix[0][0] = c;
		select_matrix[0][1] = -s;
		select_matrix[1][0] = s;
		select_matrix[1][1] = c;
		break;
	}
	
	if(TexturesLocked)
		RotateTextures(axis, deg, select_origin);

	Select_ApplyMatrix(axis, deg);
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_FitTexture(int nHeight, int nWidth)
{
	brush_t		*b;
	facesel_t	*f;
	
	if(selected_brushes.next == &selected_brushes && !selected_faces)
		return;
	
	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
	{
		Brush_FitTexture(b, nHeight, nWidth);
		Brush_Build(b);
	}
	
	for (f = selected_faces; f; f = f->next)
	{
		Face_FitTexture(f->face, nHeight, nWidth);
		Brush_Build(f->owner);
	}
	
	Sys_UpdateWindows (W_CAMERA);
}

/*
================================================================

GROUP SELECTIONS

================================================================
*/

void Select_CompleteTall (void)
{
	brush_t	*b, *next;
	int		i;
	vec3_t	mins, maxs;

	if (!QE_SingleBrush ())
		return;

	globals.d_select_mode = sel_brush;

	VectorCopy (selected_brushes.next->mins, mins);
	VectorCopy (selected_brushes.next->maxs, maxs);
	Select_Delete ();

	for (b=active_brushes.next ; b != &active_brushes ; b=next)
	{
		next = b->next;
		for (i=0 ; i<2 ; i++)
			if (b->maxs[i] > maxs[i] || b->mins[i] < mins[i])
				break;
		if (i == 2)
		{
			Brush_RemoveFromList (b);
			Brush_AddToList (b, &selected_brushes);
		}
	}
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_PartialTall (void)
{
	brush_t	*b, *next;
	int		i;
	vec3_t	mins, maxs;

	if (!QE_SingleBrush ())
		return;

	globals.d_select_mode = sel_brush;

	VectorCopy (selected_brushes.next->mins, mins);
	VectorCopy (selected_brushes.next->maxs, maxs);
	Select_Delete ();

	for (b=active_brushes.next ; b != &active_brushes ; b=next)
	{
		next = b->next;
		for (i=0 ; i<2 ; i++)
			if (b->mins[i] > maxs[i] || b->maxs[i] < mins[i])
				break;
		if (i == 2)
		{
			Brush_RemoveFromList (b);
			Brush_AddToList (b, &selected_brushes);
		}
	}
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_Touching (void)
{
	brush_t	*b, *next;
	int		i;
	vec3_t	mins, maxs;

	if (!QE_SingleBrush ())
		return;

	globals.d_select_mode = sel_brush;

	VectorCopy (selected_brushes.next->mins, mins);
	VectorCopy (selected_brushes.next->maxs, maxs);

	for (b=active_brushes.next ; b != &active_brushes ; b=next)
	{
		next = b->next;
		for (i=0 ; i<3 ; i++)
			if (b->mins[i] > maxs[i]+1 || b->maxs[i] < mins[i]-1)
				break;
		if (i == 3)
		{
			Brush_RemoveFromList (b);
			Brush_AddToList (b, &selected_brushes);
		}
	}
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_Inside (void)
{
	brush_t	*b, *next;
	int		i;
	vec3_t	mins, maxs;

	if (!QE_SingleBrush ())
		return;

	globals.d_select_mode = sel_brush;

	VectorCopy (selected_brushes.next->mins, mins);
	VectorCopy (selected_brushes.next->maxs, maxs);
	Select_Delete ();

	for (b=active_brushes.next ; b != &active_brushes ; b=next)
	{
		next = b->next;
		for (i=0 ; i<3 ; i++)
			if (b->maxs[i] > maxs[i] || b->mins[i] < mins[i])
				break;
		if (i == 3)
		{
			Brush_RemoveFromList (b);
			Brush_AddToList (b, &selected_brushes);
		}
	}
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

/*
=============
Select_Ungroup

Turn the currently selected entity back into normal brushes
=============
*/
void  Select_Ungroup (void)
{
	entity_t	*e;
	brush_t		*b, *bnext;

	e = selected_brushes.next->owner;

	if (!e || e == world_entity || e->eclass->fixedsize)
	{
		Sys_Status(0, "Not a grouped entity.");
		return;
	}

	for(b = selected_brushes.next; b != &selected_brushes; b = b->next)
	{
		if(e != b->owner)
		{
			Sys_Printf("Can't ungroup brushes from multiple entities\n");
			return;
		}
	}

	for(b = selected_brushes.next; b != &selected_brushes; b = bnext)
	{
		bnext = b->next;
		Brush_RemoveFromList (b);
		Brush_AddToList (b, &active_brushes);
		Entity_UnlinkBrush (b);
		Entity_LinkBrush (world_entity, b);
		Brush_Build( b );
		b->owner = world_entity;
	}
	if(e->brushes.onext == &e->brushes)
		Entity_Free (e);

//	for (b=e->brushes.onext ; b != &e->brushes ; b=e->brushes.onext)
//	{
//	}

	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_SetContentsFlag (int val)
{
	brush_t		*b;
	face_t		*f;
	facesel_t	*s;

	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
		for (f=b->brush_faces ; f ; f=f->next)
			f->texdef.contents |= val;
	for(s = selected_faces; s; s = s->next)
		s->face->texdef.contents |= val;
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_UnsetContentsFlag (int val)
{
	brush_t		*b;
	face_t		*f;
	facesel_t	*s;

	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
		for (f=b->brush_faces ; f ; f=f->next)
			f->texdef.contents &= ~val;
	for(s = selected_faces; s; s = s->next)
		s->face->texdef.contents &= ~val;
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_MakeLight (int value, BOOL deselect)
{
	brush_t		*b;
	face_t		*f;
	facesel_t	*s;

	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
		for (f=b->brush_faces ; f ; f=f->next)
		{
			f->texdef.flags |= SURF_LIGHT;
			f->texdef.value = value;
		}
	for(s = selected_faces; s; s = s->next)
	{
		s->face->texdef.flags |= SURF_LIGHT;
		s->face->texdef.value = value;
	}
	if(deselect)
		Select_Deselect (true);
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_SetSurfaceFlag (int val)
{
	brush_t		*b;
	face_t		*f;
	facesel_t	*s;

	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
		for (f=b->brush_faces ; f ; f=f->next)
			f->texdef.flags |= val;
	for(s = selected_faces; s; s = s->next)
		s->face->texdef.flags |= val;
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_UnsetSurfaceFlag (int val)
{
	brush_t		*b;
	face_t		*f;
	facesel_t	*s;

	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
		for (f=b->brush_faces ; f ; f=f->next)
			f->texdef.flags &= ~val;
	for(s = selected_faces; s; s = s->next)
		s->face->texdef.flags &= ~val;
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void SelectContainingTexture(BOOL brushes, texdef_t *texdef)
{
	brush_t		*b, *next;
	face_t		*f;
	char		done;

	if(brushes)
	{
		for (b=active_brushes.next ; b != &active_brushes ; b=next)
		{
			next = b->next;
			done = 0;
				for (f=b->brush_faces ; f && !done ; f=f->next)
					if(!stricmp(f->texdef.name, texdef->name))
					{
						Brush_RemoveFromList (b);
						Brush_AddToList (b, &selected_brushes);
						done = 1;
					}
		}
	}
	else
	{
		for (b=active_brushes.next ; b != &active_brushes ; b=b->next)
			for (f=b->brush_faces ; f ; f=f->next)
				if(!FaceSelected(f))
					if(!stricmp(f->texdef.name, texdef->name))
					{
						SelectFace(&selected_faces, f, b, 0);
						continue;
					}
	}
	Sys_UpdateWindows (W_ALL);
}

void SelectContainingFlag(BOOL brushes, BOOL content, int flag)
{
	brush_t		*b, *next;
	face_t		*f;
	char		done;

	if(brushes)
	{
		for (b=active_brushes.next ; b != &active_brushes ; b=next)
		{
			next = b->next;
			done = 0;
				for (f=b->brush_faces ; f && !done ; f=f->next)
				{
					if(content)
					{
						if(f->texdef.contents & flag)
						{
							Brush_RemoveFromList (b);
							Brush_AddToList (b, &selected_brushes);
							done = 1;
							continue;
						}
					}
					else
					{
						if(f->texdef.flags & flag)
						{
							Brush_RemoveFromList (b);
							Brush_AddToList (b, &selected_brushes);
							done = 1;
							continue;
						}
					}
						
				}
		}
	}
	else
	{
		for (b=active_brushes.next ; b != &active_brushes ; b=b->next)
		{
			for (f=b->brush_faces ; f ; f=f->next)
			{
				if(!FaceSelected(f))
				{
					if(content)
					{
						if(f->texdef.contents & flag)
						{
							SelectFace(&selected_faces, f, b, 0);
							continue;
						}
					}
					else
					{
						if(f->texdef.flags & flag)
						{
							SelectFace(&selected_faces, f, b, 0);
							continue;
						}
					}
				}
			}
		}
	}
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_ShiftTexture(int x, int y)
{
	brush_t		*b;
	face_t		*f;
	facesel_t	*s;

	if(selected_brushes.next == &selected_brushes && !selected_faces)
	{
		Sys_Status(0, "No selected brushes or faces");
		return;
	}
	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
	{
		for (f=b->brush_faces ; f ; f=f->next)
		{
			f->texdef.shift[0] += x;
			f->texdef.shift[1] += y;
//			f->texdef.shift[0] = (int)f->texdef.shift[0] % f->d_texture->width;
//			f->texdef.shift[1] = (int)f->texdef.shift[1] % f->d_texture->height;
			while(f->texdef.shift[0] > f->d_texture->image->width)
				f->texdef.shift[0] -= f->d_texture->image->width;
			while(f->texdef.shift[1] > f->d_texture->image->height)
				f->texdef.shift[1] -= f->d_texture->image->height;
			while(f->texdef.shift[0] < 0)
				f->texdef.shift[0] += f->d_texture->image->width;
			while(f->texdef.shift[1] < 0)
				f->texdef.shift[1] += f->d_texture->image->height;
		}
		Brush_Build(b);
	}
	for(s = selected_faces; s; s = s->next)
	{
		s->face->texdef.shift[0] += x;
		s->face->texdef.shift[1] += y;
//		s->face->texdef.shift[0] = (int)s->face->texdef.shift[0] % s->face->d_texture->width;
//		s->face->texdef.shift[1] = (int)s->face->texdef.shift[1] % s->face->d_texture->height;
		while(s->face->texdef.shift[0] > s->face->d_texture->image->width)
			s->face->texdef.shift[0] -= s->face->d_texture->image->width;
		while(s->face->texdef.shift[1] > s->face->d_texture->image->height)
			s->face->texdef.shift[1] -= s->face->d_texture->image->height;
		while(s->face->texdef.shift[0] < 0)
			s->face->texdef.shift[0] += s->face->d_texture->image->width;
		while(s->face->texdef.shift[1] < 0)
			s->face->texdef.shift[1] += s->face->d_texture->image->height;
		Brush_Build(s->owner);
	}

	Sys_Status(0, "Texture shift (%s%i, %s%i)", x>0?"+":"", x, y>0?"+":"", y);

	if(globals.d_xy.draw_mode == cd_texture)
		Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
	else
		Sys_UpdateWindows (W_CAMERA);
}

void Select_ScaleTexture(int x, int y)
{
	brush_t		*b;
	face_t		*f;
	facesel_t	*s;
	float		h = (0.10 * x), v = (0.10 * y);

	if(selected_brushes.next == &selected_brushes && !selected_faces)
	{
		Sys_Status(0, "No selected brushes or faces");
		return;
	}
	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
	{
		for (f=b->brush_faces ; f ; f=f->next)
		{
			f->texdef.scale[0] += h;
			f->texdef.scale[1] += v;
		}
		Brush_Build(b);
	}
	for(s = selected_faces; s; s = s->next)
	{
		s->face->texdef.scale[0] += h;
		s->face->texdef.scale[1] += v;
		Brush_Build(s->owner);
	}

	Sys_Status(0, "Texture scale (%s%3.2f, %s%3.2f)", h>0?"+":"", h, v>0?"+":"", v);

	if(globals.d_xy.draw_mode == cd_texture)
		Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
	else
		Sys_UpdateWindows (W_CAMERA);
}

void Select_RotateTexture(int amt)
{
	brush_t		*b;
	face_t		*f;
	facesel_t	*s;

	if(selected_brushes.next == &selected_brushes && !selected_faces)
	{
		Sys_Status(0, "No selected brushes or faces");
		return;
	}
	for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
	{
		for (f=b->brush_faces ; f ; f=f->next)
		{
			f->texdef.rotate += amt;
		}
		Brush_Build(b);
	}
	for(s = selected_faces; s; s = s->next)
	{
		s->face->texdef.rotate += amt;
		Brush_Build(s->owner);
	}

	Sys_Status(0, "Texture rotate (%s%i)", amt>0?"+":"", amt);

	if(globals.d_xy.draw_mode == cd_texture)
		Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
	else
		Sys_UpdateWindows (W_CAMERA);
}
void TextureAxisFromPlane(plane_t *pln, vec3_t xv, vec3_t yv);
void Select_DragTexture(vec3_t move)
{
	vec3_t		pvecs[2];
	facesel_t	*fs;
	float		s, t, ns, nt;
	float		ang, sinv, cosv;
	texdef_t	*td;
	int			texwidth, texheight;

	for(fs = selected_faces; fs; fs = fs->next)
	{
		TextureAxisFromPlane(&fs->face->plane, pvecs[0], pvecs[1]);
		td = &fs->face->texdef;
		ang = td->rotate / 180 * Q_PI;
		sinv = sin(ang);
		cosv = cos(ang);
		
		if (!td->scale[0])
			td->scale[0] = 1;
		if (!td->scale[1])
			td->scale[1] = 1;
		
		s = DotProduct(move, pvecs[0]);
		t = DotProduct(move, pvecs[1]);
		ns = cosv * s - sinv * t;
		nt = sinv * s +  cosv * t;
		s = ns/td->scale[0];
		t = nt/td->scale[1];
		td->shift[0] -= s;
		td->shift[1] -= t;
		texwidth = fs->face->d_texture->image->width;
		texheight = fs->face->d_texture->image->height;
		while(td->shift[0] > (texwidth-1)) td->shift[0] -= texwidth;
		while(td->shift[0] < 0) td->shift[0] += texwidth;
		while(td->shift[1] > (texheight-1)) td->shift[1] -= texheight;
		while(td->shift[1] < 0) td->shift[1] += texheight;
		Brush_Build(fs->owner);
	}
}

void Select_SetTextureName (texdef_t *texdef)
{
	brush_t		*b;
	facesel_t	*f;
	face_t		*face;

	if (selected_faces)
	{
		for(f = selected_faces; f; f = f->next)
		{
			strncpy(f->face->texdef.name, texdef->name, 32);
			Brush_Build(f->owner);
		}
	}
	else
	{
		for (b=selected_brushes.next ; b != &selected_brushes ; b=b->next)
			if (!b->owner->eclass->fixedsize)
			{
				for(face = b->brush_faces; face; face = face->next)
					strncpy(face->texdef.name, texdef->name, 32);
			}
	}
	Sys_UpdateWindows (W_ALL);
}

void SelectFacesOfSelectedBrushes(void)
{
	brush_t		*b, *b2;
	face_t		*f;

	ClearSelectedFaces(&selected_faces);

	for(b = selected_brushes.next; b != &selected_brushes;)
	{
		for(f = b->brush_faces; f; f = f->next)
		{
			SelectFace(&selected_faces, f, b, 0.0f);
		}
		b2 = b;
		b = b->next;
		Brush_RemoveFromList(b2);
		Brush_AddToList(b2, &active_brushes);
	}

	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void SelectBrushesWithSelectedFaces(void)
{
	facesel_t	*face;

	Select_Deselect(false);

	while(selected_faces)
	{
		Brush_RemoveFromList(selected_faces->owner);
		Brush_AddToList(selected_faces->owner, &selected_brushes);
		face = selected_faces;
		selected_faces = selected_faces->next;
		free(face);
	}
	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}
#define CONTENTS_SKY  128
#define CONTENTS_CLIP 256
int ContentsForTexture(char *name)
{
	if(globals.gametype == HALFLIFE)
	{
		if (!Q_strncasecmp (name, "sky",3))
			return CONTENTS_SKY;
		
		if (!Q_strncasecmp(name+1,"!lava",5))
			return CONTENTS_LAVA;
		
		if (!Q_strncasecmp(name+1,"!slime",6))
			return CONTENTS_SLIME;
		
		if (!Q_strncasecmp (name, "!cur_90",7))
			return CONTENTS_CURRENT_90;
		
		if (!Q_strncasecmp (name, "!cur_0",6))
			return CONTENTS_CURRENT_0;
		
		if (!Q_strncasecmp (name, "!cur_270",8))
			return CONTENTS_CURRENT_270;
		
		if (!Q_strncasecmp (name, "!cur_180",8))
			return CONTENTS_CURRENT_180;
		
		if (!Q_strncasecmp (name, "!cur_up",7))
			return CONTENTS_CURRENT_UP;
		
		if (!Q_strncasecmp (name, "!cur_dwn",8))
			return CONTENTS_CURRENT_DOWN;
		
		if (name[0] == '!')
			return CONTENTS_WATER;
		
		if (!Q_strncasecmp (name, "origin",6))
			return CONTENTS_ORIGIN;
		
		if (!Q_strncasecmp (name, "clip",4))
			return CONTENTS_CLIP;
		
		if( !Q_strncasecmp( name, "translucent", 11 ) )
			return CONTENTS_TRANSLUCENT;
		
		if( name[0] == '@' )
			return CONTENTS_TRANSLUCENT;
	}
	else
	{
		if(name[0] == '*')
		{
			if (!Q_strncasecmp(name+1,"lava",4))
				return CONTENTS_LAVA;
			else if (!Q_strncasecmp(name+1,"slime",5))
				return CONTENTS_SLIME;
			else
				return CONTENTS_WATER;
		}
		else
			if(!Q_strncasecmp (name, "sky",3))
				return CONTENTS_SKY;
	}

	return CONTENTS_SOLID;
}

void Select_FindBadBrushes(void)
{
	brush_t	*b, *bnext;
	face_t	*f;
	int		contents;
	int		badbrush;

	Select_Deselect(true);

	for(b = active_brushes.next; b != &active_brushes; b = bnext)
	{
		badbrush = 0;
		bnext = b->next;
		if(globals.gametype == QUAKE2)
			contents = b->brush_faces->texdef.contents;
		else
			contents = ContentsForTexture(b->brush_faces->texdef.name);
		for(f = b->brush_faces; f && !badbrush; f = f->next)
		{
			if(globals.gametype == QUAKE2)
			{
				if(contents != f->texdef.contents)
					badbrush = 1;
			}
			else
			{
				if(contents != ContentsForTexture(f->texdef.name))
					badbrush = 1;
			}
		}
		if(badbrush)
		{
			Brush_RemoveFromList(b);
			Brush_AddToList(b, &selected_brushes);
		}
	}
	Sys_UpdateWindows(W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_EntitiesByKeyVal(char *key, char *value)
{
	brush_t	*b, *bnext;
	epair_t	*ep;
	int		found;

	Select_Deselect(true);

	for(b = active_brushes.next; b != &active_brushes; b = bnext)
	{
		bnext = b->next;
		found = 0;

		for (ep=b->owner->epairs ; ep && !found ; ep=ep->next)
		{
			if((!*key || !match(key, ep->key)) && (!value || !match(value, ep->value)))
			{
				Brush_RemoveFromList(b);
				Brush_AddToList(b, &selected_brushes);
				found = 1;
			}
		}
	}
	Sys_UpdateWindows(W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
}

void Select_AllBrushes(void)
{
	brush_t		*b, *next;

	for (b=active_brushes.next ; b != &active_brushes ; b=next)
	{
		next = b->next;
		Brush_RemoveFromList (b);
		Brush_AddToList (b, &selected_brushes);
	}
	Sys_UpdateWindows (W_ALL);
}
void Select_AllFaces(void)
{
	brush_t		*b;
	face_t		*f;

	for (b=active_brushes.next ; b != &active_brushes ; b=b->next)
	{
		for (f=b->brush_faces ; f ; f=f->next)
			SelectFace(&selected_faces, f, b, 0);
	}

	Sys_UpdateWindows (W_ALL);
}
void Select_InvertSelection(void)
{
	brush_t		*b/*, templist*/;
	facesel_t	*face = NULL;
	face_t		*f, *fnext;

#if 0
	// move the entire active_brushes list to templist
	templist.next = active_brushes.next;
	templist.prev = active_brushes.prev;
	templist.next->prev = &templist;
	templist.prev->next = &templist;

	// move the entire selected_brushes list to active_brushes
	active_brushes.next = selected_brushes.next;
	active_brushes.prev = selected_brushes.prev;
	active_brushes.next->prev = &active_brushes;
	active_brushes.prev->next = &active_brushes;

	// move the entire templist to selected_brushes
	selected_brushes.next = templist.next;
	selected_brushes.prev = templist.prev;
	selected_brushes.next->prev = &selected_brushes;
	selected_brushes.prev->next = &selected_brushes;
#endif

	for(b = active_brushes.next; b != &active_brushes; b = b->next)
	{
		Sys_Printf("Inverting face selection for active_brushes...\n");
		for(f = b->brush_faces; f; f = fnext)
		{
			fnext = f->next;
			if(FaceSelected(f))
				DeselectFace(f);
			else
				SelectFace(&selected_faces, f, b, 0.0f);
		}
	}
	for(b = selected_brushes.next; b != &selected_brushes; b = b->next)
	{
		Sys_Printf("Inverting face selection for selected_brushes...\n");
		for(f = b->brush_faces; f; f = fnext)
		{
			fnext = f->next;
			if(FaceSelected(f))
				DeselectFace(f);
			else
				SelectFace(&selected_faces, f, b, 0.0f);
		}
	}
	Sys_Printf("Finished\n");

	Sys_UpdateWindows (W_ALL);
}
