#include "qe3.h"
#include "textures.h"
/*
============================================================================

TEXTURE LAYOUT

============================================================================
*/

#define	FONT_HEIGHT	10

extern qtexture_t	*current_texture;
extern qboolean	texture_showinuse;
extern char texture_directory[32];	// use if texture_showinuse is false

int			current_x, current_y, current_row;
float		non_scaledheight = 0;
float		scaledheight = 0;
static BOOL	RightClickNoMove;
int			TotalPixelHeight = 0;

void Texture_StartPos (void)
{
	current_texture = globals.d_qtextures;
	current_x = 8;
	current_y = -8;
	current_row = 0;
	non_scaledheight = 0;
	scaledheight = 0;
}

qtexture_t *Texture_NextPos (int *x, int *y)
{
	qtexture_t	*q;

	while (1)
	{
		q = current_texture;
		if (!q)
			return q;
		current_texture = current_texture->next;
		if (q->name[0] == '(')	// fake color texture
			continue;
		if (q->inuse)
			break;			// allways show in use
		if (globals.gametype == QUAKE2 && !texture_showinuse && strnicmp (q->name, texture_directory, strlen(texture_directory)))
			continue;
		break;
	}

	if (current_x + q->image->width*globals.d_texturewin.scale > globals.d_texturewin.width-8 && current_row)
	{	// go to the next row unless the texture is the first on the row
		current_x = 8;
		current_y -= current_row + FONT_HEIGHT + 4;
		scaledheight += current_row;
		current_row = 0;
		non_scaledheight += FONT_HEIGHT + 4;
	}

	*x = current_x;
	*y = current_y;

	// Is our texture larger than the row? If so, grow the 
	// row height to match it

    if (current_row < q->image->height*globals.d_texturewin.scale)
		current_row = q->image->height*globals.d_texturewin.scale;

	// never go less than 64, or the names get all crunched up
	current_x += q->image->width*globals.d_texturewin.scale < 64 ? 64 : q->image->width*globals.d_texturewin.scale;
	current_x += 8;

	return q;
}

void Texture_GetTotalHeight(void)
{
	int	x = 0, y = 0;

	qtexture_t	*q;

	Texture_StartPos();

	while(1)
	{
		q = Texture_NextPos(&x, &y);
		if(!q)
		{
			TotalPixelHeight = abs(y) + 100;
			SetScrollRange(globals.Texture.hwnd, SB_VERT, 0, TotalPixelHeight, TRUE);
			return;
		}
	}
}

void ClampTextureOrigin(void)
{
	if(globals.d_texturewin.originy > 0)
		globals.d_texturewin.originy = 0;
	else if(globals.d_texturewin.originy < -TotalPixelHeight)
		globals.d_texturewin.originy = -TotalPixelHeight;
}

/*
============================================================================

  MOUSE ACTIONS

============================================================================
*/

static	int	textures_cursorx, textures_cursory;
void Select_SetTextureName (texdef_t *texdef);

void ScrollTextureWindow(int pixels, int buttons)
{
	int scale = 1;

	if ( buttons & MK_SHIFT )
		scale = 4;

	// rbutton = drag texture origin
	if (buttons & MK_RBUTTON)
	{
		globals.d_texturewin.originy += pixels * scale;
		if (globals.d_texturewin.originy > 0)
			globals.d_texturewin.originy = 0;
		Sys_UpdateWindows (W_TEXTURE);
		ClampTextureOrigin();
		return;
	}
}
/*
============
Texture_SetTexture

============
*/
void Texture_SetTexture (texdef_t *texdef, int apply, int onlytexturename)
{
	qtexture_t	*q;
	int			x,y;

	if (texdef->name[0] == '(')
	{
		Sys_Status(0, "Can't select an entity texture");
		return;
	}

	x = (int)SendMessage(globals.d_hwndToolbarTextureCombo, CB_FINDSTRINGEXACT, (WPARAM)-1, (LPARAM)texdef->name);

	if(x != CB_ERR)
		SendMessage(globals.d_hwndToolbarTextureCombo, CB_SETCURSEL, (WPARAM)x, (LPARAM)0);

	if(apply)
	{
		globals.d_texturewin.texdef = *texdef;
		if(onlytexturename)
			Select_SetTextureName(texdef);
		else
			Select_SetTexture(texdef);
	}
	Sys_UpdateWindows (W_TEXTURE);

// scroll origin so the texture is completely on screen
	Texture_StartPos ();
	while (1)
	{
		q = Texture_NextPos (&x, &y);
		if (!q)
			break;
		if (!strcmpi(texdef->name, q->name))
		{
			if (y > globals.d_texturewin.originy)
			{
				globals.d_texturewin.originy = y;
				Sys_UpdateWindows (W_TEXTURE);
				if(apply)
				{
					Sys_Status(1, "Selected texture: %s (%dx%d)\n", texdef->name, q->image->width, q->image->height);
				}
				return;
			}

			if (y-q->image->height*globals.d_texturewin.scale-2*FONT_HEIGHT < globals.d_texturewin.originy-globals.d_texturewin.height)
			{
				globals.d_texturewin.originy = y-q->image->height*globals.d_texturewin.scale-2*FONT_HEIGHT+globals.d_texturewin.height;
				Sys_UpdateWindows (W_TEXTURE);
				if(apply)
				{
					Sys_Status(1, "Selected texture: %s (%dx%d)\n", texdef->name, q->image->width, q->image->height);
				}
				return;
			}
			if(apply)
			{
				Sys_Status(1, "Selected texture: %s (%dx%d)\n", texdef->name, q->image->width, q->image->height);
			}

			return;
		}
	}
	if(apply)
	{
		Sys_Status(1, "Selected texture: %s (?x?)\n", texdef->name);
	}
}


/*
==============
SelectTexture

  By mouse click
==============
*/
void SelectTexture (int mx, int my, int onlytexturename)
{
	int		x, y;
	qtexture_t	*q;
	texdef_t	tex;

	my += globals.d_texturewin.originy-globals.d_texturewin.height;
	
	Texture_StartPos ();
	while (1)
	{
		q = Texture_NextPos (&x, &y);
		if (!q)
			break;
		if (mx > x && mx - x < q->image->width*globals.d_texturewin.scale
			&& my < y && y - my < q->image->height*globals.d_texturewin.scale + FONT_HEIGHT)
		{
			memset (&tex, 0, sizeof(tex));
			tex.scale[0] = 1;
			tex.scale[1] = 1;
			tex.flags = q->flags;
			tex.value = q->value;
			tex.contents = q->contents;
			strcpy (tex.name, q->name);
			Texture_SetTexture (&tex, true, onlytexturename);
			return;
		}
	}

	Sys_Status(0, "Did not select a texture\n");
}

/*
==============
GetTextureUnderMouse

  Sets texture window's caption to the name and size of the texture
  under the current mouse position.
==============
*/
void GetTextureUnderMouse (int mx, int my)
{
	int			x, y;
	qtexture_t	*q;

	my += globals.d_texturewin.originy-globals.d_texturewin.height;
	
	Texture_StartPos ();
	while (1)
	{
		q = Texture_NextPos (&x, &y);
		if (!q)
			break;
		if (mx > x && mx - x < q->image->width*globals.d_texturewin.scale
			&& my < y && y - my < q->image->height*globals.d_texturewin.scale + FONT_HEIGHT)
		{
			Sys_Status(0, "%s (%d x %d)", q->name, q->image->width, q->image->height);
			return;
		}
	}

	Sys_Status(0, "");
}

/*
==============
GetTextureZoomPos

  Find first visible texture with old scale and determine the y origin
  that will place it at the top with the new scale
==============
*/
void GetTextureZoomPos (float oldscale, float newscale)
{
	int			x, y;
	int			my = globals.d_texturewin.originy-globals.d_texturewin.height;
	qtexture_t	*q, *firsttexture;

	Texture_StartPos ();
	while (1)
	{
		q = firsttexture = Texture_NextPos (&x, &y);
		if (!q)
			break;
//		if ( (y-q->height*globals.d_texturewin.scale-FONT_HEIGHT < globals.d_texturewin.originy)
//			&& (y > globals.d_texturewin.originy - 64) )
		if(oldscale < newscale)
		{
			if ( (y-q->image->height*oldscale/*-FONT_HEIGHT*/ < globals.d_texturewin.originy) && (y > globals.d_texturewin.originy - 64) )
				break;
		}
		else
		{
			if ( (y-q->image->height*oldscale-FONT_HEIGHT < globals.d_texturewin.originy) && (y > globals.d_texturewin.originy - 64) )
				break;
		}
	}
	if(!q)
		return;

	globals.d_texturewin.scale = newscale;
	Texture_StartPos();

	while(1)
	{
		q = Texture_NextPos (&x, &y);

		if(!q)
			return;

		if (q == firsttexture)
			globals.d_texturewin.originy = y;
	}
}
/*
==============
Texture_MouseDown
==============
*/
void Texture_MouseDown (int x, int y, int buttons)
{
	Sys_GetCursorPos (&textures_cursorx, &textures_cursory);

	// lbutton = select texture
	if (buttons == MK_LBUTTON )
	{
		SelectTexture (x, globals.d_texturewin.height - 1 - y, false);
		return;
	}
	if (buttons == (MK_LBUTTON | MK_CONTROL) )
	{
		SelectTexture (x, globals.d_texturewin.height - 1 - y, true);
		return;
	}
	RightClickNoMove = true;

}

/*
==============
Texture_MouseUp
==============
*/
void Texture_MouseUp (int x, int y, int buttons)
{
	if (RightClickNoMove)
	{
		Sys_Status(0, "Right-mouse no move");
	}
}

/*
==============
Texture_MouseMoved
==============
*/
void Texture_MouseMoved (int x, int y, int buttons)
{
	float scale = 1;

	if ( buttons & MK_SHIFT )
		scale = 4;

	//rbutton+control = zoom texture view
	if (buttons == (MK_CONTROL|MK_RBUTTON))
	{
		Sys_GetCursorPos (&x, &y);
		if ( y != textures_cursory)
		{
			int		i;
			float	oldscale, newscale;

			oldscale = newscale = globals.d_texturewin.scale;
			scale = 1;
			for(i = abs(y-textures_cursory); i > 0; i--)
				scale *=  y>textures_cursory?1.008:0.992063492;
			newscale *= scale;
			if (newscale > 8)
				newscale = 8;
			if (newscale < 0.125)
				newscale = (float)0.125;
			Sys_SetCursorPos (textures_cursorx, textures_cursory);
			Sys_Status(0, "Texture scale: %3.0f%%", newscale*100.0f);
//			globals.d_texturewin.originy = (globals.d_texturewin.originy-non_scaledheight) * (1/globals.d_texturewin.scale) + non_scaledheight;
			GetTextureZoomPos (oldscale, newscale);
			Sys_UpdateWindows (W_TEXTURE);
		}
		return;
	}

	// rbutton = drag texture origin
	if (buttons & MK_RBUTTON)
	{
		RightClickNoMove = false;
		Sys_GetCursorPos (&x, &y);
		if ( y != textures_cursory)
		{
			globals.d_texturewin.originy += ( y-textures_cursory) * scale;
			if (globals.d_texturewin.originy > 0)
				globals.d_texturewin.originy = 0;
			Sys_SetCursorPos (textures_cursorx, textures_cursory);
			Sys_UpdateWindows (W_TEXTURE);
		}
	}
	else
		GetTextureUnderMouse(x, globals.d_texturewin.height - 1 - y);

}


/*
============================================================================

DRAWING

============================================================================
*/

int imax(int iFloor, int i) { if (i>iFloor) return iFloor; return i; }
HFONT ghFont = NULL;
void RS_DrawQuad(rscript_t *rs, float x, float y, float h, float v);

/*
============
Texture_Draw2
============
*/
void Texture_Draw2 (int width, int height)
{
	qtexture_t	*q;
	int			x, y;
	char		*name;

	glClearColor (
		globals.d_savedinfo.colors[COLOR_TEXTUREBACK][0],
		globals.d_savedinfo.colors[COLOR_TEXTUREBACK][1],
		globals.d_savedinfo.colors[COLOR_TEXTUREBACK][2],
		0);
	glViewport (0,0,width,height);
	glClear (GL_COLOR_BUFFER_BIT);
	glDisable (GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();
	glOrtho (0, width, globals.d_texturewin.originy-height, globals.d_texturewin.originy, -100, 100);
	glEnable (GL_TEXTURE_2D);

	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	globals.d_texturewin.width = width;
	globals.d_texturewin.height = height;
	Texture_StartPos ();

	while (1)
	{
		q = Texture_NextPos (&x, &y);
		if (!q)
			break;

		// Is this texture visible?
		if ( (y-q->image->height*globals.d_texturewin.scale-FONT_HEIGHT < globals.d_texturewin.originy)
			&& (y > globals.d_texturewin.originy - height) )
		{

			// if in use, draw a background
			if (q->inuse && !texture_showinuse)
			{
				glLineWidth (1);
				glColor3f (0.5,1,0.5);
				glDisable (GL_TEXTURE_2D);

				glBegin (GL_LINE_LOOP);
				glVertex2f (x-1,y+1-FONT_HEIGHT);
				glVertex2f (x-1,y-q->image->height*globals.d_texturewin.scale-1-FONT_HEIGHT);
				glVertex2f (x+1+q->image->width*globals.d_texturewin.scale,y-q->image->height*globals.d_texturewin.scale-1-FONT_HEIGHT);
				glVertex2f (x+1+q->image->width*globals.d_texturewin.scale,y+1-FONT_HEIGHT);
				glEnd ();

				glEnable (GL_TEXTURE_2D);
			}

			// Draw the texture
			if(q->script)
			{
				RS_DrawQuad(q->script, x, y - FONT_HEIGHT, x+q->image->width*globals.d_texturewin.scale, y-FONT_HEIGHT-q->image->height*globals.d_texturewin.scale);
			}
			else
			{
				glColor3f (1,1,1);
				glBindTexture( GL_TEXTURE_2D, q->image->texture_number );
				glBegin (GL_QUADS);
				glTexCoord2f (0,0);
				glVertex2f (x,y-FONT_HEIGHT);
				glTexCoord2f (1 , 0);
				glVertex2f (x+q->image->width*globals.d_texturewin.scale,y-FONT_HEIGHT);
				glTexCoord2f (1 , 1);
				glVertex2f (x+q->image->width*globals.d_texturewin.scale,y-FONT_HEIGHT-q->image->height*globals.d_texturewin.scale);
				glTexCoord2f (0 , 1);
				glVertex2f (x,y-FONT_HEIGHT-q->image->height*globals.d_texturewin.scale);
				glEnd ();
			}

			// draw the selection border
			if (!strcmpi(globals.d_texturewin.texdef.name, q->name))
			{
				glLineWidth (3);
				glColor3f (1,0,0);
				glDisable (GL_TEXTURE_2D);

				glBegin (GL_LINE_LOOP);
				glVertex2f (x-4,y-FONT_HEIGHT+4);
				glVertex2f (x-4,y-FONT_HEIGHT-q->image->height*globals.d_texturewin.scale-4);
				glVertex2f (x+4+q->image->width*globals.d_texturewin.scale,y-FONT_HEIGHT-q->image->height*globals.d_texturewin.scale-4);
				glVertex2f (x+4+q->image->width*globals.d_texturewin.scale,y-FONT_HEIGHT+4);
				glEnd ();

				glEnable (GL_TEXTURE_2D);
				glLineWidth (1);
			}
			if (q->script)// && q->shader->type != SHADER_DISKFILE)
			{
				glLineWidth (1);
				glColor3dv(globals.d_savedinfo.colors[COLOR_RSCRIPT_OUTLINE]);
				glDisable (GL_TEXTURE_2D);

				glBegin (GL_LINE_LOOP);
				glVertex2f (x-2,y-FONT_HEIGHT+2);
				glVertex2f (x-2,y-FONT_HEIGHT-q->image->height*globals.d_texturewin.scale-2);
				glVertex2f (x+2+q->image->width*globals.d_texturewin.scale,y-FONT_HEIGHT-q->image->height*globals.d_texturewin.scale-2);
				glVertex2f (x+2+q->image->width*globals.d_texturewin.scale,y-FONT_HEIGHT+2);
				glEnd ();

				glEnable (GL_TEXTURE_2D);
			}

			// draw the texture name
			if(globals.ShowTextureNames)
			{
				glColor3dv(globals.d_savedinfo.colors[COLOR_TEXTURETEXT]);
				glDisable (GL_TEXTURE_2D);
				
				// don't draw the directory name
				for (name = q->name ; *name && *name != '/' && *name != '\\' ; name++);
				if (!*name)
					name = q->name;
				else
					name++;
				
				glPrintf(x, y-FONT_HEIGHT+2, 0, "%s", name);
				glEnable (GL_TEXTURE_2D);
			}
		}
	}

	TotalPixelHeight = abs(y) + 100;
	SetScrollRange(globals.Texture.hwnd, SB_VERT, 0, TotalPixelHeight, TRUE);
	// reset the current texture
	glBindTexture( GL_TEXTURE_2D, 0 );
	glFinish();
}

/*
============
WTexWndProc
============
*/
LONG WINAPI WTex_WndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
	int		xPos, yPos;
    RECT	rect;

    GetClientRect(hWnd, &rect);

    switch (uMsg)
    {
	case WM_CREATE:
		return 0;

	case WM_DESTROY:
		wglMakeCurrent( NULL, NULL );
		wglDeleteContext( globals.Texture.hglrc );
		ReleaseDC( hWnd, globals.Texture.hdc );
		return 0;

	case WM_PAINT:
        { 
		    PAINTSTRUCT	ps;

//			Texture_GetTotalHeight();
			ClampTextureOrigin();
			SetScrollPos(globals.Texture.hwnd, SB_VERT, -globals.d_texturewin.originy, TRUE);
		    BeginPaint(hWnd, &ps);

            if ( !wglMakeCurrent( globals.Texture.hdc, globals.Texture.hglrc ) )
				Error ("wglMakeCurrent failed");
			Texture_Draw2 (rect.right-rect.left, rect.bottom-rect.top);
			SwapBuffers(globals.Texture.hdc);

		    EndPaint(hWnd, &ps);
        }
		return 0;

	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_LBUTTONDOWN:
		SetCapture( globals.Texture.hwnd );
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		
		Texture_MouseDown (xPos, yPos, wParam);
		return 0;

	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	case WM_LBUTTONUP:
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		
		Texture_MouseUp (xPos, yPos, wParam);
		if (! (wParam & (MK_LBUTTON|MK_RBUTTON|MK_MBUTTON)))
			ReleaseCapture ();
		return 0;

	case WM_MOUSEMOVE:
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		
		Texture_MouseMoved (xPos, yPos, wParam);
		return 0;

	case WM_VSCROLL:
		{
			int	pos;

			switch(LOWORD(wParam))
			{
				case SB_BOTTOM:
					break;				
				
				case SB_ENDSCROLL:
					break;
				
				case SB_LINEDOWN:
					pos = --globals.d_texturewin.originy;
					break;
				
				case SB_LINEUP:
					pos = ++globals.d_texturewin.originy;
					break;
				
				case SB_PAGEDOWN:
					globals.d_texturewin.originy -= globals.d_texturewin.height;
					pos = globals.d_texturewin.originy;
					break;
				
				case SB_PAGEUP:
					globals.d_texturewin.originy += globals.d_texturewin.height;
					pos = globals.d_texturewin.originy;
					break;
				
				case SB_THUMBPOSITION:
				case SB_THUMBTRACK:
					{
						SCROLLINFO si;

						ZeroMemory(&si, sizeof(si));
						si.cbSize = sizeof(si);
						si.fMask = SIF_TRACKPOS;
						GetScrollInfo(hWnd, SB_VERT, &si);
						pos = globals.d_texturewin.originy = -si.nTrackPos;
						break;
					}
				
				case SB_TOP:
					break;
			}
		InvalidateRect(globals.Texture.hwnd, NULL, false);
		return 0;
		}
	}

    return DefWindowProc (hWnd, uMsg, wParam, lParam);
}



/*
==================
CreateTextureWindow

We need to create a seperate window for the textures
in the inspector window, because we can't share
gl and gdi drawing in a single window
==================
*/
#define	TEXTURE_WINDOW_CLASS	"QTEX"
void CreateTextureWindow (void)
{
    WNDCLASS   wc;

    /* Register the camera class */
	memset (&wc, 0, sizeof(wc));

    wc.style         = 0;
    wc.lpfnWndProc   = (WNDPROC)WTex_WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = globals.d_hInstance;
    wc.hIcon         = 0;
    wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = 0;
    wc.lpszClassName = TEXTURE_WINDOW_CLASS;

    if (!RegisterClass (&wc) )
        Error ("WCam_Register: failed");

	globals.Texture.hwnd = CreateWindow (TEXTURE_WINDOW_CLASS ,
		"Texture View",
		WS_BORDER|WS_CHILD|WS_VISIBLE|WS_VSCROLL,
		20,
		20,
		64,
		64,	// size

		globals.Entity.hwnd,	// parent window
		0,		// no menu
		globals.d_hInstance,
		0);
	if (!globals.Texture.hwnd)
		Error ("Couldn't create texturewindow");

	globals.Texture.hdc = GetDC(globals.Texture.hwnd);
	QEW_SetupPixelFormat(&(globals.Texture.pfd), &(globals.Texture.hdc), false);

	if ( ( globals.Texture.hglrc = wglCreateContext( globals.Texture.hdc ) ) == 0 )
		Error( "wglCreateContext in WTex_WndProc failed" );

	if (!wglMakeCurrent( globals.Texture.hdc, globals.Texture.hglrc ))
		Error ("wglMakeCurrent in WTex_WndProc failed");

	if (!wglShareLists(globals.Main.hglrc, globals.Texture.hglrc ) )
		Error( "wglShareLists in CreateTextureWindow failed" );
//	SetUpFont(&(globals.Texture));

}
