#include "qe3.h"
#include "io.h"
#include "inc.h"
#include "directory.h"
#include "mrg/gl_cin.h"
#include <il/il.h>
#define	TYP_MIPTEX	68

static unsigned		tex_palette[256];
static byte			gammatable[256];
qtexture_t			*notexture;
static qboolean		nomips;
int					MaxTextureSize;

//int		texture_mode = GL_NEAREST;
//int		texture_mode = GL_NEAREST_MIPMAP_NEAREST;
//int		texture_mode = GL_NEAREST_MIPMAP_LINEAR;
//int		texture_mode = GL_LINEAR;
//int		texture_mode = GL_LINEAR_MIPMAP_NEAREST;
int		texture_mode = GL_LINEAR_MIPMAP_LINEAR;

// current active texture directory.  if empty, show textures in use
char		texture_directory[32];	// use if texture_showinuse is false
qboolean	texture_showinuse;

// texture layout functions
qtexture_t	*current_texture;
int			texture_nummenus = 0;
#define		MAX_TEXTUREDIRS	100
char		texture_menunames[MAX_TEXTUREDIRS][64];

qboolean	g_dontuse;		// set to true to load the texture but not flag as used
GLuint		skybox_textures[6];
GLuint		skybox_display_list = 0;

void SelectTexture (int mx, int my, int onlytexturename);

void	Texture_MouseDown (int x, int y, int buttons);
void	Texture_MouseUp (int x, int y, int buttons);
void	Texture_MouseMoved (int x, int y, int buttons);

//=====================================================

void SortTextures(void)
{	
	qtexture_t	*q, *qtemp, *qhead, *qcur, *qprev;

	// standard insertion sort
	// Take the first texture from the list and
	// add it to our new list
	if ( globals.d_qtextures == NULL)
		return;	

	qhead = globals.d_qtextures;
	q = globals.d_qtextures->next;
	qhead->next = NULL;
	
	// while there are still things on the old
	// list, keep adding them to the new list
	while (q)
	{
		qtemp = q;
		q = q->next;
		
		qprev = NULL;
		qcur = qhead;

		while (qcur)
		{
			// Insert it here?
			if (strcmp(qtemp->name, qcur->name) < 0)
			{
				qtemp->next = qcur;
				if (qprev)
					qprev->next = qtemp;
				else
					qhead = qtemp;
				break;
			}
			
			// Move on

			qprev = qcur;
			qcur = qcur->next;


			// is this one at the end?

			if (qcur == NULL)
			{
				qprev->next = qtemp;
				qtemp->next = NULL;
			}
		}


	}

	globals.d_qtextures = qhead;
}

//=====================================================


/*
==============
Texture_InitPalette
==============
*/
void Texture_InitPalette (byte *pal)
{
    int		r,g,b,v;
    int		i;
	int		inf;
	float	gamma;

	gamma = globals.d_savedinfo.fGamma;

	if (gamma == 1.0)
	{
		for (i=0 ; i<256 ; i++)
			gammatable[i] = i;
	}
	else
	{
		for (i=0 ; i<256 ; i++)
		{
			inf = 255 * pow ( (i+0.5)/255.5 , gamma ) + 0.5;
			if (inf < 0)
				inf = 0;
			if (inf > 255)
				inf = 255;
			gammatable[i] = inf;
		}
	}

	if(pal)
	{
		for (i=0 ; i<256 ; i++)
		{
			r = gammatable[pal[0]];
			g = gammatable[pal[1]];
			b = gammatable[pal[2]];
			pal += 3;
			
			v = (r<<24) + (g<<16) + (b<<8) + 255;
			v = BigLong (v);
			
			tex_palette[i] = v;
		}
	}
}

void SetTexParameters (void)
{
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, texture_mode );
	
	switch ( texture_mode )
	{
	case GL_NEAREST:
	case GL_NEAREST_MIPMAP_NEAREST:
	case GL_NEAREST_MIPMAP_LINEAR:
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
		break;
	case GL_LINEAR:
	case GL_LINEAR_MIPMAP_NEAREST:
	case GL_LINEAR_MIPMAP_LINEAR:
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		break;
	}
}

/*
============
Texture_SetMode
============
*/
void Texture_SetMode(int iMenu)
{
	int			iMode;
	HMENU		hMenu;
	qboolean	texturing = true;
	image_t		*image;

	hMenu = GetMenu(globals.d_hwndMain);
	
	switch(iMenu)
	{
		case ID_VIEW_NEAREST:
			iMode = GL_NEAREST;
			break;
		case ID_VIEW_NEARESTMIPMAP:
			iMode = GL_NEAREST_MIPMAP_NEAREST;
			break;
		case ID_VIEW_LINEAR:
			iMode = GL_NEAREST_MIPMAP_LINEAR;
			break;
		case ID_VIEW_BILINEAR:
			iMode = GL_LINEAR;
			break;
		case ID_VIEW_BILINEARMIPMAP:
			iMode = GL_LINEAR_MIPMAP_NEAREST;
			break;
		case ID_VIEW_TRILINEAR:
			iMode = GL_LINEAR_MIPMAP_LINEAR;
			break;
		
		case ID_TEXTURES_WIREFRAME:
			iMode = 0;
			texturing = false;
			break;
		
		case ID_TEXTURES_FLATSHADE:
			iMode = 0;
			texturing = false;
			break;
	}
	
	CheckMenuItem(hMenu, ID_VIEW_NEAREST, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_VIEW_NEARESTMIPMAP, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_VIEW_LINEAR, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_VIEW_BILINEARMIPMAP, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_VIEW_BILINEAR, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_VIEW_TRILINEAR, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_TEXTURES_WIREFRAME, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_TEXTURES_FLATSHADE, MF_BYCOMMAND | MF_UNCHECKED);
	
	CheckMenuItem(hMenu, iMenu, MF_BYCOMMAND | MF_CHECKED);
	
	globals.d_savedinfo.iTexMenu = iMenu;
	texture_mode = iMode;
	
	if (!wglMakeCurrent( globals.Main.hdc, globals.Main.hglrc ))
		Error ("wglMakeCurrent failed");
	
	if ( texturing )
		SetTexParameters ();
	
	if ( !texturing && iMenu == ID_TEXTURES_WIREFRAME)
	{
		camera.draw_mode = cd_wire;
		Map_BuildBrushData();
		Sys_UpdateWindows (W_ALL);
		return;
	}
	else if ( !texturing && iMenu == ID_TEXTURES_FLATSHADE)
	{
		camera.draw_mode = cd_solid;
		Map_BuildBrushData();
		Sys_UpdateWindows (W_ALL);
		return;
	}
	
	for (image = globals.images ; image ; image = image->next)
	{
		glBindTexture( GL_TEXTURE_2D, image->texture_number );
		SetTexParameters ();
	}
	
	// select the default texture
	glBindTexture( GL_TEXTURE_2D, 0 );
	
	glFinish();
	
	if (camera.draw_mode != cd_texture)
	{
		camera.draw_mode = cd_texture;
		Map_BuildBrushData();
	}
	
	Sys_UpdateWindows (W_ALL);
}

void ResampleImage(int *width, int *height, unsigned **dest)
{
	unsigned	*scaledimage;
	int scaledwidth, scaledheight;

	for(scaledwidth = 1; scaledwidth < *width; scaledwidth <<= 1);
	for(scaledheight = 1; scaledheight < *height; scaledheight <<= 1);

	if(scaledwidth != *width || scaledheight != *height)
	{
		scaledimage = qmalloc (scaledwidth * scaledheight * 4);

		gluScaleImage(GL_RGBA, *width, *height, GL_UNSIGNED_BYTE, *dest, scaledwidth, scaledheight, GL_UNSIGNED_BYTE, scaledimage);
		free(*dest);
		*dest = scaledimage;
		*width = scaledwidth;
		*height = scaledheight;
	}
}

BOOL CheckTextureSize(int width, int height, char *name)
{
	if(width > MaxTextureSize)
	{
		if(globals.d_savedinfo.no_console_texture != true)
			Sys_Printf("Texture %s is wider than %d pixels (%dx%d). Ignoring...\n", name, MaxTextureSize, width, height);
		return false;
	}
	if(height > MaxTextureSize)
	{
		if(globals.d_savedinfo.no_console_texture != true)
			Sys_Printf("Texture %s is taller than %d pixels (%dx%d). Ignoring...\n", name, MaxTextureSize, width, height);
		return false;
	}
	if(width < 1)
	{
		if(globals.d_savedinfo.no_console_texture != true)
			Sys_Printf("Texture %s is less than 1 pixel wide (%dx%d). Ignoring...\n", name, width, height);
		return false;
	}
	if(height < 1)
	{
		if(globals.d_savedinfo.no_console_texture != true)
			Sys_Printf("Texture %s is less than 1 pixel tall (%dx%d). Ignoring...\n", name, width, height);
		return false;
	}

	return true;
}

void GL_Upload(int width, int height, BYTE *source, image_t *image, int usepalette)
{
	int			total[3], i, i2, count = width*height;
	unsigned	*dest = qmalloc(width*height*4);

	// The dib is upside down so we want to copy it into 
	// the buffer bottom up.

	total[0] = total[1] = total[2] = 0;

	if(usepalette)
	{
		for (i=0 ; i<count ; i++)
		{
			dest[i] = tex_palette[source[i]];
			
			total[0] += ((byte *)(dest+i))[0];
			total[1] += ((byte *)(dest+i))[1];
			total[2] += ((byte *)(dest+i))[2];
		}
	}
	else
	{
		for (i=0 ; i<count ; i++)
		{
			for(i2 = 0; i2 < 3; i2++)
			{
				((byte *)(dest+i))[i2] = gammatable[source[i*4+i2]];
				total[i2] += ((byte *)(dest+i))[i2];
			}
			((byte *)(dest+i))[3] = source[i*4+3];
		}
	}

	image->color[0] = (float)total[0]/(count*255);
	image->color[1] = (float)total[1]/(count*255);
	image->color[2] = (float)total[2]/(count*255);
	image->width = width;
	image->height = height;

	if (!wglMakeCurrent( globals.Texture.hdc, globals.Texture.hglrc ))
		Error ("wglMakeCurrent failed");

	glBindTexture( GL_TEXTURE_2D, image->texture_number );
	SetTexParameters ();

	ResampleImage(&width, &height, &dest);

	if (nomips)
		glTexImage2D(GL_TEXTURE_2D, 0, 4, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, dest);
	else
		gluBuild2DMipmaps(GL_TEXTURE_2D, 4, width, height,GL_RGBA, GL_UNSIGNED_BYTE, dest);
	free (dest);
	glBindTexture( GL_TEXTURE_2D, 0 );
}

image_t *FindImage(char *name)
{
	image_t	*image;

	for(image = globals.images; image; image = image->next)
	{
		if(!stricmp(image->name, name))
			return image;
	}
	return NULL;
}

image_t *NewImage(char *name, int is_cin)
{
	image_t *image;

	image = malloc(sizeof(image_t));
	image->next = globals.images;
	globals.images = image;
	image->name = strdup(name);

	return image;
}

image_t *Image_LoadWal(char *name, miptex_t *qtex, int *width, int *height)
{
	byte		*source;
	image_t		*image;

	*width = LittleLong(qtex->width);
	*height = LittleLong(qtex->height);

	if(!CheckTextureSize(*width, *height, qtex->name))
		return NULL;

	image = NewImage(name, false);
    source = (byte *)qtex + LittleLong(qtex->offsets[0]);
	glGenTextures (1, &image->texture_number);
	GL_Upload(*width, *height, source, image, 1);

	return image;
}

/*
=================
Texture_LoadWAL
=================
*/
qtexture_t *Texture_LoadWAL (miptex_t *qtex, char *name)
{
	int			width, height;
	qtexture_t	*q;

	q = qmalloc(sizeof(*q));
	q->flags = qtex->flags;
	q->value = qtex->value;
	q->contents = qtex->contents;
	q->image = FindImage(name);

	if(q->image)
		return q;

	q->image = Image_LoadWal(name, qtex, &width, &height);
    q->image->width = width;
    q->image->height = height;

    return q;
}

/*
=================
Texture_LoadM32
=================
*/
qtexture_t *Texture_LoadM32 (miptex32_t *qtex, char *name)
{
	byte		*source;
	int			width, height;
	qtexture_t	*q;
    
    width = qtex->width[0];
    height = qtex->height[0];

	if(!CheckTextureSize(width, height, qtex->name))
		return NULL;

	q = qmalloc(sizeof(*q));
	q->flags = qtex->flags;
	q->value = qtex->value;
	q->contents = qtex->contents;
	q->image = FindImage(name);

	if(q->image)
		return q;

	q->image = NewImage(name, false);
    q->image->width = width;
    q->image->height = height;
    source = (byte *)qtex + qtex->offsets[0];
	glGenTextures (1, &q->image->texture_number);
	GL_Upload(width, height, source, q->image, 0);

    return q;
}

/*
===============
Texture_CreateSolid

Create a single pixel texture of the apropriate color
===============
*/
qtexture_t *Texture_CreateSolid (char *name)
{
	byte	data[4];
	qtexture_t	*q;
	GLuint		texture;

    q = qmalloc(sizeof(*q));
	q->image = NewImage(name, false);
	
	sscanf (name, "(%f %f %f)", &q->image->color[0], &q->image->color[1], &q->image->color[2]);

	data[0] = q->image->color[0]*255;
	data[1] = q->image->color[1]*255;
	data[2] = q->image->color[2]*255;
	data[3] = 255;

	q->image->width = q->image->height = 1;
	glGenTextures (1, &texture);
	q->image->texture_number = texture;
	glBindTexture( GL_TEXTURE_2D, q->image->texture_number );
	SetTexParameters ();

	if (nomips)
		glTexImage2D(GL_TEXTURE_2D, 0, 3, 1, 1, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	else
		gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 1, 1,GL_RGBA, GL_UNSIGNED_BYTE, data);

	glBindTexture( GL_TEXTURE_2D, 0 );

	return q;
}


/*
=================
Texture_MakeNotexture
=================
*/
void Texture_MakeNotexture (void)
{
    qtexture_t	*q;
    byte		data[4][4];
	GLuint		textures;

	if(globals.Camera_Initialized == false)
		return;
	if (!wglMakeCurrent( globals.Camera.hdc, globals.Camera.hglrc ))
		Error ("wglMakeCurrent failed in WCam_WndProc():WM_CREATE");
	notexture = q = qmalloc(sizeof(*q));
	strcpy (q->name, "notexture");
	q->image = NewImage("notexture", false);
    q->image->width = q->image->height = 64;
    
	memset (data, 0, sizeof(data));
	data[0][2] = data[3][2] = 255;

	q->image->color[0] = 0;
	q->image->color[1] = 0;
	q->image->color[2] = 0.5;

	glGenTextures (1, &textures);
	q->image->texture_number = textures;
	q->script = NULL;
	glBindTexture( GL_TEXTURE_2D, q->image->texture_number );
	SetTexParameters ();

	if (nomips)
		glTexImage2D(GL_TEXTURE_2D, 0, 3, 2, 2, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	else
		gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 2, 2,GL_RGBA, GL_UNSIGNED_BYTE, data);

	glBindTexture( GL_TEXTURE_2D, 0 );
}

BOOL Texture_LoadFile(byte **lump, char *name)
{
	char		filename[512];

//	if(!name || !*name || !stricmp(name, "textures/"))
//		return false;

	if(globals.gametype == QUAKE2)
	{
		if(!strnicmp(name, "textures/", 9))
			setdirstr(filename, name+9, ".wal", NULL);
		else
			setdirstr(filename, name, ".wal", NULL);

		if(globals.d_savedinfo.no_console_texture != true)
			Sys_Printf ("Loading %s\n", name);

		if (qLoadTextureFile (filename, lump) == -1)
		{
			if(globals.d_savedinfo.no_console_texture != true)
				Sys_Printf ("     load failed!\n");

			return false;
		}
	}
	else if(globals.gametype == SOF)
	{
		setdirstr(filename, name, ".m32", NULL);

		if(globals.d_savedinfo.no_console_texture != true)
			Sys_Printf ("Loading %s\n", name);

		if (qLoadTextureFile (filename, lump) == -1)
		{
			if(globals.d_savedinfo.no_console_texture != true)
				Sys_Printf ("     load failed!\n");

			return false;
		}
	}
	else
	{
		if(globals.d_savedinfo.no_console_texture != true)
			Sys_Printf ("Loading %s\n", name);

		if(globals.gametype == HALFLIFE)
		{
			byte temp_palette[768];

			if(!WadLoadFile(name, lump, temp_palette))
			{
				if(globals.d_savedinfo.no_console_texture != true)
					Sys_Printf ("     load failed!\n");
				return false;
			}
			else
			{
				int		i, r, g, b, v;
				byte	*pal = (byte *)temp_palette;
		
				for (i=0 ; i<256 ; i++)
				{
					r = gammatable[pal[0]];
					g = gammatable[pal[1]];
					b = gammatable[pal[2]];
					v = (r<<24) + (g<<16) + (b<<8) + 255;
					v = BigLong (v);
					tex_palette[i] = v;
					pal += 3;
				}
			}
		}
		else
		{
			if(!WadLoadFile(name, lump, NULL))
			{
				if(globals.d_savedinfo.no_console_texture != true)
					Sys_Printf ("     load failed!\n");
				return false;
			}
		}
	}
	return true;
}

qtexture_t *LoadTextureFromDisk(char *name);

/*
===============
Texture_ForName
===============
*/
qtexture_t *Texture_ForName (char *name)
{
    byte		*lump;
	qtexture_t	*q;

	if(!name || !*name)
		return notexture;

	for (q=globals.d_qtextures ; q ; q=q->next)
    {
		if (!stricmp(name,  q->name))
		{
			if (!g_dontuse)
				q->inuse = true;
			strcpy(name, q->name);
		    return q;
		}
    }

	if (name[0] == '(')
	{
		q = Texture_CreateSolid (name);
		strncpy (q->name, name, sizeof(q->name)-1);
	}
	else
	{
		// load the file
		
		if(globals.UseExternalTextures && /*(globals.gametype == QUAKE1 || globals.gametype == HALFLIFE) &&*/ (q = LoadTextureFromDisk(name)))
			goto breakout;

		if(!Texture_LoadFile(&lump, name))
			return notexture;

		switch(globals.gametype)
		{
			case QUAKE2:
				q = Texture_LoadWAL ((miptex_t *)lump, name);
				break;
			case SOF:
				q = Texture_LoadM32 ((miptex32_t *)lump, name);
				break;
			default:
				q = Texture_LoadMip ((miptex_t1 *)lump, name);
				break;
		}
		free (lump);
		if(!q)
			return notexture;
		strncpy (q->name, name, sizeof(q->name)-1);
		StripExtension (q->name);
		setdirstr2(q->name, q->name, NULL);
	}

breakout:
	if (!g_dontuse)
		q->inuse = true;
	q->next = globals.d_qtextures;
	globals.d_qtextures = q;

    return q;
}

int LoadSkyboxElement(char *skyname, char *section, GLuint *texturenumber)
{
	int			width, height;
	byte		*imagebuf;
	char		buf[256];
	
	if(globals.gametype == HALFLIFE)
		setdirstr(buf, "gfx/env/", skyname, section, ".tga", NULL);
	else
		setdirstr(buf, "env/", skyname, section, ".tga", NULL);

	ilStart();

	if(!LoadTGA(buf, &imagebuf, &width, &height))
		return false;

	glGenTextures(1, texturenumber);

	if (!wglMakeCurrent( globals.Texture.hdc, globals.Texture.hglrc ))
		Error ("wglMakeCurrent failed");

	glBindTexture( GL_TEXTURE_2D, *texturenumber );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );

	glTexImage2D(GL_TEXTURE_2D, 0, 4, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, imagebuf);
	glBindTexture( GL_TEXTURE_2D, 0 );
	ilEnd();
	return true;
}

void LoadSkybox(void)
{
	char	*skyname = map_skyname;
	int		i=0, j;
	vec3_t sides[6][4] = {
		{ {  16,-16,-16 }, { 16, 16,-16 }, { 16, 16, 16 }, { 16,-16, 16 } }, // right
		{ { -16, 16,-16 }, {-16,-16,-16 }, {-16,-16, 16 }, {-16, 16, 16 } }, // left
		{ {  16, 16,-16 }, {-16, 16,-16 }, {-16, 16, 16 }, { 16, 16, 16 } }, // back
		{ { -16,-16,-16 }, { 16,-16,-16 }, { 16,-16, 16 }, {-16,-16, 16 } }, // front
		{ {  16,-16, 16 }, { 16, 16, 16 }, {-16, 16, 16 }, {-16,-16, 16 } }, // top
		{ { -16,-16,-16 }, {-16, 16,-16 }, { 16, 16,-16 }, { 16,-16,-16 } }  // bottom
	};
	float uv[4][2] = {{ 0, 0 }, {-1, 0 }, {-1,-1 }, { 0,-1 }};

	Sys_Printf("Loading sky: %s\n", skyname);

	if(skybox_textures[0])
	{
		glDeleteTextures(6, skybox_textures);
		glDeleteLists(skybox_display_list, 1);
	}

	skybox_textures[0] = skybox_textures[1] = skybox_textures[2] = skybox_textures[3] = skybox_textures[4] = skybox_textures[5] = 0;

	if(!LoadSkyboxElement(skyname, "rt", &skybox_textures[i++]))
		goto deinitialize;
	if(!LoadSkyboxElement(skyname, "lf", &skybox_textures[i++]))
		goto deinitialize;
	if(!LoadSkyboxElement(skyname, "bk", &skybox_textures[i++]))
		goto deinitialize;
	if(!LoadSkyboxElement(skyname, "ft", &skybox_textures[i++]))
		goto deinitialize;
	if(!LoadSkyboxElement(skyname, "up", &skybox_textures[i++]))
		goto deinitialize;
	if(!LoadSkyboxElement(skyname, "dn", &skybox_textures[i++]))
		goto deinitialize;

	if (!skybox_display_list)
		skybox_display_list = glGenLists(1);

    glNewList (skybox_display_list,  GL_COMPILE);
	for(i = 0; i < 6; i++)
	{
		glBindTexture(GL_TEXTURE_2D, skybox_textures[i]);
		glBegin(GL_QUADS);
		for(j = 0; j < 4; j++) {
			glTexCoord2fv(uv[j]);
			glVertex3dv(sides[i][j]);
		}
		glEnd();
	}
	glEndList ();

	Sys_UpdateWindows (W_CAMERA);
	return;

deinitialize:
	while(i-- >= 0)
	{
		glDeleteTextures(1, &skybox_textures[i]);
		skybox_textures[i] = 0;
	}
	Sys_Printf("Skybox load failed\n");
}

void Disk_GetTextureDirs(DIRLIST **list, char *name, char *base)
{
	struct _finddata_t fileinfo;
	int		handle;
	char	dirstring[512];
	char	dir[512];

	if(name)
	{
		setdirstr(dirstring, name, "/*.*", NULL);
	}
	else
	{
		setdirstr(dir, base, "/textures", NULL);
		SetCurrentDirectory(dir);
		setdirstr(dirstring, "*.*", NULL);
	}

	handle = _findfirst (dirstring, &fileinfo);
	if (handle != -1)
	{
		do
		{
			if (!(fileinfo.attrib & _A_SUBDIR))
				continue;
			if (fileinfo.name[0] == '.')
				continue;
			// add this directory to the list

			if(name)
				setdirstr2(dir, name, "/", fileinfo.name, NULL);
			else
				strcpy(dir, fileinfo.name);

			AddToDirListAlphabetized(list, dir, FROMDISK);
			Disk_GetTextureDirs(list, dir, base);

		} while (_findnext( handle, &fileinfo ) != -1);
		_findclose (handle);
	}

	if(!name)
		SetCurrentDirectory(globals.basepath);
}

void FillTextureTree(void);

/*
==================
FillTextureMenu

==================
*/
void FillTextureMenu (void)
{
	HMENU	hmenu;
	int		i;
	DIRLIST	*list = NULL, *temp;

	hmenu = GetSubMenu (GetMenu(globals.d_hwndMain), MENU_TEXTURE);

	// delete everything
	for (i=0 ; i<texture_nummenus ; i++)
		DeleteMenu (hmenu, CMD_TEXTUREWAD+i, MF_BYCOMMAND);

	texture_nummenus = 0;

	Rscript_GetTextureDirs(&list);

	if(globals.gametype == QUAKE2 || globals.gametype == SOF)
	{
		// add everything
		GetPackTextureDirs(&list);
		Disk_GetTextureDirs(&list, NULL, globals.basepath);

		if(globals.modpath[0])
			Disk_GetTextureDirs(&list, NULL, globals.modpath);

		for(temp = list; temp; temp = temp->next)
		{
			AppendMenu (hmenu, MF_ENABLED|MF_STRING|(texture_nummenus+1)%25?0:MF_MENUBARBREAK, CMD_TEXTUREWAD+texture_nummenus, (LPCTSTR)temp->dirname);
			strcpy (texture_menunames[texture_nummenus], temp->dirname);
			strcat (texture_menunames[texture_nummenus], "/");
			if (++texture_nummenus == MAX_TEXTUREDIRS)
				continue;
		}
		
		ClearDirList(&list);
	}
	else
	{
		char	*temp, tempwad[1024];

		if(currentwads[0])
			strcpy(tempwad, currentwads);
		else
			strcpy(tempwad, ValueForKey(globals.d_project_entity, "defaultwad"));

		for (temp = strtok(tempwad, ";"); temp; temp=strtok(0, ";"))
		{
			if(OpenWad(temp))
			{
				AppendMenu (hmenu, MF_ENABLED|MF_STRING, CMD_TEXTUREWAD+texture_nummenus, (LPCTSTR)temp);
				strcpy (texture_menunames[texture_nummenus], temp);
				texture_nummenus++;
			}
		}
	}
	ClearFileList(&availabletextures);
	UpdateListOfTextures(&availabletextures);
	FillComboBoxWithTextures(globals.d_hwndToolbarTextureCombo);
	FillTextureTree();
}


/*
==================
Texture_ClearInuse

A new map is being loaded, so clear inuse markers
==================
*/
void Texture_ClearInuse (void)
{
	qtexture_t	*q;

	for (q=globals.d_qtextures ; q ; q=q->next)
    {
		q->inuse = false;
	}
}

int Rscript_GetFileList(FILELIST **filelist, char *dir);

int ScanDirectoryForFiles(char *directory, char *subdir, char *extension, FILELIST **list)
{
	struct _finddata_t	fileinfo;
	int					handle, found = 0;
	char				dirstring[1024];
	char				name[512];
	char				*ext;

	setdirstr(dirstring, directory, "/textures/", subdir, "/*.", extension, NULL);

	handle = _findfirst (dirstring, &fileinfo);
	if (handle != -1)
	{
		do
		{
			if(fileinfo.attrib & _A_SUBDIR || fileinfo.name[0] == '.')
				continue;

			if(ext = strrchr(fileinfo.name, '.'))
				*ext = 0;

			found++;
			setdirstr2(name, subdir, "/", fileinfo.name, NULL);
			AddToFileListAlphabetized(list, name, -1, 0);
		} while (_findnext( handle, &fileinfo ) != -1);
		_findclose (handle);
	}

	return found;
}

int ScanForFiles(char *subdir, char *extension, FILELIST **list)
{
	int		retval;

	retval =  ScanDirectoryForFiles(globals.basepath, subdir, extension, list);

	if(globals.modpath[0])
		retval +=  ScanDirectoryForFiles(globals.modpath, subdir, extension, list);

	return retval;
}

void Texture_ShowDirectoryByName (char *dirname)
{
	int					found = 0;
	char				name[512];
	char				dirstring[1024];
	FILELIST			*list = NULL, *temp;

	texture_showinuse = false;
	strcpy (texture_directory, dirname);

	globals.d_texturewin.originy = 0;
	Sys_Status(0, "loading all textures");
	
	if(globals.UseExternalTextures)
	{
		found += ScanForFiles(dirname, "png", &list);
		found += ScanForFiles(dirname, "tga", &list);
		found += ScanForFiles(dirname, "jpg", &list);
		found += ScanForFiles(dirname, "pcx", &list);
	}
	if(globals.gametype == QUAKE2)
	{
		found += ScanForFiles(dirname, "wal", &list);
		setdirstr(dirstring, dirname, "*.wal", NULL);
		found += GetPackFileList(&list, dirstring);
	}
	else if(globals.gametype == SOF)
	{
		// load all .m32 files
		setdirstr(dirstring, dirname, "*.m32", NULL);
		found += ScanForFiles(dirname, "m32", &list);
		found += GetPackFileList(&list, dirstring);
	}
	else
	{
		if(globals.d_savedinfo.no_console_scanfile != true)
			Sys_Printf ("Scanning %s\n", dirname);

		found += GetWadFileList(&list, dirname);
	}

	found += Rscript_GetFileList(&list, dirname);

	if(!found)
		return;

	g_dontuse = true;
	for(temp = list; temp; temp = temp->next)
	{
		if(globals.gametype == QUAKE2 || globals.gametype == SOF)
		{
			if(temp->offset == -1)
				strcpy(name, temp->filename);
			else
				setstr(name, dirname, temp->filename, NULL, NULL);

			StripExtension (name);
			Texture_ForName (name);
		}
		else
			Texture_ForName(temp->filename);
	}

	ClearFileList(&list);

	g_dontuse = false;


	SortTextures();
	SetInspectorMode(W_TEXTURE);
	Sys_UpdateWindows(W_TEXTURE);

	sprintf (name, "Textures: %s", texture_directory);
	SetWindowText(globals.Entity.hwnd, name);

	// select the first texture in the list
	if (!globals.d_texturewin.texdef.name[0])
		SelectTexture (16, globals.d_texturewin.height -16, true);
}

/*
==============
Texture_ShowDirectory
==============
*/
void Texture_ShowDirectory (int menunum)
{
	Texture_ShowDirectoryByName(texture_menunames[menunum-CMD_TEXTUREWAD]);
}

/*
==============
Texture_ShowInuse
==============
*/
void	Texture_ShowInuse (void)
{
	char	name[1024];
	face_t	*f;
	brush_t	*b;

	texture_showinuse = true;

	globals.d_texturewin.originy = 0;
	Sys_Status(0, "Selecting active textures");
	Texture_ClearInuse ();

	for (b=active_brushes.next ; b != NULL && b != &active_brushes ; b=b->next)
		for (f=b->brush_faces ; f ; f=f->next)
			Texture_ForName (f->texdef.name);

	for (b=selected_brushes.next ; b != NULL && b != &selected_brushes ; b=b->next)
		for (f=b->brush_faces ; f ; f=f->next)
			Texture_ForName (f->texdef.name);

	SortTextures();
	SetInspectorMode(W_TEXTURE);
	Sys_UpdateWindows (W_TEXTURE);

	sprintf (name, "Textures: in use");
	SetWindowText(globals.Entity.hwnd, name);

	// select the first texture in the list
	if (!globals.d_texturewin.texdef.name[0])
		SelectTexture (16, globals.d_texturewin.height -16, true);
}

/*
==================
Texture_Init
==================
*/
void Texture_Init (void)
{
	char	name[1024];
	byte	*pal;
	
	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &MaxTextureSize);
	ilInit();
	ilEnable(IL_ORIGIN_SET);
	ilOriginFunc(IL_ORIGIN_UPPER_LEFT);

	// load the palette
	globals.images = NULL;

	if(globals.gametype == QUAKE2)
		strcpy(name, "pics/colormap.pcx");
	else if (globals.gametype == QUAKE1)
		strcpy(name, "pics/palette.pcx");

	if (globals.gametype != HALFLIFE && globals.gametype != SOF)
	{
		Load256Image (name, NULL, &pal, NULL, NULL);

		if (!pal)
			Error ("Couldn't load %s", name);
		
		Texture_InitPalette (pal);
	}
	else
	{
		Texture_InitPalette (NULL);
	}

	// create the fallback texture
	Texture_MakeNotexture ();

	globals.d_qtextures = NULL;
	ilEnd();
}

/*
==================
Texture_Flush
==================
*/
void Texture_Flush (void)
{

}

void Refresh_Textures(void)
{
}

/*
==============
Texture_ShowCategory
==============
*/
void	Texture_ShowCategory (int menunum)
{
	category_t			*category = FindCategory(menunum, categories);
	categorytextures_t	*texture;
	char		dirstring[512];
	int			userresponse = -1;
	
	if(!category)
	{
		Sys_Printf("Error: Could not find texture category for menu ID %d\n", menunum);
		return;
	}

	texture_showinuse = false;
	globals.d_texturewin.originy = 0;

	g_dontuse = true;
	for(texture = category->textures; texture; texture = texture->next)
	{
		sprintf (dirstring, "%s.wad", texture->from);
		if(globals.gametype != QUAKE2 && globals.gametype != SOF && !WadIsOpen(dirstring))
		{
			if(userresponse == -1)
				if (MessageBox (globals.d_hwndMain, "This category contains wads that aren't currently loaded. Load them and add them to the current map's \"wad\" key?", "Load Wads?", MB_YESNO|MB_APPLMODAL|MB_ICONQUESTION) == IDYES)
					userresponse = 1;
				else
					userresponse = 0;
			if(userresponse == 1)
			{
				char	wadlist[512], *currentwads;

				currentwads = ValueForKey (world_entity, "wad");
				sprintf(wadlist, "%s%s%s", dirstring, *currentwads?";":"", currentwads);
				SetKeyValue (world_entity, "wad", wadlist);
				Texture_ForName(texture->name);
			}
		}
		else
			Texture_ForName(texture->name);
	}

	g_dontuse = false;
	SortTextures();
	SetInspectorMode(W_TEXTURE);
	Sys_UpdateWindows(W_TEXTURE);
}

int LoadJPG (char *name, byte **pixels, int *width, int *height);

typedef union
{
	unsigned	data;
	byte		r, g, b, a;
}pixelrgba_t;

void SetFlags(qtexture_t *q)
{
	char	*name = q->name;

	if(globals.gametype == HALFLIFE)
	{
		if(!stricmp(name, "sky"))
			q->flags |= SURF_SKY|SURF_NODRAW;

		if(!stricmp(name, "aaatrigger"))
			q->flags |= SURF_NODRAW;		
	}
	else
	{
		if(!strnicmp(name, "sky", 3))
			q->flags |= SURF_SKY|SURF_NODRAW;
		
		if(!stricmp(name, "trigger"))
			q->flags |= SURF_NODRAW;		
	}

	if(!stricmp(name, "clip"))
	{
		q->flags |= SURF_NODRAW;
		q->contents |= (CONTENTS_PLAYERCLIP | CONTENTS_MONSTERCLIP);
	}

	if(!stricmp(name, "hint"))
		q->flags |= 0x100|SURF_NODRAW;
	


	if(name[0] == '+' || name[0] == '*')
	{
		char	*temp;

		temp = &(name[1]);
		if(!strnicmp(temp, "water", 5) || !strnicmp(temp, "slime", 5))
			q->flags |= SURF_TRANS33;
	}
}

rscript_t *RS_FindScript(char *name);
image_t *LoadTextureImage(char *name, int types)
{
	image_t			*image = NULL;
	byte			*imagebuf;
	char			buf[512], *s;
	int				type = 0, width, height;
	cinematics_t	*newcin;

	setdirstr2(buf, name, NULL);

	for(image = globals.images; image; image = image->next)
	{
		if(!stricmp(image->name, buf))
			return image;
	}

	StripExtension(buf);
	s = buf+strlen(buf);

	if (!type && (types & IMAGE_CIN))
	{
		qfile_t	*f;

		strcpy(s, ".cin");

		// Make sure its there
		if(!(f = qfopen(buf, "r", NULL)))
			return NULL;
		else
			qfclose(f);

		newcin = CIN_OpenCin(buf);

		if(newcin)
		{
			type = IMAGE_CIN;
			imagebuf = malloc(256*256*4);
			memset(imagebuf, 192, (256*256*4));
			image = NewImage(name, true);
			glGenTextures(1, &image->texture_number);
			newcin->texnum = image->texture_number;
			GL_Upload(256, 256, imagebuf, image, 0);
		}
	}

	if(!type && (types & IMAGE_PNG))
	{
		strcpy(s, ".png");
		if(LoadPNG(buf, &imagebuf, &width, &height))
			type = IMAGE_PNG;
	}

	if(!type && (types & IMAGE_TGA))
	{
		strcpy(s, ".tga");
		if(LoadTGA(buf, &imagebuf, &width, &height))
			type = IMAGE_TGA;
	}

	if(!type && (types & IMAGE_JPG))
	{
		strcpy(s, ".jpg");
		if(LoadJPG(buf, &imagebuf, &width, &height))
			type = IMAGE_JPG;
	}

	if(!type && (types & IMAGE_PCX))
	{
		byte	*pal;

		strcpy(s, ".pcx");

		if(LoadPCX(buf, &imagebuf, &pal, &width, &height))
		{
			pixelrgba_t	*temp;
			int			count = width*height, i, i2 = 0;
			byte		*btemp;

			type = IMAGE_PCX;
			temp = (pixelrgba_t *)malloc(width * height * 4);
			btemp = (byte *)temp;
			for(i = 0; i < count; i++)
			{
				i2 = imagebuf[i] * 3;
				*btemp++ = pal[i2];
				*btemp++ = pal[i2+1];
				*btemp++ = pal[i2+2];
				*btemp++ = 255;
			}
			free(imagebuf);
			imagebuf = (byte *)temp;
			free(pal);
		}
	}

	if(!type && (types & IMAGE_WAL))
	{
		miptex_t	*lump;

		strcpy(s, ".wal");

		if(Texture_LoadFile((byte **)&lump, name))
		{
			type = IMAGE_WAL;
			image = Image_LoadWal(buf, lump, &width, &height);
		}
	}

	if(!type)
		return NULL;

	if(!image)
	{
		image = qmalloc(sizeof(*image));
		image->width = width;
		image->height = height;
		image->name = strdup(name);
		glGenTextures (1, &image->texture_number);
		GL_Upload(width, height, imagebuf, image, 0);

		if(!ilEnd())
			free (imagebuf);
	}


	if(type == IMAGE_CIN)
		newcin->texnum = image->texture_number;

	return image;
}
qtexture_t *LoadTextureFromDisk(char *name)
{
	char		filename[512];
	qtexture_t	*q = NULL;
	image_t		*image = NULL;
	rscript_t	*script;


	script = RS_FindScript(name);
	if(script)
	{
		RS_ReadyScript(script);
		image = script->stage->texture;
	}
	else
	{
		setdirstr2(filename, "textures/", name, NULL);

		if(!(image = FindImage(filename)) && !(image = LoadTextureImage(filename, IMAGE_PNG|IMAGE_TGA|IMAGE_JPG|IMAGE_PCX)))
			return NULL;
	}
	
	q = qmalloc(sizeof(*q));
	q->image = image;

	if(script)
	{
		q->flags = script->surfaceflags;
		q->contents = script->contentflags;
		q->value = script->lightamount;
		q->image = script->stage->texture;
	}
	else
	{
		q->flags = 0;
		q->value = 0;
		q->contents = 0;
	}
	
	if(globals.gametype != QUAKE2)
		SetFlags(q);

	q->script = script;

	strncpy (q->name, name, sizeof(q->name)-1);
	StripExtension (q->name);
	return q;
}
