#ifndef _TEXTURES_H_
#define _TEXTURES_H_
typedef struct qtexture_s qtexture_t;
#include "mrg/gl_script.h"

typedef struct image_s
{
	char			*name;
	int				width;
	int				height;
	float			color[3];
	GLuint			texture_number;
	int				is_cin;
	struct image_s	*next;
}image_t;

typedef struct
{
	char	name[32];
	float	shift[2];
	float	rotate;
	float	scale[2];
	int		contents;
	int		flags;
	int		value;
} texdef_t;


typedef struct
{
	int			width, height;
	float		originy;
	float		scale;
	texdef_t	texdef;
} texturewin_t;

typedef struct qtexture_s
{
	struct		qtexture_s *next;
	char		name[64];		// includes partial directory and extension
	int			contents;
	int			flags;
	int			value;
	BOOL		inuse;			// true = is present on the level
	image_t		*image;
	rscript_t	*script;
} qtexture_t;

extern qtexture_t *notexture;
// a texturename of the form (0 0 0) will
// create a solid color texture


typedef struct
{
	char			name[16];	// Name of the texture
	unsigned long	width;		// Width of texture, multiple of 8
	unsigned long	height;		// Height of texture, multiple of 8
	unsigned long	offset1;	// offset to mip-level 1
	unsigned long	offset2;	// offset to mip-level 2
	unsigned long	offset4;	// offset to mip-level 3
	unsigned long	offset8;	// offset to mip-level 4
	unsigned long	dummy;
	unsigned char	palette[768];
} miptex_t1;

typedef struct miptex32_s
{
	int			version;
	char		name[128];
	char		altname[128];			// texture substitution
	char		animname[128];			// next frame in animation chain
	char		damagename[128];		// image that should be shown when damaged
	unsigned	width[16], height[16];
	unsigned	offsets[16];		
	int			flags;
	int			contents;
	int			value;
	float		scale_x, scale_y;
	int			mip_scale;

	// detail texturing info
	char		dt_name[128];		// detailed texture name
	float		dt_scale_x, dt_scale_y;
	float		dt_u, dt_v;
	float		dt_alpha;
	int			dt_src_blend_mode, dt_dst_blend_mode;

	int			flags2;
	float		damage_health;

	int			unused[18];				// future expansion to maintain compatibility with h2
} miptex32_t;

typedef struct paletteRGBA_s
{
	union
	{
		struct
		{
			byte r,g,b,a;
		};
		unsigned c;
		byte c_array[4];
	};
} paletteRGBA_t;

extern GLuint	skybox_textures[6];


qtexture_t	*Texture_LoadMip		(miptex_t1 *qtex, char *name);
void		Texture_Init			(void);
void		Texture_Flush			(void);
void		Texture_ClearInuse		(void);
void		Texture_ShowInuse		(void);
void		Texture_ShowDirectory	(int menunum);
void		Texture_ShowCategory	(int menunum);

qtexture_t	*Texture_ForName		(char *name);

void		Texture_Init			(void);
void		Texture_SetTexture		(texdef_t *texdef, int apply, int onlytexturename);

void		Texture_SetMode			(int iMenu);	// GL_TEXTURE_NEAREST, etc..
void		LoadSkybox				(void);
void		GL_Upload				(int width, int height, BYTE *source, image_t *image, int usepalette);
BOOL		CheckTextureSize		(int width, int height, char *name);
image_t		*FindImage				(char *name);
image_t		*NewImage				(char *name, int is_cin);
BOOL		Texture_LoadFile		(byte **lump, char *name);
qtexture_t	*Texture_LoadM32		(miptex32_t *qtex, char *name);
qtexture_t	*Texture_LoadWAL		(miptex_t *qtex, char *name);
image_t		*LoadTextureImage		(char *name, int types);

#define IMAGE_CIN		1
#define IMAGE_PNG		2
#define IMAGE_TGA		4
#define IMAGE_JPG		8
#define IMAGE_PCX		16
#define IMAGE_WAL		32
#define IMAGE_ALL		63
#define IMAGE_ALLBUTCIN	62
#endif