#include "qe3.h"
#include "directory.h"
#include <commctrl.h>

int			g_nOpen, g_nClosed;
HTREEITEM	TextureRoot, CategoryRoot;
dir_t		*TextureDirList = NULL;
dir_t		*CategoryDirList = NULL;

HTREEITEM AddItemToTree(HWND hwndTV, LPSTR lpszItem, int nLevel);
void Disk_GetTextureDirs(DIRLIST **list, char *name);

BOOL InitTreeViewImageLists(HWND hwndTV) 
{
	HIMAGELIST	himl;	// handle to image list
	HBITMAP		hbmp;	// handle to bitmap

	// Create the image list.
	if ((himl = ImageList_Create(13, 13, FALSE, 2, 0)) == NULL)
		return FALSE;
	
	// Add the open file, closed file, and document bitmaps.
	hbmp = LoadBitmap(globals.d_hInstance, MAKEINTRESOURCE(IDB_OPEN_BITMAP));
	g_nOpen = ImageList_Add(himl, hbmp, (HBITMAP)NULL);
	DeleteObject(hbmp);
	
	hbmp = LoadBitmap(globals.d_hInstance, MAKEINTRESOURCE(IDB_CLOSED_BITMAP));
	g_nClosed = ImageList_Add(himl, hbmp, (HBITMAP)NULL);
	DeleteObject(hbmp);
	
//	hbmp = LoadBitmap(g_hinst, MAKEINTRESOURCE(IDB_DOCUMENT));
//	g_nDocument = ImageList_Add(himl, hbmp, (HBITMAP)NULL);
//	DeleteObject(hbmp);
	
	// Fail if not all of the images were added.
	if (ImageList_GetImageCount(himl) < 2)
		return FALSE;
	
	// Associate the image list with the tree-view control.
	TreeView_SetImageList(hwndTV, himl, TVSIL_NORMAL);

	if(globals.gametype != HALFLIFE && globals.gametype != QUAKE1)
		TextureRoot = AddItemToTree(hwndTV, "Textures", 0);

	CategoryRoot = AddItemToTree(hwndTV, "Categories", 0);

	TextureDirList = malloc(sizeof(dir_t));
	CategoryDirList = malloc(sizeof(dir_t));
	TextureDirList->files = CategoryDirList->files = NULL;
	TextureDirList->subdirs = CategoryDirList->subdirs = NULL;
	
	return TRUE;
}

HTREEITEM AddItemToTree(HWND hwndTV, LPSTR lpszItem, int nLevel)
{
	TVITEM tvi;
	TVINSERTSTRUCT tvins;
	static HTREEITEM hPrev = (HTREEITEM)TVI_FIRST;
	static HTREEITEM hPrevRootItem = NULL;
	static HTREEITEM hPrevLev2Item = NULL;
	HTREEITEM hti;
	
	tvi.mask = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
//	tvi.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE;
	
	// Set the text of the item.
	tvi.pszText = lpszItem;
	tvi.cchTextMax = lstrlen(lpszItem);
	
	// Assume the item is not a parent item, so give it a
	// document image.
	tvi.iImage = g_nClosed;
	tvi.iSelectedImage = g_nOpen;
	
	// Save the heading level in the item's application-defined
	// data area.
	tvi.lParam = (LPARAM)nLevel;
	tvins.item = tvi;
	tvins.hInsertAfter = hPrev;
	
	// Set the parent item based on the specified level.
	if (nLevel == 1)
		tvins.hParent = TVI_ROOT;
	else if (nLevel == 2)
		tvins.hParent = hPrevRootItem;
	else
		tvins.hParent = hPrevLev2Item;
	
	// Add the item to the tree-view control.
	hPrev = (HTREEITEM)SendMessage(hwndTV, TVM_INSERTITEM, 0, (LPARAM)(LPTVINSERTSTRUCT)&tvins);
	
	// Save the handle to the item.
	if (nLevel == 1)
		hPrevRootItem = hPrev;
	else if (nLevel == 2)
		hPrevLev2Item = hPrev;
	
	// The new item is a child item. Give the parent item a
	// closed folder bitmap to indicate it now has child items.
	if (nLevel > 1)
	{
		hti = TreeView_GetParent(hwndTV, hPrev);
		tvi.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE;
		tvi.hItem = hti;
		tvi.iImage = g_nClosed;
		tvi.iSelectedImage = g_nClosed;
		TreeView_SetItem(hwndTV, &tvi);
	}
	
	return hPrev;
}

HTREEITEM AddItemToSubtree(HWND hwndTV, LPSTR lpszItem, HTREEITEM parent, dir_t *dir)
{
	TVITEM tvi;
	TVINSERTSTRUCT tvins;
	static HTREEITEM hPrev = (HTREEITEM)TVI_FIRST;
	static HTREEITEM hPrevRootItem = NULL;
	static HTREEITEM hPrevLev2Item = NULL;
	
	tvi.mask = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
	
	// Set the text of the item.
	tvi.pszText = lpszItem;
	tvi.cchTextMax = lstrlen(lpszItem);
	
	// Assume the item is not a parent item, so give it a
	// document image.
	tvi.iImage = g_nClosed;
	tvi.iSelectedImage = g_nOpen;
	
	// Save the heading level in the item's application-defined
	// data area.
	tvi.lParam = (LPARAM)dir;
	tvins.item = tvi;
	tvins.hInsertAfter = TVI_SORT;
	
	tvins.hParent = parent;
	
	// Add the item to the tree-view control.
	return (HTREEITEM)SendMessage(hwndTV, TVM_INSERTITEM, 0, (LPARAM)(LPTVINSERTSTRUCT)&tvins);
}

void AddSubdirsToTree(HTREEITEM parent, dir_t *dir)
{
	dir_t		*d;
	HTREEITEM	child;

	if(!dir)
		return;

	for(d = dir->subdirs; d; d = d->next)
	{
		 child = AddItemToSubtree(globals.TextureTree.hwnd, d->name, parent, d);
		 AddSubdirsToTree(child, d);
	}
}

void CreateCategoryTreeItems(HTREEITEM root, category_t *category)
{
	category_t	*currentcategory;
	HTREEITEM	thisitem;
	
	for(currentcategory = category; currentcategory; currentcategory = currentcategory->next)
	{
		thisitem = AddItemToSubtree(globals.TextureTree.hwnd, currentcategory->name, root, (dir_t *)&(currentcategory->identifier));
		if(currentcategory->subcategories)
		{
			CreateCategoryTreeItems(thisitem, currentcategory->subcategories);
		}
	}
}

void FillTextureTree(void)
{
	HTREEITEM	tree;

	if(globals.gametype != HALFLIFE && globals.gametype != QUAKE1)
	{
		while(tree = TreeView_GetChild(globals.TextureTree.hwnd, TextureRoot))
		{
			TreeView_DeleteItem(globals.TextureTree.hwnd, tree);
		}

		GetDirectories(&TextureDirList);
		AddSubdirsToTree(TextureRoot, TextureDirList);
	}

	while(tree = TreeView_GetChild(globals.TextureTree.hwnd, CategoryRoot))
	{
		TreeView_DeleteItem(globals.TextureTree.hwnd, tree);
	}

	CreateCategoryTreeItems(CategoryRoot, categories);
}

void GetFullName(dir_t *dir, char *buf)
{
	char s[512];

	if(dir->parent != TextureDirList)
	{
		GetFullName(dir->parent, buf);		
		setdirstr(s, buf, "/", dir->name, NULL);
		strcpy(buf, s);
	}
	else
		strcpy(buf, dir->name);
}

void Texture_ShowDirectoryByName (char *dirname);

HTREEITEM GetRootParent(HTREEITEM child)
{
	HTREEITEM parent = TreeView_GetParent(globals.TextureTree.hwnd,child);

	if(parent)
		return GetRootParent(parent);

	return child;
}

BOOL TreeNotifyHandler(LPNMHDR lParam)
{
	switch(lParam->code)
	{
		case TVN_SELCHANGED:
		{
			LPNMTREEVIEW	pnmtv = (LPNMTREEVIEW)lParam;
			HTREEITEM		rootparent;
			TVITEM			itemNew = pnmtv->itemNew;
			dir_t			*dir;
			char			dirname[512];

			if(itemNew.hItem == TextureRoot || itemNew.hItem == CategoryRoot)
				return 0;

			rootparent = GetRootParent(itemNew.hItem);

			SetInspectorMode(W_TEXTURE);
			Sys_BeginWait();
			if(rootparent != CategoryRoot)
			{
				dir = (dir_t *)itemNew.lParam;
				dirname[0] = 0;
				GetFullName(dir, dirname);
				setdirstr2(dirname, dirname, "/", NULL);
				Texture_ShowDirectoryByName(dirname);
			}
			else
			{
				int	i = *((int *)itemNew.lParam);

				Texture_ShowCategory(i);
			}
			break;
		}
	}

	SetFocus (globals.Camera.hwnd);
	return 0;
}

