#include "qe3.h"
#include "inc.h"

WAD2DirPtr ReadWAD2Directory(UInt32 offset, UInt16 *dirsize_r, wadlist_t *wad)
{
	WAD2DirPtr  dir;
	UInt32      pos, size;
	UInt16      max, i;

	*dirsize_r = 0;
	if (wad->filehandle == NULL)
	{
		Sys_Printf("ERROR: NULL wad file pointer\n");
		return NULL;
	}
	if(fseek(wad->filehandle, offset, SEEK_SET) < 0)
	{
		Sys_Printf("ERROR: Bad offset\n");
		return NULL;
	}
    if(globals.gametype == QUAKE1 && (ReadMagic(wad->filehandle) != FTYPE_WAD2))
	{
		Sys_Printf("ERROR: Quake 1 expects \"WAD2\"\n");
		return NULL;
	}
	if(globals.gametype == HALFLIFE && (ReadMagic(wad->filehandle) != FTYPE_WAD3))
	{
		Sys_Printf("ERROR: Half-Life expects \"WAD3\"\n");
		return NULL;
	}
    if(ReadInt32(wad->filehandle, &size) == FALSE)
	{
		Sys_Printf("ERROR: Bad size\n");
		return NULL;
	}
    if(ReadInt32(wad->filehandle, &pos) == FALSE)
	{
		Sys_Printf("ERROR: Bad pos\n");
		return NULL;
	}
    if(size == 0L)
	{
		Sys_Printf("ERROR: Empty directory\n");
		return NULL;
	}
    if(size > 65535L)
	{
		Sys_Printf("ERROR: Directory size too large\n");
		return NULL;
	}
    if(fseek(wad->filehandle, pos, SEEK_SET) < 0)
		return NULL;
	
	dir = (WAD2DirPtr)qmalloc(size * sizeof(struct WAD2Directory));
	max = (UInt16)size;
	for (i = 0; i < max; i++)
    {
		if (ReadBytes(wad->filehandle, &dir[i], sizeof(struct WAD2Directory)) == FALSE)
		{
			free(dir);
			return NULL;
		}
		dir[i].offset = SwapInt32(dir[i].offset);
		dir[i].size = SwapInt32(dir[i].size);
		dir[i].dsize = SwapInt32(dir[i].dsize);
    }
	*dirsize_r = max;
	return dir;
}

wadlist_t *OpenWad(char *filename)
{
	UInt16			i, waddirsize;
	WAD2DirPtr		dir;
	wadlist_t		*newwad;
	wadfilelist_t	*file;
	char			dirstring[1024], *temp;
	int				f_type;

	if((newwad = WadIsOpen(filename)) != NULL)
		return newwad;

	newwad = (wadlist_t *)malloc(sizeof(wadlist_t));
	newwad->files = NULL;

	temp = ValueForKey (globals.d_project_entity, "texturepath");
	SetCurrentDirectory(temp);
	strcpy(dirstring, filename);

	if((newwad->filehandle = OpenFileReadMagic(dirstring, &f_type)) == NULL)
	{
		SetCurrentDirectory(globals.WorkingDir);
		Sys_Printf("ERROR: Could not open %s", dirstring);
		free(newwad);
		return NULL;
	}
	SetCurrentDirectory(globals.WorkingDir);

	Sys_Printf("Opening wad: %s\n", filename);
	if (f_type != FTYPE_WAD2 && f_type != FTYPE_WAD3)
	{
		Sys_Printf("ERROR: %s is not a valid WAD file", dirstring);
		free(newwad);
		return NULL;
	}

	dir = ReadWAD2Directory(0, &waddirsize, newwad);

	if (dir == NULL)
	{
		Sys_Printf("Cannot read main directory from wad: %s", dirstring);
		fclose(newwad->filehandle);
		free(newwad);
		return NULL;
	}

	newwad->next = openwads.next;
	openwads.next->prev = newwad;
	openwads.next = newwad;
	newwad->prev = &openwads;

	for (i = 0; i < waddirsize; i++)
	{
		file = (wadfilelist_t *)malloc(sizeof(wadfilelist_t));
		file->loaded = 0;
		strncpy(file->name, dir[i].name, 16);
		file->name[16] = 0;
		file->offset = dir[i].offset;
		file->size = dir[i].dsize;
		file->next = newwad->files;
		file->owner = newwad;
		newwad->files = file;
	}
	free(dir);
	newwad->filename = strdup(filename);
	newwad->longname = strdup(dirstring);

	return newwad;
}

wadfilelist_t *FindWADEntry(char *name)
{
	wadlist_t		*wad;
	wadfilelist_t	*file;

	for(wad = openwads.next; wad != &openwads; wad = wad->next)
	{
		for(file = wad->files; file; file = file->next)
		{
			if(!stricmp(file->name, name))
				return file;
		}
	}
	return NULL;
}

BOOL WadLoadFile(char *name, void **bufferptr, byte *palette)
{
	void			*buffer;
	wadfilelist_t	*file;

	file = FindWADEntry(name);

	if (!file)
		return false;
		
	if (fseek(file->owner->filehandle, file->offset, SEEK_SET) < 0)
	{
		Sys_Printf("Unexpected EOF in wadfile: %s\n,", file->owner->filename);
		return false;
	}
	if((buffer = qmalloc(file->size+5)) == NULL)
		Error("Could not allocate memory");
		
	if(fread(buffer, 1, file->size, file->owner->filehandle) != file->size)
	{
		Sys_Printf("Error reading %s from wad %s\n", name, file->owner->filename);
		free(buffer);
		return false;
	}
	if(palette)
	{
		miptex_t1 *mip = (miptex_t1 *)buffer;
		int	offset = mip->offset8 + ((mip->width * mip->height) / 64) + 2;

		fseek(file->owner->filehandle, file->offset, SEEK_SET);
//		fseek(file->owner->filehandle, file->size-770, SEEK_CUR);
//		Sys_Printf("File: %s\n", file->name);
//		Sys_Printf("Width: %d\n", mip->width);
//		Sys_Printf("Height: %d\n", mip->height);
//		Sys_Printf("Offset1: %d\n", mip->offset1);
//		Sys_Printf("Offset2: %d\n", mip->offset2);
//		Sys_Printf("Offset4: %d\n", mip->offset4);
//		Sys_Printf("Offset8: %d\n", mip->offset8);
//		Sys_Printf("Real palette offset = %d\n", file->size - 770);
//		Sys_Printf("Trying %d\n", offset);
//		Sys_Printf("\n");
		fseek(file->owner->filehandle, offset, SEEK_CUR);
		if(!fread(palette, 768, 1, file->owner->filehandle))
		{
			free(buffer);
			return false;
		}
	}

	*bufferptr = buffer;
	file->loaded = 1;
	strcpy(name, file->name); //To insure right case when map is compiled
	return true;
}

void CloseWadFile(char *filename)
{
	wadlist_t		*wad, *current;
	wadfilelist_t	*file, *last;

	if(!(wad = WadIsOpen(filename)))
		return;

	for(current = openwads.next; current!= &openwads; current = current->next)
	{
		if(current == wad)
		{
			Sys_Printf("Closing %s\n", wad->filename);
			fclose(wad->filehandle);
			free(wad->filename);
			free(wad->longname);
			
			for(file = wad->files; file; last = file, file = file->next, free(last));
			
			wad->prev->next = wad->next;
			wad->next->prev = wad->prev;
			free(wad);
			return;
		}
	}
}

int GetWadFileList(FILELIST **filelist, char *filename)
{
	wadlist_t		*wad;
	wadfilelist_t	*file;
	int				found = 0;

	if (!(wad = WadIsOpen(filename)))
		return 0;
	
	for (file = wad->files; file; file = file->next)
	{
		if(!(globals.gametype == QUAKE1 && !stricmp(file->name, "palette")))
		{
			found++;
			AddToFileListAlphabetized(filelist, file->name, file->offset, 0);
		}
	}
//	Sys_Printf("Got list of textures from %s\n", wad->filename);
	return found;
}

wadlist_t *WadIsOpen(char *filename)
{
	wadlist_t *wad;

	for(wad = openwads.next; wad != &openwads; wad = wad->next)
	{
		if(!stricmp(filename, wad->filename))
			return wad;
	}
	return NULL;
}

void CloseExtinctWads(char *activewads)
{
	char		*temp, buf[1024];
	wadlist_t	*wad1, *wad2, *templist = NULL, *newwad;
	int			found;

	strcpy(buf, activewads);
	for (temp = strtok(buf, ";"); temp; temp=strtok(0, ";"))
	{
		newwad = (wadlist_t *)malloc(sizeof(wadlist_t));
		newwad->filename = strdup(temp);
		newwad->next = templist;
		templist = newwad;
	}

	for(wad2 = openwads.next; wad2 != &openwads;)
	{
		found = 0;
		for(wad1 = templist; wad1; wad1 = wad1->next)
		{
//			Sys_Printf("Checking wad: %s against %s\n", wad1->filename, wad2->filename);
			if(!stricmp(wad1->filename, wad2->filename))
			{
				found = 1;
				continue;
			}
		}
		newwad = wad2;
		wad2 = wad2->next;
		if(!found)
			CloseWadFile(newwad->filename);
	}

//	Sys_Printf("Freeing temporary list\n");
	for(wad1 = templist; wad1; wad1 = wad1->next, free(newwad))
	{		
		free(wad1->filename);
		newwad = wad1;
	}
//	Sys_Printf("Temporary list freed\n");
}
