// win_cam.c -- windows specific camera view code

#include "qe3.h"
#include "inc.h"

/*
============
CameraWndProc
============
*/
LONG WINAPI WCam_WndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
	int		fwKeys, xPos, yPos;
    RECT	rect;

    GetClientRect(hWnd, &rect);

    switch (uMsg)
    {
	case WM_CREATE:
		{

			// report OpenGL information
//			Sys_Printf ("GL_VENDOR: %s\n", glGetString (GL_VENDOR));
//			Sys_Printf ("GL_RENDERER: %s\n", glGetString (GL_RENDERER));
//			Sys_Printf ("GL_VERSION: %s\n", glGetString (GL_VERSION));
//			Sys_Printf ("GL_EXTENSIONS: %s\n", glGetString (GL_EXTENSIONS));
		}
		return 0;
	case WM_PAINT:
        { 
			PAINTSTRUCT	ps;

			xPos = 0;
			
            if (!wglMakeCurrent( globals.Camera.hdc, globals.Camera.hglrc ))
				Error ("wglMakeCurrent failed");

			SetLastError(0);
			if ( BeginPaint(globals.Camera.hwnd, &ps) )
			{
				QE_CheckOpenGLForErrors("WCam_WndProc():BeginPaint()", GL_NO_ERROR);
				Cam_Draw ();
				QE_CheckOpenGLForErrors("WCam_WndProc():Cam_Draw()", GL_NO_ERROR);

				EndPaint(globals.Camera.hwnd, &ps);
				SwapBuffers(globals.Camera.hdc);
			}
        }
		return 0;
		
	case WM_USER+267:	// benchmark
        { 
			PAINTSTRUCT	ps;
			WINDOWPLACEMENT wp;
			double	start, end;
			int		i;
			
			memset( &wp, 0, sizeof( wp ) );
			wp.length = sizeof( wp );
			GetWindowPlacement( globals.Camera.hwnd, &wp );
			
			MoveWindow( globals.Camera.hwnd, 30, 30, 400, 400, TRUE );
			
			BeginPaint(hWnd, &ps);
            if (!wglMakeCurrent( globals.Camera.hdc, globals.Camera.hglrc))
				Error ("wglMakeCurrent failed");
			glDrawBuffer (GL_FRONT);
			
			start = Sys_DoubleTime ();
			for (i=0 ; i<100 ; i++)
			{
				camera.angles[YAW] = i*4;
				Cam_Draw ();
			}
			SwapBuffers(globals.Camera.hdc);
			glDrawBuffer (GL_BACK);
			end = Sys_DoubleTime ();
			EndPaint(hWnd, &ps);
			Sys_Printf ("%5.2f seconds\n", end-start);

			SetWindowPlacement( globals.Camera.hwnd, &wp );
        }
		break;
		
	case WM_KEYDOWN:
		if ( QE_KeyDown (wParam) )
			return 0;
		else 
			return DefWindowProc( hWnd, uMsg, wParam, lParam );

//	case WM_SYSKEYDOWN:
//		Sys_Printf("%d\n", wParam);
//		return 0;

	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_LBUTTONDOWN:
		if (GetTopWindow(globals.d_hwndMain) != hWnd)
			BringWindowToTop(hWnd);
		
		SetFocus (globals.Camera.hwnd);
		SetCapture (globals.Camera.hwnd);
		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;
		Cam_MouseDown (xPos, yPos, fwKeys);
		return 0;
		
	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	case WM_LBUTTONUP:
		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;
		Cam_MouseUp (xPos, yPos, fwKeys);
		if (! (fwKeys & (MK_LBUTTON|MK_RBUTTON|MK_MBUTTON)))
			ReleaseCapture ();
		return 0;

	case WM_LBUTTONDBLCLK:
		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;
		Cam_DoubleClick(xPos, yPos, fwKeys);
		
	case WM_MOUSEMOVE:
		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;
		Cam_MouseMoved (xPos, yPos, fwKeys);
		return 0;
		
	case WM_SIZE:
		camera.width = rect.right;
		camera.height = rect.bottom;
		InvalidateRect(globals.Camera.hwnd, NULL, false);
		return 0;

	case WM_SIZING:
			if(TryDocking(hWnd, wParam, (LPRECT)lParam, 0))
				return TRUE;
			break;

	case WM_MOVING:
			if(TryDocking(hWnd, wParam, (LPRECT)lParam, 0))
				return TRUE;
			break;

	case WM_KILLFOCUS:
	case WM_SETFOCUS:
		SendMessage( hWnd, WM_NCACTIVATE, uMsg == WM_SETFOCUS, 0 );
		return 0;

	case WM_NCCALCSIZE:// don't let windows copy pixels
		DefWindowProc (hWnd, uMsg, wParam, lParam);
		return WVR_REDRAW;

	case WM_CLOSE:
		DestroyWindow (hWnd);
		return 0;
		
	case WM_DESTROY:
		QEW_StopGL( &hWnd, &globals.Camera.hglrc, &globals.Camera.hdc );
		return 0;

	}
	return DefWindowProc( hWnd, uMsg, wParam, lParam );
}


/*
==============
WCam_Create
==============
*/
void WCam_Create (HINSTANCE hInstance)
{
    WNDCLASS   wc;
	char		*title;

    /* Register the camera class */
	memset (&wc, 0, sizeof(wc));

    wc.style         = CS_OWNDC|CS_DBLCLKS;
    wc.lpfnWndProc   = (WNDPROC)WCam_WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = 0;
//    wc.hCursor       = LoadCursor (NULL,IDC_CROSS);
    wc.hCursor       = LoadCursor (globals.d_hInstance, MAKEINTRESOURCE(IDC_CURSOR1));
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = CAMERA_WINDOW_CLASS;

    if (!RegisterClass (&wc) )
        Error ("WCam_Register: failed");

	if (  globals.d_savedinfo.exclude & EXCLUDE_DETAIL )
		title = "Camera View (DETAIL EXCLUDED)";
	else
	    	title = "Camera View";

	globals.Camera.hwnd = CreateWindowEx (WS_EX_TOOLWINDOW | WS_EX_CONTEXTHELP,
		CAMERA_WINDOW_CLASS ,
		title,
		WS_CAPTION | WS_THICKFRAME | WS_MAXIMIZEBOX | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_CHILD,
		ZWIN_WIDTH,
		20,
		(int)(screen_width*CWIN_SIZE),
		(int)(screen_height*CWIN_SIZE),	// size

		globals.d_hwndMain,	// parent window
		0,		// no menu
		hInstance,
		NULL);
	if (!globals.Camera.hwnd)
		Error ("Couldn't create globals.Camera.hwnd");

	if(!LoadWindowState(globals.Camera.hwnd, globals.Multiple2DViews?"camerawindow":"camerawindow2"))
		NewWindow = true;

    ShowWindow (globals.Camera.hwnd, SW_SHOWDEFAULT);
	globals.Camera.hdc = GetDC(globals.Camera.hwnd);
	QEW_SetupPixelFormat(&(globals.Camera.pfd), &(globals.Camera.hdc), true);

	if ((globals.Camera.hglrc = wglCreateContext(globals.Camera.hdc)) == 0)
		Error ("wglCreateContext failed in WCam_WndProc():WM_CREATE");

	if (!wglMakeCurrent(globals.Camera.hdc, globals.Camera.hglrc))
		Error ("wglMakeCurrent failed in WCam_WndProc():WM_CREATE");

	if (!wglShareLists(globals.Main.hglrc, globals.Camera.hglrc ) )
		Error( "wglShareLists in WCam_Create failed" );
	Texture_SetMode(globals.d_savedinfo.iTexMenu);

}
