#include "qe3.h"
#include <process.h>
#include "mru.h"
#include "entityw.h"
#include "inc.h"
#include <mmsystem.h>

CHAR			szBuf[128];
LPTOOLTIPTEXT	lpToolTipText;
TOOLINFO		lpToolInfo;
HWND			hWndTT;
extern BOOL		SurfaceDlgActive;
extern char		IniFile[512];
char			*Surface_FlagNames[32];
char			*Content_FlagNames[32];

BOOL SaveRegistryInfo(const char *pszName, void *pvBuf, long lSize);
BOOL LoadRegistryInfo(const char *pszName, void *pvBuf, long *plSize);
void Map_RegionSet_Nonselected(void);
void DoProject(int first);
int DoFitTextures(void);

static HWND CreateMyStatusWindow(HINSTANCE hInst);
static HWND CreateToolBar(HINSTANCE hinst);

extern BOOL	TexturesLocked;
extern int	WXY_Print					(void);
void		ScrollTextureWindow			(int pixels, int buttons);
void		DoSelectEntitiesByKeyVal	(void);
void		Entity_AddToEntityGroup		(void);
void		SetDefaultWindowPositions	(void);
void		DoPreferences				(void);
/*
==============================================================================

  MENU

==============================================================================
*/

void OpenDialog (void);
void SaveAsDialog (void);
qboolean ConfirmModified (void);
void  Select_Ungroup (void);

void QE_ExpandBspString (char *bspaction, char *out, char *mapname)
{
	char	*in;
	char	src[1024];
	char	rsh[1024];
	char	base[256];

	ExtractFileName (mapname, base);
	sprintf (src, "%smaps/%s", ValueForKey(globals.d_project_entity, "remotebasepath"), base);
	StripExtension(src);
	strcpy (rsh, ValueForKey(globals.d_project_entity, "rshcmd"));

	in = ValueForKey( globals.d_project_entity, bspaction );
	while (*in)
	{
		if (in[0] == '!')
		{
			strcpy (out, rsh);
			out += strlen(rsh);
			in++;
			continue;
		}
		if (in[0] == '$')
		{
			strcpy (out, src);
			out += strlen(src);
			in++;
			continue;
		}
		if (in[0] == '@')
		{
			*out++ = '"';
			in++;
			continue;
		}
		if (in[0] == '&' && in[1] == '&')
		{
//			*out++ = 13;
			*out++ = 10;
			in++;
			in++;
			continue;
		}
		*out++ = *in++;
	}
	*out = 0;
}



void RunBsp (char *command)
{
	char	sys[1024];
	char	batpath[1024];
	char	outputpath[1024];
	char	temppath[512];
	char	name[1024];
	char	*basepath;
	FILE	*hFile;
	BOOL	ret;
	PROCESS_INFORMATION ProcessInformation;
	STARTUPINFO	startupinfo;

	SetInspectorMode (W_CONSOLE);

	if (bsp_process)
	{
		Sys_Printf ("BSP is still going...\n");
		return;
	}

	GetTempPath(512, temppath);
	sprintf (outputpath, "%sjunk.txt", temppath);

	strcpy (name, currentmap);
	if (region_active)
	{
		Map_SaveFile (name, false);
		StripExtension (name);
		strcat (name, ".reg");
	}

	Map_SaveFile (name, region_active);


	QE_ExpandBspString (command, sys, name);

	Sys_ClearPrintf ();
	Sys_Printf ("======================================\nRunning bsp command...\n");
	Sys_Printf ("\n%s\n", sys);

	//
	// write qe3bsp.bat
	//
	sprintf (batpath, "%sqe3bsp.bat", temppath);
	hFile = fopen(batpath, "w");
	if (!hFile)
		Error ("Can't write to %s", batpath);
	fprintf (hFile, sys);
	fclose (hFile);

	//
	// write qe3bsp2.bat
	//
	sprintf (batpath, "%sqe3bsp2.bat", temppath);
	hFile = fopen(batpath, "w");
	if (!hFile)
		Error ("Can't write to %s", batpath);
	fprintf (hFile, "%sqe3bsp.bat > %s", temppath, outputpath);
	fclose (hFile);

	Pointfile_Delete ();
	basepath = ValueForKey(globals.d_project_entity, "basepath");
	if(basepath && !(*basepath))
		basepath = NULL;
	GetStartupInfo (&startupinfo);

	ret = CreateProcess(
    batpath,		// pointer to name of executable module 
    NULL,			// pointer to command line string
    NULL,			// pointer to process security attributes 
    NULL,			// pointer to thread security attributes 
    FALSE,			// handle inheritance flag 
    0 /*DETACHED_PROCESS*/,		// creation flags 
    NULL,			// pointer to new environment block 
    basepath,		// pointer to current directory name 
    &startupinfo,	// pointer to STARTUPINFO 
    &ProcessInformation 	// pointer to PROCESS_INFORMATION  
   );

	if (!ret)
		Error ("CreateProcess failed");

	bsp_process = ProcessInformation.hProcess;

	Sleep (100);	// give the new process a chance to open it's window

	BringWindowToTop( globals.d_hwndMain );	// pop us back on top
	SetFocus (globals.Camera.hwnd);
}

/*
=============
DoColor

=============
*/
qboolean DoColor(int iIndex)
{
	CHOOSECOLOR	cc;
	static COLORREF	custom[16];

	cc.lStructSize = sizeof(cc);
	cc.hwndOwner = globals.d_hwndMain;
	cc.hInstance = NULL;
	cc.rgbResult = 
		(int)(globals.d_savedinfo.colors[iIndex][0]*255) +
		(((int)(globals.d_savedinfo.colors[iIndex][1]*255))<<8) +
		(((int)(globals.d_savedinfo.colors[iIndex][2]*255))<<16);
    cc.lpCustColors = custom;
    cc.Flags = CC_FULLOPEN|CC_RGBINIT; 
    //cc.lCustData; 
    //cc.lpfnHook; 
    //cc.lpTemplateName

	if (!ChooseColor(&cc))
		return false;

	globals.d_savedinfo.colors[iIndex][0] = (cc.rgbResult&255)/255.0;
	globals.d_savedinfo.colors[iIndex][1] = ((cc.rgbResult>>8)&255)/255.0;
	globals.d_savedinfo.colors[iIndex][2] = ((cc.rgbResult>>16)&255)/255.0;

	/* 
	** scale colors so that at least one component is at 1.0F 
	** if this is meant to select an entity color
	*/
	if ( iIndex == COLOR_ENTITY )
	{
		float largest = 0.0F;

		if ( globals.d_savedinfo.colors[iIndex][0] > largest )
			largest = globals.d_savedinfo.colors[iIndex][0];
		if ( globals.d_savedinfo.colors[iIndex][1] > largest )
			largest = globals.d_savedinfo.colors[iIndex][1];
		if ( globals.d_savedinfo.colors[iIndex][2] > largest )
			largest = globals.d_savedinfo.colors[iIndex][2];

		if ( largest == 0.0F )
		{
			globals.d_savedinfo.colors[iIndex][0] = 1.0F;
			globals.d_savedinfo.colors[iIndex][1] = 1.0F;
			globals.d_savedinfo.colors[iIndex][2] = 1.0F;
		}
		else
		{
			float scaler = 1.0F / largest;

			globals.d_savedinfo.colors[iIndex][0] *= scaler;
			globals.d_savedinfo.colors[iIndex][1] *= scaler;
			globals.d_savedinfo.colors[iIndex][2] *= scaler;
		}
	}

	Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);

	return true;
}


/* Copied from MSDN */

BOOL DoMru(HWND hWnd,WORD wId)
{
	char szFileName[128];
	OFSTRUCT of;
	BOOL fExist;

	GetMenuItem(globals.d_lpMruMenu, wId, TRUE, szFileName, sizeof(szFileName));

	// Test if the file exists.

	fExist = OpenFile(szFileName ,&of,OF_EXIST) != HFILE_ERROR;

	if (fExist) {

		// Place the file on the top of MRU.
		AddNewItem(globals.d_lpMruMenu,(LPSTR)szFileName);

		// Now perform opening this file !!!
		Map_LoadFile (szFileName);	
	}
	else
		// Remove the file on MRU.
		DelMenuItem(globals.d_lpMruMenu,wId,TRUE);

	// Refresh the File menu.
	PlaceMenuMRUItem(globals.d_lpMruMenu,GetSubMenu(GetMenu(hWnd),0),
			ID_FILE_EXIT);

	return fExist;
}

extern int inspector_mode;

/* handle all WM_COMMAND messages here */
LONG WINAPI CommandHandler (
    HWND    hWnd,
    WPARAM  wParam,
    LPARAM  lParam)
{
	HMENU		hMenu;

    switch (LOWORD(wParam))
    {
//
// file menu
//
    	case ID_FILE_EXIT:
            /* exit application */
//			if (!ConfirmModified())
//				return TRUE;

            PostMessage (hWnd, WM_CLOSE, 0, 0L);
			break;

		case ID_FILE_OPEN:
			if (!ConfirmModified())
				return TRUE;
			OpenDialog ();
			break;

		case ID_FILE_NEW:
			if (!ConfirmModified())
				return TRUE;
			Map_New ();
			break;
		case ID_FILE_SAVE:
			if (!strcmp(currentmap, "unnamed.map"))
				SaveAsDialog ();
			else
				Map_SaveFile (currentmap, false);	// ignore region
			break;
		case ID_FILE_SAVEAS:
			SaveAsDialog ();
			break;

		case ID_FILE_LOADPROJECT:
			if (!ConfirmModified())
				return TRUE;
			ProjectDialog ();
			break;

		case ID_FILE_EDITPROJECT:
			if (!ConfirmModified())
				return TRUE;
			DoProject(false);
			break;

		case ID_FILE_POINTFILE:
			if (globals.d_pointfile_display_list)
				Pointfile_Clear ();
			else
				Pointfile_Check ();
			break;

//
// edit menu
//
		case ID_EDIT_CREATEVISGROUP:
			DoCreateVisGroup();
			break;

		case ID_EDIT_SETCURRENTVISGROUP:
			DoSetVisGroup();
			break;
			
		case ID_EDIT_RSCRIPT_MAKESURFACEFLAGSFIELDFROMCURRENTTEXTURE:
		case ID_EDIT_RSCRIPT_MAKECONTENTFLAGSFIELDFROMCURRENTTEXTURE:
		case ID_EDIT_RSCRIPT_MAKELIGHTAMOUNTFIELDFROMCURRENTTEXTURE:
		case ID_EDIT_RSCRIPT_MAKEALL:
			RScript_CopyToClipboard(LOWORD(wParam));
			break;

//
// view menu
//
		case ID_VIEW_ENTITY:
			SetInspectorMode(W_ENTITY);
			break;
		case ID_VIEW_CONSOLE:
			SetInspectorMode(W_CONSOLE);
			break;
		case ID_VIEW_TEXTURE:
			if(inspector_mode == W_TEXTURE)
				SetInspectorMode(W_TEXTURELIST);
			else
				SetInspectorMode(W_TEXTURE);
			break;

		case ID_VIEW_100:
			globals.XY.scale = 1;
			globals.XZ.scale = 1;
			globals.YZ.scale = 1;
			Sys_UpdateWindows (W_XY|W_XZ|W_YZ);
			break;
		case ID_VIEW_ZOOMIN:
			globals.XY.scale *= 5.0/4;
			if (globals.XY.scale > 64)
				globals.XY.scale = 64;
			globals.XZ.scale *= 5.0/4;
			if (globals.XZ.scale > 64)
				globals.XZ.scale = 64;
			globals.YZ.scale *= 5.0/4;
			if (globals.YZ.scale > 64)
				globals.YZ.scale = 64;
			Sys_UpdateWindows (W_XY|W_XZ|W_YZ);
			break;
		case ID_VIEW_ZOOMOUT:
			globals.XY.scale *= (float)4.0/5;
			if (globals.XY.scale < 0.05)
				globals.XY.scale = (float)0.05;
			globals.XZ.scale *= (float)4.0/5;
			if (globals.XZ.scale < 0.05)
				globals.XZ.scale = (float)0.05;
			globals.YZ.scale *= (float)4.0/5;
			if (globals.YZ.scale < 0.05)
				globals.YZ.scale = (float)0.05;
			Sys_UpdateWindows (W_XY|W_XZ|W_YZ);
			break;

		case ID_VIEW_Z100:
			z.scale = 1;
			Sys_UpdateWindows (W_Z);
			break;
		case ID_VIEW_ZZOOMIN:
			z.scale *= 5.0/4;
			if (z.scale > 4)
				z.scale = 4;
			Sys_UpdateWindows (W_Z);
			break;
		case ID_VIEW_ZZOOMOUT:
			z.scale *= (float)4.0/5;
			if (z.scale < 0.125)
				z.scale = 0.125;
			Sys_UpdateWindows (W_Z);
			break;

		case ID_VIEW_CENTER:
			camera.angles[ROLL] = camera.angles[PITCH] = 0;
			camera.angles[YAW] = 22.5 * 
				floor( (camera.angles[YAW]+11)/22.5 );
			Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
			break;

		case ID_VIEW_UPFLOOR:
			Cam_ChangeFloor (true);
			break;
		case ID_VIEW_DOWNFLOOR:
			Cam_ChangeFloor (false);
			break;

		case ID_VIEW_TRANSPARENCY_ACTIVE:
			ToggleAndCheck(ID_VIEW_TRANSPARENCY_ACTIVE, &globals.transactive);
			Sys_UpdateWindows (W_CAMERA);
			break;

		case ID_VIEW_TRANSPARENCY_TRANS33:
			ToggleAndCheck(ID_VIEW_TRANSPARENCY_TRANS33, &globals.trans33);
			Sys_UpdateWindows (W_CAMERA);
			break;

		case ID_VIEW_TRANSPARENCY_TRANS66:
			ToggleAndCheck(ID_VIEW_TRANSPARENCY_TRANS66, &globals.trans66);
			Sys_UpdateWindows (W_CAMERA);
			break;

		case ID_VIEW_TRANSPARENCY_NODRAW:
			ToggleAndCheck(ID_VIEW_TRANSPARENCY_NODRAW, &globals.nodraw);
			Sys_UpdateWindows (W_CAMERA);
			break;

		case ID_VIEW_SHOWNAMES:
			ToggleAndCheck(ID_VIEW_SHOWNAMES, &globals.d_savedinfo.show_names);
			Map_BuildBrushData();
			Sys_UpdateWindows (W_XY | W_XZ | W_YZ);
			break;

		case ID_VIEW_SHOWCOORDINATES:
			ToggleAndCheck(ID_VIEW_SHOWCOORDINATES, &globals.d_savedinfo.show_coordinates);
			Sys_UpdateWindows (W_XY | W_XZ | W_YZ);
			break;

		case ID_VIEW_SHOWBLOCKS:
			ToggleAndCheck(ID_VIEW_SHOWBLOCKS, &globals.show_blocks);
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWLIGHTS:
			globals.d_savedinfo.exclude ^= EXCLUDE_LIGHTS;
			CheckMenu(ID_VIEW_SHOWLIGHTS,  !(globals.d_savedinfo.exclude & EXCLUDE_LIGHTS));
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWPATH:
			globals.d_savedinfo.exclude ^= EXCLUDE_PATHS;
			CheckMenu(ID_VIEW_SHOWPATH, !(globals.d_savedinfo.exclude & EXCLUDE_PATHS));
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWENT:
			globals.d_savedinfo.exclude ^= EXCLUDE_ENT;
			CheckMenu(ID_VIEW_SHOWENT, !(globals.d_savedinfo.exclude & EXCLUDE_ENT));
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWWATER:
			globals.d_savedinfo.exclude ^= EXCLUDE_WATER;
			CheckMenu(ID_VIEW_SHOWWATER, !(globals.d_savedinfo.exclude & EXCLUDE_WATER));
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWCLIP:
			globals.d_savedinfo.exclude ^= EXCLUDE_CLIP;
			CheckMenu(ID_VIEW_SHOWCLIP, !(globals.d_savedinfo.exclude & EXCLUDE_CLIP));
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWDETAIL:
			if ( ( globals.d_savedinfo.exclude ^= EXCLUDE_DETAIL ) & EXCLUDE_DETAIL )
			{
				CheckMenuItem ( GetMenu(globals.d_hwndMain), ID_VIEW_SHOWDETAIL, MF_BYCOMMAND | MF_UNCHECKED );
				SetWindowText (globals.Camera.hwnd, "Camera View (DETAIL EXCLUDED)");
			}
			else
			{
				CheckMenuItem ( GetMenu(globals.d_hwndMain), ID_VIEW_SHOWDETAIL, MF_BYCOMMAND | MF_CHECKED );
				SetWindowText (globals.Camera.hwnd, "Camera View");
			}
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWWORLD:
			globals.d_savedinfo.exclude ^= EXCLUDE_WORLD;
			CheckMenu(ID_VIEW_SHOWWORLD, !(globals.d_savedinfo.exclude & EXCLUDE_WORLD));
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWENTITYCONNECTIONS:
			ToggleAndCheck(ID_VIEW_SHOWENTITYCONNECTIONS, &globals.ShowEntityConnections);
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWSKYBOX:
			ToggleAndCheck(ID_VIEW_SHOWSKYBOX, &globals.ShowSkybox);
			Sys_UpdateWindows (W_CAMERA);
			break;

		case ID_VIEW_SHOWTEXTURENAMES:
			ToggleAndCheck(ID_VIEW_SHOWTEXTURENAMES, &globals.ShowTextureNames);
			Sys_UpdateWindows (W_TEXTURE);
			break;

		case ID_VIEW_CLIP2XYZ:
			ToggleAndCheck(ID_VIEW_CLIP2XYZ, &globals.Clip2XYZ);
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_SHOWHINT:
			ToggleAndCheck(ID_VIEW_SHOWHINT, &globals.ShowHint);
			Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_VIEW_XYVIEW_STIPPLESELECTIONOUTLINE:
			ToggleAndCheck(ID_VIEW_XYVIEW_STIPPLESELECTIONOUTLINE, &globals.StippleSelection);
			Sys_UpdateWindows (W_XY | W_XZ | W_YZ);
			break;

		case ID_VIEW_DRAWVISGROUPCOLORS:
			ToggleAndCheck(ID_VIEW_DRAWVISGROUPCOLORS, &globals.DrawVisGroupColors);
			Sys_UpdateWindows (W_XY | W_XZ | W_YZ);
			break;

		case ID_VIEW_SHOWFULLSURFACEEDITOR:
			ToggleAndCheck(ID_VIEW_SHOWFULLSURFACEEDITOR, &globals.ShowFullSurfaceEditor);
			break;

//
// grid menu
//
		case ID_GRID_1:
		case ID_GRID_2:
		case ID_GRID_4:
		case ID_GRID_8:
		case ID_GRID_16:
		case ID_GRID_32:
		case ID_GRID_64:
		{
			hMenu = GetMenu(hWnd);
			
			CheckMenuItem(hMenu, ID_GRID_1, MF_BYCOMMAND | MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_GRID_2, MF_BYCOMMAND | MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_GRID_4, MF_BYCOMMAND | MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_GRID_8, MF_BYCOMMAND | MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_GRID_16, MF_BYCOMMAND | MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_GRID_32, MF_BYCOMMAND | MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_GRID_64, MF_BYCOMMAND | MF_UNCHECKED);

			switch (LOWORD(wParam))
			{
				case ID_GRID_1: globals.d_gridsize = 0; break;
				case ID_GRID_2: globals.d_gridsize = 1; break;
				case ID_GRID_4: globals.d_gridsize = 2; break;
				case ID_GRID_8: globals.d_gridsize = 3; break;
				case ID_GRID_16: globals.d_gridsize = 4; break;
				case ID_GRID_32: globals.d_gridsize = 5; break;
				case ID_GRID_64: globals.d_gridsize = 6; break;
			}
			globals.d_gridsize = 1 << globals.d_gridsize;

			Sys_UpdateStatusBar();
			CheckMenuItem(hMenu, LOWORD(wParam), MF_BYCOMMAND | MF_CHECKED);
			Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_Z);
			break;
		}

//
// texture menu
//
		case ID_VIEW_NEAREST:
		case ID_VIEW_NEARESTMIPMAP:
		case ID_VIEW_LINEAR:
		case ID_VIEW_BILINEAR:
		case ID_VIEW_BILINEARMIPMAP:
		case ID_VIEW_TRILINEAR:
		case ID_TEXTURES_WIREFRAME:
		case ID_TEXTURES_FLATSHADE:
			Texture_SetMode (LOWORD(wParam));
			break;
			
		case ID_TEXTURES_XYVIEWRENDERMODE_WIREFRAME:
		case ID_TEXTURES_XYVIEWRENDERMODE_FLATSHADE:
		case ID_TEXTURES_XYVIEWRENDERMODE_TEXTURED:
			hMenu = GetMenu(globals.d_hwndMain);
			CheckMenuItem(hMenu, ID_TEXTURES_XYVIEWRENDERMODE_WIREFRAME, MF_BYCOMMAND | MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_TEXTURES_XYVIEWRENDERMODE_FLATSHADE, MF_BYCOMMAND | MF_UNCHECKED);
			CheckMenuItem(hMenu, ID_TEXTURES_XYVIEWRENDERMODE_TEXTURED, MF_BYCOMMAND | MF_UNCHECKED);
			
			CheckMenuItem(hMenu, LOWORD(wParam), MF_BYCOMMAND | MF_CHECKED);

			if(LOWORD(wParam) == ID_TEXTURES_XYVIEWRENDERMODE_WIREFRAME)
				globals.d_xy.draw_mode = cd_wire;
			else if (LOWORD(wParam) == ID_TEXTURES_XYVIEWRENDERMODE_FLATSHADE)
				globals.d_xy.draw_mode = cd_solid;
			else
				globals.d_xy.draw_mode = cd_texture;
			Sys_UpdateWindows(W_ALL);
			break;

		case ID_TEXTURES_SHOWINUSE:
			Sys_BeginWait ();
			Texture_ShowInuse ();
			SetInspectorMode(W_TEXTURE);
			break;

		case ID_TEXTURES_INSPECTOR:
			DoSurface ();
			break;
			
		case ID_TEXTURES_UPDATEAVAILABLETEXTURELIST:
			ClearFileList(&availabletextures);
			UpdateListOfTextures(&availabletextures);
			FillComboBoxWithTextures(globals.d_hwndToolbarTextureCombo);
			break;

//
// bsp menu
//
		case CMD_BSPCOMMAND:
		case CMD_BSPCOMMAND+1:
		case CMD_BSPCOMMAND+2:
		case CMD_BSPCOMMAND+3:
		case CMD_BSPCOMMAND+4:
		case CMD_BSPCOMMAND+5:
		case CMD_BSPCOMMAND+6:
		case CMD_BSPCOMMAND+7:
		case CMD_BSPCOMMAND+8:
		case CMD_BSPCOMMAND+9:
		case CMD_BSPCOMMAND+10:
		case CMD_BSPCOMMAND+11:
		case CMD_BSPCOMMAND+12:
		case CMD_BSPCOMMAND+13:
		case CMD_BSPCOMMAND+14:
		case CMD_BSPCOMMAND+15:
		case CMD_BSPCOMMAND+16:
		case CMD_BSPCOMMAND+17:
		case CMD_BSPCOMMAND+18:
		case CMD_BSPCOMMAND+19:
		case CMD_BSPCOMMAND+20:
		case CMD_BSPCOMMAND+21:
		case CMD_BSPCOMMAND+22:
		case CMD_BSPCOMMAND+23:
		case CMD_BSPCOMMAND+24:
		case CMD_BSPCOMMAND+25:
		case CMD_BSPCOMMAND+26:
		case CMD_BSPCOMMAND+27:
		case CMD_BSPCOMMAND+28:
		case CMD_BSPCOMMAND+29:
		case CMD_BSPCOMMAND+30:
		case CMD_BSPCOMMAND+31:
			{
				extern	char	*bsp_commands[256];

				RunBsp (bsp_commands[LOWORD(wParam-CMD_BSPCOMMAND)]);
			}
			break;

//
// misc menu
//
		case ID_MISC_BENCHMARK:
			SendMessage ( globals.Camera.hwnd,
			WM_USER+267,	0, 0);
			break;

		case ID_TEXTUREBK:
			DoColor(COLOR_TEXTUREBACK);
			Sys_UpdateWindows (W_TEXTURE);
			break;

		case ID_MISC_SELECTENTITYCOLOR:
			{
				extern int inspector_mode;
				
				if ( ( inspector_mode == W_ENTITY ) && DoColor(COLOR_ENTITY) == true )
				{
					extern void AddProp( void );
					
					char buffer[100];

					if(globals.gametype == QUAKE1)
					{
						sprintf( buffer, "%lf %lf %lf", globals.d_savedinfo.colors[COLOR_ENTITY][0]*255,
							globals.d_savedinfo.colors[COLOR_ENTITY][1]*255,
							globals.d_savedinfo.colors[COLOR_ENTITY][2]*255 );
					}
					else
					{
						sprintf( buffer, "%f %f %f", globals.d_savedinfo.colors[COLOR_ENTITY][0],
							globals.d_savedinfo.colors[COLOR_ENTITY][1],
							globals.d_savedinfo.colors[COLOR_ENTITY][2] );
					}
					
					SetWindowText( hwndEnt[EntValueField], buffer );
					SetWindowText( hwndEnt[EntKeyField], "_color" );
					AddProp();
				}
				Sys_UpdateWindows(W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			}
			break;

		case ID_MISC_PRINTXY:
			WXY_Print();
			break;

		case ID_MISC_NO_TEXTURE:
			ToggleAndCheck(ID_MISC_NO_TEXTURE, &globals.d_savedinfo.no_console_texture);
			break;

		case ID_MISC_NO_SCANFILE:
			ToggleAndCheck(ID_MISC_NO_SCANFILE, &globals.d_savedinfo.no_console_scanfile);
			break;

		case ID_MISC_FOLLOWFLOOR:
			ToggleAndCheck(ID_MISC_FOLLOWFLOOR, &globals.followfloor);

			if(globals.followfloor)
				Sys_Status (0, "Follow Floor ON");
			else
				Sys_Status (0, "Follow Floor OFF");

			Sys_UpdateWindows (W_CAMERA);
			break;

		case ID_MISC_ALWAYSUPDATECAMERA:
			ToggleAndCheck(ID_MISC_ALWAYSUPDATECAMERA, &globals.UpdateCamera);

			if(globals.UpdateCamera)
				Sys_Status (0, "Always Update Camera ON");
			else
				Sys_Status (0, "Always Update Camera OFF");

			Sys_UpdateWindows (W_CAMERA);
			break;

		case ID_MISC_ALLOWFACEANDBRUSHSELECTION:
			ToggleAndCheck(ID_MISC_ALLOWFACEANDBRUSHSELECTION, &globals.AllowFaceAndBrushSelection);

			if(globals.AllowFaceAndBrushSelection)
				Sys_Status (0, "Face+brush selection enabled");
			else
				Sys_Status (0, "Face+brush selection disabled");

			break;

		case ID_COLORS_XYBK:
			DoColor(COLOR_GRIDBACK);
			Sys_UpdateWindows (W_XY | W_XZ | W_YZ);
			break;

		case ID_COLORS_MAJOR:
			DoColor(COLOR_GRIDMAJOR);
			Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_COLORS_MINOR:
			DoColor(COLOR_GRIDMINOR);
			Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_Z);
			break;

		case ID_MISC_FINDBRUSH:
			DoFind();
			break;

		case ID_MISC_NEXTLEAKSPOT:
			Pointfile_Next();
			break;
		case ID_MISC_PREVIOUSLEAKSPOT:
			Pointfile_Prev();
			break;
		case ID_MISC_WRITEMAPFLOATS:
			ToggleAndCheck(ID_MISC_WRITEMAPFLOATS, &globals.WriteMapFloats);
			break;

		case ID_MISC_ALWAYSUSEDEFAULTWINDOWPLACEMENT:
			ToggleAndCheck(ID_MISC_ALWAYSUSEDEFAULTWINDOWPLACEMENT, &globals.AlwaysUseDefaultWindowPlacement);
			break;

		case ID_MISC_OFFSETCOPIEDBRUSHES:
			ToggleAndCheck(ID_MISC_OFFSETCOPIEDBRUSHES, &globals.OffsetCopiedBrushes);
			break;

		case ID_MISC_SETTINGS:
			DoPreferences();
			break;

		case ID_MISC_SETDEFAULTWINDOWPOSITIONSNOW:
			SetDefaultWindowPositions();
			break;

//
// brush menu
//
		case ID_BRUSH_3SIDED:
			Brush_MakeSided (3);
			break;
		case ID_BRUSH_4SIDED:
			Brush_MakeSided (4);
			break;
		case ID_BRUSH_5SIDED:
			Brush_MakeSided (5);
			break;
		case ID_BRUSH_6SIDED:
			Brush_MakeSided (6);
			break;
		case ID_BRUSH_7SIDED:
			Brush_MakeSided (7);
			break;
		case ID_BRUSH_8SIDED:
			Brush_MakeSided (8);
			break;
		case ID_BRUSH_9SIDED:
			Brush_MakeSided (9);
			break;
		case ID_BRUSH_ARBITRARYSIDED:
			DoSides ();
			break;

//
// select menu
//
		case ID_BRUSH_FLIPX:
			Select_FlipAxis (0);
			break;
		case ID_BRUSH_FLIPY:
			Select_FlipAxis (1);
			break;
		case ID_BRUSH_FLIPZ:
			Select_FlipAxis (2);
			break;
		case ID_BRUSH_ROTATEX:
			{
				int amt;
				
				if(GetKeyState(VK_CONTROL) < 0)
					amt = 45;
				else
					amt = 90;

				if(GetKeyState(VK_SHIFT) < 0)
					amt *= -1;
				
				Select_RotateAxis (0, amt);
			}
			break;
		case ID_BRUSH_ROTATEY:
			{
				int amt;
				
				if(GetKeyState(VK_CONTROL) < 0)
					amt = 45;
				else
					amt = 90;

				if(GetKeyState(VK_SHIFT) < 0)
					amt *= -1;

				Select_RotateAxis (1, amt);
			}
			break;
		case ID_BRUSH_ROTATEZ:
			{
				int amt;
				
				if(GetKeyState(VK_CONTROL) < 0)
					amt = 45;
				else
					amt = 90;

				if(GetKeyState(VK_SHIFT) < 0)
					amt *= -1;
				
				Select_RotateAxis (2, amt);
			}
			break;

		case ID_SELECTION_ARBITRARYROTATION:
			DoRotate ();
			break;

		case ID_SELECTION_UNGROUPENTITY:
			Select_Ungroup ();
			break;

		case ID_SELECTION_CONNECT:
			ConnectEntities ();
			break;

		case ID_SELECTION_DRAGVERTECIES:
			if (globals.d_select_mode == sel_vertex)
			{
				globals.d_select_mode = sel_brush;
				Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			}
			else
			{
				SetupVertexSelection ();
				if (globals.d_numpoints)
					globals.d_select_mode = sel_vertex;
			}
			break;
		case ID_SELECTION_DRAGEDGES:
			if (globals.d_select_mode == sel_edge)
			{
				globals.d_select_mode = sel_brush;
				Sys_UpdateWindows (W_CAMERA | W_XY | W_XZ | W_YZ | W_Z);
			}
			else
			{
				SetupVertexSelection ();
				if (globals.d_numpoints)
					globals.d_select_mode = sel_edge;
			}
			break;

		case ID_SELECTION_SELECTPARTIALTALL:
			Select_PartialTall ();
			break;
		case ID_SELECTION_SELECTCOMPLETETALL:
			Select_CompleteTall ();
			break;
		case ID_SELECTION_SELECTTOUCHING:
			Select_Touching ();
			break;
		case ID_SELECTION_SELECTINSIDE:
			Select_Inside ();
			break;
		case ID_SELECTION_CSGSUBTRACT:
			CSG_Subtract ();
			break;
		case ID_SELECTION_MAKEHOLLOW:
			CSG_MakeHollow ();
			break;
		case ID_SELECT_BADCONTENTS:
			Select_FindBadBrushes();
			break;

		case ID_SELECTION_CLONE:
			Select_Clone ();
			break;
		case ID_SELECTION_DELETE:
			Select_Delete ();
			break;
		case ID_SELECTION_DESELECT:
			Select_Deselect (true);
			break;

		//Contents submenu
			
		case ID_SELECTION_CONTENTFLAGS_SOLID_SET:
			Select_SetContentsFlag(CONTENTS_SOLID);
			break;
		case ID_SELECTION_CONTENTFLAGS_SOLID_UNSET:
			Select_UnsetContentsFlag(CONTENTS_SOLID);
			break;
		case ID_SELECTION_CONTENTFLAGS_WINDOW_SET:
			Select_SetContentsFlag(CONTENTS_WINDOW);
			break;
		case ID_SELECTION_CONTENTFLAGS_WINDOW_UNSET:
			Select_UnsetContentsFlag(CONTENTS_WINDOW);
			break;
		case ID_SELECTION_CONTENTFLAGS_AUX_SET:
			Select_SetContentsFlag(CONTENTS_AUX);
			break;
		case ID_SELECTION_CONTENTFLAGS_AUX_UNSET:
			Select_UnsetContentsFlag(CONTENTS_AUX);
			break;
		case ID_SELECTION_CONTENTFLAGS_LAVA_SET:
			Select_SetContentsFlag(CONTENTS_LAVA);
			break;
		case ID_SELECTION_CONTENTFLAGS_LAVA_UNSET:
			Select_UnsetContentsFlag(CONTENTS_LAVA);
			break;
		case ID_SELECTION_CONTENTFLAGS_SLIME_SET:
			Select_SetContentsFlag(CONTENTS_SLIME);
			break;
		case ID_SELECTION_CONTENTFLAGS_SLIME_UNSET:
			Select_UnsetContentsFlag(CONTENTS_SLIME);
			break;
		case ID_SELECTION_CONTENTFLAGS_WATER_SET:
			Select_SetContentsFlag(CONTENTS_WATER);
			break;
		case ID_SELECTION_CONTENTFLAGS_WATER_UNSET:
			Select_UnsetContentsFlag(CONTENTS_WATER);
			break;
		case ID_SELECTION_CONTENTFLAGS_MIST_SET:
			Select_SetContentsFlag(CONTENTS_MIST);
			break;
		case ID_SELECTION_CONTENTFLAGS_MIST_UNSET:
			Select_UnsetContentsFlag(CONTENTS_MIST);
			break;
		case ID_SELECTION_CONTENTFLAGS_PLAYERCLIP_SET:
			Select_SetContentsFlag(CONTENTS_PLAYERCLIP);
			break;
		case ID_SELECTION_CONTENTFLAGS_PLAYERCLIP_UNSET:
			Select_UnsetContentsFlag(CONTENTS_PLAYERCLIP);
			break;
		case ID_SELECTION_CONTENTFLAGS_MONSTERCLIP_SET:
			Select_SetContentsFlag(CONTENTS_MONSTERCLIP);
			break;
		case ID_SELECTION_CONTENTFLAGS_MONSTERCLIP_UNSET:
			Select_UnsetContentsFlag(CONTENTS_MONSTERCLIP);
			break;
		case ID_SELECTION_CONTENTFLAGS_CURRENT_NOCURRENT:
			Select_UnsetContentsFlag(
				CONTENTS_CURRENT_0 |
				CONTENTS_CURRENT_90 |
				CONTENTS_CURRENT_180 |
				CONTENTS_CURRENT_270 |
				CONTENTS_CURRENT_UP |
				CONTENTS_CURRENT_DOWN
				);
			break;
		case ID_SELECTION_CONTENTFLAGS_CURRENT_0:
			Select_SetContentsFlag(CONTENTS_CURRENT_0);
			break;
		case ID_SELECTION_CONTENTFLAGS_CURRENT_90:
			Select_SetContentsFlag(CONTENTS_CURRENT_90);
			break;
		case ID_SELECTION_CONTENTFLAGS_CURRENT_180:
			Select_SetContentsFlag(CONTENTS_CURRENT_180);
			break;
		case ID_SELECTION_CONTENTFLAGS_CURRENT_270:
			Select_SetContentsFlag(CONTENTS_CURRENT_270);
			break;
		case ID_SELECTION_CONTENTFLAGS_CURRENT_UP:
			Select_SetContentsFlag(CONTENTS_CURRENT_UP);
			break;
		case ID_SELECTION_CONTENTFLAGS_CURRENT_DOWN:
			Select_SetContentsFlag(CONTENTS_CURRENT_DOWN);
			break;
		case ID_SELECTION_CONTENTFLAGS_ORIGIN_SET:
			Select_SetContentsFlag(CONTENTS_ORIGIN);
			break;
		case ID_SELECTION_CONTENTFLAGS_ORIGIN_UNSET:
			Select_UnsetContentsFlag(CONTENTS_ORIGIN);
			break;
		case ID_SELECTION_CONTENTFLAGS_MONSTER_SET:
			Select_SetContentsFlag(CONTENTS_MONSTER);
			break;
		case ID_SELECTION_CONTENTFLAGS_MONSTER_UNSET:
			Select_UnsetContentsFlag(CONTENTS_MONSTER);
			break;
		case ID_SELECTION_CONTENTFLAGS_CORPSE_SET:
			Select_SetContentsFlag(CONTENTS_DEADMONSTER);
			break;
		case ID_SELECTION_CONTENTFLAGS_CORPSE_UNSET:
			Select_UnsetContentsFlag(CONTENTS_DEADMONSTER);
			break;
		case ID_SELECTION_CONTENTFLAGS_DETAIL_SET:
			Select_SetContentsFlag(CONTENTS_DETAIL);
			break;
		case ID_SELECTION_CONTENTFLAGS_DETAIL_UNSET:
			Select_UnsetContentsFlag(CONTENTS_DETAIL);
			break;
		case ID_SELECTION_CONTENTFLAGS_TRANSLUCENT_SET:
			Select_SetContentsFlag(CONTENTS_TRANSLUCENT);
			break;
		case ID_SELECTION_CONTENTFLAGS_TRANSLUCENT_UNSET:
			Select_UnsetContentsFlag(CONTENTS_TRANSLUCENT);
			break;
		case ID_SELECTION_CONTENTFLAGS_LADDER_SET:
			Select_SetContentsFlag(CONTENTS_LADDER);
			break;
		case ID_SELECTION_CONTENTFLAGS_LADDER_UNSET:
			Select_UnsetContentsFlag(CONTENTS_LADDER);
			break;
		case ID_SELECTION_CONTENTFLAGS_CLEARALLCONTENTFLAGS:
			Select_UnsetContentsFlag(
				CONTENTS_SOLID |
				CONTENTS_WINDOW |
				CONTENTS_AUX |
				CONTENTS_LAVA |
				CONTENTS_SLIME |
				CONTENTS_WATER |
				CONTENTS_MIST |
				CONTENTS_PLAYERCLIP |
				CONTENTS_MONSTERCLIP |
				CONTENTS_CURRENT_0 |
				CONTENTS_CURRENT_90 |
				CONTENTS_CURRENT_180 |
				CONTENTS_CURRENT_270 |
				CONTENTS_CURRENT_UP |
				CONTENTS_CURRENT_DOWN |
				CONTENTS_ORIGIN |
				CONTENTS_MONSTER |
				CONTENTS_DEADMONSTER |
				CONTENTS_DETAIL |
				CONTENTS_TRANSLUCENT |
				CONTENTS_LADDER
				);
			break;
		//Surface submenu
		case ID_SELECTION_MAKELIGHT:
			DoLight ();
			break;
		case ID_SELECTION_MAKENONLIGHT:
			Select_UnsetSurfaceFlag(SURF_LIGHT);
			break;
		case ID_SELECTION_SURFACEFLAGS_SETSLICKFLAG:
			Select_SetSurfaceFlag(SURF_SLICK);
			break;
		case ID_SELECTION_SURFACEFLAGS_UNSETSLICKFLAG:
			Select_UnsetSurfaceFlag(SURF_SLICK);
			break;
		case ID_SELECTION_SURFACEFLAGS_SKY_SET:
			Select_SetSurfaceFlag(SURF_SKY);
			break;
		case ID_SELECTION_SURFACEFLAGS_SKY_UNSET:
			Select_UnsetSurfaceFlag(SURF_SKY);
			break;
		case ID_SELECTION_SURFACEFLAGS_WARP_SET:
			Select_SetSurfaceFlag(SURF_WARP);
			break;
		case ID_SELECTION_SURFACEFLAGS_WARP_UNSET:
			Select_UnsetSurfaceFlag(SURF_WARP);
			break;
		case ID_SELECTION_SURFACEFLAGS_TRANS33_SET:
			Select_SetSurfaceFlag(SURF_TRANS33);
			break;
		case ID_SELECTION_SURFACEFLAGS_TRANS33_UNSET:
			Select_UnsetSurfaceFlag(SURF_TRANS33);
			break;
		case ID_SELECTION_SURFACEFLAGS_TRANS66_SET:
			Select_SetSurfaceFlag(SURF_TRANS66);
			break;
		case ID_SELECTION_SURFACEFLAGS_TRANS66_UNSET:
			Select_UnsetSurfaceFlag(SURF_TRANS66);
			break;
		case ID_SELECTION_SURFACEFLAGS_FLOWING_SET:
			Select_SetSurfaceFlag(SURF_FLOWING);
			break;
		case ID_SELECTION_SURFACEFLAGS_FLOWING_UNSET:
			Select_UnsetSurfaceFlag(SURF_FLOWING);
			break;
		case ID_SELECTION_SURFACEFLAGS_NODRAW_SET:
			Select_SetSurfaceFlag(SURF_NODRAW);
			break;
		case ID_SELECTION_SURFACEFLAGS_NODRAW_UNSET:
			Select_UnsetSurfaceFlag(SURF_NODRAW);
			break;
		case ID_SELECTION_SURFACEFLAGS_HINT_SET:
			Select_SetSurfaceFlag(0x100);
			break;
		case ID_SELECTION_SURFACEFLAGS_HINT_UNSET:
			Select_UnsetSurfaceFlag(0x100);
			break;
		case ID_SELECTION_SURFACEFLAGS_SKIP_SET:
			Select_SetSurfaceFlag(0x200);
			break;
		case ID_SELECTION_SURFACEFLAGS_SKIP_UNSET:
			Select_UnsetSurfaceFlag(0x200);
			break;
		case ID_SELECTION_SURFACEFLAGS_CLEARALLSURFACEFLAGS:
			Select_UnsetSurfaceFlag(
				SURF_LIGHT |
				SURF_SLICK |
				SURF_SKY |
				SURF_WARP |
				SURF_TRANS33 |
				SURF_TRANS66 |
				SURF_FLOWING |
				SURF_NODRAW |
				0x100 |
				0x200
				);
			break;
		//Selection->Select->Brushes menus
			//Contents
		case ID_SELECT_BRUSHES_ALLBRUSHES:
			Select_Deselect (true);
			Select_AllBrushes();
			break;
		case ID_SELECT_BRUSHES_WITHONEORMORESELECTEDFACES:
			SelectBrushesWithSelectedFaces();
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCURRENTTEXTURE:
			Select_Deselect (true);
			SelectContainingTexture(true, &globals.d_texturewin.texdef);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_SOLID:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_SOLID);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_WINDOW:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_WINDOW);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_AUX:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_AUX);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_LAVA:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_LAVA);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_SLIME:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_SLIME);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_WATER:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_WATER);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_MIST:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_MIST);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_PLAYERCLIP:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_PLAYERCLIP);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_MONSTERCLIP:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_MONSTERCLIP);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_CURRENT0:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_CURRENT_0);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_CURRENT90:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_CURRENT_90);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_CURRENT180:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_CURRENT_180);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_CURRENT270:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_CURRENT_270);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_CURRENTUP:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_CURRENT_UP);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_CURRENTDOWN:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_CURRENT_DOWN);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_ORIGIN:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_ORIGIN);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_MONSTER:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_MONSTER);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_CORPSE:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_DEADMONSTER);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_DETAIL:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_DETAIL);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_TRANSLUCENT:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_TRANSLUCENT);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGCONTENTFLAG_LADDER:
			Select_Deselect (true);
			SelectContainingFlag(true, true, CONTENTS_LADDER);
			break;
			
			//Surface
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGSURFACEFLAG_LIGHT:
			Select_Deselect (true);
			SelectContainingFlag(true, false, SURF_LIGHT);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGSURFACEFLAG_SLICK:
			Select_Deselect (true);
			SelectContainingFlag(true, false, SURF_SLICK);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGSURFACEFLAG_SKY:
			Select_Deselect (true);
			SelectContainingFlag(true, false, SURF_SKY);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGSURFACEFLAG_WARP:
			Select_Deselect (true);
			SelectContainingFlag(true, false, SURF_WARP);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGSURFACEFLAG_TRANS33:
			Select_Deselect (true);
			SelectContainingFlag(true, false, SURF_TRANS33);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGSURFACEFLAG_TRANS66:
			Select_Deselect (true);
			SelectContainingFlag(true, false, SURF_TRANS66);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGSURFACEFLAG_FLOWING:
			Select_Deselect (true);
			SelectContainingFlag(true, false, SURF_FLOWING);
			break;
		case ID_SELECTION_SELECT_BRUSHES_CONTAININGSURFACEFLAG_NODRAW:
			Select_Deselect (true);
			SelectContainingFlag(true, false, SURF_NODRAW);
			break;
		case ID_SELECT_BRUSHES_CONTAININGSURFACEFLAG_HINT:
			Select_Deselect (true);
			SelectContainingFlag(true, false, 0x100);
			break;
		case ID_SELECT_BRUSHES_CONTAININGSURFACEFLAG_SKIP:
			Select_Deselect (true);
			SelectContainingFlag(true, false, 0x200);
			break;

		//Selection->Select->Faces menus
		case ID_SELECT_FACES_ALLFACES:
			Select_Deselect (true);
			Select_AllFaces ();
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCURRENTTEXTURE:
			Select_Deselect (true);
			SelectContainingTexture(false, &globals.d_texturewin.texdef);
			break;
			//Contents
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_SOLID:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_SOLID);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_WINDOW:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_WINDOW);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_AUX:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_AUX);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_LAVA:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_LAVA);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_SLIME:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_SLIME);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_WATER:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_WATER);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_MIST:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_MIST);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_PLAYERCLIP:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_PLAYERCLIP);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_MONSTERCLIP:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_MONSTERCLIP);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_CURRENT0:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_CURRENT_0);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_CURRENT90:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_CURRENT_90);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_CURRENT180:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_CURRENT_180);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_CURRENT270:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_CURRENT_270);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_CURRENTUP:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_CURRENT_UP);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_CURRENTDOWN:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_CURRENT_DOWN);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_ORIGIN:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_ORIGIN);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_MONSTER:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_MONSTER);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_CORPSE:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_DEADMONSTER);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_DETAIL:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_DETAIL);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_TRANSLUCENT:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_TRANSLUCENT);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGCONTENTFLAG_NODRAW:
			Select_Deselect (true);
			SelectContainingFlag(false, true, CONTENTS_LADDER);
			break;

			//Surface

		case ID_SELECTION_SELECT_FACES_CONTAININGSURFACEFLAG_LIGHT:
			Select_Deselect (true);
			SelectContainingFlag(false, false, SURF_LIGHT);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGSURFACEFLAG_SLICK:
			Select_Deselect (true);
			SelectContainingFlag(false, false, SURF_SLICK);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGSURFACEFLAG_SKY:
			Select_Deselect (true);
			SelectContainingFlag(false, false, SURF_SKY);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGSURFACEFLAG_WARP:
			Select_Deselect (true);
			SelectContainingFlag(false, false, SURF_WARP);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGSURFACEFLAG_TRANS33:
			Select_Deselect (true);
			SelectContainingFlag(false, false, SURF_TRANS33);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGSURFACEFLAG_TRANS66:
			Select_Deselect (true);
			SelectContainingFlag(false, false, SURF_TRANS66);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGSURFACEFLAG_FLOWING:
			Select_Deselect (true);
			SelectContainingFlag(false, false, SURF_FLOWING);
			break;
		case ID_SELECTION_SELECT_FACES_CONTAININGSURFACEFLAG_NODRAW:
			Select_Deselect (true);
			SelectContainingFlag(false, false, SURF_NODRAW);
			break;
		case ID_SELECTION_LOCKTEXTURES:
			ToggleAndCheck(ID_SELECTION_LOCKTEXTURES, &TexturesLocked);

			if(TexturesLocked)
				Sys_Status (0, "Texture Lock ON");
			else
				Sys_Status (0, "Texture Lock OFF");

			break;

		case ID_SELECT_FACES_BELONGINGTOSELECTEDBRUSHES:
			SelectFacesOfSelectedBrushes();
			break;

		case ID_SELECT_FACES_CONTAININGSURFACEFLAG_HINT:
			Select_Deselect (true);
			SelectContainingFlag(false, false, 0x100);
			break;

		case ID_SELECT_FACES_CONTAININGSURFACEFLAG_SKIP:
			Select_Deselect (true);
			SelectContainingFlag(false, false, 0x200);
			break;

		case ID_SELECT_ENTITIESBYKEYORVALUE:
			DoSelectEntitiesByKeyVal();
			break;

		case ID_SELECTION_GROUPBRUSHESTOENTITY:
			Entity_AddToEntityGroup();
			break;

		case ID_SELECTION_INVERTSELECTION:
			Select_InvertSelection();
			break;

		case ID_SELECTION_ADDTOCURRENTVISGROUP:
			VisGroup_SetSelected();
			break;

		case ID_SELECTION_FITTEXTURES:
			DoFitTextures();
			break;

		case ID_SELECTION_CHANGESURFACECONTENTFLAGS:
			DoChangeFlags();
			break;

		case ID_SELECTION_CLIPPER_CLIP:
			DoClip();
			break;

		case ID_SELECTION_CLIPPER_SPLIT:
			globals.ClipperMode = 3;
			break;

		case ID_SELECTION_CLIPPER_FRONT:
			globals.ClipperMode = 1;
			break;

		case ID_SELECTION_CLIPPER_BACK:
			globals.ClipperMode = 2;
			break;

//
// region menu
//
		case ID_REGION_OFF:
			Map_RegionOff ();
			break;
		case ID_REGION_SETXY:
			Map_RegionXY ();
			break;
		case ID_REGION_SETXZ:
			Map_RegionXZ();
			break;
		case ID_REGION_SETYZ:
			Map_RegionYZ();
		case ID_REGION_SETTALLBRUSH:
			Map_RegionTallBrush ();
			break;
		case ID_REGION_SETBRUSH:
			Map_RegionBrush ();
			break;
		case ID_REGION_SETSELECTION:
			Map_RegionSelectedBrushes ();
			break;
		case ID_REGION_SETNONSELECTED:
			Map_RegionSet_Nonselected ();
			break;
		case ID_REGION_SWAPREGIONEDWITHNONSELECTED:
			Map_RegionSwap();
			break;
		case ID_REGION_SETCURRENTVISGROUP:
			Map_RegionVisGroup();
			break;

		case IDMRU+1:
		case IDMRU+2:
		case IDMRU+3:
		case IDMRU+4:
		case IDMRU+5:
		case IDMRU+6:
		case IDMRU+7:
		case IDMRU+8:
		case IDMRU+9:
			DoMru(hWnd,LOWORD(wParam));
			break;

//
// help menu
//

		case ID_HELP_ABOUT:
			DoAbout();
			break;
		
		case ID_TOOLBARTEXTURECOMBO:
			switch (HIWORD(wParam))
			{
				case CBN_DBLCLK:
				{
					int			i;
					char		sz[256];
					qtexture_t	*tex;
					texdef_t	t;

					i = SendMessage(globals.d_hwndToolbarTextureCombo, CB_GETCURSEL, 0, 0);
					SendMessage(globals.d_hwndToolbarTextureCombo, CB_GETLBTEXT, i, (LPARAM)sz);
					Sys_Printf("%s\n", sz);
					tex = Texture_ForName(sz);
					t.contents = tex->contents;
					t.flags = tex->flags;
					strcpy(t.name, tex->name);
					t.rotate = 0;
					t.scale[0] = t.scale[1] = 1.0f;
					t.shift[0] = t.shift[1] = 0;
					t.value = tex->value;
					if(GetKeyState(VK_SHIFT) < 0)
					{
						Texture_SetTexture (&t, true, true);
					}
					else
					{
						Texture_SetTexture (&t, true, false);
					}
				}
				break;
			}
			break;
		case ID_TOOLBARENTITYCOMBO:
			switch (HIWORD(wParam))
			{
				case CBN_DBLCLK:
				{
					int			i;
					eclass_t	*pec;
					char		sz[256];
					
					if(selected_brushes.next == &selected_brushes)
					{
						MessageBox(globals.d_hwndMain, "Must have at least one worldspawn brush selected to create an entity.", "info", 0);
						return TRUE;
					}
					i = SendMessage(globals.d_hwndToolbarEntityCombo, CB_GETCURSEL, 0, 0);
					SendMessage(globals.d_hwndToolbarEntityCombo, CB_GETLBTEXT, i, (LPARAM)sz);
					pec = Eclass_ForName(sz, false);
					if(!Entity_Create(pec))
						MessageBox(globals.d_hwndMain, "Failed to create entity.", "info", 0);
					return TRUE;
					break;
				}
				case CBN_SELCHANGE: 
				{
					int iIndex;
					
					iIndex = SendMessage(globals.d_hwndToolbarEntityCombo, CB_GETCURSEL, 0, 0);
					SendMessage(hwndEnt[EntList], LB_SETCURSEL, iIndex, 0);
					SendMessage (globals.Entity.hwnd, WM_COMMAND, (LBN_SELCHANGE<<16) + IDC_E_LIST, 0 );
					
					return TRUE; 
					break; 
				}
			}
			break;

		default:
			if(LOWORD(wParam) >= CMD_TEXTURECATEGORY && LOWORD(wParam) < CMD_TEXTURECATEGORY+256)
			{
				Sys_BeginWait ();
				Texture_ShowCategory (LOWORD(wParam));
				SetInspectorMode(W_TEXTURE);
				break;
			}

			if(LOWORD(wParam) >= CMD_TEXTUREWAD && LOWORD(wParam) <CMD_TEXTUREWAD+256)
			{
				Sys_BeginWait ();
				Texture_ShowDirectory (LOWORD(wParam));
				SetInspectorMode(W_TEXTURE);
				break;
			}

            return FALSE;
    }

    return TRUE;
}

BOOL WriteIniInt(char *Key, int IntVar)
{
	char buf[64];

	sprintf(buf, "%d", IntVar);
	if(!WritePrivateProfileString("Settings", Key, buf, IniFile))
	{
		Error("Could not write int value to INI file");
		return false;
	}

	return true;
}

int     curtime;
int Sys_Milliseconds (void)
{
	static int		base;
	static qboolean	initialized = false;
	
	if (!initialized)
	{       // let base retain 16 bits of effectively random data
		base = timeGetTime() & 0xffff0000;
		initialized = true;
	}
	curtime = timeGetTime() - base;
	
	return curtime;
}

/*
============
WMAIN_WndProc
============
*/
extern float rs_realtime;
void CIN_ProcessCins (void);

LONG WINAPI WMAIN_WndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    LONG    lRet = 1;
    RECT	rect;
	int		i;


    GetClientRect(hWnd, &rect);

    switch (uMsg)
    {
	case WM_TIMER:
		if(wParam == QE_TIMER0)
		{
			QE_CountBrushesAndUpdateStatusBar();
			QE_CheckAutoSave();
		}
		else
		{
			if(globals.AnimateCameraRScripts || globals.AnimateTextureRScripts)
			{
				rs_realtime=Sys_Milliseconds() * 0.001f;
				CIN_ProcessCins();
			}
			Sys_UpdateWindows((W_TEXTURE*globals.AnimateTextureRScripts) | (W_CAMERA*(globals.AnimateCameraRScripts|globals.AnimateSky)));
		}
		return 0;

	case WM_DESTROY:
#if 0
		SaveMruInReg(globals.d_lpMruMenu,"Software\\id\\QuakeEd4\\MRU");
#else
		if(!SaveMruInIni(globals.d_lpMruMenu,"MRU", IniFile))
			Error("Could not save recent file list to INI file");
#endif
		WritePrivateProfileStruct	("Settings","Gamma",				&globals.d_savedinfo.fGamma,	sizeof(globals.d_savedinfo.fGamma), IniFile);
		WritePrivateProfileStruct	("Settings","Colors",				&globals.d_savedinfo.colors,	sizeof(globals.d_savedinfo.colors), IniFile);
		WritePrivateProfileStruct	("Settings","Alpha",				&globals.d_savedinfo.alpha,		sizeof(globals.d_savedinfo.alpha),	IniFile);
		WritePrivateProfileString	("Settings","LastProject",			globals.d_savedinfo.szProject,	IniFile);
		WriteIniInt					("ShowNames",						globals.d_savedinfo.show_names);
		WriteIniInt					("RenderMode",						globals.d_savedinfo.iTexMenu);
		WriteIniInt					("ShowCoordinates",					globals.d_savedinfo.show_coordinates);
		WriteIniInt					("NoConsoleTexinfo",				globals.d_savedinfo.no_console_texture);
		WriteIniInt					("NoConsoleScanfile",				globals.d_savedinfo.no_console_scanfile);
		WriteIniInt					("Exclude",							globals.d_savedinfo.exclude);
		WriteIniInt					("UseDepthClipping",				globals.DepthClipping);
		WriteIniInt					("DepthClipDistance",				globals.DepthClipDistance);
		WriteIniInt					("ShowBlocks",						globals.show_blocks);
		WriteIniInt					("TransparencyActive",				globals.transactive);
		WriteIniInt					("Trans33",							globals.trans33);
		WriteIniInt					("Trans66",							globals.trans66);
		WriteIniInt					("Nodraw",							globals.nodraw);
		WriteIniInt					("TextureLock",						TexturesLocked);
		WriteIniInt					("FollowFloor",						globals.followfloor);
		WriteIniInt					("StepHeight",						globals.StepHeight);
		WriteIniInt					("UpdateCamera",					globals.UpdateCamera);
		WriteIniInt					("LockXYZ2Cam",						globals.LockXYZ2Cam);
		WriteIniInt					("Lock2DOrigins",					globals.Lock2DOrigins);
		WriteIniInt					("Lock2DZoom",						globals.Lock2DZoom);
		WriteIniInt					("LockCamTo2DOrigins",				globals.LockCamTo2DOrigins);
		WriteIniInt					("BrushAndFaceSelection",			globals.AllowFaceAndBrushSelection);
		WriteIniInt					("ShowEntityConnections",			globals.ShowEntityConnections);
		WriteIniInt					("ShowSkybox",						globals.ShowSkybox);
		WriteIniInt					("XYMode",							globals.d_xy.draw_mode);
		WriteIniInt					("ShowTextureNames",				globals.ShowTextureNames);
		WriteIniInt					("Clip2XYZ",						globals.Clip2XYZ);
		WriteIniInt					("ShowHint",						globals.ShowHint);
		WriteIniInt					("WriteMapFloats",					globals.WriteMapFloats);
		WriteIniInt					("UseExternalTextures",				globals.UseExternalTextures);
		WriteIniInt					("StippleSelection",				globals.StippleSelection);
		WriteIniInt					("AlwaysUseDefaultWindowPlacement",	globals.AlwaysUseDefaultWindowPlacement);
		WriteIniInt					("DrawVisGroupColors",				globals.DrawVisGroupColors);
		WriteIniInt					("ShowFullSurfaceEditor",			globals.ShowFullSurfaceEditor);
		WriteIniInt					("OffsetCopiedBrushes",				globals.OffsetCopiedBrushes);
		WriteIniInt					("Multiple2DViews",					globals.Multiple2DViews);
		WriteIniInt					("AnimateCameraRScripts",			globals.AnimateCameraRScripts);
		WriteIniInt					("AnimateTextureRScripts",			globals.AnimateTextureRScripts);
		WriteIniInt					("AnimateSky",						globals.AnimateSky);
		WriteIniInt					("AnimationSpeed",					globals.AnimationSpeed);
		WriteIniInt					("GridExtent",						globals.MaxGridExtent);
		WriteIniInt					("OldCameraMovement",				globals.OldCameraMovement);
		SaveClientWindowStates();
		SaveWindowState(globals.d_hwndMain,		"mainwindow");

		for(i = 0; i < 32; i++)
		{
			char temp[32];

			sprintf(temp, "%d", i);
			WritePrivateProfileString("SurfaceFlagDescriptions", temp, Surface_FlagNames[i], IniFile);
			WritePrivateProfileString("ContentFlagDescriptions", temp, Content_FlagNames[i], IniFile);
		}

		DeleteMruMenu(globals.d_lpMruMenu);
		PostQuitMessage(0);
		KillTimer( hWnd, QE_TIMER0 );
		
		if(globals.AnimateCameraRScripts || globals.AnimateTextureRScripts || globals.AnimateSky)
			KillTimer( hWnd, QE_TIMER1 );
		return 0;

	case WM_CREATE:
//	    QEW_SetupPixelFormat(maindc, false);
		globals.d_lpMruMenu = CreateMruMenuDefault();
#if 0
		LoadMruInReg(globals.d_lpMruMenu,"Software\\id\\QuakeEd4\\MRU");
#else
		LoadMruInIni(globals.d_lpMruMenu,"MRU", IniFile);
#endif
		
		// Refresh the File menu.
		PlaceMenuMRUItem(globals.d_lpMruMenu,GetSubMenu(GetMenu(hWnd),0), ID_FILE_EXIT);

		return 0;

	case WM_SIZE:
		// resize the status window
		MoveWindow( globals.d_hwndStatus, -100, 100, 10, 10, true);
		// resize the toolbar
		SendMessage(globals.d_hwndToolbar, TB_AUTOSIZE, 0, 0); 

		if(globals.AlwaysUseDefaultWindowPlacement)
			SetDefaultWindowPositions();

		return 0;

	case WM_KEYDOWN:
		return QE_KeyDown (wParam);

   	case WM_CLOSE:
        /* call destroy window to cleanup and go away */
		if (!ConfirmModified())
			return TRUE;

        DestroyWindow (hWnd);
		return 0;

    case WM_COMMAND:
		return CommandHandler (hWnd, wParam, lParam);
		return 0;
	case WM_NOTIFY:
            switch (((LPNMHDR) lParam)->code) 
            {
                case TTN_NEEDTEXT:    
                    // Display tool tip text.
			        lpToolTipText = (LPTOOLTIPTEXT)lParam;
    				LoadString (globals.d_hInstance,
	    				lpToolTipText->hdr.idFrom,    // string ID == cmd ID
		    			szBuf,
			    		sizeof(szBuf));
				    	lpToolTipText->lpszText = szBuf;
			        break;

/*                case TBN_QUERYDELETE:
                    // Toolbar customization -- can we delete this button?
                    return TRUE;
                    break;

                case TBN_GETBUTTONINFO:
                    // The toolbar needs information about a button.
                    return FALSE;
                    break;

                case TBN_QUERYINSERT:
                    // Can this button be inserted? Just say yo.
                    return TRUE;
                    break;

                case TBN_CUSTHELP:
                    // Need to display custom help.
                    MessageBox(hWnd, "This help is custom.",NULL, MB_OK);
                    break;

                case TBN_TOOLBARCHANGE:
                    // Done dragging a bitmap to the toolbar.
                    SendMessage(s_hwndToolbar, TB_AUTOSIZE, 0L, 0L);
                    break;
*/
                default:
                    return TRUE;
                    break;
            }
		break;

	default:
		if(uMsg == MOUSEWHEEL)
		{
			POINT	point;
			int		fwKeys = LOWORD(wParam);
			HWND	TargetWindow;
			short	zDelta = (short)HIWORD(wParam);

			point.x = (short) LOWORD(lParam);
			point.y = (short) HIWORD(lParam);
			TargetWindow = ChildWindowFromPoint(globals.d_hwndMain, point);
//			Sys_Printf("Mouse wheel wParam = %d",(short)HIWORD(wParam));

			if(TargetWindow == globals.Camera.hwnd)
			{
				if(fwKeys == (MK_RBUTTON|MK_SHIFT|MK_CONTROL)) // Freelook
				{
					VectorMA (camera.origin, zDelta<0?-32:32, camera.forward, camera.origin);
					Sys_UpdateWindows(W_CAMERA|W_XY|W_XZ|W_YZ|W_Z);
					return 0;
				}

				if(zDelta < 0)
				{
					if(GetKeyState(VK_SHIFT) < 0)
					{
						camera.origin[2] -= 16;
						Sys_UpdateWindows (W_CAMERA|W_Z);
					}
					else
					{
						camera.angles[0] += 22.5;
						if (camera.angles[0] > 85)
							camera.angles[0] = 85;
						Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
					}
				}
				else
				{
					if(GetKeyState(VK_SHIFT) < 0)
					{
						camera.origin[2] += 16;
						Sys_UpdateWindows (W_CAMERA|W_Z);
					}
					else
					{
						camera.angles[0] -= 22.5;
						if (camera.angles[0] < -85)
							camera.angles[0] = -85;
						Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
					}
				}
			}
			else if(TargetWindow == globals.Entity.hwnd)
			{
				fwKeys |= MK_RBUTTON;
				if(zDelta < 0)
					ScrollTextureWindow(-globals.d_texturewin.height/4, fwKeys);
				else
					ScrollTextureWindow(globals.d_texturewin.height/4, fwKeys);
			}
			else if(TargetWindow == globals.Z.hwnd)
			{
				if(zDelta < 0)
				{
					z.origin[2] += (z.height/4)/z.scale;
				}
				else
				{
					z.origin[2] -= (z.height/4)/z.scale;
				}
				Sys_UpdateWindows (W_Z);
			}
			return 0;
		}
		break;
    }

    return DefWindowProc (hWnd, uMsg, wParam, lParam);
}




/*
==============
Main_Create
==============
*/
void Main_Create (HINSTANCE hInstance)
{
    WNDCLASS	wc;
	HMENU		hMenu;

    /* Register the camera class */
	memset (&wc, 0, sizeof(wc));

    wc.style         = 0;
    wc.lpfnWndProc   = (WNDPROC)WMAIN_WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = 0;
    wc.hCursor       = LoadCursor (NULL,IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = MAKEINTRESOURCE(IDR_MENU1);
    wc.lpszClassName = "QUAKE_MAIN";

    if (!RegisterClass (&wc) )
        Error ("WCam_Register: failed");


	globals.d_hwndMain = globals.Main.hwnd = CreateWindow ("QUAKE_MAIN" ,
		"QuakeEd 5",
		WS_OVERLAPPEDWINDOW |
		WS_CLIPSIBLINGS |
		WS_CLIPCHILDREN,
		0,0,screen_width,screen_height+GetSystemMetrics(SM_CYSIZE),	// size
		0,
		0,		// no menu
		hInstance,
		NULL);
	if (!globals.d_hwndMain)
		Error ("Couldn't create main window");

	/* create a timer so that we can count brushes */
	SetTimer(globals.d_hwndMain, QE_TIMER0, 1000, NULL);
	
	LoadWindowState(globals.d_hwndMain, "mainwindow");

	if((globals.d_hwndToolbar = CreateToolBar(hInstance)) != NULL);
	{		
		hWndTT = (HWND)SendMessage(globals.d_hwndToolbar, TB_GETTOOLTIPS, 0, 0);
	}
	
	globals.d_hwndStatus = CreateMyStatusWindow(hInstance);

	selected_faces = NULL;
	SurfaceDlgActive = false;

	if ((hMenu = globals.Menu = GetMenu(globals.d_hwndMain)) != 0)
	{
		CheckMenu(ID_VIEW_SHOWNAMES,						globals.d_savedinfo.show_names);
		CheckMenu(ID_VIEW_SHOWCOORDINATES,					globals.d_savedinfo.show_coordinates);
		CheckMenu(ID_VIEW_SHOWLIGHTS,						!(globals.d_savedinfo.exclude&EXCLUDE_LIGHTS));
		CheckMenu(ID_VIEW_SHOWENT,							!(globals.d_savedinfo.exclude&EXCLUDE_ENT));
		CheckMenu(ID_VIEW_SHOWPATH,							!(globals.d_savedinfo.exclude&EXCLUDE_PATHS));
		CheckMenu(ID_VIEW_SHOWWATER,						!(globals.d_savedinfo.exclude&EXCLUDE_WATER));
		CheckMenu(ID_VIEW_SHOWWORLD,						!(globals.d_savedinfo.exclude&EXCLUDE_WORLD));
		CheckMenu(ID_VIEW_SHOWCLIP,							!(globals.d_savedinfo.exclude&EXCLUDE_CLIP));
		CheckMenu(ID_VIEW_SHOWDETAIL,						!(globals.d_savedinfo.exclude&EXCLUDE_DETAIL));
		CheckMenu(ID_VIEW_SHOWBLOCKS,						globals.show_blocks);
		CheckMenu(ID_VIEW_TRANSPARENCY_ACTIVE,				globals.transactive);
		CheckMenu(ID_VIEW_TRANSPARENCY_TRANS33,				globals.trans33);
		CheckMenu(ID_VIEW_TRANSPARENCY_TRANS66,				globals.trans66);
		CheckMenu(ID_VIEW_TRANSPARENCY_NODRAW,				globals.nodraw);
		CheckMenu(ID_SELECTION_LOCKTEXTURES,				TexturesLocked);
		CheckMenu(ID_MISC_FOLLOWFLOOR,						globals.followfloor);
		CheckMenu(ID_MISC_ALWAYSUPDATECAMERA,				globals.UpdateCamera);
		CheckMenu(ID_MISC_ALLOWFACEANDBRUSHSELECTION,		globals.AllowFaceAndBrushSelection);
		CheckMenu(ID_VIEW_SHOWENTITYCONNECTIONS,			globals.ShowEntityConnections);
		CheckMenu(ID_VIEW_SHOWSKYBOX,						globals.ShowSkybox);
		CheckMenu(ID_TEXTURES_XYVIEWRENDERMODE_WIREFRAME,	globals.d_xy.draw_mode == cd_wire);
		CheckMenu(ID_TEXTURES_XYVIEWRENDERMODE_FLATSHADE,	globals.d_xy.draw_mode == cd_solid);
		CheckMenu(ID_TEXTURES_XYVIEWRENDERMODE_TEXTURED,	globals.d_xy.draw_mode == cd_texture);
		CheckMenu(ID_VIEW_SHOWTEXTURENAMES,					globals.ShowTextureNames);
		CheckMenu(ID_VIEW_CLIP2XYZ,							globals.Clip2XYZ);
		CheckMenu(ID_VIEW_SHOWHINT,							globals.ShowHint);
		CheckMenu(ID_MISC_WRITEMAPFLOATS,					globals.WriteMapFloats);
		CheckMenu(ID_VIEW_XYVIEW_STIPPLESELECTIONOUTLINE,	globals.StippleSelection);
		CheckMenu(ID_MISC_ALWAYSUSEDEFAULTWINDOWPLACEMENT,	globals.AlwaysUseDefaultWindowPlacement);
		CheckMenu(ID_VIEW_DRAWVISGROUPCOLORS,				globals.DrawVisGroupColors);
		CheckMenu(ID_VIEW_SHOWFULLSURFACEEDITOR,			globals.ShowFullSurfaceEditor);
		CheckMenu(ID_MISC_OFFSETCOPIEDBRUSHES,				globals.OffsetCopiedBrushes);
		CheckMenu(ID_MISC_NO_TEXTURE,						globals.d_savedinfo.no_console_texture);
		CheckMenu(ID_MISC_NO_SCANFILE,						globals.d_savedinfo.no_console_scanfile);
	}

	ShowWindow (globals.d_hwndMain, SW_SHOWMAXIMIZED);
	globals.Main.hdc = GetDC(globals.d_hwndMain);
	QEW_SetupPixelFormat(&(globals.Main.pfd), &(globals.Main.hdc), false);

	if ((globals.Main.hglrc = wglCreateContext(globals.Main.hdc)) == 0)
		Error ("wglCreateContext failed in Main_Create()");

	if (!wglMakeCurrent(globals.Main.hdc, globals.Main.hglrc))
		Error ("wglMakeCurrent failed in Main_Create()");
	if(!globals.DepthClipping)
		Sys_Status(2, "Depth Clip OFF");
	else
		Sys_Status(2, "Clip Distance: %d", globals.DepthClipDistance);
	SetUpFont(&(globals.Main));
}


/*
=============================================================

REGISTRY INFO

=============================================================
*/

BOOL SaveRegistryInfo(const char *pszName, void *pvBuf, long lSize)
{
	LONG lres;
	DWORD dwDisp;
	HKEY  hKeyId;

	lres = RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\id\\QuakeEd4", 0, NULL, 
			REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKeyId, &dwDisp);
    
	if (lres != ERROR_SUCCESS)
		return FALSE;

	lres = RegSetValueEx(hKeyId, pszName, 0, REG_BINARY, pvBuf, lSize);
    
	RegCloseKey(hKeyId);

	if (lres != ERROR_SUCCESS)
		return FALSE;

	return TRUE;
}

BOOL LoadRegistryInfo(const char *pszName, void *pvBuf, long *plSize)
{
	HKEY  hKey;
	long lres, lType, lSize;

	if (plSize == NULL)
		plSize = &lSize;

	lres = RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\id\\QuakeEd4", 0, KEY_READ, &hKey);
    
	if (lres != ERROR_SUCCESS)
		return FALSE;

	lres = RegQueryValueEx(hKey, pszName, NULL, &lType, pvBuf, plSize);

	RegCloseKey(hKey);

	if (lres != ERROR_SUCCESS)
		return FALSE;

	return TRUE;
}

BOOL SaveWindowState(HWND hWnd, const char *pszName)
{
	RECT rc;
	
	GetWindowRect(hWnd, &rc);
	if (hWnd != globals.d_hwndMain)
		MapWindowPoints(NULL, globals.d_hwndMain, (POINT *)&rc, 2);
//	return SaveRegistryInfo(pszName, &rc, sizeof(rc));
	return WritePrivateProfileStruct("WindowPositions",pszName,&rc,sizeof(rc),IniFile);
}

void SaveClientWindowStates(void)
{
	if(globals.Multiple2DViews)
	{
		SaveWindowState(globals.XY.hwnd,		"xywindow");
		SaveWindowState(globals.YZ.hwnd,		"yzwindow");
		SaveWindowState(globals.XZ.hwnd,		"xzwindow");
		SaveWindowState(globals.Camera.hwnd,	"camerawindow");
		SaveWindowState(globals.Z.hwnd,			"zwindow");
		SaveWindowState(globals.Entity.hwnd,	"EntityWindow");
	}
	else
	{
		SaveWindowState(globals.XY.hwnd,		"2Dwindow");
		SaveWindowState(globals.Camera.hwnd,	"camerawindow2");
		SaveWindowState(globals.Z.hwnd,			"zwindow2");
		SaveWindowState(globals.Entity.hwnd,	"EntityWindow2");
	}
}

BOOL LoadWindowState(HWND hWnd, const char *pszName)
{
	RECT rc;
	LONG lSize = sizeof(rc);

//	if (LoadRegistryInfo(pszName, &rc, &lSize))
	if (GetPrivateProfileStruct("WindowPositions",pszName,&rc,sizeof(rc),IniFile))
	{
		if (rc.left < 0)
			rc.left = 0;
		if (rc.top < 0)
			rc.top = 0;
		if (rc.right < rc.left + 16)
			rc.right = rc.left + 16;
		if (rc.bottom < rc.top + 16)
			rc.bottom = rc.top + 16;

		MoveWindow(hWnd, rc.left, rc.top, rc.right - rc.left, 
				rc.bottom - rc.top, FALSE);
		return TRUE;
	}

	return FALSE;
}

/*
===============================================================

  STATUS WINDOW

===============================================================
*/

void Sys_UpdateStatusBar( void )
{
	extern int   g_numbrushes, g_numentities;

	Sys_Status(3, "Brushes: %d Entities: %d", g_numbrushes, g_numentities );
	Sys_Status(4, "Grid size: %d", globals.d_gridsize);
}

void Sys_Status(int part, char *format, ...)
{
	va_list argptr;
	char	buf[512];
	char	*out;

	va_start (argptr, format);
	vsprintf (buf, format, argptr);
	va_end (argptr);

	out = TranslateString (buf);

	SendMessage(globals.d_hwndStatus, SB_SETTEXT, part, (LPARAM)out);
}

static HWND CreateMyStatusWindow(HINSTANCE hInst)
{
	HWND hWnd;
	int partsize[5] = { 150, 400, 550, 700,	1100 };

	hWnd = CreateWindowEx( WS_EX_TOPMOST,	// no extended styles
            STATUSCLASSNAME,				// status bar
            "",								// no text 
            SBARS_SIZEGRIP | WS_CHILD |
			WS_BORDER | WS_VISIBLE,			// styles
            -100, -100, 10, 10,				// x, y, cx, cy
            globals.d_hwndMain,			// parent window
            (HMENU)100,						// window ID
            hInst,							// instance
            NULL);							// window data

	SendMessage( hWnd, SB_SETPARTS, 5, ( long ) partsize );
    
	return hWnd;
}

//==============================================================

LRESULT (CALLBACK* lpfnDefComboTexture) (HWND, UINT, WPARAM, LPARAM);
LRESULT (CALLBACK* lpfnDefComboEntity) (HWND, UINT, WPARAM, LPARAM);

LRESULT CALLBACK ToolbarTextureComboWndProc (HWND hWnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
	// Handle tooltips for the toolbar comboboxes and double-clicks and keyboard enters
	switch (uMessage)
	{
		case WM_MOUSEMOVE:
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		{
			MSG msg;
			msg.lParam = lParam;
			msg.wParam = wParam;
			msg.message = uMessage;
			msg.hwnd = hWnd;
			SendMessage (hWndTT, TTM_RELAYEVENT, 0, (LPARAM)(LPMSG)&msg);
			break;
		}
		case WM_KEYDOWN:
			if (LOWORD(wParam) == VK_RETURN)
			{
				SendMessage ( globals.d_hwndMain, WM_COMMAND, (CBN_DBLCLK<<16) + ID_TOOLBARTEXTURECOMBO, 0 );
				return 0;
			}
			break;
	}
	return CallWindowProc (lpfnDefComboTexture, hWnd, uMessage, wParam,lParam);
} 

LRESULT CALLBACK ToolbarEntityComboWndProc (HWND hWnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
	// Handle tooltips for the toolbar comboboxes and double-clicks and keyboard enters
	switch (uMessage)
	{
		case WM_MOUSEMOVE:
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		{
			MSG msg;
			msg.lParam = lParam;
			msg.wParam = wParam;
			msg.message = uMessage;
			msg.hwnd = hWnd;
			SendMessage (hWndTT, TTM_RELAYEVENT, 0, (LPARAM)(LPMSG)&msg);
			break;
		}
		case WM_KEYDOWN:
			if (LOWORD(wParam) == VK_RETURN)
			{
				SendMessage ( globals.d_hwndMain, WM_COMMAND, (CBN_DBLCLK<<16) + ID_TOOLBARENTITYCOMBO, 0 );
				return 0;
			}
			break;
	}
	return CallWindowProc (lpfnDefComboEntity, hWnd, uMessage, wParam,lParam);
} 

#define NUMBUTTONS 23
HWND CreateToolBar(HINSTANCE hinst)
{ 
    HWND hwndTB, hWndTT;
    TBADDBITMAP tbab;
    TBBUTTON tbb[NUMBUTTONS];
    
     // Ensure that the common control DLL is loaded. 

	    InitCommonControls(); 
 
    // Create a toolbar that the user can customize and that has a 
    // tooltip associated with it. 

    hwndTB = CreateWindowEx(0, TOOLBARCLASSNAME, (LPSTR) NULL, 
        WS_CHILD | TBSTYLE_TOOLTIPS /*| CCS_ADJUSTABLE*/ | WS_BORDER, 
        0, 0, 0, 0, globals.d_hwndMain, (HMENU) IDR_TOOLBAR1, hinst, NULL); 
 
    // Send the TB_BUTTONSTRUCTSIZE message, which is required for 
    // backward compatibility. 

    SendMessage(hwndTB, TB_BUTTONSTRUCTSIZE, (WPARAM) sizeof(TBBUTTON), 0);
 
    // Add the bitmap containing button images to the toolbar. 

    tbab.hInst = hinst; 
    tbab.nID   = IDR_TOOLBAR1; 
    SendMessage(hwndTB, TB_ADDBITMAP, (WPARAM)NUMBUTTONS, (WPARAM) &tbab); 
 
    // Fill the TBBUTTON array with button information, and add the 
    // buttons to the toolbar. 

    tbb[0].iBitmap = 15;
    tbb[0].idCommand = ID_FILE_NEW;
    tbb[0].fsState = TBSTATE_ENABLED;
    tbb[0].fsStyle = TBSTYLE_BUTTON;
    tbb[0].dwData = 0;
    tbb[0].iString = 0;
 
    tbb[1].iBitmap = 16;
    tbb[1].idCommand = ID_FILE_OPEN;
    tbb[1].fsState = TBSTATE_ENABLED;
    tbb[1].fsStyle = TBSTYLE_BUTTON;
    tbb[1].dwData = 0;
    tbb[1].iString = 0;
 
    tbb[2].iBitmap = 17;
    tbb[2].idCommand = ID_FILE_SAVE;
    tbb[2].fsState = TBSTATE_ENABLED;
    tbb[2].fsStyle = TBSTYLE_BUTTON;
    tbb[2].dwData = 0;
    tbb[2].iString = 0;
 
    tbb[3].iBitmap = 0;
    tbb[3].idCommand = 0;
    tbb[3].fsState = TBSTATE_ENABLED;
    tbb[3].fsStyle = TBSTYLE_SEP;
    tbb[3].dwData = 0;
    tbb[3].iString = 0;
 
    tbb[4].iBitmap = 0;
    tbb[4].idCommand = ID_BRUSH_FLIPX;
    tbb[4].fsState = TBSTATE_ENABLED;
    tbb[4].fsStyle = TBSTYLE_BUTTON;
    tbb[4].dwData = 0;
    tbb[4].iString = 0;
 
    tbb[5].iBitmap = 2; 
    tbb[5].idCommand = ID_BRUSH_FLIPY; 
    tbb[5].fsState = TBSTATE_ENABLED; 
    tbb[5].fsStyle = TBSTYLE_BUTTON; 
    tbb[5].dwData = 0; 
    tbb[5].iString = 0;
 
    tbb[6].iBitmap = 4; 
    tbb[6].idCommand = ID_BRUSH_FLIPZ; 
    tbb[6].fsState = TBSTATE_ENABLED;
    tbb[6].fsStyle = TBSTYLE_BUTTON; 
    tbb[6].dwData = 0; 
    tbb[6].iString = 0; 
 
    tbb[7].iBitmap = 0;
    tbb[7].idCommand = 0;
    tbb[7].fsState = TBSTATE_ENABLED;
    tbb[7].fsStyle = TBSTYLE_SEP;
    tbb[7].dwData = 0;
    tbb[7].iString = 0;
 
    tbb[8].iBitmap = 1; 
    tbb[8].idCommand = ID_BRUSH_ROTATEX; 
    tbb[8].fsState = TBSTATE_ENABLED; 
    tbb[8].fsStyle = TBSTYLE_BUTTON; 
    tbb[8].dwData = 0; 
    tbb[8].iString = 0; 

    tbb[9].iBitmap = 3; 
    tbb[9].idCommand = ID_BRUSH_ROTATEY; 
    tbb[9].fsState = TBSTATE_ENABLED; 
    tbb[9].fsStyle = TBSTYLE_BUTTON; 
    tbb[9].dwData = 0; 
    tbb[9].iString = 0; 

    tbb[10].iBitmap = 5; 
    tbb[10].idCommand = ID_BRUSH_ROTATEZ; 
    tbb[10].fsState = TBSTATE_ENABLED; 
    tbb[10].fsStyle = TBSTYLE_BUTTON; 
    tbb[10].dwData = 0; 
    tbb[10].iString = 0; 

    tbb[11].iBitmap = 0;
    tbb[11].idCommand = 0;
    tbb[11].fsState = TBSTATE_ENABLED;
    tbb[11].fsStyle = TBSTYLE_SEP;
    tbb[11].dwData = 0;
    tbb[11].iString = 0;
 
    tbb[12].iBitmap = 6; 
    tbb[12].idCommand = ID_SELECTION_SELECTCOMPLETETALL; 
    tbb[12].fsState = TBSTATE_ENABLED; 
    tbb[12].fsStyle = TBSTYLE_BUTTON; 
    tbb[12].dwData = 0; 
    tbb[12].iString = 0; 

    tbb[13].iBitmap = 7; 
    tbb[13].idCommand = ID_SELECTION_SELECTTOUCHING; 
    tbb[13].fsState = TBSTATE_ENABLED; 
    tbb[13].fsStyle = TBSTYLE_BUTTON; 
    tbb[13].dwData = 0; 
    tbb[13].iString = 0; 

    tbb[14].iBitmap = 8; 
    tbb[14].idCommand = ID_SELECTION_SELECTPARTIALTALL; 
    tbb[14].fsState = TBSTATE_ENABLED; 
    tbb[14].fsStyle = TBSTYLE_BUTTON; 
    tbb[14].dwData = 0; 
    tbb[14].iString = 0; 


    tbb[15].iBitmap = 9; 
    tbb[15].idCommand = ID_SELECTION_SELECTINSIDE; 
    tbb[15].fsState = TBSTATE_ENABLED; 
    tbb[15].fsStyle = TBSTYLE_BUTTON; 
    tbb[15].dwData = 0; 
    tbb[15].iString = 0; 

    tbb[16].iBitmap = 0;
    tbb[16].idCommand = 0;
    tbb[16].fsState = TBSTATE_ENABLED;
    tbb[16].fsStyle = TBSTYLE_SEP;
    tbb[16].dwData = 0;
    tbb[16].iString = 0;
 
    tbb[17].iBitmap = 10; 
    tbb[17].idCommand = ID_SELECTION_CSGSUBTRACT; 
    tbb[17].fsState = TBSTATE_ENABLED; 
    tbb[17].fsStyle = TBSTYLE_BUTTON; 
    tbb[17].dwData = 0; 
    tbb[17].iString = 0; 


    tbb[18].iBitmap = 11;
    tbb[18].idCommand = ID_SELECTION_MAKEHOLLOW; 
    tbb[18].fsState = TBSTATE_ENABLED; 
    tbb[18].fsStyle = TBSTYLE_BUTTON; 
    tbb[18].dwData = 0; 
    tbb[18].iString = 0; 

    tbb[19].iBitmap = 0;
    tbb[19].idCommand = 0;
    tbb[19].fsState = TBSTATE_ENABLED;
    tbb[19].fsStyle = TBSTYLE_SEP;
    tbb[19].dwData = 0;
    tbb[19].iString = 0;
 
    tbb[20].iBitmap = 12;
    tbb[20].idCommand = ID_TEXTURES_WIREFRAME; 
    tbb[20].fsState = TBSTATE_ENABLED; 
    tbb[20].fsStyle = TBSTYLE_BUTTON; 
    tbb[20].dwData = 0; 
    tbb[20].iString = 0;

    tbb[21].iBitmap = 13;
    tbb[21].idCommand = ID_TEXTURES_FLATSHADE;
    tbb[21].fsState = TBSTATE_ENABLED;
    tbb[21].fsStyle = TBSTYLE_BUTTON;
    tbb[21].dwData = 0;
    tbb[21].iString = 0;

    tbb[22].iBitmap = 14;
    tbb[22].idCommand = ID_VIEW_TRILINEAR;
    tbb[22].fsState = TBSTATE_ENABLED;
    tbb[22].fsStyle = TBSTYLE_BUTTON;
    tbb[22].dwData = 0;
    tbb[22].iString = 0;

    SendMessage(hwndTB, TB_ADDBUTTONS, (WPARAM)NUMBUTTONS, (LPARAM) (LPTBBUTTON) &tbb);
 
	globals.d_hwndToolbarTextureCombo = CreateWindowEx (0L,								// no extended styles
													"COMBOBOX",							// class name
													"", // default text
													WS_CHILD | WS_BORDER | WS_VISIBLE | WS_VSCROLL |
													CBS_AUTOHSCROLL | CBS_DROPDOWNLIST,	// styles and defaults
													465, 0, 250, 250,					// size and position
													hwndTB,								// parent window
													(HMENU)ID_TOOLBARTEXTURECOMBO,		// ID
													hinst,								// current instance
													NULL);								// no class data
	globals.d_hwndToolbarEntityCombo = CreateWindowEx (0L,								// no extended styles
													"COMBOBOX",							// class name
													"", // default text
													WS_CHILD | WS_BORDER | WS_VISIBLE | WS_VSCROLL |
													CBS_AUTOHSCROLL | CBS_DROPDOWNLIST,	// styles and defaults
													725, 0, 250, 250,					// size and position
													hwndTB,								// parent window
													(HMENU)ID_TOOLBARENTITYCOMBO,		// ID
													hinst,								// current instance
													NULL);								// no class data
    ShowWindow(hwndTB, SW_SHOW);
	lpfnDefComboTexture = (WNDPROC) GetWindowLong (globals.d_hwndToolbarTextureCombo, GWL_WNDPROC);
	lpfnDefComboEntity = (WNDPROC) GetWindowLong (globals.d_hwndToolbarEntityCombo, GWL_WNDPROC);
	SetWindowLong (globals.d_hwndToolbarTextureCombo, GWL_WNDPROC, (LONG)ToolbarTextureComboWndProc);
	SetWindowLong (globals.d_hwndToolbarEntityCombo, GWL_WNDPROC, (LONG)ToolbarEntityComboWndProc);
	
	// Get the handle to the ToolTip window.
	hWndTT = (HWND) SendMessage (hwndTB, TB_GETTOOLTIPS, 0, 0);
	
	if (hWndTT)
	{
		// Fill out the TOOLINFO structure.
		lpToolInfo.cbSize = sizeof (lpToolInfo);
		// The uID is the handle of the tool (the combo box).
		lpToolInfo.uFlags = TTF_IDISHWND | TTF_CENTERTIP;
		// The string ID in the resource
		lpToolInfo.lpszText = (LPSTR)ID_TOOLBARTEXTURECOMBO;
		// The window that gets the ToolTip messages
		lpToolInfo.hwnd = hwndTB;
		// The tool
		lpToolInfo.uId = (UINT)globals.d_hwndToolbarTextureCombo;
		// The instance that owns the string resource
		lpToolInfo.hinst = hinst;
		
		// Set up the ToolTip for the combo box.
		SendMessage (hWndTT, TTM_ADDTOOL, 0, (LPARAM)(LPTOOLINFO)&lpToolInfo);
		lpToolInfo.lpszText = (LPSTR)ID_TOOLBARENTITYCOMBO;
		lpToolInfo.uId = (UINT)globals.d_hwndToolbarEntityCombo;
		SendMessage (hWndTT, TTM_ADDTOOL, 0, (LPARAM)(LPTOOLINFO)&lpToolInfo);
	}
    return hwndTB; 
} 
