#include "qe3.h"
#include "mru.h"
#include "inc.h"
#include "directory.h"
#include <zmouse.h>

int			screen_width;
int			screen_height;
qboolean	have_quit;
int			update_bits;
extern BOOL	TexturesLocked;
int			NewWindow = false;
HANDLE		bsp_process;
char		IniFile[512];
UINT		MOUSEWHEEL;
#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL 0x020A
#endif
//===========================================

void Texture_MakeNotexture(void);
BOOL LoadRegistryInfo(const char *pszName, void *pvBuf, long *plSize);
void DoProject(int first);
void SaveAsDialog (void);

char *DefaultSurfaceDescriptions_Quake2[32] = {
	"light",
	"slick",
	"sky",
	"warp",
	"trans33",
	"trans66",
	"flowing",
	"nodraw",
	"hint",
	"skip",
	"400",
	"800",
	"1000",
	"2000",
	"4000",
	"8000",
	"10000",
	"20000",
	"40000",
	"80000",
	"100000",
	"200000",
	"400000",
	"800000",
	"1000000",
	"2000000",
	"4000000",
	"8000000",
	"10000000",
	"20000000",
	"40000000",
	"80000000"
};
char *DefaultSurfaceDescriptions_SOF[32] = {
	"light",
	"slick",
	"sky",
	"warp",
	"trans33",
	"trans66",
	"flowing",
	"nodraw",
	"hint",
	"skip",
	"400",
	"Alpha Texture",
	"Anim Speed",
	"Undulate",
	"Sky Reflect",
	"8000",
	"10000",
	"20000",
	"40000",
	"80000",
	"100000",
	"200000",
	"400000",
	"800000",
	"Type=Metal",
	"Type=Stone",
	"Type=Wood",
	"8000000",
	"10000000",
	"20000000",
	"40000000",
	"80000000"
};
char *DefaultContentDescriptions_Quake2[32] = {
	"solid",
	"window",
	"aux",
	"lava",
	"slime",
	"water",
	"mist",
	"80",
	"100",
	"200",
	"400",
	"800",
	"1000",
	"2000",
	"4000",
	"8000",
	"playerclip",
	"monsterclip",
	"current_0",
	"current_90",
	"current_180",
	"current_270",
	"current_up",
	"current_dn",
	"origin",
	"monster",
	"corpse",
	"detail",
	"translucent",
	"ladder",
	"40000000",
	"80000000"
};
char *DefaultContentDescriptions_SOF[32] = {
	"solid",
	"window",
	"Pushpull",
	"lava",
	"slime",
	"water",
	"mist",
	"80",
	"100",
	"200",
	"400",
	"800",
	"1000",
	"2000",
	"4000",
	"Areaportal",
	"playerclip",
	"monsterclip",
	"current_0",
	"current_90",
	"current_180",
	"current_270",
	"current_up",
	"current_dn",
	"origin",
	"monster",
	"corpse",
	"detail",
	"translucent",
	"ladder",
	"40000000",
	"80000000"
};

void SetWindowRect(HWND hwnd, RECT *rc)
{
	MoveWindow(hwnd, rc->left, rc->top, rc->right - rc->left, rc->bottom - rc->top, TRUE);
//	SetWindowPos(hwnd, HWND_TOP, rc->left, rc->top, rc->right - rc->left, rc->bottom - rc->top, SWP_SHOWWINDOW);
}

void SetDefaultWindowPositions(void)
{
	RECT	parent, status, toolbar;
	RECT	z, xy, xz, yz, cam, texture;
	int		midy, midx;

	GetWindowRect(globals.d_hwndToolbar, &toolbar);
	GetWindowRect(globals.d_hwndStatus, &status);
	GetClientRect(globals.d_hwndMain, &parent);
	parent.top = toolbar.bottom - toolbar.top;
	parent.bottom -= status.bottom - status.top;
	midy = parent.top + (parent.bottom - parent.top) * 0.5;

	if(globals.Multiple2DViews)
	{
		// Set Z rect
		z.top = parent.top;
		z.bottom = parent.bottom;
		z.left = parent.left;
		z.right = ZWIN_WIDTH;
		SetWindowRect(globals.Z.hwnd, &z);
		
		// Set Texture rect
		texture.top = parent.top;
		texture.bottom = parent.bottom;
		texture.left = parent.right-300;
		texture.right = parent.right;
		SetWindowRect(globals.Entity.hwnd, &texture);
		
		midx = z.right + (texture.left - z.right) * 0.5;
		
		// Set XY rect
		xy.top = midy;
		xy.bottom = parent.bottom;
		xy.left = z.right;
		xy.right = midx;
		SetWindowRect(globals.XY.hwnd, &xy);

		// Set XZ rect
		xz.top = midy;
		xz.bottom = parent.bottom;
		xz.left = midx;
		xz.right = texture.left;
		SetWindowRect(globals.XZ.hwnd, &xz);
		
		// Set YZ rect
		yz.top = parent.top;
		yz.bottom = midy;
		yz.left = midx;
		yz.right = texture.left;
		SetWindowRect(globals.YZ.hwnd, &yz);
		
		// Set Camera rect
		cam.top = parent.top;
		cam.bottom = midy;
		cam.left = z.right;
		cam.right = midx;
		SetWindowRect(globals.Camera.hwnd, &cam);
	}
	else
	{
		// Set Z rect
		z.top = parent.top;
		z.bottom = parent.bottom;
		z.left = parent.left;
		z.right = ZWIN_WIDTH;
		SetWindowRect(globals.Z.hwnd, &z);
		
		// Set Texture rect
		texture.top = parent.top;
		texture.bottom = parent.bottom;
		texture.left = parent.right-300;
		texture.right = parent.right;
		SetWindowRect(globals.Entity.hwnd, &texture);
		
		midx = z.right + (texture.left - z.right) * 0.5;
		
		// Set XY rect
		xy.top = midy;
		xy.bottom = parent.bottom;
		xy.left = z.right;
		xy.right = texture.left;
		SetWindowRect(globals.XY.hwnd, &xy);

		// Set Camera rect
		cam.top = parent.top;
		cam.bottom = midy;
		cam.left = z.right;
		cam.right = texture.left;
		SetWindowRect(globals.Camera.hwnd, &cam);
	}
}

void LoadRegistrySettings(void)
{
	int		i;

	GetPrivateProfileString	("Settings","LastProject",	"",		globals.d_savedinfo.szProject, 255, IniFile);

	if(!GetPrivateProfileStruct	("Settings","Gamma",				&globals.d_savedinfo.fGamma, sizeof(globals.d_savedinfo.fGamma), IniFile))
		globals.d_savedinfo.fGamma = 1.0;

	if(!GetPrivateProfileStruct	("Settings","Colors",	&globals.d_savedinfo.colors, sizeof(globals.d_savedinfo.colors), IniFile))
	{
		for (i=0 ; i<3 ; i++)
		{
			globals.d_savedinfo.colors[COLOR_TEXTUREBACK][i] = 0.25;
			globals.d_savedinfo.colors[COLOR_GRIDBACK][i] = 1.0;
			globals.d_savedinfo.colors[COLOR_GRIDMINOR][i] = 0.75;
			globals.d_savedinfo.colors[COLOR_GRIDMAJOR][i] = 0.5;
			globals.d_savedinfo.colors[COLOR_CAMERABACK][i] = 0.25;
			globals.d_savedinfo.colors[COLOR_TEXTURETEXT][i] = 0.0;
			globals.d_savedinfo.colors[COLOR_GRIDTEXT][i] = 0.0;
			globals.d_savedinfo.colors[COLOR_WORLDSPAWNLINES][i] = 0.0;
		}
		globals.d_savedinfo.colors[COLOR_ZICON][0] = 0.0;
		globals.d_savedinfo.colors[COLOR_ZICON][1] = 0.0;
		globals.d_savedinfo.colors[COLOR_ZICON][2] = 1.0;
		globals.d_savedinfo.colors[COLOR_CAMERAICON][0] = 0.0;
		globals.d_savedinfo.colors[COLOR_CAMERAICON][1] = 0.0;
		globals.d_savedinfo.colors[COLOR_CAMERAICON][2] = 1.0;
	}
	
	if(!GetPrivateProfileStruct	("Settings","Alpha",	&globals.d_savedinfo.alpha, sizeof(globals.d_savedinfo.alpha), IniFile))
	{
		globals.d_savedinfo.alpha[COLOR_SEL_COLOR] = 0.3f;
	}
	
	globals.d_savedinfo.iTexMenu			= GetPrivateProfileInt("Settings","RenderMode",						ID_VIEW_NEAREST,	IniFile);
	globals.d_savedinfo.show_names			= GetPrivateProfileInt("Settings","ShowNames",						true,				IniFile);
	globals.d_savedinfo.show_coordinates	= GetPrivateProfileInt("Settings","ShowCoordinates",				true,				IniFile);
	globals.d_savedinfo.no_console_texture	= GetPrivateProfileInt("Settings","NoConsoleTexinfo",				true,				IniFile);
	globals.d_savedinfo.no_console_scanfile	= GetPrivateProfileInt("Settings","NoConsoleScanfile",				true,				IniFile);
	globals.d_savedinfo.exclude				= GetPrivateProfileInt("Settings","Exclude",						0,					IniFile);
	globals.DepthClipping					= GetPrivateProfileInt("Settings","UseDepthClipping",				false,				IniFile);
	globals.DepthClipDistance				= GetPrivateProfileInt("Settings","DepthClipDistance",				512,				IniFile);
	globals.show_blocks						= GetPrivateProfileInt("Settings","ShowBlocks",						false,				IniFile);
	globals.transactive						= GetPrivateProfileInt("Settings","TransparencyActive",				false,				IniFile);
	globals.trans33							= GetPrivateProfileInt("Settings","Trans33",						true,				IniFile);
	globals.trans66							= GetPrivateProfileInt("Settings","Trans66",						true,				IniFile);
	globals.nodraw							= GetPrivateProfileInt("Settings","Nodraw",							true,				IniFile);
	TexturesLocked							= GetPrivateProfileInt("Settings","TextureLock",					false,				IniFile);
	globals.followfloor						= GetPrivateProfileInt("Settings","FollowFloor",					false,				IniFile);
	globals.StepHeight						= GetPrivateProfileInt("Settings","StepHeight",						24,					IniFile);
	globals.UpdateCamera					= GetPrivateProfileInt("Settings","UpdateCamera",					true,				IniFile);
	globals.LockXYZ2Cam						= GetPrivateProfileInt("Settings","LockXYZ2Cam",					false,				IniFile);
	globals.Lock2DOrigins					= GetPrivateProfileInt("Settings","Lock2DOrigins",					false,				IniFile);
	globals.Lock2DZoom						= GetPrivateProfileInt("Settings","Lock2DZoom",						false,				IniFile);
	globals.LockCamTo2DOrigins				= GetPrivateProfileInt("Settings","LockCamTo2DOrigins",				false,				IniFile);
	globals.AllowFaceAndBrushSelection		= GetPrivateProfileInt("Settings","BrushAndFaceSelection",			false,				IniFile);
	globals.ShowEntityConnections			= GetPrivateProfileInt("Settings","ShowEntityConnections",			true,				IniFile);
	globals.ShowSkybox						= GetPrivateProfileInt("Settings","ShowSkybox",						true,				IniFile);
	globals.d_xy.draw_mode					= GetPrivateProfileInt("Settings","XYMode",							cd_wire,			IniFile);
	globals.ShowTextureNames				= GetPrivateProfileInt("Settings","ShowTextureNames",				true,				IniFile);
	globals.Clip2XYZ						= GetPrivateProfileInt("Settings","Clip2XYZ",						false,				IniFile);
	globals.ShowHint						= GetPrivateProfileInt("Settings","ShowHint",						false,				IniFile);
	globals.StippleSelection				= GetPrivateProfileInt("Settings","StippleSelection",				true,				IniFile);
	globals.UseExternalTextures				= GetPrivateProfileInt("Settings","UseExternalTextures",			false,				IniFile);
	globals.AlwaysUseDefaultWindowPlacement	= GetPrivateProfileInt("Settings","AlwaysUseDefaultWindowPlacement",false,				IniFile);
	globals.WriteMapFloats					= GetPrivateProfileInt("Settings","WriteMapFloats",					false,				IniFile);
	globals.DrawVisGroupColors				= GetPrivateProfileInt("Settings","DrawVisGroupColors",				false,				IniFile);
	globals.ShowFullSurfaceEditor			= GetPrivateProfileInt("Settings","ShowFullSurfaceEditor",			-1,					IniFile);
	globals.OffsetCopiedBrushes				= GetPrivateProfileInt("Settings","OffsetCopiedBrushes",			true,				IniFile);
	globals.Multiple2DViews					= GetPrivateProfileInt("Settings","Multiple2DViews",				1,					IniFile);
	globals.AnimateCameraRScripts			= GetPrivateProfileInt("Settings","AnimateCameraRScripts",			1,					IniFile);
	globals.AnimateTextureRScripts			= GetPrivateProfileInt("Settings","AnimateTextureRScripts",			1,					IniFile);
	globals.AnimateSky						= GetPrivateProfileInt("Settings","AnimateSky",						1,					IniFile);
	globals.AnimationSpeed					= GetPrivateProfileInt("Settings","AnimationSpeed",					33,					IniFile);
	globals.MaxGridExtent					= GetPrivateProfileInt("Settings","GridExtent",						4096,				IniFile);
	globals.OldCameraMovement				= GetPrivateProfileInt("Settings","OldCameraMovement",				0,					IniFile);
}
void LoadSurfaceFlagDescriptions(void)
{
	int		i;
	char	s[128], temp[32];

	for(i = 0; i < 32; i++)
	{
		sprintf(temp, "%d", i);
		GetPrivateProfileString	("SurfaceFlagDescriptions", temp, "", s, 127, IniFile);

		if(!*s)
		{
			if(globals.gametype == SOF)
				Surface_FlagNames[i] = strdup(DefaultSurfaceDescriptions_SOF[i]);
			else
				Surface_FlagNames[i] = strdup(DefaultSurfaceDescriptions_Quake2[i]);
		}
		else
			Surface_FlagNames[i] = strdup(s);

		GetPrivateProfileString	("ContentFlagDescriptions", temp, "", s, 127, IniFile);

		if(!*s)
		{
			if(globals.gametype == SOF)
				Content_FlagNames[i] = strdup(DefaultContentDescriptions_SOF[i]);
			else
				Content_FlagNames[i] = strdup(DefaultContentDescriptions_Quake2[i]);
		}
		else
			Content_FlagNames[i] = strdup(s);
	}
}

void Sys_SetTitle (char *text)
{
	char	buf[512];
	int		i, len;

	switch(globals.gametype)
	{
		case QUAKE2:
			if(!globals.UseExternalTextures)
				len = sprintf(buf, "QuakeEd 5 [%s] (Quake 2 Mode)", text);
			else
				len = sprintf(buf, "QuakeEd 5 [%s] (Modified Quake 2 Mode)", text);
			break;
		case QUAKE1:
			if(!globals.UseExternalTextures)
				len = sprintf(buf, "QuakeEd 5 [%s] (Quake 1 Mode))", text);
			else
				len = sprintf(buf, "QuakeEd 5 [%s] (Modified Quake 1 Mode))", text);
			break;
		case HALFLIFE:
			if(!globals.UseExternalTextures)
				len = sprintf(buf, "QuakeEd 5 [%s] (Half-Life Mode)", text);
			else
				len = sprintf(buf, "QuakeEd 5 [%s] (Modified Half-Life Mode)", text);
			break;
		case SOF:
			len = sprintf(buf, "QuakeEd 5 [%s] (Soldier of Fortune Mode)", text);
			break;
	}

	for(i = 0; i < len; i++) // Make pathnames look more "Win32" (HI HEFFO!)
		if(buf[i] == '/')
			buf[i] = '\\';

	SetWindowText (globals.d_hwndMain, buf);
}

HCURSOR	waitcursor;

void Sys_BeginWait (void)
{
	waitcursor = SetCursor (LoadCursor (NULL, IDC_WAIT));
}

void Sys_EndWait (void)
{
	if (waitcursor)
	{
		SetCursor (waitcursor);
		waitcursor = NULL;
	}
}


void Sys_GetCursorPos (int *x, int *y)
{
	POINT lpPoint;

	GetCursorPos (&lpPoint);
	*x = lpPoint.x;
	*y = lpPoint.y;
}

void Sys_SetCursorPos (int x, int y)
{
	SetCursorPos (x, y);
}

void Sys_UpdateWindows (int bits)
{
//	Sys_Printf("updating 0x%X\n", bits);
	update_bits |= bits;
//update_bits = -1;
}


void Sys_Beep (void)
{
	MessageBeep (MB_ICONASTERISK);
}

char	*TranslateString (char *buf)
{
	static	char	*buf2 = NULL;
	int		i, l, l2 = 1;
	char	*out;

	if(!buf)
		return "\r\n";

	l = strlen(buf);

	if(!l)
		return "\r\n";

	for (i=0 ; i<l ; i++)
	{
		if (buf[i] == '\n')
			l2++;
	}

	if(buf2)
		free(buf2);

	out = buf2 = (char *)malloc(l2 + l);

	for (i=0 ; i<l ; i++)
	{
		if (buf[i] == '\n')
		{
			*out++ = '\r';
			*out++ = '\n';
		}
		else
			*out++ = buf[i];
	}
	*out++ = 0;


	return buf2;
}

void Sys_ClearPrintf (void)
{
	char	text[4];

	text[0] = 0;

	SendMessage (globals.Edit.hwnd,
		WM_SETTEXT,
		0,
		(LPARAM)text);
}

void Sys_Printf (char *text, ...)
{
	va_list argptr;
	char	buf[32768];
	char	*out;

	va_start (argptr,text);
	vsprintf (buf, text,argptr);
	va_end (argptr);

	out = TranslateString (buf);

#ifdef LATER
	Sys_Status(out);
#else
	SendMessage (globals.Edit.hwnd, EM_REPLACESEL, 0, (LPARAM)out);
	SendMessage (globals.Edit.hwnd, EM_SCROLLCARET, 0, 0);
#endif

}

void glPrintf (float x, float y, float z, char *text, ...)
{
	va_list		argptr;
	static char	buf[32768];

	va_start (argptr,text);
	vsprintf (buf, text,argptr);
	va_end (argptr);

	glPushAttrib(GL_LIST_BIT);
	glListBase(globals.Main.fontlist);
	glRasterPos3f (x, y, z);
	glCallLists (strlen(buf), GL_UNSIGNED_BYTE, buf);
	glPopAttrib();
}

double Sys_DoubleTime (void)
{
	return clock()/ 1000.0;
}

void PrintPixels (HDC hDC)
{
	int		i;
	PIXELFORMATDESCRIPTOR p[64];

	printf ("### flags color layer\n");
	for (i=1 ; i<64 ; i++)
	{
		if (!DescribePixelFormat ( hDC, i, sizeof(p[0]), &p[i]))
			break;
		printf ("%3i %5i %5i %5i\n", i,
			p[i].dwFlags,
			p[i].cColorBits,
			p[i].bReserved);
	}
	printf ("%i modes\n", i-1);
}



//==========================================================================

void QEW_StopGL(HWND *hWnd, HGLRC *hGLRC, HDC *hDC)
{
	wglMakeCurrent( NULL, NULL );
	wglDeleteContext( *hGLRC );
	ReleaseDC( *hWnd, *hDC );
}
		
int QEW_SetupPixelFormat(PIXELFORMATDESCRIPTOR *pfd, HDC *hDC, qboolean zbuffer )
{
	int pixelformat;
	int n = GetPixelFormat(*hDC);

    DescribePixelFormat(*hDC, n, sizeof(PIXELFORMATDESCRIPTOR), pfd);

	memset(pfd,0, sizeof(PIXELFORMATDESCRIPTOR)) ;

	pfd->nSize = sizeof(PIXELFORMATDESCRIPTOR);   
	pfd->nVersion =		1;						// Version number
	pfd->dwFlags =		PFD_DOUBLEBUFFER |		// Use double buffer
						PFD_SUPPORT_OPENGL |	// Use OpenGL
						PFD_DRAW_TO_WINDOW;		// Pixel format is for a window.
	pfd->iPixelType =	PFD_TYPE_RGBA ;
	pfd->cColorBits =	24;						// 32-bit color
	pfd->cDepthBits = zbuffer?32:0;				// 32-bit depth buffer
	pfd->iLayerType =	PFD_MAIN_PLANE;			// Layer type

	if ((pixelformat = ChoosePixelFormat(*hDC, pfd)) == 0)
	{
//		printf("%d",GetLastError());
		Error ("ChoosePixelFormat failed");
	}
	
	if (!SetPixelFormat(*hDC, pixelformat, pfd))
		Error ("SetPixelFormat failed");

	return pixelformat;
}

/*
=================
Error

For abnormal program terminations
=================
*/
void Error (char *error, ...)
{
	va_list argptr;
	char	text[1024];
	char	text2[1024];
	int		err;

	err = GetLastError ();

	va_start (argptr,error);
	vsprintf (text, error,argptr);
	va_end (argptr);

	sprintf (text2, "%s\nGetLastError() = %i", text, err);
	if(MessageBox(globals.d_hwndMain, text2, "Error", MB_OKCANCEL|MB_ICONERROR /* MB_OK */ ) == IDOK)
	{
		if(modified && MessageBox(globals.d_hwndMain, "Map has been modified. Do you want to save your changes?", "Error", MB_YESNO|MB_ICONQUESTION ) == IDYES)
			SaveAsDialog();
		
		exit(1);
	}
}

/*
======================================================================

FILE DIALOGS

======================================================================
*/
 
qboolean ConfirmModified (void)
{
	if (!modified)
		return true;

	if (MessageBox (globals.d_hwndMain, "This will lose changes to the map"
		, "warning", MB_OKCANCEL) == IDCANCEL)
		return false;
	return true;
}

static OPENFILENAME ofn;       /* common dialog box structure   */ 
static char szDirName[MAX_PATH];    /* directory string              */ 
static char szFile[260];       /* filename string               */ 
static char szFileTitle[260];  /* file title string             */ 
static char szFilter[260] =     /* filter string                 */ 
	"QuakeEd file (*.map)\0*.map\0\0";
static char szProjectFilter[260] =     /* filter string                 */ 
	"QuakeEd project (*.qe4)\0*.qe4\0\0";
static char chReplace;         /* string separator for szFilter */ 
static int i, cbString;        /* integer count variables       */ 
static HANDLE hf;              /* file handle                   */ 

void OpenDialog (void)
{
	/* 
	 * Obtain the system directory name and 
	 * store it in szDirName. 
	 */ 
 
	strcpy (szDirName, ValueForKey (globals.d_project_entity, "basepath") );
//	if(szDirName[strlen(szDirName)-1] == '\\')
//		szDirName[strlen(szDirName)-1] = '\0';
	strcat (szDirName, "maps");

	/* Place the terminating null character in the szFile. */ 
 
	szFile[0] = '\0'; 
 
	/* Set the members of the OPENFILENAME structure. */ 
 
	ofn.lStructSize = sizeof(OPENFILENAME); 
	ofn.hwndOwner = globals.Camera.hwnd;
	ofn.lpstrFilter = szFilter; 
	ofn.nFilterIndex = 1; 
	ofn.lpstrFile = szFile; 
	ofn.nMaxFile = sizeof(szFile); 
	ofn.lpstrFileTitle = szFileTitle; 
	ofn.nMaxFileTitle = sizeof(szFileTitle); 
	ofn.lpstrInitialDir = szDirName; 
	ofn.Flags = OFN_SHOWHELP | OFN_PATHMUSTEXIST | 
		OFN_FILEMUSTEXIST; 

	/* Display the Open dialog box. */ 
 
	if (!GetOpenFileName(&ofn))
		return;	// canceled
 
	// Add the file in MRU.
	AddNewItem( globals.d_lpMruMenu, ofn.lpstrFile);

	// Refresh the File menu.
	PlaceMenuMRUItem(globals.d_lpMruMenu,GetSubMenu(GetMenu(globals.d_hwndMain),0), 
			ID_FILE_EXIT);

	/* Open the file. */ 
 
	Map_LoadFile (ofn.lpstrFile);	
}

void ProjectDialog (void)
{
	/* 
	 * Obtain the system directory name and 
	 * store it in szDirName. 
	 */ 
 
	strcpy (szDirName, ValueForKey(globals.d_project_entity, "basepath") );
	strcat (szDirName, "\\scripts");

	/* Place the terminating null character in the szFile. */ 
 
	szFile[0] = '\0'; 
 
	/* Set the members of the OPENFILENAME structure. */ 
 
	ofn.lStructSize = sizeof(OPENFILENAME); 
	ofn.hwndOwner = globals.Camera.hwnd;
	ofn.lpstrFilter = szProjectFilter; 
	ofn.nFilterIndex = 1; 
	ofn.lpstrFile = szFile; 
	ofn.nMaxFile = sizeof(szFile); 
	ofn.lpstrFileTitle = szFileTitle; 
	ofn.nMaxFileTitle = sizeof(szFileTitle); 
	ofn.lpstrInitialDir = szDirName; 
	ofn.Flags = OFN_SHOWHELP | OFN_PATHMUSTEXIST | 
		OFN_FILEMUSTEXIST; 

	/* Display the Open dialog box. */ 
 
	if (!GetOpenFileName(&ofn))
		return;	// canceled
 
	// Refresh the File menu.
	PlaceMenuMRUItem(globals.d_lpMruMenu,GetSubMenu(GetMenu(globals.d_hwndMain),0), 
			ID_FILE_EXIT);

	/* Open the file. */ 
	if (!QE_LoadProject(ofn.lpstrFile))
		Error ("Couldn't load project file");
}


void SaveAsDialog (void)
{ 
	strcpy (szDirName, ValueForKey (globals.d_project_entity, "basepath") );
//	if(szDirName[strlen(szDirName)-1] == '\\')
//		szDirName[strlen(szDirName)-1] = '\0';
	strcat (szDirName, "maps");

	/* Place the terminating null character in the szFile. */ 
 
	szFile[0] = '\0'; 
 
	/* Set the members of the OPENFILENAME structure. */ 
 
	ofn.lStructSize = sizeof(OPENFILENAME); 
	ofn.hwndOwner = globals.Camera.hwnd;
	ofn.lpstrFilter = szFilter; 
	ofn.nFilterIndex = 1; 
	ofn.lpstrFile = szFile; 
	ofn.nMaxFile = sizeof(szFile); 
	ofn.lpstrFileTitle = szFileTitle; 
	ofn.nMaxFileTitle = sizeof(szFileTitle); 
	ofn.lpstrInitialDir = szDirName; 
	ofn.Flags = OFN_SHOWHELP | OFN_PATHMUSTEXIST | 
		OFN_FILEMUSTEXIST | OFN_OVERWRITEPROMPT; 

	/* Display the Open dialog box. */ 
 
	if (!GetSaveFileName(&ofn))
		return;	// canceled
  
	DefaultExtension (ofn.lpstrFile, ".map");
	strcpy (currentmap, ofn.lpstrFile);

	// Add the file in MRU.
	AddNewItem(globals.d_lpMruMenu, ofn.lpstrFile);

	// Refresh the File menu.
	PlaceMenuMRUItem(globals.d_lpMruMenu,GetSubMenu(GetMenu(globals.d_hwndMain),0),
			ID_FILE_EXIT);

	Map_SaveFile (ofn.lpstrFile, false);	// ignore region
}

/*
=======================================================

Menu modifications

=======================================================
*/

/*
==================
FillBSPMenu

==================
*/
char	*bsp_commands[256];

void FillBSPMenu (void)
{
	HMENU	hmenu;
	epair_t	*ep;
	int		i;
	static int count;

	hmenu = GetSubMenu (GetMenu(globals.d_hwndMain), MENU_BSP);

	for (i=0 ; i<count ; i++)
		DeleteMenu (hmenu, CMD_BSPCOMMAND+i, MF_BYCOMMAND);
	count = 0;

	i = 0;
	for (ep = globals.d_project_entity->epairs ; ep ; ep=ep->next)
	{
		if (ep->key[0] == 'b' && ep->key[1] == 's' && ep->key[2] == 'p')
		{
			bsp_commands[i] = ep->key;
			AppendMenu (hmenu, MF_ENABLED|MF_STRING,
			CMD_BSPCOMMAND+i, (LPCTSTR)ep->key);
			i++;
		}
	}
	count = i;
}

//==============================================

/*
===============
CheckBspProcess

See if the BSP is done yet
===============
*/
void CheckBspProcess (void)
{
	char	outputpath[1024];
	char	temppath[512];
	DWORD	exitcode;
	char	*out;
	BOOL	ret;

	if (!bsp_process)
		return;

	ret = GetExitCodeProcess (bsp_process, &exitcode);
	if (!ret)
		Error ("GetExitCodeProcess failed");
	if (exitcode == STILL_ACTIVE)
		return;

	bsp_process = 0;

	GetTempPath(512, temppath);
	sprintf (outputpath, "%sjunk.txt", temppath);

	LoadFile (outputpath, (void *)&out);
	Sys_Printf ("%s", out);
	Sys_Printf ("\ncompleted.\n");
	free (out);
	Sys_Beep ();

	Pointfile_Check ();
}

extern int	cambuttonstate;

void SetUpFont(WInfo_t *winfo)
{
	//
	// create GL font
	//
	winfo->font = CreateFont(
		10,	// logical height of font 
		5,	// logical average character width 
		0,	// angle of escapement 
		10,	// base-line orientation angle 
		0,	// font weight 
		0,	// italic attribute flag 
		0,	// underline attribute flag 
		0,	// strikeout attribute flag 
		0,	// character set identifier 
		0,	// output precision 
		0,	// clipping precision 
		0,	// output quality 
		0,	// pitch and family 
		0 	// pointer to typeface name string 
		);

	if (!winfo->font)
		Error("couldn't create font");

	wglMakeCurrent(winfo->hdc, winfo->hglrc );
	SelectObject (winfo->hdc, winfo->font);
	if ( (winfo->fontlist = glGenLists (256) ) == 0 )
		Error( "couldn't create font dlists" );

	// create the bitmap display lists
	// we're making images of glyphs 0 thru 255
#if 0
	if (wglUseFontBitmaps(winfo->hdc, 1, 255, winfo->fontlist))
	// indicate start of glyph display lists
		glListBase (1);
	else if (wglUseFontBitmaps(winfo->hdc, 0, 255, winfo->fontlist))
	// indicate start of glyph display lists
		glListBase (0);
	else
		Error("wglUseFontBitmaps failed");
#else
	if (!wglUseFontBitmaps(winfo->hdc, 0, 255, winfo->fontlist))
		Error("wglUseFontBitmaps failed");
#endif
}

UINT GetMouseWheelMsg( void )
{
    OSVERSIONINFO osvi;

    osvi.dwOSVersionInfoSize = sizeof(osvi);

    if ( !GetVersionEx(&osvi) )
        return WM_MOUSEWHEEL;            // Got a better idea?

    // NT 4 and later supports WM_MOUSEWHEEL
    if ( VER_PLATFORM_WIN32_NT == osvi.dwPlatformId )
        if ( osvi.dwMajorVersion >= 4 )
            return WM_MOUSEWHEEL;

    // Win98 Supports WM_MOUSEWHEEL, assume that anything above it
    // does too
    if ( (osvi.dwMajorVersion >= 5) || (osvi.dwMajorVersion == 4 &&
    		osvi.dwMinorVersion >= 10 &&
    		osvi.dwBuildNumber >= 1998))
        return WM_MOUSEWHEEL;

    // Hmmm... an older version.  The mouse driver support app should
    // have registered a window message for it.  By registering the
    // same message, we should get back the same message number.
    // Note that "MSWHEEL_ROLLMSG" below is a #define taken from ZMOUSE.H,
    // which is from the "Intellimouse SDK".

    return RegisterWindowMessage( "MSWHEEL_ROLLMSG" );
}

void SetupXYOrientation(int Orientation);

/*
==================
WinMain

==================
*/
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance
					,LPSTR lpCmdLine, int nCmdShow)
{
	double		time, oldtime, delta;
    MSG			msg;
	HACCEL		accelerators;

	globals.d_hInstance = hInstance;
	globals.gametype = QUAKE2;

	InitCommonControls ();

	screen_width = GetSystemMetrics (SM_CXFULLSCREEN);
	screen_height = GetSystemMetrics (SM_CYFULLSCREEN);
	GetCurrentDirectory(512, globals.WorkingDir);

	// hack for broken NT 4.0 dual screen
	if (screen_width > 2*screen_height)
		screen_width /= 2;

	accelerators = LoadAccelerators (hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR1));
	if (!accelerators)
		Error ("LoadAccelerators failed");

	_fullpath(IniFile, "scripts/qe4.ini", 511);
	LoadRegistrySettings();

	// the project file can be specified on the command line,
	// or implicitly found in the scripts directory
	if (lpCmdLine && strlen(lpCmdLine))
	{
		ParseCommandLine (lpCmdLine);
		if (!QE_LoadProject(argv[1]))
			Error ("Couldn't load %s project file", argv[1]);
	}
	else if (!QE_LoadProject("scripts/quake.qe5"))
	{
		DoProject(true);
		if (!QE_LoadProject("scripts/quake.qe5"))
			Error ("Couldn't load scripts/quake.qe5 project file");
	}
	if(!stricmp(ValueForKey(globals.d_project_entity, "gametype"), "Quake1"))
		globals.gametype = QUAKE1;
	else if(!stricmp(ValueForKey(globals.d_project_entity, "gametype"), "HalfLife"))
		globals.gametype = HALFLIFE;
	else if(!stricmp(ValueForKey(globals.d_project_entity, "gametype"), "SoF"))
		globals.gametype = SOF;

	strcpy(globals.basepath, ValueForKey(globals.d_project_entity, "basepath"));
	strcpy(globals.modpath, ValueForKey(globals.d_project_entity, "modpath"));
	SetHeightsForGame();
	LoadSurfaceFlagDescriptions();
	globals.Camera_Initialized = false;
	Main_Create (hInstance);
	CreateEntityWindow(hInstance);
	WCam_Create (hInstance);
	WZ_Create (hInstance);
	WXY_Create (hInstance, "XY View", globals.Multiple2DViews?"xywindow":"2Dwindow", &globals.XY);

	if(globals.Multiple2DViews)
		CreateExtra2DViews();

	if(globals.AnimateCameraRScripts || globals.AnimateTextureRScripts || globals.AnimateSky)
		SetTimer(globals.d_hwndMain, QE_TIMER1, globals.AnimationSpeed, NULL);


	globals.XY.Orientation = 0;
	globals.XZ.Orientation = 1;
	globals.YZ.Orientation = 2;
	VisGroups_Init();
	globals.Camera_Initialized = true;

	if(NewWindow)
		SetDefaultWindowPositions();

	FillBSPMenu ();
	categories = NULL;
	DirectoryInit();

	if(ValueForKey(globals.d_project_entity, "defaultpaks"))
		OpenMultiplePakFiles(ValueForKey(globals.d_project_entity, "defaultpaks"));

	openwads.next = &openwads;
	openwads.prev = &openwads;
	SetupXYOrientation(0);
	currentwads[0] = 0;
	skybox_textures[0] = skybox_textures[1] = skybox_textures[2] = skybox_textures[3] = skybox_textures[4] = skybox_textures[5] = 0;
	InitRscript();
	ParseTextureCategoryFile();
	FillTextureMenu ();
	Texture_MakeNotexture();
	QE_Init ();
	Map_New ();
	camera.viewdistance = 256.0;
	globals.d_texturewin.scale = 1.0f;
	MOUSEWHEEL = GetMouseWheelMsg();
	SetClipperMode(3);

	if(globals.ShowFullSurfaceEditor == -1)
	{
		if(globals.gametype == SOF || globals.gametype == QUAKE2)
			globals.ShowFullSurfaceEditor = true;
		else
			globals.ShowFullSurfaceEditor = false;
	}
	
	InitModels();
	Map_RegionOff(); // Make sure grid extent is set
	Sys_Printf ("Entering message loop\n");
	oldtime = Sys_DoubleTime ();

	while (!have_quit)
	{
		Sys_EndWait ();		// remove wait cursor if active

		while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (!TranslateAccelerator(globals.d_hwndMain, accelerators, &msg) )
			{
      			TranslateMessage (&msg);
      			DispatchMessage (&msg);
			}
			if (msg.message == WM_QUIT)
				have_quit = true;
		}


		CheckBspProcess ();

		time = Sys_DoubleTime ();
		delta = time - oldtime;
		oldtime = time;
		if (delta > 0.2)
			delta = 0.2;

		// run time dependant behavior
		Cam_MouseControl (delta);

		// update any windows now
		if (update_bits & W_CAMERA)
		{
			InvalidateRect(globals.Camera.hwnd, NULL, false);
			UpdateWindow (globals.Camera.hwnd);
		}
		if (update_bits & W_Z)
		{
			InvalidateRect(globals.Z.hwnd, NULL, false);
			UpdateWindow (globals.Z.hwnd);
		}
			
		if (update_bits & W_TEXTURE)
		{
			InvalidateRect(globals.Texture.hwnd, NULL, false);
			UpdateWindow (globals.Texture.hwnd);
		}
	
		if (update_bits & W_XY)
		{
			InvalidateRect(globals.XY.hwnd, NULL, false);
			UpdateWindow (globals.XY.hwnd);
		}
		if(globals.Multiple2DViews)
		{
			if (update_bits & W_XZ)
			{
				InvalidateRect(globals.XZ.hwnd, NULL, false);
				UpdateWindow (globals.XZ.hwnd);
			}
			if (update_bits & W_YZ)
			{
				InvalidateRect(globals.YZ.hwnd, NULL, false);
				UpdateWindow (globals.YZ.hwnd);
			}
		}
		update_bits = 0;

		if (!cambuttonstate && !have_quit)
		{	// if not driving in the camera view, block
			WaitMessage ();
		}

	}

    /* return success of application */
    return TRUE;

}

