// win_xy.c -- windows specific xy view code

#include "qe3.h"
#include "inc.h"

void SetupXYOrientation(int Orientation);
void Select_Ungroup(void);
static unsigned s_stipple[32] =
{
	0xaaaaaaaa, 0x55555555,0xaaaaaaaa, 0x55555555,
	0xaaaaaaaa, 0x55555555,0xaaaaaaaa, 0x55555555,
	0xaaaaaaaa, 0x55555555,0xaaaaaaaa, 0x55555555,
	0xaaaaaaaa, 0x55555555,0xaaaaaaaa, 0x55555555,
	0xaaaaaaaa, 0x55555555,0xaaaaaaaa, 0x55555555,
	0xaaaaaaaa, 0x55555555,0xaaaaaaaa, 0x55555555,
	0xaaaaaaaa, 0x55555555,0xaaaaaaaa, 0x55555555,
	0xaaaaaaaa, 0x55555555,0xaaaaaaaa, 0x55555555,
};

#define MENU_CREATEPOINTENTITY	1
#define MENU_MAKEBRUSHENTITY	2
#define MENU_MAKEPOINTENTITY	4
#define MENU_MAKEANYENTITY		8
#define MENU_REVERTTOWORLDSPAWN	16
#define MENU_DELETE				32
#define MENU_DESELECT			64

#define SELECTION_BRUSHES		1
#define SELECTION_BRUSHENT		2
#define	SELECTION_POINTENT		4

int GetSelectionInfo(void)
{
	brush_t	*b;
	int		retval = 0;

	for(b = selected_brushes.next; b != &selected_brushes; b = b->next)
	{
		if(b->owner->eclass->fixedsize)
			retval |= SELECTION_POINTENT;
		else if(!stricmp(b->owner->eclass->name, "worldspawn"))
			retval |= SELECTION_BRUSHES;
		else
			retval |= SELECTION_BRUSHENT;
	}

	return retval;
}

void DoCreateEntity(int pointbased, int brushbased, int sel, vec3_t origin);
void XY_ToGridPoint (int x, int y, vec3_t point, WInfo_t *winfo);

void DoXYPopupMenu(int x, int y, WInfo_t *winfo)
{
	HMENU	menu = CreatePopupMenu();
	BOOL	retval;
	POINT	point;
	int		brushesselected = selected_brushes.next != &selected_brushes;
	int		selected = GetSelectionInfo();

	GetCursorPos(&point);

	if(!selected)
		AppendMenu(menu, MF_ENABLED, MENU_CREATEPOINTENTITY, (LPCTSTR)"Create Point Entity");
	else
	{
		if(selected == SELECTION_BRUSHES)
		{
			AppendMenu(menu, MF_ENABLED, MENU_MAKEANYENTITY, (LPCTSTR)"Make Brushes Into Entity");
			AppendMenu(menu, MF_ENABLED, MENU_MAKEBRUSHENTITY, (LPCTSTR)"Make Brushes Into Brush-based Entity");
			AppendMenu(menu, MF_ENABLED, MENU_MAKEPOINTENTITY, (LPCTSTR)"Make Brushes Into Point Entity");
		}
		else if(selected == SELECTION_BRUSHENT)
			AppendMenu(menu, MF_ENABLED, MENU_REVERTTOWORLDSPAWN, (LPCTSTR)"Unlink Entity");

		AppendMenu(menu, MF_ENABLED, MENU_DESELECT, (LPCTSTR)"Deselect");
		AppendMenu(menu, MF_ENABLED, MENU_DELETE, (LPCTSTR)"Delete Selection");
	}

	retval = TrackPopupMenu(menu, TPM_LEFTALIGN|TPM_NONOTIFY|TPM_RETURNCMD|TPM_RIGHTBUTTON, point.x, point.y, 0, globals.XY.hwnd, NULL);

	switch(retval)
	{
		vec3_t	origin;
		case MENU_CREATEPOINTENTITY:

			XY_ToGridPoint(x, y, origin, winfo);
			DoCreateEntity(true, false, false, origin);
			break;

		case MENU_MAKEBRUSHENTITY:
			DoCreateEntity(false, true, true, origin);
			break;
			
		case MENU_MAKEPOINTENTITY:
			DoCreateEntity(true, false, true, origin);
			break;
			
		case MENU_MAKEANYENTITY:
			DoCreateEntity(true, true, true, origin);
			break;

		case MENU_REVERTTOWORLDSPAWN:
			Select_Ungroup();
			break;
			
		case MENU_DELETE:
			Select_Delete();
			break;

		case MENU_DESELECT:
			Select_Deselect(true);
			break;
	}
	DestroyMenu(menu);
}

extern int DraggingClipPoint;
int	moved, mousex, mousey;
/*
============
WXY_WndProc
============
*/
LONG WINAPI WXY_WndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
	int			fwKeys, xPos, yPos;
    RECT		rect;
	WInfo_t		*winfo;

    GetClientRect(hWnd, &rect);

    switch (uMsg)
    {
	case WM_CREATE:


		return 0;

	case WM_DESTROY:
		if(hWnd == globals.XY.hwnd)
			QEW_StopGL( &hWnd, &globals.XY.hglrc, &globals.XY.hdc );
		
		if(hWnd == globals.YZ.hwnd)
			QEW_StopGL( &hWnd, &globals.YZ.hglrc, &globals.YZ.hdc );
		
		if(hWnd == globals.XZ.hwnd)
			QEW_StopGL( &hWnd, &globals.XZ.hglrc, &globals.XZ.hdc );

		return 0;

	case WM_PAINT:
    {
		PAINTSTRUCT	ps;
		
		BeginPaint(hWnd, &ps);
		
		if(hWnd == globals.XY.hwnd)
		{
			if (!wglMakeCurrent( globals.XY.hdc, globals.XY.hglrc ))
				Error ("wglMakeCurrent failed");
			
			if(globals.Multiple2DViews)
				SetupXYOrientation(0);
			else
				SetupXYOrientation(globals.XY_Orientation[3]);

			QE_CheckOpenGLForErrors("WXY_WndProc()1", GL_INVALID_ENUM);
			XY_Draw (&globals.XY);
			QE_CheckOpenGLForErrors("WXY_WndProc()2", GL_NO_ERROR);
			
			SwapBuffers(globals.XY.hdc);
		}
		if(hWnd == globals.XZ.hwnd)
		{
			if (!wglMakeCurrent( globals.XZ.hdc, globals.XZ.hglrc ))
				Error ("wglMakeCurrent failed");
			
			SetupXYOrientation(1);
			QE_CheckOpenGLForErrors("WXY_WndProc()1", GL_INVALID_ENUM);
			XY_Draw (&globals.XZ);
			QE_CheckOpenGLForErrors("WXY_WndProc()2", GL_NO_ERROR);
			
			SwapBuffers(globals.XZ.hdc);
		}
		if(hWnd == globals.YZ.hwnd)
		{
			if (!wglMakeCurrent( globals.YZ.hdc, globals.YZ.hglrc ))
				Error ("wglMakeCurrent failed");
			
			SetupXYOrientation(2);
			QE_CheckOpenGLForErrors("WXY_WndProc()1", GL_INVALID_ENUM);
			XY_Draw (&globals.YZ);
			QE_CheckOpenGLForErrors("WXY_WndProc()2", GL_NO_ERROR);
			
			SwapBuffers(globals.YZ.hdc);
		}
		EndPaint(hWnd, &ps);
	}
	return 0;

	case WM_KEYDOWN:
		return QE_KeyDown (wParam);
		
	case WM_RBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_LBUTTONDOWN:
		if ( GetTopWindow( globals.d_hwndMain ) != hWnd)
			BringWindowToTop(hWnd);
		SetFocus(hWnd);
		SetCapture(hWnd);

		if(hWnd == globals.XY.hwnd)
		{
			winfo = &globals.XY;

			if(globals.Multiple2DViews)
				SetupXYOrientation(0);
			else
				SetupXYOrientation(globals.XY_Orientation[3]);
		}
		else if(hWnd == globals.XZ.hwnd)
		{
			winfo = &globals.XZ;
			SetupXYOrientation(1);
		}
		else if(hWnd == globals.YZ.hwnd)
		{
			winfo = &globals.YZ;
			SetupXYOrientation(2);
		}

		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;
		XY_MouseDown (xPos, yPos, fwKeys, winfo);
		if(uMsg == WM_RBUTTONDOWN)
		{
			moved = 0;
			mousex = xPos;
			mousey = yPos;
		}
		return 0;

	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	case WM_LBUTTONUP:
		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;

		if(DraggingClipPoint)
			moved = 1;

		XY_MouseUp (xPos, yPos, fwKeys);

		if (! (fwKeys & (MK_LBUTTON|MK_RBUTTON|MK_MBUTTON)))
			ReleaseCapture ();

		if(uMsg == WM_RBUTTONUP && !moved)
		{
			if(hWnd == globals.XY.hwnd)
			{
				winfo = &globals.XY;
				
				if(globals.Multiple2DViews)
					SetupXYOrientation(0);
				else
					SetupXYOrientation(globals.XY_Orientation[3]);
			}
			else if(hWnd == globals.XZ.hwnd)
			{
				winfo = &globals.XZ;
				SetupXYOrientation(1);
			}
			else if(hWnd == globals.YZ.hwnd)
			{
				winfo = &globals.YZ;
				SetupXYOrientation(2);
			}

			DoXYPopupMenu(xPos, yPos, winfo);
		}
		return 0;

	case WM_MOUSEMOVE:
		if(hWnd == globals.XY.hwnd)
		{
			winfo = &globals.XY;

			if(globals.Multiple2DViews)
				SetupXYOrientation(0);
			else
				SetupXYOrientation(globals.XY_Orientation[3]);
		}
		else if(hWnd == globals.XZ.hwnd)
		{
			winfo = &globals.XZ;
			SetupXYOrientation(1);
		}
		else if(hWnd == globals.YZ.hwnd)
		{
			winfo = &globals.YZ;
			SetupXYOrientation(2);
		}

		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;
		XY_MouseMoved (xPos, yPos, fwKeys, winfo);

		if(!moved && (mousex != xPos || mousey != yPos))
		{
			moved = 1;
//			mousex = xPos;
//			mousey = yPos;
		}
		return 0;

    case WM_SIZE:
		if(hWnd == globals.XY.hwnd)
		{
			globals.XY.width = rect.right;
			globals.XY.height = rect.bottom;
		}
		else if(hWnd == globals.XZ.hwnd)
		{
			globals.XZ.width = rect.right;
			globals.XZ.height = rect.bottom;
		}
		else if(hWnd == globals.YZ.hwnd)
		{
			globals.YZ.width = rect.right;
			globals.YZ.height = rect.bottom;
		}
		InvalidateRect(hWnd, NULL, false);
		return 0;

	case WM_SIZING:
			if(TryDocking(hWnd, wParam, (LPRECT)lParam, 0))
				return TRUE;
		break;

	case WM_MOVING:
			if(TryDocking(hWnd, wParam, (LPRECT)lParam, 0))
				return TRUE;
			break;

	case WM_NCCALCSIZE:// don't let windows copy pixels
		DefWindowProc (hWnd, uMsg, wParam, lParam);
		return WVR_REDRAW;

	case WM_KILLFOCUS:
	case WM_SETFOCUS:
		SendMessage( hWnd, WM_NCACTIVATE, uMsg == WM_SETFOCUS, 0 );
		return 0;

   	case WM_CLOSE:
        DestroyWindow (hWnd);
		return 0;
    }

	return DefWindowProc (hWnd, uMsg, wParam, lParam);
}


/*
==============
WXY_Create
==============
*/
void WXY_Create (HINSTANCE hInstance, char *WindowName, char *LoadName, WInfo_t *winfo)
{
	static int	ClassInitialized = false;
    WNDCLASS	wc;

	if(!ClassInitialized)
	{
		/* Register the camera class */
		memset (&wc, 0, sizeof(wc));
		
		wc.style         = CS_OWNDC;
		wc.lpfnWndProc   = (WNDPROC)WXY_WndProc;
		wc.cbClsExtra    = 0;
		wc.cbWndExtra    = 0;
		wc.hInstance     = hInstance;
		wc.hIcon         = 0;
		wc.hCursor       = LoadCursor (NULL,IDC_CROSS);
		wc.hbrBackground = NULL; //(HBRUSH)(COLOR_WINDOW+1);
		wc.lpszMenuName  = NULL;
		wc.lpszClassName = XY_WINDOW_CLASS;
		
		if (!RegisterClass (&wc) )
			Error ("RegisterClass: failed");

		ClassInitialized = true;
	}

	winfo->hwnd = CreateWindowEx (WS_EX_TOOLWINDOW,
		XY_WINDOW_CLASS ,
//		"XY View",
		WindowName,
		WS_CAPTION | WS_THICKFRAME | WS_MAXIMIZEBOX  | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_CHILD ,
  		ZWIN_WIDTH,
		(int)(screen_height*CWIN_SIZE)-20,
		screen_width-ZWIN_WIDTH,
		(int)(screen_height*(1.0-CWIN_SIZE)-38),	// size

		globals.d_hwndMain,	// parent
		0,		// no menu
		hInstance,
		NULL);

	if (!winfo->hwnd )
		Error ("Couldn't create %s", WindowName);

	if(!LoadWindowState(winfo->hwnd, LoadName))
		NewWindow = true;

    ShowWindow(winfo->hwnd, SW_SHOWDEFAULT);
	winfo->hdc = GetDC(winfo->hwnd);

	QEW_SetupPixelFormat(&(winfo->pfd), &(winfo->hdc), false);

	if ( ( winfo->hglrc = wglCreateContext( winfo->hdc ) ) == 0 )
		Error( "wglCreateContext in WXY_Create failed" );

	if (!wglMakeCurrent( winfo->hdc, winfo->hglrc ))
		Error ("wglMakeCurrent failed");

	if (!wglShareLists(globals.Main.hglrc, winfo->hglrc ) )
		Error( "wglShareLists in WXY_Create failed" );

	glPolygonStipple ((char *)s_stipple);
	glLineStipple (3, 0xaaaa);
//	SetUpFont(&(globals.XY));
}

void SetDefaultWindowPositions(void);

void DestroyExtra2DViews(void)
{
	NewWindow = false;
	if(globals.Multiple2DViews)
	{
		SaveWindowState(globals.XY.hwnd,		"xywindow");
		SaveWindowState(globals.YZ.hwnd,		"yzwindow");
		SaveWindowState(globals.XZ.hwnd,		"xzwindow");
		SaveWindowState(globals.Camera.hwnd,	"camerawindow");
		SaveWindowState(globals.Z.hwnd,			"zwindow");
		SaveWindowState(globals.Entity.hwnd,	"EntityWindow");
	}
	SendMessage(globals.XZ.hwnd, WM_CLOSE, 0, 0);
	SendMessage(globals.YZ.hwnd, WM_CLOSE, 0, 0);
	if(!LoadWindowState(globals.XY.hwnd, "2Dwindow"))
		NewWindow = true;
	if(!LoadWindowState(globals.Camera.hwnd, "camerawindow2"))
		NewWindow = true;
	if(!LoadWindowState(globals.Entity.hwnd, "EntityWindow2"))
		NewWindow = true;
	if(!LoadWindowState(globals.Z.hwnd, "zwindow2"))
		NewWindow = true;

	if(NewWindow)
	{
		SetDefaultWindowPositions();
		NewWindow = 0;
		Sys_Printf("New Window\n");
	}
	SetupXYOrientation(globals.XY_Orientation[3]);
	InvalidateRect(globals.Main.hwnd, NULL, false);
}

void CreateExtra2DViews(void)
{
	NewWindow = false;
	if(!globals.Multiple2DViews)
	{
		SaveWindowState(globals.XY.hwnd,		"2Dwindow");
		SaveWindowState(globals.Camera.hwnd,	"camerawindow2");
		SaveWindowState(globals.Z.hwnd,			"zwindow2");
		SaveWindowState(globals.Entity.hwnd,	"EntityWindow2");
	}
	WXY_Create (globals.d_hInstance, "XZ View", "xzwindow", &globals.XZ);
	WXY_Create (globals.d_hInstance, "YZ View", "yzwindow", &globals.YZ);
	SetWindowText(globals.XY.hwnd, "XY View");
	if(!LoadWindowState(globals.XY.hwnd, "xywindow"))
		NewWindow = true;
	if(!LoadWindowState(globals.Camera.hwnd, "camerawindow"))
		NewWindow = true;
	if(!LoadWindowState(globals.Entity.hwnd, "EntityWindow"))
		NewWindow = true;
	if(!LoadWindowState(globals.Z.hwnd, "zwindow"))
		NewWindow = true;

	if(NewWindow)
	{
		SetDefaultWindowPositions();
		NewWindow = 0;
		Sys_Printf("New Window\n");
	}
	SetupXYOrientation(0);
	globals.XY.Orientation = 0;
	InvalidateRect(globals.Main.hwnd, NULL, false);
}

static void WXY_InitPixelFormat( PIXELFORMATDESCRIPTOR *pPFD )
{
	memset( pPFD, 0, sizeof( *pPFD ) );

	pPFD->nSize    = sizeof( PIXELFORMATDESCRIPTOR );
	pPFD->nVersion = 1;
	pPFD->dwFlags  = PFD_DOUBLEBUFFER | PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW;
	pPFD->iPixelType = PFD_TYPE_RGBA;
	pPFD->cColorBits = 24;
	pPFD->cDepthBits = 32;
	pPFD->iLayerType = PFD_MAIN_PLANE;

}

void WXY_Print( void )
{
	DOCINFO di;

	PRINTDLG pd;

	/*
	** initialize the PRINTDLG struct and execute it
	*/
	memset( &pd, 0, sizeof( pd ) );
	pd.lStructSize = sizeof( pd );
	pd.hwndOwner = globals.XY.hwnd;
	pd.Flags = PD_RETURNDC;
	pd.hInstance = 0;
	if ( !PrintDlg( &pd ) || !pd.hDC )
	{
		MessageBox( globals.d_hwndMain, "Could not PrintDlg()", "QE4 Print Error", MB_OK | MB_ICONERROR );
		return;
	}

	/*
	** StartDoc
	*/
	memset( &di, 0, sizeof( di ) );
	di.cbSize = sizeof( di );
	di.lpszDocName = "QE4";
	if ( StartDoc( pd.hDC, &di ) <= 0 )
	{
		MessageBox( globals.d_hwndMain, "Could not StartDoc()", "QE4 Print Error", MB_OK | MB_ICONERROR );
		return;
	}

	/*
	** StartPage
	*/
	if ( StartPage( pd.hDC ) <= 0 )
	{
		MessageBox( globals.d_hwndMain, "Could not StartPage()", "QE4 Print Error", MB_OK | MB_ICONERROR );
		return;
	}

	/*
	** read pixels from the XY window
	*/
	{
		int bmwidth = 320, bmheight = 320;
		int pwidth, pheight;

		RECT r;

		GetWindowRect( globals.XY.hwnd, &r );

		bmwidth  = r.right - r.left;
		bmheight = r.bottom - r.top;

		pwidth  = GetDeviceCaps( pd.hDC, PHYSICALWIDTH ) - GetDeviceCaps( pd.hDC, PHYSICALOFFSETX );
		pheight = GetDeviceCaps( pd.hDC, PHYSICALHEIGHT ) - GetDeviceCaps( pd.hDC, PHYSICALOFFSETY );

		StretchBlt( pd.hDC,
			0, 0,
			pwidth, pheight,
			globals.XY.hdc,
			0, 0,
			bmwidth, bmheight,
			SRCCOPY );
	}

	/*
	** EndPage and EndDoc
	*/
	if ( EndPage( pd.hDC ) <= 0 )
	{
		MessageBox( globals.d_hwndMain, "QE4 Print Error", "Could not EndPage()", MB_OK | MB_ICONERROR );
		return;
	}

	if ( EndDoc( pd.hDC ) <= 0 )
	{
		MessageBox( globals.d_hwndMain, "QE4 Print Error", "Could not EndDoc()", MB_OK | MB_ICONERROR );
		return;
	}
}
