// win_cam.c -- windows specific camera view code

#include "qe3.h"
#include "inc.h"

void SetDefaultWindowPositions(void);

/*
============
WZ_WndProc
============
*/
LONG WINAPI WZ_WndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
	int		fwKeys, xPos, yPos;
    RECT	rect;

    GetClientRect(hWnd, &rect);

    switch (uMsg)
    {

	case WM_DESTROY:
		QEW_StopGL( &hWnd, &globals.Z.hglrc, &globals.Z.hdc );
		return 0;

	case WM_CREATE:
		return 0;

	case WM_PAINT:
        { 
		    PAINTSTRUCT	ps;

		    BeginPaint(hWnd, &ps);

            if ( !wglMakeCurrent( globals.Z.hdc, globals.Z.hglrc ) )
				Error ("wglMakeCurrent failed");
			QE_CheckOpenGLForErrors("WZ_WndProc()", GL_NO_ERROR);

			Z_Draw ();
		    SwapBuffers(globals.Z.hdc);

			EndPaint(hWnd, &ps);
        }
		return 0;


	case WM_KEYDOWN:
		QE_KeyDown (wParam);
		return 0;

	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_LBUTTONDOWN:
		if (GetTopWindow(globals.d_hwndMain) != hWnd)
			BringWindowToTop(hWnd);

		SetFocus( globals.Z.hwnd );
		SetCapture( globals.Z.hwnd );
		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;
		Z_MouseDown (xPos, yPos, fwKeys);
		return 0;

	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	case WM_LBUTTONUP:
		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;
		Z_MouseUp (xPos, yPos, fwKeys);
		if (! (fwKeys & (MK_LBUTTON|MK_RBUTTON|MK_MBUTTON)))
			ReleaseCapture ();
		return 0;

	case WM_GETMINMAXINFO:
	{
		MINMAXINFO *pmmi = (LPMINMAXINFO) lParam;

		pmmi->ptMinTrackSize.x = ZWIN_WIDTH;
		return 0;
	}

	case WM_MOUSEMOVE:
		fwKeys = wParam;        // key flags 
		xPos = (short)LOWORD(lParam);  // horizontal position of cursor 
		yPos = (short)HIWORD(lParam);  // vertical position of cursor 
		yPos = (int)rect.bottom - 1 - yPos;
		Z_MouseMoved (xPos, yPos, fwKeys);
		return 0;

    case WM_SIZE:
		z.width = rect.right;
		z.height = rect.bottom;
		InvalidateRect( globals.Z.hwnd, NULL, false);
		return 0;

	case WM_SIZING:
			if(TryDocking(hWnd, wParam, (LPRECT)lParam, ZWIN_WIDTH))
				return TRUE;
		break;

	case WM_MOVING:
			if(TryDocking(hWnd, wParam, (LPRECT)lParam, ZWIN_WIDTH))
				return TRUE;
			break;

	case WM_NCCALCSIZE:// don't let windows copy pixels
		DefWindowProc (hWnd, uMsg, wParam, lParam);
		return WVR_REDRAW;

	case WM_KILLFOCUS:
	case WM_SETFOCUS:
		SendMessage( hWnd, WM_NCACTIVATE, uMsg == WM_SETFOCUS, 0 );
		return 0;

   	case WM_CLOSE:
        /* call destroy window to cleanup and go away */
        DestroyWindow (hWnd);
		return 0;
    }

	return DefWindowProc (hWnd, uMsg, wParam, lParam);
}


/*
==============
WZ_Create
==============
*/
void WZ_Create (HINSTANCE hInstance)
{
    WNDCLASS   wc;

    /* Register the camera class */
	memset (&wc, 0, sizeof(wc));

    wc.style         = CS_OWNDC;
    wc.lpfnWndProc   = (WNDPROC)WZ_WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = 0;
    wc.hCursor       = LoadCursor (NULL,IDC_CROSS);
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = Z_WINDOW_CLASS;

    if (!RegisterClass (&wc) )
        Error ("WCam_Register: failed");

	globals.Z.hwnd = CreateWindowEx (WS_EX_TOOLWINDOW,
		Z_WINDOW_CLASS ,
		"Z",
		WS_CAPTION | WS_THICKFRAME | WS_MAXIMIZEBOX  | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_CHILD,
		0,20,ZWIN_WIDTH,screen_height-38,	// size
		globals.d_hwndMain,	// parent
		0,		// no menu
		hInstance,
		NULL);
	if (!globals.Z.hwnd)
		Error ("Couldn't create zwindow");

	if(!LoadWindowState(globals.Z.hwnd, globals.Multiple2DViews?"zwindow":"zwindow2"))
		NewWindow = true;

    ShowWindow (globals.Z.hwnd, SW_SHOWDEFAULT);
	globals.Z.hdc = GetDC(globals.Z.hwnd);
	QEW_SetupPixelFormat(&(globals.Z.pfd), &(globals.Z.hdc), false);
	if ( ( globals.Z.hglrc = wglCreateContext( globals.Z.hdc ) ) == 0 )
		Error( "wglCreateContext in WZ_Create failed" );

	if (!wglMakeCurrent( globals.Z.hdc, globals.Z.hglrc ))
		Error ("wglMakeCurrent in WZ_Create failed");

	if (!wglShareLists( globals.Main.hglrc, globals.Z.hglrc ) )
		Error( "wglShareLists in WZ_Create failed" );
//	SetUpFont(&(globals.Z));
}
