
#include "qe3.h"

#define	PAGEFLIPS	2

void Face_DrawXY( face_t *face );

int Orientations[3][3] =
{
	{0, 1, 2},
	{0, 2, 1},
	{1, 2, 0}
};

int *Orientation;

static char *WindowDescriptions[3] =
{
	"XY View",
	"XZ View",
	"YZ View"
};

/*
============
XY_Init
============
*/
void XY_Init (void)
{
	globals.XY.origin[0] = globals.XZ.origin[0] = globals.YZ.origin[0] = 0;
	globals.XY.origin[1] = globals.XZ.origin[1] = globals.YZ.origin[1] = 20;
	globals.XY.origin[2] = globals.XZ.origin[2] = globals.YZ.origin[2] = 46;

	globals.XY.scale = globals.XZ.scale = globals.YZ.scale = 1;
}


int DraggingClipPoint = 0;
int WhichClipPoint;

/*
============================================================================

  MOUSE ACTIONS

============================================================================
*/

static	int	cursorx, cursory;
static	int	buttonstate;
static	int	pressx, pressy;
static	vec3_t	pressdelta;
static	qboolean	press_selection;

void XY_ToPoint (int x, int y, vec3_t point, WInfo_t *winfo)
{
	point[globals.XY_Orientation[0]] = winfo->origin[globals.XY_Orientation[0]] + (x - winfo->width/2)/winfo->scale;
	point[globals.XY_Orientation[1]] = winfo->origin[globals.XY_Orientation[1]] + (y - winfo->height/2)/winfo->scale;
	point[globals.XY_Orientation[2]] = 0;
}

void XY_ToGridPoint (int x, int y, vec3_t point, WInfo_t *winfo)
{
	point[globals.XY_Orientation[0]] = winfo->origin[globals.XY_Orientation[0]] + (x - winfo->width/2)/winfo->scale;
	point[globals.XY_Orientation[1]] = winfo->origin[globals.XY_Orientation[1]] + (y - winfo->height/2)/winfo->scale;
	point[globals.XY_Orientation[2]] = 0;
	point[globals.XY_Orientation[0]] = floor(point[globals.XY_Orientation[0]]/globals.d_gridsize+0.5)*globals.d_gridsize;
	point[globals.XY_Orientation[1]] = floor(point[globals.XY_Orientation[1]]/globals.d_gridsize+0.5)*globals.d_gridsize;
}

void SnapPointToGrid(vec3_t pt)
{
	pt[0] = floor(pt[0]/globals.d_gridsize+0.5)*globals.d_gridsize;
	pt[1] = floor(pt[1]/globals.d_gridsize+0.5)*globals.d_gridsize;
	pt[2] = floor(pt[2]/globals.d_gridsize+0.5)*globals.d_gridsize;
}

/*
==============
XY_MouseDown
==============
*/
void XY_MouseDown (int x, int y, int buttons, WInfo_t *winfo)
{
	vec3_t	point;
	vec3_t	origin, dir, right, up;

	buttonstate = buttons;
	pressx = x;
	pressy = y;
	VectorCopy (vec3_origin, pressdelta);

	XY_ToPoint (x, y, point, winfo);

	VectorCopy (point, origin);
	origin[globals.XY_Orientation[2]] = globals.MaxGridExtent + 1024;

	dir[globals.XY_Orientation[0]] = 0; dir[globals.XY_Orientation[1]] = 0; dir[globals.XY_Orientation[2]] = -1;
	right[globals.XY_Orientation[0]] = 1/winfo->scale; right[globals.XY_Orientation[1]] = 0; right[globals.XY_Orientation[2]] = 0;
	up[globals.XY_Orientation[0]] = 0; up[globals.XY_Orientation[1]] = 1/winfo->scale; up[globals.XY_Orientation[2]] = 0;

	press_selection = (selected_brushes.next != &selected_brushes);

	Sys_GetCursorPos (&cursorx, &cursory);

	// lbutton = manipulate selection
	// shift-LBUTTON = select
	if ( (buttons == MK_LBUTTON)
		|| (buttons == (MK_LBUTTON | MK_SHIFT))
		|| (buttons == (MK_LBUTTON | MK_CONTROL))
		|| (buttons == (MK_LBUTTON | MK_CONTROL | MK_SHIFT)) )
	{	
		Drag_Begin (x, y, buttons, 
			right, up,
			origin, dir);
		return;
	}

	// control mbutton = move camera
	if ((buttonstate == (MK_CONTROL|MK_MBUTTON)) || (buttonstate == (MK_CONTROL|MK_LBUTTON|MK_RBUTTON)))
	{	
		camera.origin[globals.XY_Orientation[0]] = point[globals.XY_Orientation[0]];
		camera.origin[globals.XY_Orientation[1]] = point[globals.XY_Orientation[1]];
		Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
	}

	// mbutton = angle camera
	if ((buttonstate == MK_MBUTTON) || (buttonstate == (MK_LBUTTON|MK_RBUTTON)))
	{	
		VectorSubtract (point, camera.origin, point);
		if (point[globals.XY_Orientation[1]] || point[globals.XY_Orientation[0]])
		{
			camera.angles[YAW] = 180/Q_PI*atan2 (point[1], point[0]);
			Sys_UpdateWindows (W_CAMERA|W_XY | W_XZ | W_YZ);
		}
	}

	// shift mbutton = move z checker
	if ((buttonstate == (MK_SHIFT|MK_MBUTTON)) || (buttonstate == (MK_SHIFT|MK_LBUTTON|MK_RBUTTON)))
	{
		XY_ToPoint (x, y, point, winfo);
		z.origin[globals.XY_Orientation[0]] = point[globals.XY_Orientation[0]];
		z.origin[globals.XY_Orientation[1]] = point[globals.XY_Orientation[1]];
		Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_Z);
		return;
	}

	if(buttonstate == (MK_SHIFT|MK_RBUTTON))
	{
		DraggingClipPoint = 1;
		WhichClipPoint = FindClipperPoint(origin, dir);

		if(WhichClipPoint == -1)
		{
			origin[globals.XY_Orientation[2]] = globals.d_new_brush_top[globals.XY_Orientation[2]];
			SnapPointToGrid(origin);
			WhichClipPoint = Clipper_AddPoint(origin, winfo->Orientation);
		}
		else
			SnapPointToGrid(globals.Clipper[WhichClipPoint]);

		Sys_UpdateWindows(W_XY|W_XZ|W_YZ);
	}
}

/*
==============
XY_MouseUp
==============
*/
void XY_MouseUp (int x, int y, int buttons)
{
	Drag_MouseUp ();
	DraggingClipPoint = 0;

	if (!press_selection)
		Sys_UpdateWindows (W_ALL);
	else if(!globals.UpdateCamera)
		Sys_UpdateWindows (W_CAMERA);
	buttonstate = 0;
}

qboolean DragDelta (int x, int y, vec3_t move, WInfo_t *winfo)
{
	vec3_t	xvec, yvec, delta;
	int		i;

	xvec[globals.XY_Orientation[0]] = 1/winfo->scale;
	xvec[globals.XY_Orientation[1]] = xvec[globals.XY_Orientation[2]] = 0;
	yvec[globals.XY_Orientation[1]] = 1/winfo->scale;
	yvec[globals.XY_Orientation[0]] = yvec[globals.XY_Orientation[2]] = 0;

	for (i=0 ; i<3 ; i++)
	{
		delta[globals.XY_Orientation[i]] = xvec[globals.XY_Orientation[i]]*(x - pressx) + yvec[globals.XY_Orientation[i]]*(y - pressy);
		delta[globals.XY_Orientation[i]] = floor(delta[globals.XY_Orientation[i]]/globals.d_gridsize+0.5)*globals.d_gridsize;
	}
	VectorSubtract (delta, pressdelta, move);
	VectorCopy (delta, pressdelta);

	if (move[0] || move[1] || move[2])
		return true;
	return false;
}

/*
==============
NewBrushDrag
==============
*/
void NewBrushDrag (int x, int y, WInfo_t *winfo)
{
	vec3_t	mins, maxs, junk;
	int		i, h = 0, v = 0;
	float	temp;
	brush_t	*n;

	if (!DragDelta (x,y, junk, winfo))
		return;
	// delete the current selection
	if (selected_brushes.next != &selected_brushes)
		Brush_Free (selected_brushes.next);
	XY_ToGridPoint (pressx, pressy, mins, winfo);
	mins[globals.XY_Orientation[2]] = globals.d_gridsize * ((int)(globals.d_new_brush_bottom[globals.XY_Orientation[2]]/globals.d_gridsize));
	XY_ToGridPoint (x, y, maxs, winfo);
	maxs[globals.XY_Orientation[2]] = globals.d_gridsize * ((int)(globals.d_new_brush_top[globals.XY_Orientation[2]]/globals.d_gridsize));
	if (maxs[globals.XY_Orientation[2]] <= mins[globals.XY_Orientation[2]])
		maxs[globals.XY_Orientation[2]] = mins[globals.XY_Orientation[2]] + globals.d_gridsize;

	for (i=0 ; i<3 ; i++)
	{
		if (mins[i] == maxs[i])
			return;	// don't create a degenerate brush
		if (mins[i] > maxs[i])
		{
			temp = mins[i];
			mins[i] = maxs[i];
			maxs[i] = temp;
		}
	}

	n = Brush_Create (mins, maxs, &globals.d_texturewin.texdef);
	if (!n)
		return;

	Brush_AddToList (n, &selected_brushes);

	Entity_LinkBrush (world_entity, n);

	Brush_Build( n );

//	Sys_UpdateWindows (W_ALL);
	if(x<0)
	{
		h = x;
		winfo->origin[globals.XY_Orientation[0]] += x;
	}
	else if(x>winfo->width)
	{
		h = x-winfo->width;
		winfo->origin[globals.XY_Orientation[0]] += h;
	}
	if(y<0)
	{
		v = y;
		winfo->origin[globals.XY_Orientation[1]] += y;
	}
	else if(y>winfo->height)
	{
		v = y-winfo->height;
		winfo->origin[globals.XY_Orientation[1]] += v;
	}
	if(h || v)
	{
		Sys_GetCursorPos(&x, &y);
		Sys_SetCursorPos(x-h, y+v);
		pressx -= h*winfo->scale;
		pressy -= v*winfo->scale;
	}

	if(globals.UpdateCamera)
		Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_CAMERA);
	else
		Sys_UpdateWindows (W_XY | W_XZ | W_YZ);
}

extern int SnapPlanepts;
/*
==============
XY_MouseMoved
==============
*/
void XY_MouseMoved (int x, int y, int buttons, WInfo_t *winfo)
{
	vec3_t		point;
	float		scale;
	int			i;
	int			AltPressed = GetKeyState(VK_MENU) < 0;
	int			Window2Update;

	if(!globals.Multiple2DViews)
		Window2Update = W_XY;
	else if(winfo->hwnd == globals.XY.hwnd)
		Window2Update = W_XY;
	else if(winfo->hwnd == globals.XZ.hwnd)
		Window2Update = W_XZ;
	else
		Window2Update = W_YZ;

	if (!buttonstate)
	{
		point[globals.XY_Orientation[0]] = winfo->origin[globals.XY_Orientation[0]] + (x - (winfo->width/2))/winfo->scale;
		point[globals.XY_Orientation[1]] = winfo->origin[globals.XY_Orientation[1]] + (y - (winfo->height/2))/winfo->scale;
		point[globals.XY_Orientation[2]] = winfo->origin[globals.XY_Orientation[2]];
		Sys_Status(0, "%7.3f %7.3f %7.3f\n", point[0], point[1], point[2]);
		return;
	}

	// lbutton without selection = drag new brush
	if (buttonstate == MK_LBUTTON && !press_selection)
	{
		NewBrushDrag (x, y, winfo);
		return;
	}

	// lbutton (possibly with control and or shift)
	// with selection = drag selection
	if (buttonstate & MK_LBUTTON && !(buttonstate&MK_RBUTTON))
	{
		Drag_MouseMoved (x, y, buttons);
		if(globals.UpdateCamera)
			Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_CAMERA);
		else
		{
			Sys_UpdateWindows (W_XY|W_XZ|W_YZ);
			update_bits &= ~W_CAMERA;
		}
		return;
	}

	// control mbutton = move camera
	if ((buttonstate == (MK_CONTROL|MK_MBUTTON)) || (buttonstate == (MK_CONTROL|MK_LBUTTON|MK_RBUTTON)))
	{
		XY_ToPoint (x, y, point, winfo);
		camera.origin[globals.XY_Orientation[0]] = point[globals.XY_Orientation[0]];
		camera.origin[globals.XY_Orientation[1]] = point[globals.XY_Orientation[1]];
		Sys_UpdateWindows (W_XY | W_XZ | W_YZ | W_CAMERA);
		return;
	}

	// shift mbutton = move z checker
	if ((buttonstate == (MK_SHIFT|MK_MBUTTON)) || (buttonstate == (MK_SHIFT|MK_LBUTTON|MK_RBUTTON)))
	{
		XY_ToPoint (x, y, point, winfo);
		z.origin[globals.XY_Orientation[0]] = point[globals.XY_Orientation[0]];
		z.origin[globals.XY_Orientation[1]] = point[globals.XY_Orientation[1]];
		Sys_UpdateWindows (W_XY | W_XZ | W_YZ |W_Z);
		return;
	}

	// mbutton = angle camera
	if (buttonstate == MK_MBUTTON || buttonstate == (MK_LBUTTON|MK_RBUTTON))
	{
		int n1 = winfo->Orientation == 0 ? 1 : 2;
		int	n2 = winfo->Orientation == 2 ? 1 : 0;
	    int	nAngle =  winfo->Orientation == 0 ? YAW : PITCH;
		XY_ToPoint (x, y, point, winfo);
		VectorSubtract (point, camera.origin, point);
		
		if (point[n1] || point[n2])
		{
			camera.angles[nAngle] = 180/Q_PI*atan2 (point[n1], point[n2]);
			Sys_UpdateWindows (W_CAMERA|W_XY|W_XZ|W_YZ);
		}

		return;
	}

	if(DraggingClipPoint)
	{
		if(DragDelta(x, y, point, winfo))
		{
			VectorAdd(point, globals.Clipper[WhichClipPoint], globals.Clipper[WhichClipPoint]);
			
			if(globals.ClipperNumPoints == 2)
			{
				VectorCopy(globals.Clipper[1], globals.Clipper[2]);
				globals.Clipper[2][globals.XY_Orientation[2]] += 3;
			}
		}
		



		Sys_UpdateWindows(W_XY|W_XZ|W_YZ);
		return;
	}

	// rbutton = drag xy origin
	if (buttonstate == MK_RBUTTON)
	{
		Sys_GetCursorPos (&x, &y);
		if(!AltPressed)
		{
			if (x != cursorx || y != cursory)
			{
//				winfo->origin[globals.XY_Orientation[0]] -= (x-cursorx)/winfo->scale;
//				winfo->origin[globals.XY_Orientation[1]] += (y-cursory)/winfo->scale;
				winfo->origin[Orientations[winfo->Orientation][0]] -= (x-cursorx)/winfo->scale;
				winfo->origin[Orientations[winfo->Orientation][1]] += (y-cursory)/winfo->scale;
				Sys_SetCursorPos (cursorx, cursory);
				Sys_UpdateWindows (Window2Update);
				if(globals.Clip2XYZ)
					Sys_UpdateWindows (W_Z|W_CAMERA|globals.Multiple2DViews?W_XY|W_XZ|W_YZ:0);
				Sys_Status(0, "Origin: %7.3f, %7.3f, %7.3f\n", winfo->origin[0], winfo->origin[1], winfo->origin[2]);

				if(globals.Multiple2DViews && globals.Lock2DOrigins)
				{
					for(i = 0; i < 2; i++)
					{
						globals.XY.origin[Orientations[0][i]] = winfo->origin[i];
						globals.XZ.origin[Orientations[0][i]] = winfo->origin[i];
						globals.YZ.origin[Orientations[0][i]] = winfo->origin[i];

						if(globals.LockCamTo2DOrigins)
						{
							camera.origin[Orientations[0][i]] = winfo->origin[i];
							Sys_UpdateWindows(W_CAMERA);
						}
					}
					Sys_UpdateWindows(W_XY|W_XZ|W_YZ);
				}
			}
		}
		else
		{
			SnapPlanepts = 0;
			Select_RotateAxis (Orientations[winfo->Orientation][2], winfo->Orientation == 1?(cursorx-x):(x-cursorx));
			SnapPlanepts = 1;
			cursorx=x;
		}
		return;
	}

	// control rbutton = zoom xy view
	if (buttonstate == (MK_CONTROL|MK_RBUTTON))
	{
		Sys_GetCursorPos (&x, &y);
		if (y != cursory)
		{
			scale = 1;
			for(i = abs(y-cursory); i > 0; i--)
				scale *=  y>cursory?1.01:0.99;
			winfo->scale *= scale;
			if (winfo->scale > 64)
				winfo->scale = 64;
			else if (winfo->scale < 0.05)
				winfo->scale = (float)0.05;
			Sys_SetCursorPos (cursorx, cursory);
			Sys_UpdateWindows (Window2Update);
			if(globals.Clip2XYZ)
				Sys_UpdateWindows (W_Z|W_CAMERA|globals.Multiple2DViews?W_XY|W_XZ|W_YZ:0);
			if(globals.Multiple2DViews && globals.Lock2DZoom)
			{
				globals.XY.scale = winfo->scale;
				globals.XZ.scale = winfo->scale;
				globals.YZ.scale = winfo->scale;
				Sys_UpdateWindows(W_XY|W_XZ|W_YZ);
			}
		}
		return;
	}
}


/*
============================================================================

DRAWING

============================================================================
*/


/*
==============
XY_DrawGrid
==============
*/
void XY_DrawGrid (WInfo_t *winfo)
{
	float	x, y, xb, xe, yb, ye;
	int		w, h;

	glDisable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);
	glLineWidth(1);

	w = winfo->width /2.0f / winfo->scale;
	h = winfo->height/2.0f / winfo->scale;

	xb = winfo->origin[globals.XY_Orientation[0]] - w;
	if (xb < region_mins[globals.XY_Orientation[0]])
		xb = region_mins[globals.XY_Orientation[0]];
	xb = 64 * floor (xb/64);

	xe = winfo->origin[globals.XY_Orientation[0]] + w;
	if (xe > region_maxs[globals.XY_Orientation[0]])
		xe = region_maxs[globals.XY_Orientation[0]];
	xe = 64 * ceil (xe/64);

	yb = winfo->origin[globals.XY_Orientation[1]] - h;
	if (yb < region_mins[globals.XY_Orientation[1]])
		yb = region_mins[globals.XY_Orientation[1]];
	yb = 64 * floor (yb/64);

	ye = winfo->origin[globals.XY_Orientation[1]] + h;
	if (ye > region_maxs[globals.XY_Orientation[1]])
		ye = region_maxs[globals.XY_Orientation[1]];
	ye = 64 * ceil (ye/64);

	// draw major blocks

	glColor3dv(globals.d_savedinfo.colors[COLOR_GRIDMAJOR]);

	if ( globals.d_showgrid )
	{		
		glBegin (GL_LINES);
		
		for (x=xb ; x<=xe ; x+=64)
		{
			glVertex2f (x, yb);
			glVertex2f (x, ye);
		}
		for (y=yb ; y<=ye ; y+=64)
		{
			glVertex2f (xb, y);
			glVertex2f (xe, y);
		}
		
		glEnd ();
		
	}

	// draw minor blocks
	if ( globals.d_showgrid && globals.d_gridsize*winfo->scale >= 4)
	{
		glColor3dv(globals.d_savedinfo.colors[COLOR_GRIDMINOR]);

		glBegin (GL_LINES);
		for (x=xb ; x<xe ; x += globals.d_gridsize)
		{
			if ( ! ((int)x & 63) )
				continue;
			glVertex2f (x, yb);
			glVertex2f (x, ye);
		}
		for (y=yb ; y<ye ; y+=globals.d_gridsize)
		{
			if ( ! ((int)y & 63) )
				continue;
			glVertex2f (xb, y);
			glVertex2f (xe, y);
		}
		glEnd ();
	}

	// draw coordinate text if needed

	if ( globals.d_savedinfo.show_coordinates)
	{
		glColor3dv(globals.d_savedinfo.colors[COLOR_GRIDTEXT]);

		for (x=xb ; x<xe ; x+=64)
		{
			glPrintf(x, winfo->origin[globals.XY_Orientation[1]] + h - 6/winfo->scale, 0, "%i", (int)x);
		}
		for (y=yb ; y<ye ; y+=64)
		{
			glPrintf (winfo->origin[globals.XY_Orientation[0]] - w + 1, y, 0, "%i",(int)y);
		}
	}
}

/*
==============
XY_DrawBlockGrid
==============
*/
void XY_DrawBlockGrid (WInfo_t *winfo)
{
	float	x, y, xb, xe, yb, ye;
	int		w, h;
	char	text[32];

	glDisable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);

	w = winfo->width/2 / winfo->scale;
	h = winfo->height/2 / winfo->scale;

	xb = winfo->origin[globals.XY_Orientation[0]] - w;
	if (xb < region_mins[globals.XY_Orientation[0]])
		xb = region_mins[globals.XY_Orientation[0]];
	xb = 1024 * floor (xb/1024);

	xe = winfo->origin[globals.XY_Orientation[0]] + w;
	if (xe > region_maxs[globals.XY_Orientation[0]])
		xe = region_maxs[globals.XY_Orientation[0]];
	xe = 1024 * ceil (xe/1024);

	yb = winfo->origin[globals.XY_Orientation[1]] - h;
	if (yb < region_mins[globals.XY_Orientation[1]])
		yb = region_mins[globals.XY_Orientation[1]];
	yb = 1024 * floor (yb/1024);

	ye = winfo->origin[globals.XY_Orientation[1]] + h;
	if (ye > region_maxs[globals.XY_Orientation[1]])
		ye = region_maxs[globals.XY_Orientation[1]];
	ye = 1024 * ceil (ye/1024);

	// draw major blocks

	glColor3f(0,0,1);
	glLineWidth (2);

	glBegin (GL_LINES);
	
	for (x=xb ; x<=xe ; x+=1024)
	{
		glVertex2f (x, yb);
		glVertex2f (x, ye);
	}
	for (y=yb ; y<=ye ; y+=1024)
	{
		glVertex2f (xb, y);
		glVertex2f (xe, y);
	}
	
	glEnd ();
	glLineWidth (1);

	// draw coordinate text if needed

	for (x=xb ; x<xe ; x+=1024)
		for (y=yb ; y<ye ; y+=1024)
		{
			glRasterPos2f (x+512, y+512);
			sprintf (text, "%i,%i",(int)floor(x/1024), (int)floor(y/1024) );
			glCallLists (strlen(text), GL_UNSIGNED_BYTE, text);
		}

	glColor4f(0, 0, 0, 0);
}


void DrawCameraIcon (WInfo_t *winfo)
{
	float	x, y, a, xadd = 48;

	x = camera.origin[globals.XY_Orientation[0]];
	y = camera.origin[globals.XY_Orientation[1]];
	if(!globals.XY_Orientation[3])
		a = camera.angles[YAW]/180*Q_PI;
	else
	{
		a = camera.angles[PITCH]/180*Q_PI;

		if(camera.angles[YAW] < -180)
			camera.angles[YAW] += 360; 

		if(camera.angles[YAW] > 180)
			camera.angles[YAW] -= 360;

		if(winfo->Orientation == 1)
		{
			if(fabs(camera.angles[YAW]) > 90)
				xadd = -xadd;
		}
		else
		{
			if(camera.angles[YAW] < 0)
				xadd = -xadd;
		}
//		Sys_Printf("camera.angles[YAW] = %lf\n", camera.angles[YAW]);
	}

	glColor4f(globals.d_savedinfo.colors[COLOR_CAMERAICON][0],
			  globals.d_savedinfo.colors[COLOR_CAMERAICON][1],
			  globals.d_savedinfo.colors[COLOR_CAMERAICON][2],
			  (float)0.5);
	glBegin(GL_LINE_STRIP);
	glVertex3f (x-16,y,0);
	glVertex3f (x,y+8,0);
	glVertex3f (x+16,y,0);
	glVertex3f (x,y-8,0);
	glVertex3f (x-16,y,0);
	glVertex3f (x+16,y,0);
	glEnd ();
	
	glBegin(GL_LINE_STRIP);
	glVertex3f (x+xadd*cos(a+Q_PI/4), y+48*sin(a+Q_PI/4), 0);
	glVertex3f (x, y, 0);
	glVertex3f (x+xadd*cos(a-Q_PI/4), y+48*sin(a-Q_PI/4), 0);
	glEnd ();

}

void DrawZIcon (void)
{
	float	x, y;

//	if(globals.XY_Orientation[3])
	{
		x = z.origin[globals.XY_Orientation[0]];
		y = z.origin[globals.XY_Orientation[1]];
		
		glEnable (GL_BLEND);
		glDisable (GL_TEXTURE_2D);
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
		glDisable (GL_CULL_FACE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//		glColor4f (0.0, 0.0, 1.0, 0.25);
		glColor4f(
			globals.d_savedinfo.colors[COLOR_ZICON][0],
			globals.d_savedinfo.colors[COLOR_ZICON][1],
			globals.d_savedinfo.colors[COLOR_ZICON][2],
			0.25
			);
		glBegin(GL_QUADS);
		glVertex3f (x-8,y-8,0);
		glVertex3f (x+8,y-8,0);
		glVertex3f (x+8,y+8,0);
		glVertex3f (x-8,y+8,0);
		glEnd ();
		glDisable (GL_BLEND);

//		glColor4f (0.0, 0.0, 1.0, 1);
		glColor4f(
			globals.d_savedinfo.colors[COLOR_ZICON][0],
			globals.d_savedinfo.colors[COLOR_ZICON][1],
			globals.d_savedinfo.colors[COLOR_ZICON][2],
			1.0
			);
		
		glBegin(GL_LINE_LOOP);
		glVertex3f (x-8,y-8,0);
		glVertex3f (x+8,y-8,0);
		glVertex3f (x+8,y+8,0);
		glVertex3f (x-8,y+8,0);
		glEnd ();
		
		glBegin(GL_LINE_STRIP);
		glVertex3f (x-4,y+4,0);
		glVertex3f (x+4,y+4,0);
		glVertex3f (x-4,y-4,0);
		glVertex3f (x+4,y-4,0);
		glEnd ();
	}
}


/*
==================
FilterBrush
==================
*/
BOOL FilterBrush(brush_t *pb)
{
	if (!pb->owner)
		return FALSE;		// during construction

	if (globals.d_savedinfo.exclude & EXCLUDE_CLIP)
	{
		if(globals.gametype == QUAKE2 || globals.gametype == SOF)
		{
			if (pb->brush_faces->texdef.contents & (CONTENTS_PLAYERCLIP|CONTENTS_MONSTERCLIP))
				return TRUE;
		}
		else
		{
			if (!strnicmp(pb->brush_faces->texdef.name, "clip", 4))
				return TRUE;
		}
	}

	if (globals.d_savedinfo.exclude & EXCLUDE_WATER)
	{
		if(globals.gametype == QUAKE2 || globals.gametype == SOF)
		{
			if (pb->brush_faces->texdef.contents & CONTENTS_WATER)
				return TRUE;
		}
		else
		{
			if (pb->brush_faces->texdef.name[0] == '*')
				return TRUE;
		}
	}

	if (globals.d_savedinfo.exclude & EXCLUDE_DETAIL)
	{
		if (pb->brush_faces->texdef.contents & CONTENTS_DETAIL)
			return TRUE;
	}

	if (pb->owner == world_entity)
	{
		if (globals.d_savedinfo.exclude & EXCLUDE_WORLD)
			return TRUE;
	}
	else if (globals.d_savedinfo.exclude & EXCLUDE_ENT)
		return TRUE;
	
	if (globals.d_savedinfo.exclude & EXCLUDE_LIGHTS)
	{
		if (!strncmp(pb->owner->eclass->name, "light", 5))
			return TRUE;
	}

	if (globals.d_savedinfo.exclude & EXCLUDE_PATHS)
	{
		if (!strncmp(pb->owner->eclass->name, "path", 4))
			return TRUE;
	}

	if(globals.gametype == HALFLIFE)
	{
		if(!globals.ShowHint && !stricmp(pb->brush_faces->texdef.name, "hint"))
			return TRUE;

		if(globals.ShowSkybox && !stricmp(pb->brush_faces->texdef.name, "sky"))
			return TRUE;
	}
	else if(globals.gametype == QUAKE1)
	{
		if(globals.ShowSkybox && !strnicmp(pb->brush_faces->texdef.name, "sky", 3))
			return TRUE;
	}
	else
	{
		if(globals.ShowSkybox && pb->brush_faces->texdef.flags & SURF_SKY)
			return TRUE;

		if(!globals.ShowHint && pb->brush_faces->texdef.flags & 0x100)
			return TRUE;
	}


	if(globals.Clip2XYZ)
	{
		vec3_t	mins, maxs;

		mins[0] = globals.XY.origin[globals.XY_Orientation[0]] - (globals.XY.width/2)/globals.XY.scale;
		mins[1] = globals.XY.origin[globals.XY_Orientation[1]] - (globals.XY.height/2)/globals.XY.scale;
		mins[2] = z.origin[2] - (z.height/2)/z.scale;
		maxs[0] = globals.XY.origin[0] + (globals.XY.width/2)/globals.XY.scale;
		maxs[1] = globals.XY.origin[1] + (globals.XY.height/2)/globals.XY.scale;
		maxs[2] = z.origin[2] + (z.height/2)/z.scale;

		if(pb->mins[0] > maxs[0] ||
			pb->mins[1] > maxs[1] ||
			pb->mins[2] > maxs[2] ||
			pb->maxs[0] < mins[0] ||
			pb->maxs[1] < mins[1] ||
			pb->maxs[2] < mins[2])
			return TRUE;
	}

	return FALSE;
}

/*
=============================================================

  PATH LINES

=============================================================
*/

/*
==================
DrawPathLines

Draws connections between entities.
Needs to consider all entities, not just ones on screen,
because the lines can be visible when neither end is.
Called for both camera view and xy view.
==================
*/
void DrawPathLines (void)
{
	int			i, j, k;
	vec3_t		mid, mid1;
	entity_t	*se, *te;
	brush_t		*sb, *tb;
	char		*psz;
	vec3_t		dir, s1, s2;
	vec_t		len, f;
	int			arrows;
	int			num_entities;
	char		*ent_target[MAX_MAP_ENTITIES];
	entity_t	*ent_entity[MAX_MAP_ENTITIES];
	epair_t		*ep;


	num_entities = 0;
	for (te = entities.next ; te != &entities && num_entities != MAX_MAP_ENTITIES ; te = te->next)
	{
		ent_target[num_entities] = ValueForKey (te, "target");
		if (ent_target[num_entities][0])
		{
			ent_entity[num_entities] = te;
			num_entities++;
			continue;
		}

		ent_target[num_entities] = ValueForKey (te, "killtarget");
		if (ent_target[num_entities][0])
		{
			ent_entity[num_entities] = te;
			num_entities++;
			continue;
		}
		// Half-Life's multi_manager can have a lot of targets
		if(globals.gametype != HALFLIFE)
			continue;

		if(!strcmp(ValueForKey(te, "classname"), "multi_manager"))
		{
			for (ep=te->epairs ; ep && num_entities != MAX_MAP_ENTITIES ; ep=ep->next)
			{
				if (!strcmp (ep->key, "classname") ) continue;
				if (!strcmp (ep->key, "spawnflags") ) continue;
				if (!strcmp (ep->key, "origin") ) continue;
				if (!strcmp (ep->key, "targetname") ) continue;
				ent_target[num_entities] = ep->key;
				ent_entity[num_entities++] = te;
			}
		}
	}

	for (se = entities.next ; se != &entities ; se = se->next)
	{
		psz = ValueForKey(se, "targetname");
	
		if (psz == NULL || psz[0] == '\0')
			continue;
		
		sb = se->brushes.onext;
		if (sb == &se->brushes)
			continue;

		for (k=0 ; k<num_entities ; k++)
		{
			if (strcmp (ent_target[k], psz))
				continue;

			te = ent_entity[k];
			tb = te->brushes.onext;
			if (tb == &te->brushes)
				continue;

			for (i=0 ; i<3 ; i++)
				mid[i] = (sb->mins[i] + sb->maxs[i])*0.5; 

			for (i=0 ; i<3 ; i++)
				mid1[i] = (tb->mins[i] + tb->maxs[i])*0.5; 

			VectorSubtract (mid1, mid, dir);
			len = VectorNormalize (dir);
			s1[0] = -dir[1]*8 + dir[0]*8;
			s2[0] = dir[1]*8 + dir[0]*8;
			s1[1] = dir[0]*8 + dir[1]*8;
			s2[1] = -dir[0]*8 + dir[1]*8;

			glColor3f (se->eclass->color[0], se->eclass->color[1], se->eclass->color[2]);

			glBegin(GL_LINES);
			glVertex3dv(mid);
			glVertex3dv(mid1);

			arrows = (int)(len / 256) + 1;

			for (i=0 ; i<arrows ; i++)
			{
				f = len * (i + 0.5) / arrows;

				for (j=0 ; j<3 ; j++)
					mid1[j] = mid[j] + f*dir[j];
				glVertex3dv (mid1);
				glVertex3d (mid1[0] + s1[0], mid1[1] + s1[1], mid1[2]);
				glVertex3dv (mid1);
				glVertex3d (mid1[0] + s2[0], mid1[1] + s2[1], mid1[2]);
			}

			glEnd();
		}
	}

	return;
}

//=============================================================

void Brush_DrawXY_Flat(brush_t *b); // Flat-shaded draw mode
void Brush_DrawXY_Textured(brush_t *b); // Textured draw mode
void Face_DrawXY_Solid(face_t *face );

void DrawListWire(brush_t *head, int *drawn)
{
	brush_t	*b;

	for (b = head->next; b != head; b = b->next)
	{
		if(drawn)
			*drawn++;
		
		Brush_DrawXY(b);
	}
}

void XY_DrawSelWire(int *drawn)
{
	if(globals.StippleSelection)
	{
		glEnable (GL_LINE_STIPPLE);
		glLineStipple (3, 0xaaaa);
	}
	
	if(!globals.ClipperNumPoints)
	{
		DrawListWire(&selected_brushes, drawn);
	}
	else
	{
		if(globals.ClipperMode == 1)
		{
			glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
			DrawListWire(&clipper_brushes_back, NULL);
			glColor3f(1.0, 1.0, 0.0);
			DrawListWire(&clipper_brushes_front, NULL);
		}
		else if(globals.ClipperMode == 2)
		{
			glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
			DrawListWire(&clipper_brushes_front, NULL);
			glColor3f(1.0, 1.0, 0.0);
			DrawListWire(&clipper_brushes_back, NULL);
		}
		else
		{
			glColor3f(1.0, 1.0, 0.0);
			DrawListWire(&clipper_brushes_front, NULL);
			DrawListWire(&clipper_brushes_back, NULL);
		}		
	}
}

void DrawListSolid(brush_t *list)
{
	brush_t	*b;
	face_t	*f;

		for (b = list->next; b != list; b = b->next)
		{
			for (f = b->brush_faces; f; f = f->next)
				Face_DrawXY_Solid(f);
		}
}

void XY_DrawSelSolid(int *drawn)
{
	brush_t	*b;
	
	for (b = selected_brushes.next; b != &selected_brushes; b = b->next)
	{
		*drawn++;
		Brush_DrawXY_Flat(b);
	}
	
	glColor4f(globals.d_savedinfo.colors[COLOR_SEL_COLOR][0], globals.d_savedinfo.colors[COLOR_SEL_COLOR][1], globals.d_savedinfo.colors[COLOR_SEL_COLOR][2], globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
	glEnable (GL_BLEND);
	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_TEXTURE_2D);
	
	if(!globals.ClipperNumPoints)
	{
		DrawListSolid(&selected_brushes);
	}
	else
	{
		if(globals.ClipperMode == 1)
		{
			glColor4f(globals.d_savedinfo.colors[COLOR_SEL_COLOR][0], globals.d_savedinfo.colors[COLOR_SEL_COLOR][1], globals.d_savedinfo.colors[COLOR_SEL_COLOR][2], globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_back);
			glColor4f(1.0, 1.0, 0.0, globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_front);
		}
		else if(globals.ClipperMode == 2)
		{
			glColor4f(globals.d_savedinfo.colors[COLOR_SEL_COLOR][0], globals.d_savedinfo.colors[COLOR_SEL_COLOR][1], globals.d_savedinfo.colors[COLOR_SEL_COLOR][2], globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_front);
			glColor4f(1.0, 1.0, 0.0, globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_back);
		}
		else
		{
			glColor4f(1.0, 1.0, 0.0, globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_front);
			DrawListSolid(&clipper_brushes_back);
		}		
	}
	
	glDisable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
	glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
	
	if(!globals.ClipperNumPoints)
	{
		DrawListSolid(&selected_brushes);
	}
	else
	{
		if(globals.ClipperMode == 1)
		{
			glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
			DrawListWire(&clipper_brushes_back, NULL);
			glColor3f(1.0, 1.0, 0.0);
			DrawListWire(&clipper_brushes_front, NULL);
		}
		else if(globals.ClipperMode == 2)
		{
			glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
			DrawListWire(&clipper_brushes_front, NULL);
			glColor3f(1.0, 1.0, 0.0);
			DrawListWire(&clipper_brushes_back, NULL);
		}
		else
		{
			glColor3f(1.0, 1.0, 0.0);
			DrawListWire(&clipper_brushes_front, NULL);
			DrawListWire(&clipper_brushes_back, NULL);
		}		
	}
}

void DrawListTextured(brush_t *list)
{
	brush_t	*b;
	face_t	*f;
	
	for (b = list->next; b != list; b = b->next)
	{
		for (f = b->brush_faces; f ;f = f->next)
			Face_DrawXY_Solid(f);
	}
}

void XY_DrawSelTextured(int *drawn)
{
	brush_t	*b;

	for (b = selected_brushes.next; b != &selected_brushes ; b = b->next)
	{
		*drawn++;
		Brush_DrawXY_Textured(b);
	}
	
	glColor4f(globals.d_savedinfo.colors[COLOR_SEL_COLOR][0], globals.d_savedinfo.colors[COLOR_SEL_COLOR][1], globals.d_savedinfo.colors[COLOR_SEL_COLOR][2], globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
	glEnable (GL_BLEND);
	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_TEXTURE_2D);
	
	if(!globals.ClipperNumPoints)
	{
		DrawListSolid(&selected_brushes);
	}
	else
	{
		if(globals.ClipperMode == 1)
		{
			glColor4f(globals.d_savedinfo.colors[COLOR_SEL_COLOR][0], globals.d_savedinfo.colors[COLOR_SEL_COLOR][1], globals.d_savedinfo.colors[COLOR_SEL_COLOR][2], globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_back);
			glColor4f(1.0, 1.0, 0.0, globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_front);
		}
		else if(globals.ClipperMode == 2)
		{
			glColor4f(globals.d_savedinfo.colors[COLOR_SEL_COLOR][0], globals.d_savedinfo.colors[COLOR_SEL_COLOR][1], globals.d_savedinfo.colors[COLOR_SEL_COLOR][2], globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_front);
			glColor4f(1.0, 1.0, 0.0, globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_back);
		}
		else
		{
			glColor4f(1.0, 1.0, 0.0, globals.d_savedinfo.alpha[COLOR_SEL_COLOR]);
			DrawListSolid(&clipper_brushes_front);
			DrawListSolid(&clipper_brushes_back);
		}
		
	}
	
	glDisable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
	glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
	
	if(!globals.ClipperNumPoints)
	{
		DrawListWire(&selected_brushes, NULL);
	}
	else
	{
		if(globals.ClipperMode == 1)
		{
			glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
			DrawListWire(&clipper_brushes_back, NULL);
			glColor3f(1.0, 1.0, 0.0);
			DrawListWire(&clipper_brushes_front, NULL);
		}
		else if(globals.ClipperMode == 2)
		{
			glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
			DrawListWire(&clipper_brushes_front, NULL);
			glColor3f(1.0, 1.0, 0.0);
			DrawListWire(&clipper_brushes_back, NULL);
		}
		else
		{
			glColor3f(1.0, 1.0, 0.0);
			DrawListWire(&clipper_brushes_front, NULL);
			DrawListWire(&clipper_brushes_back, NULL);
		}
		
	}
}

/*
==============
XY_Draw
==============
*/
void XY_Draw (WInfo_t *winfo)
{
    brush_t		*brush;
	float		w, h;
	entity_t	*e;
	double		start, end;
	vec3_t		mins, maxs;
	int			drawn, culled;
	int			i;
	facesel_t	*currentface;
	visgroup_t	*g;

	if (!active_brushes.next)
		return;	// not valid yet

	if (globals.d_xy.timing)
		start = Sys_DoubleTime ();

	Orientation = Orientations[winfo->Orientation];

	//
	// clear
	//
	globals.d_xy.d_dirty = false;

	glViewport(0, 0, winfo->width, winfo->height);
	glClearColor (
		globals.d_savedinfo.colors[COLOR_GRIDBACK][0],
		globals.d_savedinfo.colors[COLOR_GRIDBACK][1],
		globals.d_savedinfo.colors[COLOR_GRIDBACK][2],
		0);

    glClear(GL_COLOR_BUFFER_BIT);

	//
	// set up viewpoint
	//
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();

	w = winfo->width/2 / winfo->scale;
	h = winfo->height/2 / winfo->scale;
	mins[Orientation[0]] = winfo->origin[Orientation[0]] - w;
	maxs[Orientation[0]] = winfo->origin[Orientation[0]] + w;
	mins[Orientation[1]] = winfo->origin[Orientation[1]] - h;
	maxs[Orientation[1]] = winfo->origin[Orientation[1]] + h;
	mins[Orientation[2]] = -(globals.MaxGridExtent + 1024);
	maxs[Orientation[2]] = globals.MaxGridExtent + 1024;


	glOrtho (mins[globals.XY_Orientation[0]], maxs[globals.XY_Orientation[0]], mins[globals.XY_Orientation[1]], maxs[globals.XY_Orientation[1]], mins[globals.XY_Orientation[2]], maxs[globals.XY_Orientation[2]]);
//	glOrtho (mins[0], maxs[0], mins[1], maxs[1], mins[2], maxs[2]);
	//
	// now draw the grid
	//
	XY_DrawGrid (winfo);

	//
	// draw stuff
	//

	switch (globals.d_xy.draw_mode)
	{
		case cd_wire:
		    glShadeModel	(GL_FLAT);
			glPolygonMode	(GL_FRONT_AND_BACK, GL_LINE);
			glDisable		(GL_TEXTURE_2D);
			glDisable		(GL_TEXTURE_1D);
			glDisable		(GL_BLEND);
			glDisable		(GL_DEPTH_TEST);
			glColor3f		(1.0, 1.0, 1.0);
//			glEnable (GL_LINE_SMOOTH);
			break;
		
		case cd_solid:
			glCullFace		(GL_FRONT);
			glEnable		(GL_CULL_FACE);
			glShadeModel	(GL_FLAT);
			glPolygonMode	(GL_FRONT_AND_BACK, GL_FILL);
			glDisable		(GL_TEXTURE_2D);
			glDisable		(GL_BLEND);
			glEnable		(GL_DEPTH_TEST);
			glDepthFunc		(GL_LEQUAL);
//			glMatrixMode	(GL_TEXTURE);
			break;
		
		case cd_texture:
			glCullFace		(GL_FRONT);
			glEnable		(GL_CULL_FACE);		
			glShadeModel	(GL_FLAT);
			glPolygonMode	(GL_FRONT_AND_BACK, GL_FILL);
			glEnable		(GL_TEXTURE_2D);
			glTexEnvf		(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glDisable		(GL_BLEND);
			glEnable		(GL_DEPTH_TEST);
			glDepthFunc		(GL_LEQUAL);
//			glMatrixMode	(GL_TEXTURE);
			break;		
	}

//	glEnable(GL_TEXTURE_2D);
//	glEnable(GL_TEXTURE_1D);
//	glDisable(GL_DEPTH_TEST);
//	glDisable(GL_BLEND);
//		glEnable (GL_LINE_SMOOTH);

	if(winfo->Orientation)
	{
		glPushMatrix();
		
		if(winfo->Orientation == 1)
			glRotatef(-90, 1, 0, 0);
		else
		{
			glRotatef(-90, 1, 0, 0);
			glRotatef(-90, 0, 0, 1);
		}
	}
	drawn = culled = 0;

	e = NULL;
	g = NULL;
    glClear(GL_DEPTH_BUFFER_BIT);
	
	for (brush = active_brushes.next ; brush != &active_brushes ; brush=brush->next)
	{
		if (brush->mins[0] > maxs[0] || brush->mins[1] > maxs[1] || brush->mins[2] > maxs[2] || brush->maxs[0] < mins[0] || brush->maxs[1] < mins[1] || brush->maxs[2] < mins[2])
		{
			culled++;
			continue;		// off screen
		}
		
		if (FilterBrush (brush))
			continue;
		
		drawn++;
		
		switch(globals.d_xy.draw_mode)
		{
			case cd_wire:
				if(!globals.DrawVisGroupColors || brush->group == visgroups.next) // Don't use default group's color
				{
					if (brush->owner != e)
					{
						e = brush->owner;
						glColor3fv(e->eclass->color);
					}
				}
				else
				{
					if(brush->group != g)
					{
						g = brush->group;
						glColor3dv(g->color);
					}
				}
				
				Brush_DrawXY( brush );
				break;
			
			case cd_solid:
				Brush_DrawXY_Flat( brush );
				break;
			
			case cd_texture:
				if (brush->owner != e)
				{
					e = brush->owner;
					if(globals.d_xy.draw_mode == cd_wire)
						glColor3fv(e->eclass->color);
				}

				Brush_DrawXY_Textured( brush );
				break;
		}
	}

	if(globals.ShowEntityConnections)
		DrawPathLines ();

	//
	// draw pointfile
	//
	if ( globals.d_pointfile_display_list)
		glCallList (globals.d_pointfile_display_list);

	//
	// draw block grid
	//
	if (globals.show_blocks)
	{
		if(winfo->Orientation)
			glPopMatrix();

		XY_DrawBlockGrid (winfo);

		if(winfo->Orientation)
		{
			glPushMatrix();
			
			if(winfo->Orientation == 1)
				glRotatef(-90, 1, 0, 0);
			else
			{
				glRotatef(-90, 1, 0, 0);
				glRotatef(-90, 0, 0, 1);
			}
		}
	}

	//
	// now draw selected brushes
	//
	glTranslatef( globals.d_select_translate[0], globals.d_select_translate[1], globals.d_select_translate[2]);

	glColor3dv(globals.d_savedinfo.colors[COLOR_SEL_OUTLINE]);
	glLineWidth (2);

	switch(globals.d_xy.draw_mode)
	{
		case cd_wire:
			XY_DrawSelWire(&drawn);
			break;

		case cd_solid:
			XY_DrawSelSolid(&drawn);
			break;

		case cd_texture:
			XY_DrawSelTextured(&drawn);
			break;
	}
	for(currentface = selected_faces; currentface; currentface = currentface->next)
			Face_DrawXY(currentface->face);

	glDisable (GL_LINE_STIPPLE);
	glLineWidth (1);

	// edge / vertex flags

	if (globals.d_select_mode == sel_vertex)
	{
		glPointSize (4);
		glColor3f (0,1,0);
		glBegin (GL_POINTS);
		for (i=0 ; i<globals.d_numpoints ; i++)
			glVertex3dv (globals.d_points[i]);
		glEnd ();
		glPointSize (1);
	}
	else if (globals.d_select_mode == sel_edge)
	{
		vec_t	*v1, *v2;

		glPointSize (4);
		glColor3f (0,0,1);
		glBegin (GL_POINTS);
		for (i=0 ; i<globals.d_numedges ; i++)
		{
			v1 = globals.d_points[globals.d_edges[i].p1];
			v2 = globals.d_points[globals.d_edges[i].p2];
			glVertex3f ( (v1[0]+v2[0])*0.5,(v1[1]+v2[1])*0.5,(v1[2]+v2[2])*0.5);
		}
		glEnd ();
		glPointSize (1);
	}

	if(globals.ClipperNumPoints)
	{
		glPointSize (4);
		glColor3f (1,0,0);
		glBegin (GL_POINTS);
		for (i=0 ; i<globals.ClipperNumPoints ; i++)
		{
			glVertex3dv (globals.Clipper[i]);
		}
		glEnd ();
		glPointSize (1);
	}

	glTranslatef (-globals.d_select_translate[0], -globals.d_select_translate[1], -globals.d_select_translate[2]);

	//
	// now draw camera point
	//
	if(winfo->Orientation)
		glPopMatrix();

	DrawCameraIcon (winfo);
	DrawZIcon ();

    glFinish();
	QE_CheckOpenGLForErrors("XY_Draw()", GL_NO_ERROR);

	if (globals.d_xy.timing)
	{
		end = Sys_DoubleTime ();
		Sys_Printf ("xy: %i ms\n", (int)(1000*(end-start)));
	}
}

void SetupXYOrientation(int Orientation)
{
	globals.XY_Orientation[0] = Orientations[Orientation][0];
	globals.XY_Orientation[1] = Orientations[Orientation][1];
	globals.XY_Orientation[2] = Orientations[Orientation][2];
	globals.XY_Orientation[3] = Orientation;

	if(!globals.Multiple2DViews)
	{
		if(globals.XY.Orientation != Orientation)
		{
			SetWindowText(globals.XY.hwnd, WindowDescriptions[globals.XY_Orientation[3]]);
			globals.XY.Orientation = Orientation;
		}
	}
}
